 /**
 *  Java binding for libGlass
 *  Copyright (C) 2003,2004
 *  Bruno Barberi Gnecco and Marcelo de Paiva Guimaraes
 *  Copyright (C) 2005-2010 Corollarium Tecnologia
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <string>
#include "Glass_Alias.h"
#include "glassjava.h"

static Alias<std::string> *getAlias(JNIEnv *jenv, jobject jobj) {
	jfieldID fid;
	jclass cls = jenv->GetObjectClass(jobj);

	fid = jenv->GetFieldID(cls, "pointer", "J");
	if (fid == NULL) {
		DEBUG("ERROR" << __LINE__);
		return NULL;
	}
	return (Alias<std::string> *)jenv->GetLongField(jobj, fid);
}


/*
 * Class:     Glass_Alias
 * Method:    createAlias
 * Signature: (Ljava/lang/String;)J
 */
JNIEXPORT jlong JNICALL Java_Glass_Alias_createAlias
  (JNIEnv *jenv, jobject jobj, jstring name, jstring data, jint propagation) {
	Alias<std::string> *alias;

	const char *strname = jenv->GetStringUTFChars(name, 0);
	const char *strdata = jenv->GetStringUTFChars(data, 0);

	try {
		propagationMode mode = static_cast<propagationMode>(propagation);
		if (mode != Local && mode != Global) {
			throw Exception("Invalid propagation mode");
		}

		alias = new Alias<std::string>(strname, strdata, mode);
	}
	catch (Exception e) {
		jenv->ReleaseStringUTFChars(name, strname);
		jenv->ReleaseStringUTFChars(name, strdata);
		jclass excCls = jenv->FindClass("java/lang/Exception");
		if (excCls != 0) {
			jenv->ThrowNew(excCls, e.getMessage().c_str());
		}
		return 0;
	}

	jenv->ReleaseStringUTFChars(name, strname);
	jenv->ReleaseStringUTFChars(name, strdata);
	return (jlong)alias;
}

/*
 * Class:     Glass_Alias
 * Method:    destroyAlias
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_Glass_Alias_destroyAlias
  (JNIEnv *jenv, jobject jobj) {
	Alias<std::string> *alias = getAlias(jenv, jobj);
	delete alias;
}


/*
 * Class:     Glass_Alias
 * Method:    _addAlias
 * Signature: (Ljava/lang/String;Ljava/lang/String;)Z
 */
JNIEXPORT jboolean JNICALL Java_Glass_Alias__1addAlias
(JNIEnv *jenv, jobject jobj, jstring name, jstring data) {
	Alias<std::string> *alias = getAlias(jenv, jobj);
	const char *strname = jenv->GetStringUTFChars(name, 0);
	const char *strdata = jenv->GetStringUTFChars(data, 0);
	bool ret = alias->addAlias(strname, strdata);
	jenv->ReleaseStringUTFChars(name, strname);
	jenv->ReleaseStringUTFChars(name, strdata);
	return ret;
}

/*
 * Class:     Glass_Alias
 * Method:    deleteAlias
 * Signature: (Ljava/lang/String;)Z
 */
JNIEXPORT jboolean JNICALL Java_Glass_Alias_deleteAlias
  (JNIEnv *jenv, jobject jobj, jstring name) {
	Alias<std::string> *alias = getAlias(jenv, jobj);
	const char *str = jenv->GetStringUTFChars(name, 0);
	bool ret = alias->deleteAlias(str);
	jenv->ReleaseStringUTFChars(name, str);
	return ret;
}

/*
 * Class:     Glass_Alias
 * Method:    checkAlias
 * Signature: ()Z
 */
JNIEXPORT jboolean JNICALL Java_Glass_Alias_checkAlias
  (JNIEnv *jenv, jobject jobj) {
	Alias<std::string> *alias = getAlias(jenv, jobj);
	return alias->checkAlias();
}

/*
 * Class:     Glass_Alias
 * Method:    associate
 * Signature: (ILjava/lang/String;)Z
 */
JNIEXPORT jboolean JNICALL Java_Glass_Alias_associate
  (JNIEnv *jenv, jobject jobj, jint nodeid, jstring name) {
	Alias<std::string> *alias = getAlias(jenv, jobj);
	const char *strname = jenv->GetStringUTFChars(name, 0);
	bool ret = alias->associate((nodeId) nodeid, strname); // TODO
	jenv->ReleaseStringUTFChars(name, strname);
	return ret;
}

/*
 * Class:     Glass_Alias
 * Method:    _getData
 * Signature: ()Ljava/lang/String;
 */
JNIEXPORT jstring JNICALL Java_Glass_Alias__1getData
  (JNIEnv *jenv, jobject jobj) {
	Alias<std::string> *alias = getAlias(jenv, jobj);

	std::string data = alias->getData();
	jstring jdata = jenv->NewStringUTF (data.c_str());

	return jdata;
}
