/**
 *  Java binding for libGlass
 *  Copyright (C) 2003-2010
 *  Corollarium Technologies
 * 
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  @author      Corollarium Technologies
 *  @version     %I%, %G%
 *  @since       1.0
 */

package Glass;

import java.util.HashMap;

/**
 * 
 * @author corollarium
 *
 * @param <T> The variable type.
 */
public class Alias<T> extends Plugin {
	/**
	 * Map of string to object, used to convert from an alias to an object.
	 */
	private HashMap<String, T> map;
	
	/**
	 * A wrapper for the C++ constructor.
	 * 
	 * @param name The alias name
	 * @param defaultValue The default value key
	 * @return A pointer to the C++ instance.
	 */
	private native long createAlias(String name, String defaultValue, int propagation);
	
	/**
	 * A wrapper for the C++ destructor.
	 */
	private native void destroyAlias();

	/**
	 * 	
	 * @see #Alias(String , T, int)
	 */
	public Alias(String aliasname, T value) throws GlassException {
		this(aliasname, value, Glass.propagationLocal);
	}

	/**
	 * Constructor. Creates an alias.
	 * 
	 * @param aliasname The name.
	 * @param value The default value for the alias.
	 * @throws GlassException
	 */
	public Alias(String aliasname, T value, int propagation) throws GlassException {
		this.map = new HashMap<String, T>();
		this.map.put("___DEFAULT", value);

		try {
			this.pointer = createAlias(aliasname, "___DEFAULT", propagation);
		}
		catch (Exception e) {
			throw new GlassException(e);
		}
	}

	protected void finalize() throws Throwable {
		try {
			destroyAlias();
		}
		finally {
			super.finalize();
		}
	}

	/**
	 * Wrapper to the C++ method.
	 * @param name The alias.
	 * @param value The key to data.
	 * @return true on sucess, false otherwise.
	 */
	public native boolean _addAlias(String name, String value);

	/**
	 * Add an alias to function.
	 * @param name The alias.
	 * @param value The data.
	 * @return true If successful. false Otherwise.
	 */
	public boolean addAlias(String name, T value) {
		this.map.put(name, value);
		return this._addAlias(name, name);
	}

	/**
	 * Deletes an alias to function.
	 * 
	 * @param name The alias.
	 * @return true If successful. false Alias doesn't exist.
	 */
	public native boolean deleteAlias(String name);

	/**
	 * Checks if there's data associated with the current alias.
	 * @return true If there is. false If there is not.
	 */
	public native boolean checkAlias();
	
	/**
	 * Associates a given node to an alias, propagating to other nodes.
	 * 
	 * @param id The node id.
	 * @param name Alias association
	 * @return True if message sent successfully. Doesn't guarantee either
	 *  that node "id" exists or that it has an associated alias "name".
	*/
	public native boolean associate(int id, String name);
	
	/**
	 * Returns associated to the current alias.
	 * 
	 * @return The associated data, or null if nothing found.
	 */
	public T getData() {
		String n = this._getData();
		return this.map.get(n);
	}

	/**
	 * Wrapper to the C++ call.
	 * @return The alias key.
	 */
	private native String _getData(); 
}
