 /**
 *  Java binding for libGlass
 *  Copyright (C) 2003,2004
 *  Bruno Barberi Gnecco and Marcelo de Paiva Guimaraes
 *  Copyright (C) 2005-2010 Corollarium Tecnologia
 * 
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "Glass_Barrier.h"
#include "glassjava.h"

Barrier* getBarrier(JNIEnv *jenv, jobject jobj) {
	jclass cls = jenv->GetObjectClass(jobj);
	static jfieldID fid = NULL;
	if (fid == NULL) {
		fid = jenv->GetFieldID(cls, "pointer", "J");
		if (fid == NULL) {
			DEBUG("ERROR" << __LINE__);
			// error
			return NULL;
		}
	}
	return (Barrier *)jenv->GetLongField(jobj, fid);
}

/*
 * Class:     Glass_Barrier
 * Method:    createBarrier
 * Signature: (II)I
 */
JNIEXPORT jlong JNICALL Java_Glass_Barrier_createBarrier
  (JNIEnv *jenv, jobject jobj, jint barriernum, jint barriergroup, jboolean start, jint propagation) {
	boost::mutex::scoped_lock l(glassmutex);
	Barrier *b;
	
	try {
		propagationMode mode = static_cast<propagationMode>(propagation);
		if (mode != Local && mode != Global) {
			throw Exception("Invalid propagation mode");
		}

		b = new Barrier(barriernum, barriergroup, start, mode);
	}
	catch (Exception e) {
		jclass excCls = jenv->FindClass("java/lang/Exception");
		if (excCls != 0) {
			jenv->ThrowNew(excCls, e.getMessage().c_str());
		}
		return 0;
	}

	return (jlong)b;
}

/*
 * Class:     Glass_Barrier
 * Method:    destroyBarrier
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_Glass_Barrier_destroyBarrier
  (JNIEnv *jenv, jobject jobj) {
	boost::mutex::scoped_lock l(glassmutex);
	Barrier *b = getBarrier(jenv, jobj);
	delete b;
}

/*
 * Class:     Glass_Barrier
 * Method:    getNumber
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_Glass_Barrier_getNumber
  (JNIEnv *jenv, jobject jobj) {
	boost::mutex::scoped_lock l(glassmutex);
	Barrier *b = getBarrier(jenv, jobj);
	return b->getNumber();
}
  	

/*
 * Class:     Glass_Barrier
 * Method:    getGroup
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_Glass_Barrier_getGroup
  (JNIEnv *jenv, jobject jobj) {
	boost::mutex::scoped_lock l(glassmutex);
	Barrier *b;
	jfieldID fid;
	jclass cls = jenv->GetObjectClass(jobj);

	fid = jenv->GetFieldID(cls, "pointer", "J");
	if (fid == NULL) {
		DEBUG("ERROR" << __LINE__);
		// error
		return INT_ERROR;
	}
	b = (Barrier *)jenv->GetLongField(jobj, fid);

	return b->getGroup();
}

/*
 * Class:     Glass_Barrier
 * Method:    sync
 * Signature: ()Z
 */
JNIEXPORT jboolean JNICALL Java_Glass_Barrier_sync
  (JNIEnv *jenv, jobject jobj) {
	boost::mutex::scoped_lock l(glassmutex);
	Barrier *b = getBarrier(jenv, jobj);
	return b->sync();
}

/*
 * Class:     Glass_Barrier
 * Method:    enable
 * Signature: ()Z
 */
JNIEXPORT jboolean JNICALL Java_Glass_Barrier_enable
  (JNIEnv *jenv, jobject jobj) {
	boost::mutex::scoped_lock l(glassmutex);
	Barrier *b = getBarrier(jenv, jobj);

	return b->enable();
}

/*
 * Class:     Glass_Barrier
 * Method:    disable
 * Signature: ()Z
 */
JNIEXPORT jboolean JNICALL Java_Glass_Barrier_disable
  (JNIEnv *jenv, jobject jobj) {
	boost::mutex::scoped_lock l(glassmutex);
	Barrier *b = getBarrier(jenv, jobj);

	return b->disable();
}
