/**
 *  Java binding for libGlass
 *  Copyright (C) 2003-2010
 *  Corollarium Technologies
 * 
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  @author      Corollarium Technologies
 *  @version     %I%, %G%
 *  @since       1.0
 */

package Glass;

/**
 * Barrier class
 * 
 * @author corollarium
 *
 */
public class Barrier extends Plugin {

	/**
	 * Internal wrapper to native C++ constructor.
	 *  
	 * @param number
	 * @param group
	 * @param active
	 * @param propagation 
	 * @return The pointer, cast to long.
	 */
	private native long createBarrier(int number, int group, boolean active, int propagation);

	/**
	 * Internal wrapper to native C++ destructor.
	 */
	private native void destroyBarrier();

	/**
	 * @see #Barrier(int, int, boolean, int)
	 */
	public Barrier(int number) throws GlassException {
		try {
			this.pointer = createBarrier(number, 0, false, Glass.propagationLocal);
		}
		catch (Exception e) {
			throw new GlassException(e);
		}
	}
	/**
	 * @see #Barrier(int, int, boolean, int)
	 */
	public Barrier(int number, int group) throws GlassException {
		this(number, group, false, Glass.propagationLocal);
	}
	/**
	 * @see #Barrier(int, int, boolean, int)
	 */
	public Barrier(int number, int group, boolean active) throws GlassException {
		this(number, group, active, Glass.propagationLocal);
	}
	/**
	 * @see #Barrier(int, int, boolean, int)
	 */
	public Barrier(int number, int group, int propagation) throws GlassException {
		this(number, group, false, propagation);
	}

	/**
	 * The Barrier constructor.
	 * 
	 * @param number The barrier number
	 * @param group The barrier group
	 * @param active Should the barrier be considered active before its
	 * first synchronization?
	 * @throws GlassException
	 */
	public Barrier(int number, int group, boolean active, int propagation) throws GlassException {
		try {
			this.pointer = createBarrier(number, group, active, propagation);
		}
		catch (Exception e) {
			throw new GlassException(e);
		}
	}
	
	protected void finalize() throws Throwable {
		try {
			destroyBarrier();
		}
		finally {
			super.finalize();
		}
	}

	/**
	 * Returns the number of this barrier.
	 * @return The number of this barrier.
	 */
	public native int getNumber();

	/**
	 * Returns the number of this barrier.
	 * @return The number of this barrier.
	 */
	public native int getGroup();

	/**
	 * Synchronizes. Blocks until all nodes owning this barrier have called 
	 * this method.
	 * 
	 * @return true on sucess, false otherwise. 
	 */
	public native boolean sync();

	/** 
	 * Synchronization method.
	 * 
	 * Blocks until all nodes that own this barrier synchronize.
	 * This function is reentrant and can be used by several threads as
	 * a synchronization point.
	 * 
	 * @return True If the synchronization was successful.
	 * False Not successfull. The node probably did not synchronize.
	 * @see #disable()
	 * @see #sync()
	 */
	public native boolean enable();

	/**
	 * Enables this barrier temporarily. Useful if this node should skip
	 * a few barriers, without locking everybody else.
	 * 
	 * @return Boolean, true on success 
	 * @see #sync()
	 * @see #disable()
	 */
	public native boolean disable();
}
