/**
 *  Java binding for libGlass
 *  Copyright (C) 2003,2004
 *  Bruno Barberi Gnecco and Marcelo de Paiva Guimaraes
 *  Copyright (C) 2005-2010 Corollarium Tecnologia
 * 
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "Glass_Event.h"
#include "glassjava.h"

static Event<char *, rawPack, rawUnpack>* getEventPointer(JNIEnv *jenv, jobject jobj) {
	// TODO: cache ids

	jfieldID fid;
	jclass cls = jenv->GetObjectClass(jobj);

	fid = jenv->GetFieldID(cls, "pointer", "J");
	if (fid == NULL) {
		return JNI_FALSE;
	}
	return (Event<char *, rawPack, rawUnpack> *)jenv->GetLongField(jobj, fid);
}


/*
 * Class:     Glass_Event
 * Method:    createEvent
 * Signature: (Ljava/lang/String;)I
 */
JNIEXPORT jlong JNICALL Java_Glass_Event_createEvent
(JNIEnv *jenv, jobject jobj, jstring name, jint propagation) {
	boost::mutex::scoped_lock l(glassmutex);
	Event<char *, rawPack, rawUnpack> *e;
	const char *str;

	// convert the Java String to use it in C
	str = jenv->GetStringUTFChars(name, 0);

	try {
		propagationMode mode = static_cast<propagationMode>(propagation);
		if (mode != Local && mode != Global) {
			throw Exception("Invalid propagation mode");
		}
		e = new Event<char *, rawPack, rawUnpack>(str, NULL, mode);
	}
	catch (Exception e) {
		jenv->ReleaseStringUTFChars(name, str);  
		jclass except = jenv->FindClass("java/lang/Exception");
		if (except != 0) {
			jenv->ThrowNew(except, e.getMessage().c_str());
		}
		return 0;
	}

	jenv->ReleaseStringUTFChars(name, str);

	return (jlong)e;
}

/*
 * Class:     Glass_Event
 * Method:    destroyEvent
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_Glass_Event_destroyEvent
(JNIEnv *jenv, jobject jobj) {
	boost::mutex::scoped_lock l(glassmutex);
	Event<char *, rawPack, rawUnpack> *e = getEventPointer(jenv, jobj);
	delete e;
}

/*
 * Class:     Glass_Event
 * Method:    clear
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_Glass_Event_clear
(JNIEnv *jenv, jobject jobj) {
	boost::mutex::scoped_lock l(glassmutex);
	Event<char *, rawPack, rawUnpack> *e = getEventPointer(jenv, jobj);
	e->clear();
}

/*
 * Class:     Glass_Event
 * Method:    enqueueEvent
 * Signature: (Ljava/io/Serializable;LGlass/nodeId;)Z
 */
//JNIEXPORT jboolean JNICALL Java_Glass_Event_enqueueEvent
//  (JNIEnv *jenv, jobject jobj, jbyteArray data, jint target) {
JNIEXPORT jboolean JNICALL Java_Glass_Event_enqueueEvent
(JNIEnv *jenv, jobject jobj, jbyteArray data, jint target, jint len, jboolean sendtoself) {
	boost::mutex::scoped_lock l(glassmutex);
	Event<char *, rawPack, rawUnpack> *e = getEventPointer(jenv, jobj);

	//jsize len = jenv->GetArrayLength(data);
	
	jbyte *jBytes = jenv->GetByteArrayElements(data, 0);

	bool ret = e->enqueueEvent((char *)jBytes, (nodeId)target, len, sendtoself);

	jenv->ReleaseByteArrayElements(data, jBytes, JNI_ABORT);

	return ret;
}

/*
 * Class:     Glass_Event
 * Method:    getEvent
 * Signature: (LGlass/nodeId;)Ljava/io/Serializable;
 */
JNIEXPORT jbyteArray JNICALL Java_Glass_Event__1getEvent
(JNIEnv *jenv, jobject jobj, jint id) {
	boost::mutex::scoped_lock l(glassmutex);
	try {
		Event<char *, rawPack, rawUnpack> *e = getEventPointer(jenv, jobj);

		unsigned int len = 0;
		char *data = e->getEvent((nodeId)id, &len);
		jbyteArray jvalue = jenv->NewByteArray(len);
		jenv->SetByteArrayRegion(jvalue, 0, len, (jbyte *)data);
		delete[] data;
		return jvalue;
	}
	catch (EmptyException e) {
		jclass Exception = jenv->FindClass("Glass/EmptyException");
		jenv->ThrowNew(Exception, "");
	}
	return 0;
}

/*
 * Class:     Glass_Event
 * Method:    isEmpty
 * Signature: ()Z
 */
JNIEXPORT jboolean JNICALL Java_Glass_Event_isEmpty
(JNIEnv *jenv, jobject jobj) {
	boost::mutex::scoped_lock l(glassmutex);
	Event<char *, rawPack, rawUnpack> *e = getEventPointer(jenv, jobj);
	return e->isEmpty();
}

/*
 * Class:     Glass_Event
 * Method:    queueSize
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_Glass_Event_queueSize
(JNIEnv *jenv, jobject jobj) {
	boost::mutex::scoped_lock l(glassmutex);
	Event<char *, rawPack, rawUnpack> *e = getEventPointer(jenv, jobj);
	return e->queueSize();
}
