/**
 *  Java binding for libGlass
 *  Copyright (C) 2003-2010
 *  Corollarium Technologies
 * 
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  @author      Corollarium Technologies
 *  @version     %I%, %G%
 *  @since       1.0
 */

package Glass;
import java.io.*;

/**
 * Asynchronous events.
 * 
 * @author corollarium
 *
 * @param <T> The object type.
 */
public class Event<T extends Serializable> extends Plugin {

	private int len = 1;
	private boolean sendtoself = false;

	/**
	 * A wrapper for the C++ constructor.
	 * 
	 * @param name
	 * @return A pointer to the C++ instance.
	 */
	private native long createEvent(String name, int propagation);

	/**
	 * A wrapper for the C++ destructor.
	 */
	private native void destroyEvent();

	/**
	 *  Constructor
	 */
	public Event(String name) throws GlassException {
		this(name, Glass.propagationLocal);
	}

	public Event(String name, int propagationMode) throws GlassException {
		try {
			this.pointer = createEvent(name, propagationMode);
		}
		catch (Exception e) {
			throw new GlassException(e);
		}
	}

	/**
	 * Destructor
	 */
	protected void finalize() throws Throwable {
		try {
			destroyEvent();
		}
		finally {
			super.finalize();
		}
	}

	/**
	 * Clears queue.
	 */
	public native void clear();

	/**
	 * Enqueues data.
	 * 
	 * @param o The object to enqueue.
	 * @param target The target node id.
	 * @return True if successfully sent. False if an error occurred.
	 */
	public boolean enqueueEvent(T o, nodeId target, int len, boolean sendtoself) throws IOException {
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		ObjectOutputStream objstream = new ObjectOutputStream(baos);
		objstream.writeObject(o);
		objstream.close();
		byte[] x = baos.toByteArray();
		return this.enqueueEvent(x, target.getId(), len, sendtoself);
	}

	/**
	 * @param nick
	 */
	public boolean enqueueEvent(T o, String nick, int len, boolean sendtoself) throws IOException {
		// TODO: return this->enqueueEvent(o, this->glass->NickNameToId(nick));
		return this.enqueueEvent(o, nodeId.everybodyId, len, sendtoself);
	}

	/**
	 * Native call to Glass, enqueues event as an array of bytes.
	 * 
	 * @param o The object, serialized to a string.
	 * @param target Target node
	 * @return True if successful
	 */
	public synchronized native boolean enqueueEvent(byte[] o, int target, int len, boolean sendtoself);

	/**
	 * Native call to Glass, enqueues event as an array of bytes.
	 * 
	 * @param o The object, serialized to a string.
	 * @return True if successful
	 * @throws IOException
	 */
	public boolean enqueueEvent(T o) throws IOException {
		return this.enqueueEvent(o, nodeId.everybodyId, len, sendtoself);
	}

	/**
	 * Gets an event sent by a specific node.
	 * 
	 * @param id The node id to get an event for.
	 * @return The event object.
	 * @throws IOException If cannot unserialize object.
	 * @throws EmptyException If no events in queue. 
	 * @throws ClassNotFoundException If object is of a non-existent class.
	 */
	@SuppressWarnings("unchecked")
	public synchronized T getEvent(int id) 
	throws IOException, EmptyException, ClassNotFoundException {
		byte[] data = this._getEvent(id);
		ByteArrayInputStream bais = new ByteArrayInputStream(data);
		ObjectInputStream ois = new ObjectInputStream(bais);
		T s = (T)ois.readObject();
		ois.close();
		return s;
	}

	/**
	 * Wrapper for C++ native call.
	 * 
	 * @param id
	 * @return The object serialized as bytes.
	 * @throws EmptyException If queue is empty.
	 */
	protected native byte[] _getEvent(int id) throws EmptyException;

	/**
	 * Gets an event sent by any node.
	 * 
	 * @return The event object.
	 * @throws IOException If cannot unserialize object.
	 * @throws EmptyException If no events in queue. 
	 * @throws ClassNotFoundException If object is of a non-existent class.
	 */
	public T getEvent() throws IOException, EmptyException, ClassNotFoundException {
		return this.getEvent(nodeId.everybodyId.getId());
	}


	/**
	 * Checks queue for emptiness.
	 * @return True if empty. False if not empty.
	 */
	public native boolean isEmpty();

	/**
	 * Returns the number of entries in queue.
	 * 
	 * @return The number of entries in queue.
	 */
	public native int queueSize();
}
