 /**
  *  Java binding for libGlass
  *  Copyright (C) 2003-2010
  *  Corollarium Technologies
  * 
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public
  *  License as published by the Free Software Foundation; either
  *  version 2 of the License, or (at your option) any later version.
  *
  *  This library is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  *  Lesser General Public License for more details.
  *
  *  You should have received a copy of the GNU Lesser General Public
  *  License along with this library; if not, write to the Free Software
  *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  *
  *  @author      Corollarium Technologies
  *  @version     %I%, %G%
  *  @since       1.0
  */

package Glass;

/**
 * Abstract base class for behavior classes (Server, Client)
 * 
 * @author corollarium
 *
 */
public abstract class Glass {
	/**
	 * Relationship between nodes: none
	 */
	public static final int nodeRelationNone = 0;
	
	/**
	 * Relationship between nodes: server 
	 */
	public static final int nodeRelationServer = 1;
	
	/**
	 * Relationship between nodes: client
	 */
	public static final int nodeRelationClient = 2;

	/**
	 * Relationship between nodes: peer
	 */
	public static final int nodeRelationPeer = 3;
	
	/**
	 * Propagation mode
	 */
	public static final int propagationLocal = 0;

	/**
	 * Propagation mode
	 */
	public static final int propagationGlobal = 1;
	
	static { // load the JNI library
		System.loadLibrary("glassjava");
	}

	/**
	 * pointer to C++ class instance.
	 */
	protected long pointer;
	
	/**
	 * Constructor. Empty, this is abstract.
	 * 
	 * @throws GlassException
	 */
	Glass() throws GlassException {
		
	}

	/**
	 * Returns this node's Id.
	 * @return This node's Id.
	 */
	public native int getNodeId();

	/**
	 * Returns this node's nickname
	 * @return This node's nickname.
	 */
	public native String getNodeNickname();
	
	/**
	 * Returns the relationship with a given node.
	 
	 * @param id The other node's id.
	 * @return The relationship of the node id with this node. That
	 *   is, if this node is the Server and id is the Client, returns
	 *   Client.
	*/
	public native int getNodeRelation(int id);

	/**
	 * Sets this Glass instance as the current one.
	 * @see #lockCurrent()
	 * @see #unlockCurrent()
	 */
	public native void setCurrent();

	/**
	 * @see #setCurrent()
	 * @see #unlockCurrent()
	 */
	public native void lockCurrent();
	
	/**
	 * @see #lockCurrent()
	 * @see #setCurrent()
	 */
	public native void unlockCurrent();

	/**
	 * Returns the number of nodes connected to the network.
	 * 
	 * This is a virtual function. The default implementation returns 0.
	 * Most derived class will override this function.
	 * 
	 * This function may be slow, requiring a network transmission. The
	 * number of nodes may change at any time.
	 * 
	 * Important: The actual implementation in the derived class may
	 * choose to return the number of nodes of a certain type. For example,
	 * the client/server architecture returns the number of CLIENTS, which
	 * makes sense, since the server is not doing any computation on the
	 * application. So, refer to the documentation of the derived class to
	 * know exactly what you are getting.
	 * 
	 * @return The number of nodes connected to the network, or zero if
	 * could not get it. See the derived class documentation for a more
	 * complete description.
	 */
	public native int getTotalNodes();
}
