 /**
  *  Java binding for libGlass
  *  Copyright (C) 2003-2010
  *  Corollarium Technologies
  * 
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public
  *  License as published by the Free Software Foundation; either
  *  version 2 of the License, or (at your option) any later version.
  *
  *  This library is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  *  Lesser General Public License for more details.
  *
  *  You should have received a copy of the GNU Lesser General Public
  *  License along with this library; if not, write to the Free Software
  *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  *
  *  @author      Corollarium Technologies
  *  @version     %I%, %G%
  *  @since       1.0
  */

package Glass;

import java.io.*;

public class Shared<T extends Serializable> extends Plugin {

	/**
	 * Internal wrapper to native C++ constructor.
	 *  
	 * @param name
	 * @return The pointer, cast to long.
	 */
	private native long createShared(String name, int propagation);
	
	/**
	 *  Internal wrapper to native C++ destructor.
	 */
	private native void destroyShared();
	
	/**
	 * The current data value.
	 */
	private T data;

	/**
	 * @see #Shared(String, int)
	 */
	public Shared(String name, T value) throws GlassException {
		this(name, value, Glass.propagationLocal);
	}
	
	/**
	 * Constructor.
	 * 
	 * @param name
	 */
	public Shared(String name, T value, int propagation) throws GlassException {
		try {
			this.pointer = createShared(name, propagation);
			this.setData(value);
		}
		catch (Exception e) {
			throw new GlassException(e);
		}
	}

	protected void finalize() throws Throwable {
		try {
			destroyShared();
		}
		finally {
			super.finalize();
		}
	}

	/**
	 * Has the value changed since last update?
	 * @return True if changed.False otherwise.
	 */
	public native boolean hasChanged();

	/**
	 * Wrapper to C++ call.
	 * @param o The object.
	 * @return true on success
	 */
	public native boolean sendUpdate(byte[] o);

	private String getHexString(byte[] b) throws Exception {
		String result = "";
		for (int i=0; i < b.length; i++) {
			result +=
				Integer.toString( ( b[i] & 0xff ) + 0x100, 16).substring( 1 );
		}
		return result;
	}

	
	/**
	 * Sends new value to other nodes.
	 *  
	 * @throws IOException 
	 */
	public boolean sendUpdate() throws IOException {
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
        ObjectOutputStream objstream = new ObjectOutputStream(baos);
        objstream.writeObject(this.data);
        objstream.close();
        byte[] x = baos.toByteArray();
        try {
			// DEBUG: System.out.println("Out:" + this.getHexString(x));
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		return this.sendUpdate(x);
	}
	
	private native boolean _getUpdate();

	/**
	 * Synchronizes this variable.
	 * @return True If successful. false Otherwise.
	 */
	public boolean getUpdate() {
		return this._getUpdate();
	}
	
	private native byte[] _getData();

	/**
	 * Returns the variable value, checking for new values that may have 
	 * arrived from other nodes.
	 * 
	 * @return The variable value.
	 */
	@SuppressWarnings("unchecked") // suppress compile warnings about unchecked generic cast
	public synchronized T getData() throws ClassNotFoundException {
		try {
			byte[] data = this._getData();
		    ByteArrayInputStream bais = new ByteArrayInputStream(data);
		    ObjectInputStream ois = new ObjectInputStream(bais);
		    T s = (T)ois.readObject();
		    ois.close();
		    this.data = s;
		}
		catch (IOException e) {
			; // nothing
		}
		return this.data;
	}
	
	/**
	 * Returns the current value. No check is made to see if new values 
	 * may have arrived from other nodes.
	 */
	public synchronized T getCurrentData() {
		return this.data;
	}
	
	/**
	 * Sets the current variable value. You need to call sendUpdate to propagate
	 * it to other nodes.
	 * 
	 */
	public synchronized void setData(T data) {
		this.data = data;
	}
}
