/*  Glass - a distributed computing library
    Copyright (C) 2003-2009 Bruno Barberi Gnecco
    Copyright (C) 2009-2010 Corollarium Technologies


    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* This program illustrates how tiled rendering can be done with libglass. 
The screen is subdivided in N vertical stripes, where N is the number of
clients. If a new client is added, the screen is subdivided again. The 
rendering is bogus, each client paints its tile of a random color. */

#include <math.h>
#include <stdlib.h>
#include <iostream>

#include <time.h>
#include <sys/types.h>
#include <unistd.h>

#ifdef MACOSX
 #include <OpenGL/gl.h>
 #include <OpenGL/glu.h>
 #include <GLUT/glut.h>
#else
 #include <GL/glut.h>
#endif
#include "libglass.h"

#define WIDTH	400
#define HEIGHT	200

using std::endl;
using namespace libglass;

typedef struct {
	unsigned int		width, height;
	unsigned char		*buffer;
} FrameBuffer;

typedef struct {
	unsigned int		beginx, beginy;
	unsigned int		width, height;
} RenderData;

bool framepack(chain &c, FrameBuffer * const &fb, unsigned int nmemb, 
		void *extra) {
	c.append(reinterpret_cast<const char *>(&fb->width), sizeof(unsigned int));
	c.append(reinterpret_cast<const char *>(&fb->height), sizeof(unsigned int));
	c.append(reinterpret_cast<const char *>(fb->buffer), fb->width*fb->height*3);
	return sizeof(unsigned int)*2+ fb->width*fb->height*3;
}

bool frameunpack(const chain &c, unsigned int offset, FrameBuffer * &fb, 
		unsigned int &used, unsigned int nmemb, void *extra) {
	unsigned int width, height;

	/* get image width, height */
	if (c.substrcopy(reinterpret_cast<char *>(&width), offset, offset+3) != true || 
	    c.substrcopy(reinterpret_cast<char *>(&height), offset+4, offset+7) != true) {
		return false;
	}

	std::cerr << "Got image: " << width << "x" << height << std::endl;

	// TODO: INSTEAD OF MALLOCING, GET A POINTER (possibly with a chain
	// to do GC

	/* alloc memory if necessary */
	if (fb->width != width || fb->height != height) {
		fb->width = width;
		fb->height = height;
		if (fb->buffer != NULL) {
			delete fb->buffer;
		}
		fb->buffer = new unsigned char[width*height*3];
	}
	else {
		if (fb->buffer == NULL) {
			fb->buffer = new unsigned char[width*height*3];
		}
	}

	used = 2*sizeof(unsigned int)+width*height*3;
	return c.substrcopy(reinterpret_cast<char *>(fb->buffer), offset+8, used-1);
}

GlassClient *g;
Event<FrameBuffer *, framepack, frameunpack> *tiles;
//Shared<int, rawPack, rawUnpack> *w; // this tile width
//Alias<RenderData> *renderdata;
Barrier *b;

FrameBuffer f;
unsigned char color[3];
float anim;

void gldraw(void) {
/*	unsigned int totalnodes = g->getNodes();
	int df = f.width/totalnodes;
*/
	glClear(GL_COLOR_BUFFER_BIT);
	glLoadIdentity();

	unsigned int totalnodes = 1;
	while (totalnodes) {
		while (tiles->isEmpty()) {
			// nanosleep
			usleep(1000);
		}

//		glRasterPos2i(tiles->getNode()*df, 0);
//		glDrawPixels(df, f.height, GL_RGB, GL_FLOAT, tiles->getEvent());
		glRasterPos2i(0, 0);
		glDrawPixels(f.width, f.height, GL_RGB, GL_FLOAT, tiles->getEvent()->buffer);
		// would glTexture(tiles->getEvent())/glBegin(GL_QUADS) be faster?
		totalnodes--;
	}
	glutSwapBuffers();
        glutPostRedisplay();
}

void glreshape(int w, int h) {
	// update f && shared
	
	// TODO
	
	glViewport(0, 0, w, h);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluOrtho2D(0, w, 0, h);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

void render(void) {
#ifdef LATER
	unsigned int 		i, j;
	unsigned char		*p = f.buffer;
	unsigned int		c[3];

	/* animate the colors */
	c[0] = (unsigned char)((float)color[0] * anim);
	c[1] = (unsigned char)((float)color[1] * anim);
	c[2] = (unsigned char)((float)color[2] * anim);

	for (j = 0; j < f.height; j++) {
		for (i = 0; i < f.width; i++) {
			*p++ = c[0];
			*p++ = c[1];
			*p++ = c[2];
		}
	}
#endif
	
	unsigned char		*p = f.buffer;
		for (unsigned int j = 0; j < f.width * f.height; j++) {
			*p++ = 150;
			p++;
			*p++ = 150;
		}
   
	tiles->enqueueEvent(&f);
}

static void resize(void) {
#ifdef TODO
	width->getUpdate();
	height->getUpdate();
	if (f.width != width->getData() || f.height != height->getData()) {
		f.width = width->getData();
		f.height = height->getData();
		delete f.buffer;
		f.buffer = new unsigned char[f.width * f.height * 3];
	}
#endif
}

static int args (int argc, char **argv) {
	int i = 1;
	while (i < argc) {
		if (*argv[i] != '-') {
			return i;
		}
		switch (*(argv[i]+1)) {
		   case 'w':
			f.width = atoi(argv[++i]);
			break;
		   case 'h':
			f.height = atoi(argv[++i]);
			break;
		   default:
			std::cout <<"Arguments are:" << endl <<
				"-w [n]\twidth" << endl <<
				"-h [n]\theight" << endl <<
				"-n [x y z]\tnumber of cubes" << endl;
			exit(1);
			break;
		}
		i++;
	}
	return 0;
}

int main(int argc, char **argv) {
//	glutInit(&argc, argv);

	int server = args(argc, argv);
	try {
		if (!server) {
			new GlassServer(new TCP());
		}
		else {
			g = new GlassClient(new TCP(), argv[server]);
		}
	}
	catch (Exception e) {
		std::cerr << e.getMessage() << std::endl;
		exit(1);
	}

	f.width = WIDTH;
	f.height = HEIGHT;
	f.buffer = new unsigned char[f.width * f.height * 3];

	/* create our distributed data */
	tiles = new Event<FrameBuffer *, framepack, frameunpack>("tiles");
//	renderdata = new Alias<RenderData>("renderdata");
	b = new Barrier(0,false);

	if (g->isMaster()) {
//		*width = f.width;
//		*height = f.height;
	}
	else {
//		width->getUpdate();
//		height->getUpdate();
	}

	srandom(time(NULL)+getpid());
	color[0] = 1 + (int)(255.0*random()/(RAND_MAX+1.0));
	color[1] = 1 + (int)(255.0*random()/(RAND_MAX+1.0));
	color[2] = 1 + (int)(255.0*random()/(RAND_MAX+1.0));

	if (g->isMaster()) {
		glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE);
		glutInitWindowPosition(0, 0);
		glutInitWindowSize(300, 300);
		glutCreateWindow("libGlass Tiled Rendering example");
		glutDisplayFunc(gldraw);
		glutReshapeFunc(glreshape);
		glutMainLoop();
	}
	else while(1) {
		resize();
		render();
		// sync()?
	}

	return 0;
}
