/*  Glass - a distributed computing library
    Copyright (C) 2003-2009 Bruno Barberi Gnecco
    Copyright (C) 2009-2010 Corollarium Technologies

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef _GLASS_PLUGIN_ALIAS_HH_
#define _GLASS_PLUGIN_ALIAS_HH_

#include <map>
#include <utility>
#include <set>
#include "plugin.h"
#include "exception.h"

namespace libglass {

class _Alias;
class AliasPacket;

/*! 
  Alias plugin.

  These are valid actions in client->server packets:

  \li 'C'[name] = Created
  \li 'D'[name] = Deleted
  \li 'A'[targetid][name][alias] = Associated
  \li 'I' = dIssociated

  These are valid actions in server->client packets:

  \li 'A'[id][name] = set alias of this node as name.
  \li 'I'[id] = dIssociated

  \sa Barrier.
*/
class AliasBase : public PluginBase {
	/*! 
	  Plugin creation. 
	*/
	PLUGIN(AliasBase);
private:
	std::map<const string, _Alias *> aliases;

protected:
	friend class _Alias;


	/*!
	  Registers an alias in the list.
	  \param a A pointer to the alias.
	  \retval true If successful.
	  \retval false Otherwise.
	*/	
	bool registerAlias(_Alias *a);

	/*!
	  Unregisters an alias, taking it off the list.
	  \param a A pointer to the alias.
	  \retval true If successful.
	  \retval false Otherwise.
	*/
	bool unregisterAlias(_Alias *a);

	/**********************************************************************
	// Destructor should be PROTECTED to ensure singleton behaviour
	// However, Microsoft compilers prior to VC v7.0 .NET are buggy and
	// generate the following error
	//   error C2248: 'AliasBase::~AliasBase' : cannot access protected member declared in class 'libglass::AliasBase'
	// The only workaround is to make the destructor public - which kind
	// of destroys the whole reason for implementing the singleton class */
#if defined(_MSC_VER) && (_MSC_VER < 1300)  // earlier than .NET compiler (VC 7.0)
public:
#else
protected:
#endif
	/*********************************************************************/
	/*!
	  Destructor.
	*/
	virtual ~AliasBase();

public:
};



#if defined(_MSC_VER)
// Instantiate std::set<T> template containers used by class _Alias
//LIBGLASS_IMP_TEMPLATE_STL_SET(nodeId)
// Instantiate std::map<Ta,Tb> template containers used by class _Alias
//LIBGLASS_IMP_TEMPLATE_STL_MAP(libglass::nodeId, libglass::string)
#endif // _MSC_VER

/*!
*/
class LIBGLASS_API _Alias : public PluginInterface {
private:
	void init(const string &name) throw(Exception);

protected:
	friend class AliasBase;
	friend class AliasPacket;

	AliasBase		*ab;		/**< our owner */
	
	/* client */
	string			name;		/**< our name */
	bool			associated;	/**< are we associated? */
	string			current_association;	/**< our current associations */
//	string 			(*func)(nodeId);

	/* server */
	std::set<nodeId>	owners;		/**< owner list */
	std::map<nodeId, string> associations;	/**< */

	_Alias(const char *aliasname, propagationMode mode = Local) throw(Exception);
	_Alias(const string &, propagationMode mode = Local) throw(Exception);

	/*!
	  Destructor.
	*/
	virtual ~_Alias();

	/*!
	  Associates a given node to an alias locally.

	  \param name
	  \retval True 
	  \retval False
	*/
	bool _associate(const string &association);

	void unregisterNode(nodeId id);

public:
	/*!
	  Returns this instance's name.

	  \return This instance's name.
	*/
	string getName(void) const;

	/*!
	  Associates a given node to an alias, propagating to other nodes.

	  \param id The node id.
	  \param name Alias association
	  \retval True if message sent successfully. Doesn't guarantee either
	    that node "id" exists or that it has an associated alias "name".
	  \retval False
	*/
	bool associate(nodeId id, string &name);

	/*!
	  \overload void associate(nodeId id, string &name)
	  \param id The node id.
	  \param name Alias association
	*/
	bool associate(nodeId id, const char *name);

	/*!
	  \overload void associate(nodeId id, string &name)
	  \param map A map of nodes and aliases.
	*/
	bool associate(std::map<nodeId, string> &map);

#ifdef TODO //how to propagate to other nodes?
	/*!	
	  \overload void associate(nodeId id, string &name)
	  \param func A function that returns an alias given a nodeId.
	*/
	void associate(string (*func)(nodeId));
#endif

	bool dissociate(nodeId id);
};

/*!
  Alias template class.

  It's usually more efficient to store pointers to data than the data 
  itself; that is, prefer Alias<yourclass *> to Alias<yourclass>.

*/
template<typename T>
class Alias : public _Alias {
private:
	std::map<string, T>	aliases;
	T			default_value;

public:
	/*! 
	  
	*/
	Alias(const char *aliasname, T default_value, propagationMode mode = Local) throw(Exception) :
		_Alias(aliasname, mode), default_value(default_value) {
	}

	/*! 
	  Destructor.
	*/
	~Alias() {
	}

	/*!
	  Add an alias to function.
	  
	  \param name The alias.
	  \param t The data.
	  \retval true If successful.
	  \retval false Otherwise.
	*/
	bool addAlias(string &name, T t)  {
		aliases[name] = t;
		return true;
	}

	/*!
	  \overload bool addAlias(string &name, T t)
	*/
	bool addAlias(const char *name, T t)  {
		if (!name)
			return false;
		aliases[string(name)] = t;
		return true;
	}

	/*!
	  Deletes an alias to function.
	  
	  \param name The alias.
	  \retval true If successful.
	  \retval false Alias doesn't exist.
	*/
	bool deleteAlias(const char *name) {
		// todo
		// if (data == alias[name] ) ???
		return false;
	}

	/*!
	  Checks if there's data associated with the current alias.
	  \retval true If there is.
	  \retval false If there is not.
	*/
	bool checkAlias(void) const {
		return associated;
	}

	/*!
	  Returns associated to the current alias.

	  \return The associated data.
	  */
	T getData(void) {
		if (associated == true) {
			typename std::map<string, T>::iterator it;
			it = aliases.find(current_association);
			if (it == aliases.end()) {
				/* we don't have a T associated with our alias */
				return default_value;
			}
			return it->second;
		}
		return default_value;
	}
};

}//namespace


#endif // _GLASS_PLUGIN_ALIAS_HH_
