/*  Glass - a distributed computing library
    Copyright (C) 2003-2009 Bruno Barberi Gnecco
    Copyright (C) 2009-2010 Corollarium Technologies


    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef _GLASS_CHAIN_HH_
#define _GLASS_CHAIN_HH_

#include "msvc.h"
#include "types.h"
#include "exception.h"
#include <string.h>
#include <list>
#include <iostream>
#include <limits.h>

namespace libglass {

class chain;

/*!
  Record class.

  These are individual items of a chain.


  \sa chain.
*/
class LIBGLASS_API record {
#if defined(_MSC_VER)
public:
#else
private:
#endif // _MSC_VER
	/*! Private constructor */
	record();
	/*! Private assignment operator */
	record *operator=(const record *s);

protected:
	/*! record structure. */
	struct _record {
		unsigned int	references;	/**< number of references */
		bool		allocated;	/**< should we free it */
		unsigned int	size;		/**< total bytes, in memory */
		const char	*data;		/**< the data itself */
	};
	_record		*r;			/**< pointer to the _record data of this instance */
	
	int 		offset;			/**< offset of r->data */
	int 		roffset;		/**< right offset of r->data */

public:

	/*! 
	  Construtor.
	  \param p The data array.
	  \param size Data size.
	  \param allocated True if data should be deleted when there are no
	    more references to it.
	  \param allocateddata The beginning of the allocated data; if NULL,
	    the beginning is at p.
	*/
	record(const char *p, unsigned int size, bool allocated, 
		const char *allocateddata = NULL);

	/*! 
	  Construtor. Updates references.
	  
	  \param r The record to copy from.
	  \param offset Offset from beginning of data.
	  \param roffset Offset from end of data.
	*/
	explicit record(const record *r, int offset = 0, int roffset = 0);

	explicit record(const record &r, int offset = 0, int roffset = 0);

	~record();

	record &operator=(const record &s);

	/*! 
	  Returns the data, taking offsets in consideration.
	*/
	const char *getData(void) const;

	/*! 
	  Returns the data length, taking offsets in consideration.
	*/
	unsigned int getLength(void) const;

	/*! 
	  Debug.
	*/
	void print() const;
};

class chain_iterator {
private:
	std::list<record>::iterator it;		/**< list of records */

protected:
	friend class chain;

	void setiterator(std::list<record>::iterator it) {
		this->it = it;
	}

public:
	chain_iterator() { }

	const chain_iterator& operator++() { // Prefix
		++it;
		return *this;
	}

	const chain_iterator operator++(int) { // Postfix
		it++;
		return *this;
	}

	const chain_iterator& operator--() { // Prefix
		--it;
		return *this;
	}

	const chain_iterator operator--(int) { // Postfix
		it--;
		return *this;
	}

	bool operator==(const chain_iterator &i) const {
		return (this->it == i.it);
	}

	bool operator!=(const chain_iterator &i) const {
		return (this->it != i.it);
	}

	const char *data(void) const {
		return it->getData();
	}

	unsigned int size(void) const {
		return it->getLength();
	}
};

/*!
  chain is a string-like class. It was written to avoid memory copying, and
  therefore be efficient when working with large arrays. It's basic usage
  is to concatenate several std::string or char *.

  The internal implementation uses instances of the record class to keep
  track of the different "links" (or pieces) or the chain. The
  implementation uses a reference counting system, and can shared records
  or parts of a record between different chains. 
  \sa record, chain_iterator
*/

#if defined(_MSC_VER)
// Instantiate std::list<T> template containers used by class chain
//LIBGLASS_IMP_TEMPLATE_STL_LIST(libglass::record)
#endif // _MSC_VER

// class chain {
class LIBGLASS_API chain {
public:
	// iterators
	typedef chain_iterator iterator;

#if defined(_MSC_VER) && (_MSC_VER < 1300)  // earlier than .NET compiler (VC 7.0)
	// Microsoft Knowledge Base Article - 241569 
	// Workaround for BUG: C2258 and C2252 on in Place Initialization of Static Const Members
	enum{ npos = UINT_MAX };
#else
	static const unsigned int npos = UINT_MAX;
#endif

private: 
	const char *getPosition(unsigned int position) const throw(Exception);

protected:
	friend class chain_iterator;

	std::list<record>	pieces;

public:
	/*! 
	  Default construtor.
	*/
	chain();

	/*! 
	  Construtor.
	  \param p The data array.
	  \param s Data size. If zero, the length of p is calculated with 
	  	strlen().
	*/
	explicit chain(const char *p, unsigned int s = 0);

	/*! 
	  Construtor.

	  Note that unlike the const char* constructor, STL strings don't end
	  with '\0' (unless created from raw data or something).

	  \param s The string.
	*/
	explicit chain(std::string &s);

	/*! 
	  Destructor.
	*/
	~chain();

	/*!
	  Clears this chain.
	*/
	void clear(void);

	/*! 
	  Size (in bytes) of this chain.
	  \sa length.
	*/
	unsigned int size(void) const;

	/*! 
	  Size (in bytes) of this chain.
	  \sa size.
	*/
	unsigned int length(void) const;

	/*! 
	  Returns the character in a given position.
	
	  \param position The position.
	  \return A reference to the character in that position, if the position
	  is valid. 
	*/
	char &operator[](unsigned int position) throw(Exception);

	/*! 
	  Assignment operator.
	 
	  \param s
	  \return 
	*/
	const chain &operator=(const chain &s);

	/*! 
	  Assignment operator.
	 
	  \param p The data array to assign.
	  \param s size
	  \param free If true, this array will be automatically deleted when
	  no references to it exist anymore.
	  \param allocatedbegin If you want the data to be automatically
	  deleted, but the beginning of data starts at some other address
	  (for example, you are ignoring a header), pass the correct address
	  here.
	  \return This chain.
	*/
	const chain &assign(char *p, unsigned int s = 0, bool free = false,
			char *allocatedbegin = NULL);

	const chain &assign(std::string &s);

	const chain &assign(const chain &s);

	/*! 
	  Comparison operator.
	 
	  \param c The chain to compare to.
	  \return true if this chain is less than c.
	*/
	const bool operator<(const chain &c) const;

	/*! 
	  Append operator. Exactly equivalent to append().

	  \param s The chain to append.
	  \return .
	  \sa append, appendcopy.
	*/
	const chain &operator+=(const chain &s);

	/*! 
	  Append operator. Exactly equivalent to append().

	  \param p The data array to append. Size is calculated with strlen().
	  \return This chain.
	  \sa append, appendcopy.
	*/
	const chain &operator+=(char *p);
	
	/*! 
	  Append operator. Exactly equivalent to append().

	  \param s The data array to append.
	  \return This chain.
	  \sa append, appendcopy.
	*/
	const chain &operator+=(std::string &s);

#ifdef TODO
	const chain operator+(const chain &s) const;
#endif

	/*! 
	  Equality operator. 

	  \param s The chain to compare to.
	  \return True if equal, false otherwise.
	*/
	bool operator==(const chain &s) const;

	/*! 
	  Inequality operator. 

	  \param s The chain to compare to.
	  \return False if equal, true otherwise.
	*/
	bool operator!=(const chain &s) const;

	/*! 
	  Append operator. 

	  Appends a chain to this chain. Memory is \e shared.

	  \param s The chain to append.
	  \sa appendcopy.
	*/
	void append(const chain &s);

	/*! 
	  Append operator.
	
	  Appends a data arra to this chain. Memory is \e shared.

	  \param p The data array to append. 
	  \param s Array length. If zero, length is calculated with strlen().
	  \param free If true, this array will be automatically deleted when
	  no references to it exist anymore.
	  \param allocatedbegin The start address of the allocated block
	  \sa appendcopy.
	*/
	void append(const char *p, unsigned int s = 0, bool free = false,
				char *allocatedbegin = NULL);

	void append(std::string &s);

	/*! 
	  Append operator. 

	  Appends a chain to this chain. Memory is \e duplicated.

	  \param s The chain to append.
	  \sa append.
	*/
	void appendcopy(chain &s);

	/*! 
	  Append operator.
	
	  Appends a data arra to this chain. Memory is \e duplicated.

	  \param p The data array to append. 
	  \param s Array length. If zero, length is calculated with strlen().
	  \sa appendcopy.
	*/
	void appendcopy(const char *p, unsigned int s = 0);

	void appendcopy(std::string &s);

#ifdef DEPRECATED
	/*! 
	  Returns a new chain with the contents of a range of records of this
	  chain, but merged to a single record.
	  
	  \param begin The first RECORD.
	  \param end The last RECORD.
	  \return A new chain, composed of a single record. If the parameters 
	  are invalid, returns NULL.
	*/
	chain *concatenate(int begin = 0, int end = -1) const;
#endif

	/*! 
	  Returns a character array with the contents of a range of records 
	  of this chain, but merged to a single record.

	  \param begin The first RECORD.
	  \param end The last RECORD.
	  \return A character array with the same contents of this chain. You
	  are responsible for delete[]ing this array. If the parameters are
	  invalid, returns NULL.
	*/
	char *toArray(unsigned int begin = 0, unsigned int end = npos) 
			const throw(Exception);

	/*! 
	  Returns a std::string with the contents of a range of this chain.

	  \param begin The index of the first character.
	  \param end The index of the last character.
	  \return A std::string with the contents of a range of this chain.
	*/
	std::string toString(unsigned int begin = 0, unsigned int end = npos) 
			const throw(Exception);

	/*! 
	  Returns a std::string with the contents of a range of this chain up
	  to the first \\0 character or end.

	  \param begin The index of the first character.
	  \param end The index of the maximum character to search for.
	  \return A std::string with the contents of a range of this chain.
	*/
	std::string toStringNull(unsigned int begin = 0, unsigned int end = npos) const throw(Exception);

	/*!
	  Returns a 4-byte integer astroed at a given position.

	  \param begin The start of the data.
	  \return A integer.
	  \sa charToNum.
	*/
	int toInt(unsigned int begin = 0) const throw(Exception);

	/*! 
	  Substring.
	
	  Creates a new chain, which is a substring to the original one. The
	  data is shared, though, so modifications to one will result in
	  modifications to the other.

	  \param begin The first position of the substring.
	  \param end The last position of the substring. Use npos to get the
	    last character of the original string.
	  \return A chain containing the substring.
	*/
	chain substr(unsigned int begin, unsigned int end) const throw(Exception);

	/*! 
	  Substring copy.
	
	  Copies the substring to a given buffer. The contents are duplicated,
	  so this function is slower than substr. User is responsible for
	  providing a valid buffer at least (end-begin+1) bytes long.

	  \param dest The destination buffer
	  \param begin The first position of the substring.
	  \param end The last position of the substring. Use npos to get the
	    last character of the original string.
	  \retval True if successful.
	  \retval False otherwise.
	  \sa substr, duplicate.
	*/
	bool substrcopy(char *dest, unsigned int begin, unsigned int end) const
			throw(Exception);

	/*! 
	  Duplicates this string, copying the contents.
	  
	  Memory is not shared, but copied to new buffers. Use only when it's
	  absolutely necessary to copy the chain contents, since this function
	  is much slower than assign().
	  
	  \return The duplicated string.
	  \sa substrcopy.
	*/
	chain duplicate(void);

	/*! 
	  Finds the first occurrence of a given character.

	  \param c The character to search for.
	  \param initial The position to start the search.
	  \return The position of the first occurrence of c, or npos if not
	  found.
	*/
	unsigned int find(char c, unsigned int initial = 0) const;

	// iterator
	chain_iterator begin(void);

	chain_iterator end(void);

	void erase(chain_iterator it);

	void debug(void) const;

	friend std::ostream& operator<<(std::ostream &o, chain &p);
};

extern std::ostream& operator<<(std::ostream &o, chain &p);

} //namespace


#endif // _GLASS_CHAIN_HH_
