/*  Glass - a distributed computing library
    Copyright (C) 2003-2009 Bruno Barberi Gnecco
    Copyright (C) 2009-2010 Corollarium Technologies


    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef _GLASS_PLUGIN_EVENT_HH_
#define _GLASS_PLUGIN_EVENT_HH_

#include <map>
#include <queue>
#include <set>
#include "types.h"
#include "plugin.h"
#include "exception.h"
#include "packager.h"
#include <boost/type_traits.hpp>
#include <boost/thread.hpp>
#include "serial.h"

namespace libglass {

class _Event;

/*!
  Event plugin.

  \sa _Event.
*/
class EventBase : public PluginBase {
	/*!
	  Plugin creation.
	*/
	PLUGIN(EventBase);
private:
	/* client */
	std::map<const string, _Event *> 		events; /**< the list of events. */

	/**
	 * Store
	 */
	struct _EventBaseData {
		std::set<nodeId>	owners; /**< list of owners */
		propagationMode		mode; /**< Propagation mode for this event */
	};
	std::map<string, _EventBaseData> eventdata;	/**< Event owners. */
	boost::mutex				events_lock; /**< avoids deleting an event that is being accessed */

protected:
	friend class _Event;

	/*!
	  Called only in client mode.

	  Registers a event in clientmap.

	  \param e A pointer to the event object.
	  \retval true If successful.
	  \retval false If failed: event number already exists.
	  \sa unregisterEvent().
	*/
	bool registerEvent(_Event *e);

	/*!
	  Called only in client mode.

	  Unregisters a event in clientmap.

	  \param e A pointer to the event object.
	  \retval true If successful.
	  \retval false If failed: event number does not exist.
	  \sa registerEvent().
	*/
	bool unregisterEvent(_Event *e);

	// *********************************************************************
	// Destructor should be PROTECTED to ensure singleton behaviour
	// However, Microsoft compilers prior to VC v7.0 .NET are buggy and
	// generate the following error
	//   error C2248: 'AliasBase::~AliasBase' : cannot access protected member declared in class 'libglass::AliasBase'
	// The only workaround is to make the destructor public - which kind
	// of destroys the whole reason for implementing the singleton class
#if defined(_MSC_VER) && (_MSC_VER < 1300)  // earlier than .NET compiler (VC 7.0)
public:
#else
protected:
#endif
	// *********************************************************************
	/*!
	  Destructor.
	*/
	~EventBase();

public:

	/**
	 * Returns the instance of _Event with a given name or NULL if it does
	 * not exist.
	 * @param name
	 * @return
	 */
	_Event *getByName(const string &name) const;
};

class LIBGLASS_API _Event : public PluginInterface {
protected:
	EventBase 		*eb;		/**< EventBase owner. */
	string			name;		/**< Event name. */

	friend class EventBase;
	friend class EventPacket;

	/*!
	  Constructor.
	*/
	_Event(const char *name, propagationMode mode) throw(Exception);

	/*!
	  Destructor.
	*/
	virtual ~_Event();

	/*!
	  Enqueues an event, sending it to other nodes.

	  The event is not enqueued locally.

	  \param s The packed event.
	  \param target The target node.
	  \retval True if successfully sent.
	  \retval False if an error occurred.
	*/
	bool _enqueueEvent(chain &s, nodeId target);

	/*!
	  \overload bool _enqueueEvent(chain &s, nodeId target)
	  \param s The packed event.
	  \param targets The list of target nodes.
	  \param target The target node.
	*/
	bool _enqueueEvent(chain &s, std::vector<nodeId> &targets);

	/*!
	  Enqueues locally an event received by the network.

	  \param s The packed event.
	  \param sender The node which created the event.
	  \retval True if successfully queued.
	  \retval False if an error occurred.
	*/
	virtual bool pushEvent(chain &s, nodeId sender) = 0;
public:
};

/**
 * Class to hold event data
 */
template<typename T>
class _EventData {
public:
	T		data;
	nodeId		sender;
	unsigned int used;

	_EventData() {}
	~_EventData() {}
	void free(void) {}
};

/**
 * Specialization for arrays
 */
template<typename T>
class _EventData<T*> {
public:
	T *		data;
	nodeId		sender;
	unsigned int used;

	_EventData() : data(NULL) {}
	~_EventData() {
		/* we don't delete because the user will get the data. If we are throwing data
		away, call free() explicitly */
	}
	void free(void) {
		if (data) delete[] data;
	}
};

/*!
  Event class.
*/
template<typename T,
	GlassSerializer(*packf, T) = toString<T>,
	GlassUnserializer(*unpackf, T) = fromString<T> >
class Event : public _Event {
private:
	std::queue< _EventData<T> >	q;		/**< Event queue. */
	void 			(*handler)(T &);/**< Handler callback. */

protected:


	/*!
	  Pushes
	  \param c The serialized event data.
	  \retval True if successfully pushed to the stack.
	  \retval False otherwise.
	*/
	bool pushEvent(chain &c, nodeId sender) {
		_EventData<T>	d;
		unsigned int 	initial = 0;

		/* get the data */
		// TODO: used
		if (unpackf(c, initial, d.data, d.used, 1, (void *)1 /* delete */) == false) {
			return false;
		}

		d.sender = sender;

		if (handler) {
			handler(d.data);
		}
		else {
			q.push(d);
		}
		return true;
	}
public:

	/**
	 * Factory works better than explicit constructors, since it returns a valid
	 * variable if it has been created before, which the constructor will throw an
	 * exception saying that the variable already exists.
	 *
	 * @param name The variable name.
	 * @param data The variable initial data.
	 * @param len If it is an array, its length. Otherwise 1.
	 * @param mode Controls the propagation of this variable.
	 * @return A pointer to the existing variable if it exists,
	 */
	static Event *factory(const char *name, void (*func)(T &data) = NULL, propagationMode mode = Local) {
		const PluginManager *pm = PluginInterface::getPluginManager();
		if (!pm) {
			return NULL;
		}
		EventBase *sb(dynamic_cast<EventBase *>(pm->getPlugin("EventBase")));
		_Event *s = sb->getByName(name);
		if (s) {
			return dynamic_cast<Event *>(s);
		}
		return new Event(name, func, mode);
	}

	static Event *factory(const string &name, void (*func)(T &data) = NULL, propagationMode mode = Local) {
		return Event::factory(name.c_str(), func, mode);
	}

	Event(const char *name, void (*func)(T &data) = NULL, propagationMode mode = Local) throw(Exception) :
		_Event(name, mode), handler(func) {
	}

	~Event() {
	}

	/*!
	  Clears queue locally.
	*/
	void clear(void) {
		while (!q.empty()) {
			_EventData<T> d(q.front());
			d.free();
			q.pop();
		}
	}

	bool enqueueEvent(const T &data, nodeNickName nick, unsigned int len = 1, bool sendtoself = false) {
		return this->enqueueEvent(data, this->getGlass()->NickNameToId(nick), len, sendtoself);
	}

	/*!
	  Enqueues data.

	  \param data The data to enqueue.
	  \param target The target node.
	  \param len Length of the data, if appliable. If you are transferring
	  arrays, set this to the array length.
	  \param sendtoself Enqueues this event in the node that created it too.
	  \retval true if successfully sent.
	  \retval false if an error occurred.
	*/
	bool enqueueEvent(const T &data, nodeId target = everybodyId, unsigned int len = 1, bool sendtoself = false) {
		chain c;
		if (packf(c, data, len, NULL) == false) {
			return false;
		}
		bool r = _enqueueEvent(c, target);
		if (r == true && sendtoself == true && (target == everybodyId || target == this->getGlass()->getNodeId())) {
			// TODO: sendtoself is a hack. Should it be handled by send? The logic here for target == is imperfect.
			_EventData<T> ed;
			ed.data = data;
			ed.sender = this->eb->getGlass()->getNodeId();
			ed.used = 0;
			this->q.push(ed);
		}
		return r;
	}

	/*!
	  \overload bool enqueueEvent(T data, nodeId target = everybodyId)
	  \param data The data to enqueue.
	  \param targets The target node.
	  \param size Length of the data, if appliable. If you are transferring
	  arrays, set this to the array length.
	  \param sendtoself Enqueues this event in the node that created it too.
	  \retval true if successfully sent.
	  \retval false if an error occurred.
	*/
	bool enqueueEvent(T data, std::set<nodeId> &targets, unsigned int len = 1, bool sendtoself = false) {
		chain c;
		if (packf(c, data, len, NULL) == false) {
			return false;
		}
		bool r = _enqueueEvent(c, targets);
		if (r == true && sendtoself == true &&
				(targets.find(everybodyId) != targets.end() ||
				 targets.find(this->getGlass()->getNodeId()) != targets.end())) {
			// TODO: sendtoself is a hack. Should it be handled by send? The logic here for target == is imperfect.
			_EventData<T> ed;
			ed.data = data;
			ed.sender = this->getGlass()->getNodeId();
			ed.used = 0;
			this->q.push(ed);
		}
		return r;
	}

	/*!
	  Returns first event in queue. If this is dynamically allocated, user must free it.

	  \return First event in queue.
	*/
	// TODO: we should be able to return by reference
	T getEvent(nodeId id = everybodyId, unsigned int *used = 0) throw(EmptyException) {
		if (q.empty()) {
			throw EmptyException();
		}

		if (id == everybodyId) {
			_EventData<T> d(q.front());
			q.pop();
			if (used) *used = d.used;
			return d.data;
		}
		else {
			// TODO!!!!
			throw EmptyException();
		}
	}

	/*!
	  Checks queue for emptiness.

	  \retval True if empty.
	  \retval False if not empty.
	*/
	inline bool isEmpty(void) {
		return q.empty();
	}

	/*!
	  Sets a handler.

	  If the queue is not empty, the handler is called for every event
	  in queue.

	  \param func The handler function. If NULL, event changes to queueing
	    mode.
	  \retval True if successfull.
	  \retval False otherwise.
	*/
	bool setHandler(void (*func)(T &data)) {
		this->handler = func;

		if (!func) { // queue mode
			return true;
		}

		while (!q.empty()) {
			handler(q.front());
			q.pop();
		}

		return true;
	}

	/*!
	  Returns the number of entries in queue.
	  \return The number of entries in queue.
	*/
	unsigned int queueSize(void) {
		return q.size();
	}

};

class EventSerializable : public _Event {
private:
	std::queue<SerializableObject *>	q;	/**< Event queue. */
	void			(*handler)(SerializableObject *o);

protected:
	bool pushEvent(chain &s, nodeId sender);

public:

	EventSerializable(const char *name,
		void (*func)(SerializableObject *o) = NULL,
		propagationMode mode = Local) throw(Exception);

	~EventSerializable();

	/*!
	  Clears queue.
	*/
	void clear(void);

	/*!
	  Enqueues data.

	  \param o The object to enqueue.
	  \param target The target node id.
	  \retval True if successfully sent.
	  \retval False if an error occurred.
	*/
	bool enqueueEvent(SerializableObject &o, nodeId target = everybodyId);

#ifdef TODO
	bool enqueueEvent(SerializableObject &o, std::vector<nodeId> &targets);
#endif

	/*!
	  Get first event in queue.

	  User is responsible for freeing the memory.

	  \return First event in queue.
	*/
	SerializableObject *getEvent(void) throw(EmptyException);

	/*!
	  Checks queue for emptiness.

	  \retval True if empty.
	  \retval False if not empty.
	*/
	bool isEmpty(void);

	/*!
	  Sets a handler.

	  If the queue is not empty, the handler is called for every event
	  in queue.

	  The handler is responsible for freeing the object.

	  \retval True if successfull.
	  \retval False otherwise.
	*/
	bool setHandler(void (*func)(SerializableObject *o));

	/*!
	  Returns the number of entries in queue.
	  \return The number of entries in queue.
	*/
	unsigned int queueSize(void);
};

}//namespace


#endif // _GLASS_PLUGIN_EVENT_HH_
