/*  Glass - a distributed computing library
    Copyright (C) 2003-2009 Bruno Barberi Gnecco
    Copyright (C) 2009-2010 Corollarium Technologies


    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef _GLASS_HH_
#define _GLASS_HH_

#include <boost/thread.hpp>
// #include <boost/bimap.hpp>

#include "types.h"
#include "exception.h"

namespace libglass {

class PluginManager;
class PluginInterface;
class PluginBase;
class Protocol;
class Packet;

class LIBGLASS_API Glass {
protected:
	/* Friend classes. */
	friend class GlassManager;
	friend class PluginInterface;
	friend class PluginManager;
	friend class Protocol;

#ifdef BIMAP
	// this is some code that worked for 1:1 relations between nodeId and nodeAddress
	// since we may have two nodeIds with the same address (perhaps empty), we're
	// using standard std::map.
	struct Address  {};
	struct Id    {};

	typedef boost::bimap<
			boost::bimaps::tagged<nodeAddress, Address>,
			boost::bimaps::tagged<nodeId, Id>
		> AddressIdMap;

	typedef AddressIdMap::value_type AddressIdMapItem;

	AddressIdMap addridmap;  /**< store all ids and addresses. */

	/**
	 * Debug function
	 */
	void printAddrmap(void) const;
#endif

	typedef std::map<nodeId, nodeAddress> AddressIdMap;
	AddressIdMap addridmap;

	/**
	 * our basic mutex and lock types. Recursive mutexes are not elegant, but
	 * this class can be extended and overriden methods may call their parents.
	 */
	typedef boost::recursive_mutex mutex;
	typedef boost::recursive_mutex::scoped_lock lock;

	mutable mutex mutexdata; /// data mutex, for data that can change asynchronously.

	Protocol 			*protocol; 	/**< the protocol to use */
	nodeId				id;		/**< node id. */
	PluginManager		*pm;		/**< our internal plugin manager */
	boost::thread::id	pt_id;		/**< thread id (used in multi-threaded programs) */
	nodeNickName		nickname;

	/*!
	  This is a virtual destructor.
	  
	  It exists to circumvent a C++ behaviour. What happens is this: 
	  when a Glass-derived class (let's call it GD) calls a destructor,
	  it in turn calls Glass' destructor. So far so good; the problem is
	  that Glass' destructor accesses some virtual functions that may
	  have been overriden by GD. But, by this time, GD has been destroyed
	  (or, at least, is not "seen" by Glass' destructor). So we have to
	  do this "virtual destructor idiom".
	  
	  Clean does not delete the protocol, so you may still send messages
	  on your destructor after calling clean. This is useful for a
	  swan's song.
	  
	  You \e MUST call clean() in your GD destructor.
	*/
	void clean(void);

	/**
	 * Sets protocol owner.
	 */
	void setProtocolOwner(const Glass *g);

	/*!
	  Forwards the packet to the pluginManager, guaranteeing protectedness.

	  \param p The packet to send.
	  \param rootOnly Only process root packets; other packets are ignored.
	  \retval true If packet was correctly processed.
	  \retval false If an error occurred while processing packet or if a
	  	non-root packet was ignored.
	*/
	virtual bool processPacket(Packet &p, bool rootOnly = false);
	
	/*!
	  Forwards the packet to the Protocol, guaranteeing protectedness. It
	  is also responsible for translating the Packet's nodeIds to nodeAddresses.

	  \param p The packet to send.
	  \retval True on success.
	  \retval False if packet couldn't be sent for some reason.
	*/
	virtual bool sendPacket(Packet &p);

	/*!
	  Responsible for processing 

	  \param p The packet to process.
	  \retval True on success.
	  \retval False .
	*/
	virtual bool rootPacket(Packet &p) = 0;

	/*!
	  Returns a plugin given its id.

	  \param id The plugin id.
	  \return The plugin, or NULL if not found.
	  \sa getPluginManager().
	*/
	PluginBase *getPlugin(PluginId id) const;

	/*!
	  Returns our plugin manager.

	  \return Our plugin manager.
	  \sa getPlugin().
	*/
	PluginManager *getPluginManager(void) const;

	/*!
	  Converts a nodeId to a nodeAddress.

	  \param id The node id.
	  \return The associated nodeAddress.
	  \sa FDToId
	  \sa IdToAddress
	*/
	virtual const nodeAddress &IdToAddress(nodeId id) const;

	/**
	 * Returns a nodeAddress structure given a fd.
	 *
	 * If there is no nodeAddress currently registered with this fd,
	 * a new nodeAddress is returned that is == nobodyAddress,
	 * except for the fd field, which is set to the fd argument.
	 *
	 * @param fd The fd to search
	 */
	virtual const nodeAddress FDToAddress(nodeFD fd) const;
	

	/**
	 * Converts a nodeAddress to a nodeId.
	 *
	 * @param addr The node address.
	 * @return The associated nodeId, or nobodyId if none found.
	 * @sa IdToFD.
	*/
	virtual nodeId AddressToId(const nodeAddress &addr) const;

	/*!
	  Registers a node.

	  \param fd A valid node descriptor.
	  \param id A nodeId, that possibly came from the other node. Derived class can choose
	  whether to use or ignore. May be == nobodyId.
	  \return A new node Id, or nobodyId if an error occurred.
	  \sa unregisterNode().
	*/
	virtual nodeId registerNode(nodeAddress &addr, nodeId id, nodeRelation r, bool force) = 0;

	/*!
	  Associates a given fd and id to each other, and the relationship.

	  \param fd The node fd.
	  \param id The node id.
	  \param r The node relationship.
	  \param force If true, associates even if the id or fd were already
	     associated to others. Does not change the previous association.
	  \retval true if association was ok.
	  \retval false if an error ocurred.
	  \sa registerNode(), setRelation().
	*/
	virtual bool setNode(nodeAddress &addr, nodeId id, nodeRelation r, bool force = false);

	/*!
	  Changes a node relationship.

	  \param id The node id.
	  \param r The relationship of the node.
	  \retval true If successful.
	  \retval false If an error occurred (id or r invalid).
	  \sa registerNode(), unregisterNode(), setNode().
	*/
	virtual bool setRelation(nodeId id, nodeRelation r);

	/*!
	  Unregisters a id/fd association.
	  
	  Unregisters the relationship too.

	  \param id The node id.
	  \retval True if successful.
	  \retval False if not in list.
	*/
	virtual bool unregisterNode(nodeId id);

public:
	explicit Glass(Protocol *p) throw(Exception);
	virtual ~Glass();

	/*!
	  Returns this node's Id.
	 
	  \return This node's Id.
	*/
	nodeId getNodeId(void) const;

	/*!
	  Returns this node's nickname

	  \return This node's nicmake
	*/
	nodeNickName getNodeNickname(void) const;

	/*!
	  Returns the relationship with a given node.
	 
	  \param id The other node's id.
	  \return The relationship of \e the node id with this one. That
	    is, if this node is the Server and id is the Client, returns
	    Client.
	*/
	virtual nodeRelation getNodeRelation(nodeId id) const;

	/*!
	  Sets this Glass instance as the current one.
	  \sa lockCurrent(), unlockCurrent().
	*/
	void setCurrent(void);

	/*!
	  \sa unlockCurrent(), setCurrent().
	*/
	void lockCurrent(void);
	
	/*!
	  \sa lockCurrent(), setCurrent().
	*/
	void unlockCurrent(void);

	/*!
	  Returns the number of nodes connected to the network.
	  
	  This is a virtual function. The default implementation returns 0.
	  Most derived class will override this function.
	  
	  This function may be slow, requiring a network transmission. The
	  number of nodes may change at any time.

	  \e Important: The actual implementation in the derived class may
	  choose to return the number of nodes of a certain type. For example,
	  the client/server architecture returns the number of CLIENTS, which
	  makes sense, since the server is not doing any computation on the
	  application. So, refer to the documentation of the derived class to
	  know exactly what you are getting.

	  \return The number of nodes connected to the network, or zero if
	  could not get it. See the derived class documentation for a more
	  complete description.
	*/
	virtual unsigned int getTotalNodes(void);

	/**
	 * @param nick The nickname
	 * @return The id, or everybodyId if this nick is not registered.
	 */
	virtual nodeId NickNameToId(const nodeNickName &nick) const;

};

} // namespace

#endif // _GLASS_HH_
