/*  Glass - a distributed computing library
    Copyright (C) 2003-2009 Bruno Barberi Gnecco
    Copyright (C) 2009-2010 Corollarium Technologies


    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef _GLASS_CLIENT_HH_
#define _GLASS_CLIENT_HH_

#include <set>

#include "glass.h"
#include "types.h"
#include "exception.h"
#include "thread.h"

namespace libglass {

class Protocol;

class LIBGLASS_API GlassClient : public Glass, protected GlassThread {
private:
	bool			initialized;	/**< is initialization complete? */
	bool			initsuccess;	/// have we initialized successfully?
	bool			quitting;	/**< is this class being destroyed? */
	bool			tryuntilconnect;
	bool			ismaster;	/**< is this node the master? */

protected:
	nodeAddress		serveraddr;
	nodeId			serverid;	/**< */

	boost::condition_variable cond; /**< initialization condition variable */
	boost::mutex mut;		/**< initialization mutex */

	/*! 
	  Thread method. 
	*/
	void execute(void);

	/*! 
	  Thread method. 
	*/
	void cleanup(void);

	bool sendPacket(Packet &p);

	virtual bool rootPacket(Packet &p);

	virtual nodeId registerNode(nodeAddress &addr, nodeId id, nodeRelation r, bool force = false);

public:
	/*! 
	  Constructor.
	  \param p The protocol to use.
	  \param server The server name or IP.
	  \param nickname A string that can be used to address this node. Unlike
	  nodeId, this string does not change upon each connection, but there is
	  OTOH no guarantee that it will be unique -- user should ensue this.
	  \param tryuntilconnect If true, try to connect until successful (and
	  only return if successful!). If false, quit after trying to connect
	  once and throw an Exception.
	  \throws Exception If not successful
	*/
	GlassClient(Protocol *p, const std::string &server,
			const std::string &nickname = "", bool tryuntilconnect = false) throw(Exception);

	virtual ~GlassClient();
	
	/*!
	  Returns the number \e CLIENT nodes connected to the network.
	  
	  This function returns the number \e CLIENT nodes connected to the 
	  network. There's only one server around.
	  
	  The returned number is not guaranteed to be valid, as nodes may
	  join or leave at any time. 

	  \bug As it is, there may be a racing condition if a packet containing
	  a new list arrives at the precise moment this function is called. A
	  mutex could solve the problem in theory, but it would have to be 
	  called after return or require an extra copy to a temporary variable.
	  
	  \return The number of \e CLIENT nodes connected to the network, or 
	  zero if an error occurred.
	*/
	virtual unsigned int getTotalNodes(void);

	/*!
	  Returns the list of all client nodes.
	  
	  The returned value is not guaranteed to be valid, as nodes may
	  join or leave at any time. 

	  \bug As it is, there may be a racing condition if a packet containing
	  a new list arrives at the precise moment this function is called. A
	  mutex could solve the problem in theory, but it would have to be 
	  called after return or require an extra copy to a temporary variable.
	  
	  \return A set with all valid nodeIds at the moment the call 
	  was executed, or an empty list if an error occurred. We return a copy of
	  the list, so this is race-safe.
	*/
	std::map<nodeId, nodeNickName> getClientList(void) const;

	/*!
	  Is this node the master?

	  \bug As it is, there may be a racing condition if a packet containing
	  a new list arrives at the precise moment this function is called. A
	  mutex could solve the problem in theory, but it would have to be 
	  called after return or require an extra copy to a temporary variable.
	  
	  \retval true if this is the node with lowest id.
	  \retval false otherwise.
	*/
	bool isMaster(void);

	
	/*!
	  Close all nodes, including this one.
	*/
	void killAll(void);

};

} // namespace

#endif // _GLASS_CLIENT_HH_
