/*  Glass - a distributed computing library
    Copyright (C) 2003-2009 Bruno Barberi Gnecco
    Copyright (C) 2009-2010 Corollarium Technologies


    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef _GLASS_MEMORYMANAGER_HH
#define _GLASS_MEMORYMANAGER_HH

#include "types.h"
#include <new>
#include <boost/smart_ptr.hpp>

namespace libglass {

class MemoryManager;

/**
 * Lock class. This does nothing, but a derived class may implement a
 * real lock of some kind.
 */
class _Lock {
public:
	const MemoryManager &mm;
	_Lock(MemoryManager &mm) : mm(mm) { }

	/**
	 * Unlocks on destruction
	 */
	virtual ~_Lock() { }

	/**
	 * Unlocks
	 */
	virtual void unlock(void) { }
};

typedef boost::shared_ptr<_Lock> Lock;

/**
 * Abstract base class.
 *
 */
class LIBGLASS_API MemoryManager {
public:

	MemoryManager() { }

	virtual ~MemoryManager() {}

	/**
	 * Allocates memory. Identifier, if used, should be unique; see the
	 * child class documentation for more information.
	 *
	 * @param nbytes Total bytes to allocated. Implementation may allocate more
	 * bytes at its discretion.
	 * @param identifier The (probably unique) identifier for this region. Optional,
	 * see the actual implementation docs.
	 * @return A pointer to the allocated region.
	 */
	virtual void* allocate(size_t nbytes, const std::string &identifier = "") = 0;

	/**
	 * Deallocates a memory that was previously allocated by this class.
	 */
	virtual void deallocate(void *p) = 0;

	/**
	 * Returns a scoped lock object. This is essentially a mutex for this
	 * instance. allocate() and deallocate() call this automatically, so
	 * you shouldn't.
	 */
	virtual Lock scoped_lock() {
		_Lock *l(new _Lock(*this));
		return Lock(l);
	}
};

/**
 * Default memory manager. Uses plain malloc.
 */
class LIBGLASS_API DefaultMemoryManager : public MemoryManager {
public:

	DefaultMemoryManager();

	virtual ~DefaultMemoryManager();

	virtual void* allocate(size_t nbytes, const std::string &identifier = "");

	virtual void deallocate(void *p);
};

} //namespace

/**
 * Placement new.
 *
 * @param nbytes
 * @param pool The memory manager instance to use.
 * @param identifier The identifier.
 */
extern void* operator new(size_t nbytes, libglass::MemoryManager& pool, const std::string &identifier = "");

/**
 * Placement delete.
 *
 * @param p
 * @param pool
 */
extern void operator delete(void* p, libglass::MemoryManager& pool);


#endif // _GLASS_MEMORYMANAGER_HH
