/*  Glass - a distributed computing library
    Copyright (C) 2003-2009 Bruno Barberi Gnecco
    Copyright (C) 2009-2010 Corollarium Technologies


    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef _GLASS_PLUGIN_MESSENGER_HH_
#define _GLASS_PLUGIN_MESSENGER_HH_

#include <map>
#include <set>
#include <queue>
#include "plugin.h"
#include "exception.h"
#include <boost/thread.hpp>

namespace libglass {

class _Messenger;

/*! 
  Messenger plugin.
  
  Provides point to point communication.

  \sa Messenger.
*/
class MessengerBase : public PluginBase {
	/*! 
	  Plugin creation. 
	*/
	PLUGIN(MessengerBase);

private:
	std::map<const string, _Messenger *>
				messengers;		/**< List of messengers */
	boost::mutex				msg_lock; /**< avoids deleting an event that is being accessed */

protected:
	friend class _Messenger;

	/*!
	  Destructor.
	*/
	~MessengerBase();

	/*!
	  Registers a messenger in the list.
	  \param v A pointer to the messenger.
	  \retval true If successful.
	  \retval false Otherwise.
	*/
	bool registerMessenger(_Messenger *v);

	/*!
	  Unregisters a messenger, taking it off the list.
	  \param v A pointer to the messenger.
	  \retval true If successful.
	  \retval false Otherwise.
	*/
	bool unregisterMessenger(_Messenger *v);
public:
};

/*!
  Messenger class.


*/
class _Messenger : public PluginInterface { 
private:
	MessengerBase 		*mb;		/**< MessengerBase owner. */
	string			name;

	std::set<nodeId>	owners;

protected:
	friend class MessengerBase;

	/*! 
	  Constructor.
	*/
	_Messenger(const char *name) throw(Exception);

	/*! 
	  Destructor.
	*/
	virtual ~_Messenger();

	/*! 
	  Sends a message to a given node.

	  \param id
	  \param message
	  \retval true if successful. Does not guarantee that the target node
	  received the message, but that it was sent across the network.
	  \retval false otherwise.
	  \sa getMessage().
	*/
	bool _sendMessage(nodeId id, chain &message);

	/*! 
	  Enqueues locally a messaged received by the network.

	  \param s The packed message.
	  \retval True if successfully queued.
	  \retval False if an error occurred.
	*/
	virtual bool pushMessage(chain &) = 0;
public:
};

/*!
  Message class
*/
template<typename T,
	GlassSerializer(*packf, T) = toString<T>,
	GlassUnserializer(*unpackf, T) = fromString<T> >
class Messenger : public _Messenger {
protected:
	std::queue<string>	q;
	void 			(*handler)(T &);

	bool pushMessage(chain &c) {
		
	}
public:

	Messenger(const char *name, void (*func)(T &data) = NULL) throw(Exception) :
		_Messenger(name), handler(func) {
	}

	/*!
	  Gets a message from a given node.

	  \param id The id of the node that you want to get a message from.
	  \param message A string where the message will be stored.
	  \param block If true, the call will block until a message arrives.
	  \retval true if a package was available and successfully stored in
	    message.
	  \retval false otherwise.
	  \sa sendMessage().
	*/
	bool getMessage(nodeId id, string &message, bool block = false);


	bool setMessageHandler(void (*func)(string &message));

	/*!
	  Sets a handler.
	  
	  If the queue is not empty, the handler is called for every event
	  in queue. 

	  The handler is responsible for freeing the object.

	  \retval True if successfull.
	  \retval False otherwise.
	*/
	bool setHandler(void (*func)(SerializableObject *o));
};

}//namespace

#endif
