/*  Glass - a distributed computing library
    Copyright (C) 2003-2009 Bruno Barberi Gnecco
    Copyright (C) 2009-2010 Corollarium Technologies


    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef _GLASS_MSVC_HH_
#define _GLASS_MSVC_HH_


/*! \file msvc.h
    \brief Support for MSVC compiler
*/

#if defined(_MSC_VER)


#define MSG_NOSIGNAL 0
typedef int socklen_t;


// --- Only applicable to Microsoft C compiler ---

// The following ifdef block is the standard way of creating macros which make exporting 
// from a DLL simpler. All files within this DLL are compiled with the LIBGLASS_EXPORTS
// symbol automatically defined on the command line if using Microsoft Visual Studio. This
// symbol should not be defined on any project that imports this DLL. This way any other
// project whose source files use this DLL see LIBGLASS_API functions as being _imported_
// from a DLL, wheras this DLL sees symbols defined with this macro as being _exported_

// Also, see Microsoft Knowledge Base Article 168958:
// How To Exporting STL Components Inside & Outside of a Class

#  if defined(LIBGLASS_DLL)
#    if defined(LIBGLASS_EXPORTS)
#      define LIBGLASS_API __declspec(dllexport)
#      define LIBGLASS_IMP_TEMPLATE
#    else
#      define LIBGLASS_API __declspec(dllimport)
#      define LIBGLASS_IMP_TEMPLATE extern
#    endif // LIBGLASS_EXPORTS
#  else /* LIBGLASS_DLL */
#    define LIBGLASS_API
#    define LIBGLASS_IMP_TEMPLATE
#  endif // LIBGLASS_DLL


// --- Export/Import all STL classes used by the libGlass library ---

// Note that you may not export a generalized template. The template must be instantiated;
// that is, all of the template parameters must be supplied and must be completely defined
// types at the point of instantiation. For instance "stack<int>;" instantiates the STL stack
// class. The instantiation forces all members of class stack<int> to be generated.



// --- Workarounds for bugs in the Microsoft Visual C compiler ---


// The VC compiler can't handle symbols longer than 255 characters and
// automatically truncates (with associated warning message).
// Unfortunately the Standard Template Library (STL) often creates symbols
// longer than this limit so we need to disable these warning messages.
#  pragma warning(disable:4786)
// Similarly VC compiler can't handle decorated names longer than 247
// characters and automatically truncates (with associated warning message)
#  pragma warning(disable:4503)
// Disable warning message 4290: C++ Exception Specification ignored
#  pragma warning( disable: 4290 )


// Standard template library (STL) header files
// STL not optimized for level 4 warnings so we wrap the header files in
// a pragma which temporarily sets the debug level to 3
#  pragma warning( push, 3 )
#  include <iostream>
#  include <utility>
#  include <vector>
#  include <list>
#  include <algorithm>
#  include <fstream>
#  include <memory>
#  include <map>
// Restore the previous warning level
#  pragma warning( pop )


#  if defined(LIBGLASS_DLL)

// Push the current warning state so that we can restore later
#    pragma warning( push )
// Disable warning if template class is already instantiated
#    pragma warning(disable:4660)


// We need to define user-defined types here so that they can be used  in
// template instantiation below
#include "types.h"

// See Microsoft Knowledge Base Article 168958: How To Exporting STL Components Inside & Outside of a Class...
// Note that you may not export a generalized template. The template must be instantiated; that is, all of the
// template parameters must be supplied and must be completely defined types at the point of instantiation.
// For instance "stack<int>;" instantiates the STL stack class. The instantiation forces all members of class
// stack<int> to be generated.
// Note also that base classes will need to be instantiated.

/*!
 Macro for instantiating std::map containers (i.e. creating a specialisation 
 of the template class)
 @deprecated MSVC now supports exporting STL containers
*/
#    define LIBGLASS_IMP_TEMPLATE_STL_MAP(Ta,Tb)
/*
  LIBGLASS_IMP_TEMPLATE template struct LIBGLASS_API _STL::less<Ta>; \
  LIBGLASS_IMP_TEMPLATE template struct LIBGLASS_API _STL::less<Tb>; \
  LIBGLASS_IMP_TEMPLATE template class LIBGLASS_API _STL::_STLP_alloc_proxy<_STL::_rbT__B<_STL::map<Ta,Tb>::value_type,_STL::allocator<_STL::pair<const Ta,Tb> > >::_Node *,_STL::_rbT__B<_STL::map<Ta,Tb>::value_type,_STL::allocator<_STL::pair<const Ta,Tb> > >::_Node,_STL::_rbT__B<_STL::map<Ta,Tb>::value_type,_STL::allocator<_STL::pair<const Ta,Tb> > >::_M_node_allocator_type>; \
  LIBGLASS_IMP_TEMPLATE template class LIBGLASS_API _STL::_Rb_tree<_STL::map<Ta,Tb>::key_type,_STL::map<Ta,Tb>::value_type,_STL::_S1st<_STL::map<Ta,Tb>::value_type>,_STL::map<Ta,Tb>::key_compare,_STL::allocator<_STL::pair<const Ta,Tb> > >;  \
  LIBGLASS_IMP_TEMPLATE template class LIBGLASS_API _STL::map<Ta,Tb>
*/

/*!
 Macro for instantiating std::list containers (i.e. creating a specialisation of the 
 template class)
 @deprecated MSVC now supports exporting STL containers
*/
#define LIBGLASS_IMP_TEMPLATE_STL_LIST(T) 
/*
	LIBGLASS_IMP_TEMPLATE template class LIBGLASS_API _STL::_STLP_alloc_proxy<_STL::_List_base<T,_STL::allocator<T> >::_Node *,_STL::_List_base<T,_STL::allocator<T> >::_Node,_STL::_List_base<T,_STL::allocator<T> >::_Node_allocator_type>; \
	LIBGLASS_IMP_TEMPLATE template class LIBGLASS_API _STL::list<T>
*/

  /*!
 Macro for instantiating std::set containers (i.e. creating a specialisation of 
 the template class)
 @deprecated MSVC now supports exporting STL containers
 */
#define LIBGLASS_IMP_TEMPLATE_STL_SET(T)
/*	LIBGLASS_IMP_TEMPLATE template struct LIBGLASS_API _STL::less<T>; \
	LIBGLASS_IMP_TEMPLATE template class LIBGLASS_API _STL::_STLP_alloc_proxy<_STL::_rbT__B<_STL::set<T>::value_type,_STL::allocator<T> >::_Node *,_STL::_rbT__B<_STL::set<T>::value_type,_STL::allocator<T> >::_Node,_STL::_rbT__B<_STL::set<T>::value_type,_STL::allocator<T> >::_M_node_allocator_type>; \
	LIBGLASS_IMP_TEMPLATE template class LIBGLASS_API _STL::_Rb_tree<_STL::set<T>::key_type,_STL::set<T>::value_type,_STL::_Identity<_STL::set<T>::value_type>,_STL::set<T>::key_compare,_STL::allocator<T> >; \
	LIBGLASS_IMP_TEMPLATE template class LIBGLASS_API _STL::set<T>
*/

// LIBGLASS_IMP_TEMPLATE template class LIBGLASS_API _STL::basic_string<char,_STL::char_traits<char>,_STL::allocator<char> >;

// Restore compiler warning state to previous state
#    pragma warning( pop )


#  endif // LIBGLASS_DLL


#else // _MSC_VER

// --- For all other compilers ---

#  define LIBGLASS_API

#endif // _MSC_VER




#endif // _GLASS_MSVC_HH_
