/*  Glass - a distributed computing library
    Copyright (C) 2003-2009 Bruno Barberi Gnecco
    Copyright (C) 2009-2010 Corollarium Technologies


    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef _GLASS_PACKAGER_HH_
#define _GLASS_PACKAGER_HH_

#include <iostream>
#include "chain.h"
#include "types.h"
#include <sstream>
#include <boost/archive/text_oarchive.hpp>
#include <boost/archive/text_iarchive.hpp>
#include <boost/archive/binary_oarchive.hpp>
#include <boost/archive/binary_iarchive.hpp>
#include <boost/iostreams/stream.hpp>
#include <boost/serialization/serialization.hpp>
#include <boost/serialization/map.hpp>
#include <boost/serialization/list.hpp>
#include <boost/serialization/string.hpp>
#include <boost/serialization/vector.hpp>

// using namespace libglass;

namespace libglass {

#define GlassSerializer(name, type) \
	bool (name)(chain &, type const &, unsigned int, void *)
#define GlassUnserializer(name, type) \
	bool (name)(const chain &, unsigned int, type &, unsigned int &, \
		unsigned int, void *)

/*
 * Interoperable using STL
 */

/*! 
   Serializes

   \param t The data.
   \param c The chain to append to.
   \param nmemb Number of elements. Use with arrays.
   \param extra Any extra data to be used by this function.
   \return The length of the data appended to the chain, or -1 if an error
   occurred.
*/
template<typename T>
bool toString(chain &c, T const &t, unsigned int nmemb = 1, void *extra = NULL) {
	std::ostringstream		s;
	s << t;

	string st = s.str(); 

#if defined(_MSC_VER)
	// Push the current warning state so that we can restore later */
#  pragma warning( push )
	// Disable warning message 4267: conversion from 'size_t' to 'unsigned int', possible loss of data
#  pragma warning( disable: 4267 )
#endif // _MSC_VER
	unsigned int size = st.size(); // *** Possible loss of data here ***
#if defined(_MSC_VER)
	// Restore compiler warning state to previous state
#  pragma warning( pop )
#endif // _MSC_VER

	// TODO: use a special allocator in string st to avoid this copy/malloc
	char *p = new char[size];
	memcpy(p, st.data(), size);
	c.append(p, size, true);
	return true;
}

template<typename T>
bool toString(chain &c, T *const &t, unsigned int nmemb = 1, void *extra = NULL) {
	std::ostringstream		s;
	T 				*ptr = t;

	while (nmemb-- > 0) {
		s << *ptr++ << " ";
	}
	string st = s.str();
	int size = st.size();
	char *p = new char[size];
	memcpy(p, st.data(), size);
//	std::cerr << "p " << p << endl;
	c.append(p, size, true);
	return true;
}

/*! 
   Unserializes

   \param c The chain containing data to unserialize.
   \param offset The start of the data to unserialize.
   \param data A reference to unserialize to.
   \param used A reference to a integer which will contain the number of used
      characters by the unserializer.
   \param nmemb Number of elements. Use with arrays.
   \param extra Any extra data to be used by this function.
   \retval True if successful.
   \retval False otherwise.
*/
template<typename T>
bool fromString(const chain &c, unsigned int offset, T &data, unsigned int &used,
			unsigned int nmemb = 1, void *extra = NULL) {
	try {
		std::istringstream is(c.toString(offset));
		is >> data;
		used = is.tellg();
	}
	catch (Exception e) {
		return false;
	}
	return true;
} 

template<typename T>
bool fromString(const chain &c, unsigned int offset, T * const &data, unsigned int &used,
			unsigned int nmemb = 1, void *extra = NULL) {
	T 						*ptr = data;

	try {
		std::istringstream is(c.toString(offset));
		while (nmemb-- > 0) {
			is >> *ptr++;
			used += is.tellg();
		}
	}
	catch (Exception e) {
		return false;
	}
	return true;
} 

/*
 * RAW
 */
template<typename T>
bool rawPack(chain &c, T const &t, unsigned int nmemb = 1, void *extra = NULL) {
	c.append(reinterpret_cast<const char *>(&t), sizeof(T));
	return true;
}
template<typename T>
bool rawPack(chain &c, T *const &t, unsigned int nmemb = 1, void *extra = NULL) {
	unsigned int used = sizeof(T)*nmemb;
	c.append(reinterpret_cast<const char *>(t), used);
	return true;
}

template<typename T>
bool rawUnpack(const chain &c, unsigned int offset, T &data, unsigned int &used,
			unsigned int nmemb = 1, void *extra = NULL) {
	used = sizeof(T);
	try {
		c.substrcopy(reinterpret_cast<char *>(&data), offset, offset+used-1);
	}
	catch (Exception e) {
		return false;
	}
	return true;
}
template<typename T>
bool rawUnpack(const chain &c, unsigned int offset, T * const &data, unsigned int &used,
			unsigned int nmemb = 1, void *extra = NULL) {
	used = sizeof(T)*nmemb;
	try {
		c.substrcopy(reinterpret_cast<char *>(data), offset, offset+used-1);
	}
	catch (Exception e) {
		return false;
	}
	return true;
}

/*
 *
 * BOOST
 *
 */
// TODO: test.
template<typename _T>
bool binaryPackBoost(chain &c, const _T & s, unsigned int nmemb = 1, void *extra = NULL) {
	std::stringstream ss(std::ios_base::binary);
	boost::archive::binary_oarchive oa(ss);
	oa << s;

	string st(ss.str());
	c.append(st.data(), st.length());
	return true;
}
// TODO: test.
template<typename _T>
bool binaryUnpackBoost(const chain &c, unsigned int offset,
		_T &s, unsigned int &used, unsigned int nmemb = 1, void *extra = NULL) {
	try {
		std::string cs(c.toString(offset));
		std::istringstream ifs(cs, std::ios_base::binary);
		boost::archive::binary_iarchive archive(ifs);
		archive >> s;
		used = ifs.tellg();
	}
	catch (Exception e) {
		return false;
	}
	catch (boost::archive::archive_exception e) {
		std::cerr << e.what() << std::endl;
		return false;
	}
	return true;
}


template<typename _T>
bool textPackBoost(chain &c, const _T & s, unsigned int nmemb = 1, void *extra = NULL) {
	std::stringstream ss;
	boost::archive::text_oarchive oa(ss);
	oa << s;

	string st(ss.str());
	c.append(st.data(), st.length());
	return true;
}
template<typename _T>
bool textUnpackBoost(const chain &c, unsigned int offset,
		_T &s, unsigned int &used, unsigned int nmemb = 1, void *extra = NULL) {
	try {
		std::string cs(c.toString(offset));
		std::istringstream ifs(cs);
		boost::archive::text_iarchive archive(ifs);
		archive >> s;
		used = ifs.tellg();
	}
	catch (Exception e) {
		return false;
	}
	catch (boost::archive::archive_exception e) {
		std::cerr << e.what() << std::endl;
		return false;
	}
	return true;
}



/* 
 * Specializations
 */
#ifdef TODO
template<>
bool toString<int>(chain &c, const int & t, unsigned int nmemb = 1, 
		void *extra = NULL);
template<>
bool toString<int>(chain &c, int *const &t, unsigned int nmemb = 1, 
		void *extra = NULL);

template<>
bool fromString(const chain &c, unsigned int offset, int &data, 
		unsigned int &used, unsigned int nmemb = 1, void *extra = NULL);
template<>
bool fromString(const chain &c, unsigned int offset, int * &data, 
		unsigned int &used, unsigned int nmemb = 1, void *extra = NULL);


template<>
bool toString<float>(chain &c, const T& t);
template<>
bool fromString<float>(const chain &c, unsigned int offset, T &data, unsigned int &used);
#endif

template<>
bool fromString<std::string>(const chain &c, unsigned int offset, 
		std::string &data, unsigned int &used, unsigned int nmemb, 
		void *extra);

template<>
bool rawPack<std::string>(chain &c, std::string const &t, 
		unsigned int nmemb, void *extra);
template<>
bool rawUnpack<std::string>(const chain &c, unsigned int offset, 
		std::string &data, unsigned int &used, unsigned int nmemb, 
		void *extra);

template<>
bool rawPack<char *>(chain &c, char* const &t, unsigned int nmemb, void *extra);
template<>
bool rawUnpack<char *>(const chain &c, unsigned int offset,
		char * &data, unsigned int &used, unsigned int nmemb,
		void *extra);


} // namespace

#endif
