/*  Glass - a distributed computing library
    Copyright (C) 2003-2009 Bruno Barberi Gnecco
    Copyright (C) 2009-2010 Corollarium Technologies


    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef _GLASS_PACKET_HH_
#define _GLASS_PACKET_HH_

#include <ostream>
#include <set>
#include <vector>
#include "types.h"
#include "chain.h"

namespace libglass {

class Glass;
const PluginId rootBaseId = PluginId("RootBase");

class Packet {
protected:
	nodeId			from;		/**< who sent it */
	nodeAddress		fromaddress;
	string			identifier;	/**< unique identifier for this data. May be used by some protocols */

	chain			data;		/**< the data. */
	PluginId		pid;		/**< the plugin id */
	bool			reliable;	/**< must be sent by a reliable protocol? */
	unsigned int	maximumDataLength; /**< some protocols may need this to create buffers. Value should be constant for any packets with this identifier. Default is zero: unknown. */
	int				timeout;	/**< timeout in milliseconds. If message is not send within this time, it's thrown away. */

	Glass			*g;

	std::set<nodeId>	targetIds;	/**< destination ids */
	std::set<nodeAddress>	targetAddresses;	/**< destination addresses */

	std::set<nodeId>	blackIds;	/**< destination ids that should not be sent to */
	std::set<nodeId>	whiteIds;	/**< destination ids that may be sent to */


protected:
	friend class Glass;
	friend class Protocol;

public:
	/*!
	  Constructor.
	*/
	explicit Packet(Glass *g);

	/*!
	  Destructor.
	*/
	virtual ~Packet();

	/*!
	 *
	 * @return The Glass instance
	 */
	Glass *getGlass(void) const;

	/*!
	  Sets the data field.
	  \param data The data. May contain '\0's.
	  \param length The length of data.
	*/
	void setData(unsigned char *data, unsigned int length, bool free = false, unsigned char *allocatedbegin = NULL);

	/*!
	  Sets the data field.
	  \param data The data. May contain '\\0's.
	  \param length The length of data.
	*/
	void setData(const char *data, unsigned int length, bool free = false, unsigned char *allocatedbegin = NULL);

	/*!
	  \overload void setData(unsigned char *data, unsigned int length)
	  \param s The data.
	*/
	void setData(chain &s);

	/*!
	  Appends to the data field.
	  \param data The data. May contain '\0's.
	  \param length The length of data.
	  \sa setData().
	*/
	void appendData(unsigned char *data, unsigned int length, bool free = false, unsigned char *allocatedbegin = NULL);

	/*!
	  Appends to the data field.
	  \param data The data. May contain '\\0's.
	  \param length The length of data.
	  \sa setData().
	*/
	void appendData(const char *data, unsigned int length, bool free = false, char *allocatedbegin = NULL);

	/*!
	  \overload void appendData(unsigned char *data, unsigned int length)
	  \param s The data.
	*/
	void appendData(string &s);

	/*!
	  \overload void appendData(unsigned char *data, unsigned int length)
	  \param s The data.
	*/
	void appendData(chain &s);

	/*!
	  Gets data length.
	  \sa setData(), appendData(), getData().
	  \return The data length.
	*/
	unsigned int getDataLength(void) const;

	unsigned int getMaximumDataLength(void) const { return this->maximumDataLength; }

#ifdef DEPRECATED
	/*!
	  Gets data.

	  User is responsible for delete [] the returned pointer.

	  \sa setData(), appendData(), getDataLength().
	  \return The data, an array of getDataLength() characters.
	*/
	unsigned char *getData(void) const;
#endif

	chain &getData(void);

	/*!
	  Gets data.

	  \sa setData(), appendData(), getDataLength().
	*/
	void getData(chain &s);

	/*!
	  Sets the plugin id field.
	  \param id The plugin id.
	  \sa getPluginId().
	*/
	void setPluginId(PluginId id);

	/*!
	  Gets the plugin id field.
	  \return The plugin id.
	  \sa setPluginId().
	*/
	PluginId getPluginId(void) const;

	const string &getIdentifier(void) const { return identifier; }

	/*!
	  Sets the target node id field.
	  \param id The target node id.
	  \sa getTargetNodes().
	*/
	void setTargetNodes(nodeId id);

	/*!
	  Sets the target node id field.
	  \param id_set The target node ids set.
	  \sa getTargetNodes().
	*/
	void setTargetNodes(std::set<nodeId> &id_set);

	/*!
	  Sets the target node id field.
	  \param id_set The target node ids set.
	  \sa getTargetNodes().
	*/
	void setTargetNodes(std::vector<nodeId> &id_set);

	/*!
	  Adds the node to a mask. This packet won't be sent to this node, even
	  if everybodyId is present.

	  \param id The target node id.
	  \sa
	*/
	void addBlackMask(nodeId id);

	void addBlackMask(std::set<nodeId> &id_set);

	void setBlackMask(nodeId id);

	void setBlackMask(std::set<nodeId> &id_set);

	void clearBlackMask(void);

	std::set<nodeId> &getBlackMask(void);

	/*
	 * Unlike the more standard use, black lists have precedence over white lists here.
	 */
	void addWhiteMask(nodeId id);

	void addWhiteMask(std::set<nodeId> &id_set);

	void setWhiteMask(nodeId id);

	void setWhiteMask(std::set<nodeId> &id_set);

	void clearWhiteMask(void);

	std::set<nodeId> &getWhiteMask(void);


	/*!
	  Adds the target node id field.
	  \param id The target node id.
	  \sa getTargetNodes().
	*/
	void addTargetNodes(nodeId id);

	/*!
	  Sets the target node id field.
	  \param id_set The target node ids set.
	  \sa getTargetNodes().
	*/
	void addTargetNodes(std::set<nodeId> &id_set);

	/*!
	  Gets the target node list.
	  \return The target node id.
	  \sa setTargetNodes(), addTargetNodes().
	*/
	std::set<nodeId> &getTargetNodes(void);

	/*!
	  Clears the list of target nodes.
	  \sa setTargetNodes(), addTargetNodes().
	*/
	void clearTargetNodes(void);

	/*!
	  Gets the sender node id field.
	  \return The target node id.
	  \sa setSenderNodes().
	*/
	nodeId getSenderNode(void) const;

	/*!
	  Sets the sender node id field.
	  \param id The sender node id.
	  \sa setSenderNode().
	*/
	void setSenderNode(nodeId id);

	/*!
	  Gets the sender node id field.
	  \return The target node id.
	  \sa setSenderNodes().
	*/
	const nodeAddress getSenderNodeAddress(void) const;

	/*!
	  Sets the sender node id field.
	  \param id The sender node id.
	  \sa setSenderNode().
	*/
	void setSenderNodeAddress(const nodeAddress &addr);

	/*!
	  Sets if this packet must be sent reliably (i.e., guaranteeing
	  delivery) or not.

	  See the protocol documentation to see if this field is supported. Not
	  all protocols may support this flag.

	  \return true if is a reliable packet, false otherwise.
	*/
	bool isReliable(void) const;

	/*!
	  Sets if this packet must be sent reliably (i.e., guaranteeing
	  delivery) or not.

	  See the protocol documentation to see if this field is supported.

	  \param reliable True if it must, false otherwise.
	*/
	void setReliable(bool reliable);

	/*!
	  \param fd
	*/
	void setTargetAddresses(const nodeAddress &fd);

	/*!
	  \param fds
	*/
	void setTargetAddresses(std::set<nodeAddress> &fds);

	std::set<nodeAddress> &getTargetAdresses(void);

	/*!
	  Encodes the packet to a string that can be sent over the network.

	  The from field (which you can get with getSenderNode()) is
	  automatically set. You must set the target node, since the
	  default is #nobodyId.
	  \sa decode().
	  \param len A pointer to a variable.
	  \return The encoded data, and the total length in the len argument,
	    if not NULL.
	*/
	virtual chain encode(unsigned int *len = NULL);

	/*!
	  Decodes a string to a packet.

	  The from field (which you can get with getSenderNode()) is set with
	  the id of the sender.

	  The target field (getTargetNode()) is NOT set, and its value is
	  meaningless. This behaviour may change in future versions.

	  The caller should always pass a valid packet, but checks are made
	  to avoid buffer security attacks. If the packet does not pass these
	  tests, false is returned. Obviously, no checks are made on the data
	  field.

	  \param s The data to decode.
	  \retval true If correctly decoded. This doesn't mean that the values
	   are meaningful.
	  \retval false Otherwise.
	  \sa encode().
	*/
	bool decode(chain &s);

	/*!
	  \overload bool decode(chain &s)
	  \param buffer The buffer to decode
	  \param length The buffer length.
	*/
	bool decode(unsigned char *buffer, unsigned int length);

	/*!
	  \overload bool decode(chain &s)
	  \param header The packet header. User is responsible for freeing this buffer.
	  \param headerlen The header length.
	  \param databuf The data packet. This buffer is freed by libGlass.
	  \param datalen The data packet length.
	*/
	bool decode(unsigned char *header, unsigned int headerlen,
		unsigned char *databuf, unsigned int datalen);

	/*!
	  A nice outputter, for debugging purposes.
	*/
	virtual std::ostream& print(std::ostream &o);

	friend std::ostream& operator<<(std::ostream &o, Packet &p);
};

extern std::ostream& operator<<(std::ostream &o, Packet &p);

} //namespace

#endif
