/*  Glass - a distributed computing library
    Copyright (C) 2003-2009 Bruno Barberi Gnecco
    Copyright (C) 2009-2010 Corollarium Technologies


    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef _GLASS_PLUGIN_HH_
#define _GLASS_PLUGIN_HH_

#include "exception.h"
#include "types.h"
#include <vector>
#include <map>
#include <string>

namespace libglass {

class Glass;
class Packet;
class PluginBase;
class PluginManager;

/* TODO!!! Plugins registered with PluginManagerBase are not registered in
PluginManager instances already created. Although this situation should never
happen currently, since all plugins register themselves before a PluginManager
class can be instantiated, it precludes registration of new plugins during
code execution. To overcome this, create a list of all the PluginManager's
the factory creates, and when a new plugin is registered, register in all of
them. */

/*!
  All plugins should register with PluginManagerBase (this should be done 
  automatically, and is explained in the corresponding section). 

  This class is a singleton factory. Every Glass has its own PluginManager,
  which is built by this class.

  \sa Glass, PluginManager.
 */
class PluginManagerBase {
private:
	/*! 
	  List of plugins. 
	*/
	std::map<PluginId, PluginBase *> pluginmap;

	/*! 
	  Private constructor. This class is a pseudo-singleton. 
	*/
	PluginManagerBase();

	/*! 
	  Operator= overloading. This class is a pseudo-singleton.
	*/
	void operator=(PluginManagerBase &);

	/*! 
	  This class is a pseudo-singleton.
	*/
	PluginManagerBase(const PluginManagerBase &) { }

protected:
	~PluginManagerBase();

	friend class PluginBase;
	friend class Glass;

	/*! 
	  Creates a new, clean instance of PluginManager. This is a virtual
	  constructor.

	  \param g Our owner.
	  \return A brand new PluginManager, with brand new PluginBases.
	*/
	PluginManager *createPluginManager(Glass *g);

	/*! 
	  Registers a plugin.
	*/
	bool registerPlugin(PluginBase *base);
public:

	/*! 
	  Returns a handle to this class. This class is a pseudo-singleton.

	  \return A handle to this class.
	*/
	static PluginManagerBase& getHandle(void);
};

/*!
  PluginManager is the bridge between the Glass derived class (server, client,
  etc) and the plugins, which are responsible for all Glass functionality from
  the user's point of view. PluginManager guarantees that any message received
  by Glass is forwarded to the correct plugin.
  There's only one instance of this class for each Glass instance. This class
  can only be constructed by PluginManagerBase.
  \sa PluginManagerBase, Glass.
  */
class PluginManager {
private:
	/*! 
	  List of plugins. 
	*/
	std::map<PluginId, PluginBase *> *pluginmap;

	/*!
	  Glass class that owns us.
	*/
	Glass 				*glass;

	/*! 
	  Operator= overloading. This class is a pseudo-singleton.
	*/
	void operator=(PluginManager &);

	/*! 
	  Only this class can clone. Friend classes should call instance().
	  \sa instance().
	*/
	PluginManager(const PluginManager &);

protected:
	friend class PluginManagerBase;
	friend class PluginBase;
	friend class PluginInterface;
	friend class Glass;

	/*! 
	  Private constructor. This class is a pseudo-singleton. 
	*/
	PluginManager(Glass *g, std::map<PluginId, PluginBase *> *pluginmap);

	/*! 
	  Packet processing. 
	  
	  This function is called only by #Glass. 

	  \param p The packet to process.
	  \param rootOnly Only process root packets; other packets are ignored.
	  \retval true If packet was correctly processed.
	  \retval false If an error occurred while processing packet or if a
	  	non-root packet was ignored.
	  \sa sendPacket().
	*/
	bool processPacket(Packet &p, bool rootOnly = false);

	/*! 
	  Send a packet. This function is called by PluginBase.
	  \param p The packet to process.
	  \retval true If packet was correctly sent.
	  \retval false Otherwise.
	  \sa processPacket().
	*/
	bool sendPacket(Packet &p);

	/*! 
	  Clean up system.

  	  Guarantees that a node quit event is properly handled by the plugins,
	  avoiding potential deadlocks.

	  Do not call sendPacket() from this function to send any data to id,
	  since it's quite likely that the connection to the node is already 
	  dead. You can send to other nodes or to everybodyId, however.

	  \param id The nodeId of the node.
	*/
	void unregisterNode(nodeId id);

	/*!
	  Returns our owner.
	  \return The Glass class that owns us.
	*/
	Glass *getGlass(void) const;

public:
	/*!
	   Destructor.
	*/
	~PluginManager();

	/*!
	  Returns a plugin given its id.
	  \param id The plugin id.
	  \return A plugin given its id, or NULL if none found.
	*/
	PluginBase *getPlugin(PluginId id) const;

	void printPlugins(void);
};


/*! 
  Base class for all plugins.

  Let me explain in the philosophy of PluginBase and derived classes. This is a
  singleton class (per PluginManager instance, that is) that should be unknown
  to the user. This is a good thing (TM), because the user doesn't have to use
  static methods to get a handle and all that kind of stuff. This class has
  a virtual method processPacket() that is called for every packet that arrives
  for this plugin.
  
  To let the user interact with the plugin, you should provide classes derived
  from PluginInterface. These classes are not singletons, and they provide a
  method to talk to PluginBase. You can keep track of all these objects, calling
  the PluginBase class from the constructor, etc. 
  
  Since you're probably not following this lousy explanation, let's see an
  example. I'll use the Barrier plugin to illustrate this discussion. First,
  you have the PluginBase derived class, which is BarrierBase. This class is
  responsible for processing all barrier packets. 
  
  If the user wants to setup a barrier, he'll create a instance of another 
  class, Barrier (which is derived from PluginInterface). To him, BarrierBase
  doesn't exist. When a Barrier is instanced, it calls BarrierBase, which keeps
  track of all Barrier objects. This way BarrierBase can release barriers when
  the appropriate packets arrive, etc.

*/
class PluginBase { 
private:
	/*! 
	  Plugin id. Must be unique.
	*/
	PluginId id;

protected:
	friend class PluginManagerBase;
	friend class PluginManager;

	/*!
	  Our boss and owner.
	*/
	PluginManager			*pm_parent;

	/*! 
	  Callback from PluginManager. 

	  This prototype is automatically added by PLUGIN(cl).

	  \param p The packet to process.
	  \retval true If packet was correctly processed.
	  \retval false Otherwise.
	  \sa sendPacket().
	*/
	virtual bool processPacket(Packet &p) = 0;

	/**
	 * Synchronizes data with another node. This function is used
	 * on new connections and reconnections after loss of packets.
	 *
	 * Implementation is obviously up to the plugin, and may do
	 * nothing at all.
	 *
	 * @param id The remote node id
	 * @param r Our relation to it.
	 * @return
	 */
	virtual void synchronize(nodeId id, nodeRelation r) = 0;

	/*! 
	  Virtual constructor, used by PluginManagerBase::createPluginManager.

	  This function is automatically created by PLUGIN(cl).

	  \param parent Our parent PluginManager, to fill #pm_parent.
	  \return A brand new PluginBase.
	*/
	virtual PluginBase *instance(PluginManager *parent) = 0;

	/*! 
	  Clean up system.

  	  Guarantees that a node quit event is properly handled by the plugin,
	  avoiding potential deadlocks.

	  This prototype is automatically added by PLUGIN(cl).

	  \param id The nodeId of the node.
	*/
	virtual void unregisterNode(nodeId id) = 0;

	/*! 
	  Wrapper to send a packet. You can override if desired.
	  \param p The packet to send.
	  \retval true If packet was correctly sent.
	  \retval false Otherwise.
	  \sa processPacket().
	*/
	virtual bool sendPacket(Packet &p);

	/*! 
	  Get node type wrapper.
	  \return Our node type.
	*/
	nodeRelation getNodeRelation(nodeId id) const; 

public:
	/*! 
	  Constructor. 
	*/
	PluginBase(const PluginId id, bool *registered);

	/*!
	  Destructor.
	*/
	virtual ~PluginBase() = 0;

	/*! 
	  Returns this plugin's id.
	*/
	PluginId getPluginId(void) const;


	/*!
	  Get Glass wrapper.
	  \return Our Glass owner.
	 */
	Glass *getGlass(void) const;

};


/* And now, the plugins themselves */

/*! 
  This define does everything you need to create a plugin. Just run it with
  your plugin class name inside the plugin class. If you need special behaviour,
  you may want to manually copy the define and change as it suits you.
*/
#define PLUGIN(cl)					\
private:						\
	cl();						\
	void operator=(cl &);				\
	static bool registered; /* avoid duplicate id */\
	static cl singleton; /* for auto-registering */	\
protected:						\
	/* prototypes for virtual functions */		\
	bool processPacket(Packet &p);			\
	void unregisterNode(nodeId id);			\
	void synchronize(nodeId id, nodeRelation r); \
	/* the virtual constructor */			\
	PluginBase *instance(PluginManager *parent) {	\
		cl *inst = new cl();			\
		inst->pm_parent = parent;		\
		return inst;				\
	}						\
	friend class PluginManagerBase;			\
public:							\
	static cl & getHandle(void) {			\
		return singleton;			\
	}

/*! 
  This is the static declaration of your plugin AND the constructor. It will 
  ensure that the plugin registers itself automatically. Use it as you'd
  normally use the declaration of the constructor, except that if you use
  initialization lists, you should start with a comma ',' instead of a double 
  colon ':'. Here's an example:

  \code
  	PLUGIN_CONSTRUCTOR(MyClass), init1(), init2() {
		some code;
	}
  \endcode

  Consider a concrete example that uses the PLUGIN_CONSTRUCTOR macro:
  In alias.cpp we have...
  
  \code
   PLUGIN_CONSTRUCTOR(AliasBase) { // add your constructor code here }
  \endcode

  This gets expanded to (note additional comments added in expanded code)...

  \code
   // Initialisation of class static member
   // Note: This is performed during runtime initialisation before calling main()
   bool AliasBase::registered = false;

   // Initialisation of class static member object
   // Note: This is performed during runtime initialisation before calling main()
   // Will call default constructor for the class (see below), which calls
   // the PluginBase constructor during initialisation as...
   // PluginBase(const PluginId id, bool *registered)
   AliasBase AliasBase::singleton;

   // Define the default constructor
   AliasBase::AliasBase() : PluginBase("AliasBase", &registered) { // add your constructor code here }
  \endcode

  You can't have any parameters in the constructors. If you desperatedly need
  parameters, let us developers know, because I can't think of any reason.
  Remember that the user will *not* call this constructor, *ever*, so it's
  impossible to pass a parameter! If you wish to get/set parameters after the
  constructor returns, it's suggested to use an objected derived from 
  PluginInterface, so the user will have access to those functions.

  Note that the PLUGIN_CONSTRUCTOR macro uses the gnu c-preprocessor syntax for
  _stringification_ (or stringization as Microsoft call it) of the argument.
  That is to say #c1  gets expanded (once) to a constant string.
  For example,

  \code
    #define FOO(bar) format(#bar) 
    FOO(hello)
  \endcode

  would produce
  
  \code
    format("hello")
  \endcode

  \sa PluginBase.
*/
#define PLUGIN_CONSTRUCTOR(cl)				\
  bool cl::registered = false;				\
  cl cl::singleton;					\
  cl::cl() : PluginBase(#cl, &registered)

/*!
  This is an interface that should be used by all classes that are related to
  plugins. There is only one PluginBase per Glass, which should not be directly
  accessed by the user. This class provides useful methods to find the type
  of node (so you can set your behaviour accordingly) and to get your plugin
  base. Use the plugin base to do any kind of interaction with the %Glass 
  library (such as PluginBase::sendPacket()).
  
  Note that this interface is not bound to only one PluginBase; so you can use
  it to access more than one, if you can find some reason to do this.
*/
class PluginInterface {
private:
	Glass				*g;
	propagationMode		mode;

protected:
	/*!
	  Constructor.
	  @param mode The propagation mode for this plugin instance.
	*/
	PluginInterface(propagationMode mode = Local);

	/*!
	  Destructor.
	*/
	virtual ~PluginInterface() = 0;

	/*!
	  Gets a plugin base given its id.
	  \param id The plugin id.
	  \return A pointer to the PluginBase.
	*/
	PluginBase *getPluginBase(PluginId id);

	/*!
	  Gets our node type. Useful if you have different behaviours.
	  \return The node type.
	*/
	propagationMode getMode(void) const;

	/*!
	  Gets the current Glass. This is useful for static methods.
	  \return The current PluginManager.
	 */
	static const PluginManager *getPluginManager(void);

	const Glass *getGlass(void) const;
};

} //namespace

#endif
