/*  Glass - a distributed computing library
    Copyright (C) 2003-2009 Bruno Barberi Gnecco
    Copyright (C) 2009-2010 Corollarium Technologies


    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef _GLASS_PROTOCOL_HH_
#define _GLASS_PROTOCOL_HH_

#include "types.h"

namespace libglass {

class Packet;
class Glass;

#define PROTOCOL_INVALID_PACKET 3
#define PROTOCOL_TIMEOUT 2
#define PROTOCOL_CONNECTION_CLOSED 1
#define PROTOCOL_OK 0
#define PROTOCOL_NOT_INITIALIZED -1
#define PROTOCOL_ALREADY_INITIALIZED -1
#define PROTOCOL_UNREGISTERED_NODE -2
#define PROTOCOL_NETWORK_ERROR -2
#define PROTOCOL_NOT_SERVER -1


/*!
  The Protocol abstract base class. All protocols should be derived from
  this class. 
 */
class Protocol {
protected:
	friend class Glass;

	Glass			*owner;		/**< our owner. */

	/*!
	  Sets our owner. 
	  
	  Called by Glass(). The reason this is not passed as argument to the
	  constructor is that the Protocol is not constructed by the Glass
	  class.
	  
	  \param g The owner.
	*/
	void setOwner(const Glass *g);

	/*!
	  Converts a nodeId to a nodeAddress

	  Wrapper to the real call, which is a member of Glass().

	  \param id The node id.
	  \return The associated nodeAddress
	  \sa Glass::IdToAddress().
	*/
	nodeAddress IdToAddress(nodeId id) const;

	nodeAddress FDToAddress(nodeFD fd) const;

	/*!
	  Converts a nodeAddress to a nodeId.

	  Wrapper to the real call, which is a member of Glass().

	  \param addr The node address.
	  \return The associated nodeId.
	*/
	nodeId AddressToId(const nodeAddress &addr) const;

#ifdef deprecated
	/*!
	  Registers a node, by getting its descriptor. 
	  
	  Wrapper to the real call, which is a member of Glass().

	  \param fd A valid node descriptor.
	  \return A new node Id, or nobodyId if an error occurred.
	  \sa Glass::unregisterNode().
	*/
	nodeId registerNode(nodeAddress &addr, nodeId id);
#endif

	/*!
	  Unregisters a id/fd association.
	  
	  Wrapper to the real call, which is a member of Glass().

	  \param id The node id.
	  \retval true
	*/
	bool unregisterNode(nodeId id);

#ifdef TODO
	virtual bool registerGroup(XXXX groupname, std::set<nodeAddress> &list) = 0;
	virtual bool addToGroup(groupname, FD) = 0;
	virtual bool removeFromGroup(groupname, FD) = 0;
	virtual bool deleteGroup();
#endif


	/*!
	  This is the derived class implementation of listen().
	  \sa int listen(unsigned int listenport)
	*/
	virtual int _listen(unsigned int port) = 0;

	/*!
	  This is the derived class implementation of accept().
	  \sa nodeAddress accept(void)
	*/
	virtual nodeAddress _accept(void) = 0;
	
	/*!
	  This is the derived class implementation of connect().
	  \sa nodeAddress connect(const char *host, unsigned int listenport);
	*/
	virtual nodeAddress _connect(const char *host, unsigned int port) = 0;
	virtual nodeAddress _connect(const nodeAddress &address) = 0;

	/*!
	  This is the derived class implementation of disconnect().
	  \sa void disconnect(nodeAddress node)
	*/
	virtual void _disconnect(const nodeAddress &node) = 0;

	/*!
	  This is the derived class implementation of send().
	  \sa bool _send(Packet &p)
	*/
	virtual bool _send(Packet &p) = 0;

	/*!
	  This is the derived class implementation of recv().
	  \sa int recv(Packet &p)
	*/
	virtual int _recv(Packet &p) = 0;

public:

	/*!
	  Constructor.
	*/
	Protocol();

	/*!
	  Virtual destructor.
	  Responsible for closing all connections.
	*/
	virtual ~Protocol();

	/*!
	  Opens a listenport for listening to connections.

	  Must not block. If you need to block, do it in accept().

	  \param listenport The listenport to listen to.
	  \sa accept(), connect().
	  \retval PROTOCOL_OK Ok.
	  \retval PROTOCOL_ALREADY_INITIALIZED Already initialized.
	  \retval PROTOCOL_NETWORK_ERROR Network error.
	*/
	int listen(unsigned int port = 0);

	/*!
	  Waits for a connection.

	  When the connection is successful, you must register the new FD by
	  calling registerNode().

	  \sa listen(), connect().
	  \retval PROTOCOL_NOT_SERVER if protocol type is not server.
	  \retval PROTOCOL_NETWORK_ERROR if could not open connection.
	  \retval >=0 otherwise, a glass node file descriptor to the node 
	    connected.
	*/
	nodeAddress accept(void);
	
	/*!
	  Connects to a given host and listenport.

	  \param host The host to contact.
	  \param listenport The listenport to connect.
	  \sa listen(), accept().
	  \retval -1 Already initialized or NULL argument.
	  \retval -2 Network error.
	  \retval >=0 A glass node file descriptor to the node.
	*/
	nodeAddress connect(const char *host, unsigned int port = 0);

	nodeAddress connect(const nodeAddress &address);

	/*!
	  Closes the connection to a given node. Derived classes may choose 
	    exactly what to do.
 
	  \param node The glass node descriptor 
	*/
	void disconnect(const nodeAddress &node);

	/*!
	  Sends a message.

	  This function must be reentrant, that is, it may be called 
	  simultaneously by two or more threads. It's up to the implementation
	  to make it truly concurrent (the preferred alternative) or use some 
	  kind of locking system, such as a mutex.

	  \param p The packet to send.
	  \sa recv().
	  \retval true Ok.
	  \retval false Protocol not initialized or transmission error.
	*/
	bool send(Packet &p);

	/*!
	  Receives a message.
	  
	  This function gets the first packet in queue, and fills the Packet
	  structure with the data. 
	  
	  \param p The packet to store the message.
	  \sa send().
	  \retval 3 Invalid packet format.
	  \retval 2 Timeout (may not be implemented by the derived class). This
	    is \e not an error, and ought to be handled properly by the caller.
	  \retval 1 Connection closed.
	  \return 0 OK.
	  \retval -1 Protocol not initialized, or error reading.
	  \retval -2 Unregistered node.
	*/
	int recv(Packet &p);

	/**
	 * Returns the port this protocols listens to.
	 *
	 * @return
	 */
	virtual unsigned int getListenPort(void) {
		return 0;
	}

	/**
	 * Gets the local address.
	 *
	 * @return
	 */
	virtual nodeAddress getLocalAddress(void) {
		return nobodyAddress;
	}

	/**
	 * Gets the local hostname.
	 *
	 * @return
	 */
	virtual string getLocalHostname(void) {
		return "";
	}

#ifdef FUTURE
	/**
	 *
	 * @return
	 */
	virtual nodeId getUniqueId(void) = 0;
#endif

};

}// namespace

#endif
