/*  Glass - a distributed computing library
    Copyright (C) 2003-2009 Bruno Barberi Gnecco
    Copyright (C) 2009-2010 Corollarium Technologies


    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef _GLASS_PLUGIN_REMOTE_VAR_HH_
#define _GLASS_PLUGIN_REMOTE_VAR_HH_

#include "exception.h"
#include "packager.h"
#include "types.h"
#include "plugin.h"
#include <boost/thread.hpp>
#include <map>
#include <set>

namespace libglass {

class _Remote;

/*!
  Remote variables plugin.

  Provides asynchronous remote variables among the nodes.

  \sa _Remote, Remote.
*/
class RemoteBase : public PluginBase {
	PLUGIN(RemoteBase);
private:
	std::map<const string, _Remote *> vars; /**< the list of variables. */

protected:
	friend class _Remote;

	/*!
	  Registers a remote variable in the list.
	  \param v A pointer to the variable.
	  \retval true If successful.
	  \retval false Otherwise.
	*/
	bool registerRemote(_Remote *v);

	/*!
	  Unregisters a remote variable, taking it off the list.
	  \param v A pointer to the variable.
	  \retval true If successful.
	  \retval false Otherwise.
	*/
	bool unregisterRemote(_Remote *v);

	// *********************************************************************
	// Destructor should be PROTECTED to ensure singleton behaviour
	// However, Microsoft compilers prior to VC v7.0 .NET are buggy and
	// generate the following error
	//   error C2248: 'AliasBase::~AliasBase' : cannot access protected member declared in class 'libglass::AliasBase'
	// The only workaround is to make the destructor public - which kind
	// of destroys the whole reason for implementing the singleton class
#if defined(_MSC_VER) && (_MSC_VER < 1300)  // earlier than .NET compiler (VC 7.0)
public:
#else
protected:
#endif
	// *********************************************************************
	/*!
	  Destructor.
	*/
	virtual ~RemoteBase();


public:

	/**
	 * Returns the instance of _Remote with a given name or NULL if it does
	 * not exist.
	 * @param name
	 * @return
	 */
	_Remote *getByName(const string &name) const;
};


/*!
   Abstract base class for remote variables.

   \todo Prefetch new values, presend updates.   
*/
class LIBGLASS_API _Remote : public PluginInterface {
protected:
	friend class RemoteBase;
	friend class RemotePacket;

	RemoteBase		*rb;		/**< our owner plugin. */
	string			name;		/**< the variable name. */

	/* client: */
	boost::mutex		mutex;		/**< Mutex for atomic operations */

	/* server: */
	std::set<nodeId>	owners;		/**< owner list */

	/*!
	  Constructor.
	  \param name The variable name.
	  \throws Exception.
	*/
	_Remote(const string &name) throw(Exception);

	/*!
	  Destructor.
	*/
	virtual ~_Remote();

	/*!
	  Sets the value of this variable in a remote node.

	  \param c The packed value.
	  \param id The remote node.
	  \retval true If successful.
	  \retval false Otherwise.
	*/
	bool _setValue(const chain &c, nodeId id);

	/*!
	  Gets the packed value of this variable in a remote node.

	  \param c The packed data.
	  \param id The remote node.
	  \retval true If successful.
	  \retval false Otherwise.
	*/
	bool _getValue(chain &c, nodeId id);


	/*!
	  Pack function. Serializes the variable to a string.
	  \return The variable, serialized to a string.
	*/
	virtual chain pack(void) { return chain(); }

	/*!
	  Unpack function. Unserializes the string to a variable.
	  \param s The string.
	  \param used The number of bytes used by the unserialization.
	  \return
	*/
	virtual bool unpack(chain &s, unsigned int &used) { return false; }

public:
	
	/*!
	  Updates the remote variable with the current local value.

	  \param id The remote node.
	  \retval true If successful.
	  \retval false Otherwise.
	*/
	bool copyDataToRemote(nodeId id);

	/*!
	  Updates this variable with its value in another node.
	  \param id The remote node.
	  \retval true If successful.
	  \retval false Otherwise.
	*/
	bool copyDataFromRemote(nodeId id);
};

/*!
  Remote variables base template.
  \sa _Remote.
*/
template<typename T,
	GlassSerializer(*packf, T) = rawPack<T>,
	GlassUnserializer(*unpackf, T) = rawUnpack<T> >
class Remote : public _Remote {
protected:
	T			data;
	unsigned int len; // length argument for pack

	virtual chain pack(void) {
		chain c;
		packf(c, data, len, NULL); // Provide default values for args 3 and 4 as required by MSVC
		return c;
	}

	virtual bool unpack(chain &c, unsigned int &used) {
		unsigned int 	initial = 0;
		bool		b;

		mutex.lock();
		try {
			b = unpackf(c, initial, data, used, 1, NULL); // Provide default values for args 5 and 6 as required by MSVC
			this->len = used - initial;
		}
		catch (...) {
			mutex.unlock();
			throw;
		}
		mutex.unlock();
		return b;
	}

public:

	/**
	 * Factory works better than explicit constructors, since it returns a valid
	 * variable if it has been created before, which the constructor will throw an
	 * exception saying that the variable already exists.
	 *
	 * @param name The variable name.
	 * @param data The variable initial data.
	 * @param len If it is an array, its length. Otherwise 1.
	 * @param mode Controls the propagation of this variable.
	 * @return A pointer to the existing variable if it exists,
	 */
	static Remote *factory(const char *name, T data = T(), unsigned int len = 1) {
		const PluginManager *pm = PluginInterface::getPluginManager();
		if (!pm) {
			return NULL;
		}
		RemoteBase *sb(dynamic_cast<RemoteBase *>(pm->getPlugin("RemoteBase")));
		_Remote *s = sb->getByName(name);
		if (s) {
			return dynamic_cast<Remote *>(s);
		}
		return new Remote(name, data, len);
	}

	static Remote *factory(const string &name, T data = T(), unsigned int len = 1, propagationMode mode = Local) {
		return Remote::factory(name.c_str(), data, len, mode);
	}

	/*!
	  Constructor.
	  \param name The variable name.
	  \param data The variable data.
	*/
	Remote(const char *name, T data = T(), unsigned int len = 1) throw(Exception) :
		_Remote(name), data(data), len(len) {
	
	}

	Remote(const std::string name, T data = T(), unsigned int len = 1) throw(Exception) :
		_Remote(name), data(data), len(len) {

	}


	/*!
	  Assignment operator.
	  \param b The value to assign to this.
	  \return A reference to this variable value.
	*/
/*	const T &operator=(const T &b) { } */

	const Remote &operator=(const T &b) {
		data = b;
		return *this;
	}

	/*!
	  Assignment operator.
	  \param b The Shared to be assign to this.
	  \return A reference to this instance.
	*/
	const Remote &operator=(const Remote &b) {
		data = b.data;
		return *this;
	}

	const unsigned int &getLen(void) const {
		return len;
	}

	/*!
	  Assignment operator.
	  \param b The value to assign.
	  \param id The node to assign to.
	  \return
	*/
	bool setValue(nodeId id, const T &b, unsigned int len = 1, bool free = false) {
		if (id == nobodyId || id == this->rb->getGlass()->getNodeId()) {
			data = b;
			return true;
		}
		else {
			chain c;
			packf(c, b, len, NULL);
			return this->_setValue(c, id);
		}
	}

	/*!
	  Returns the variable value.
	  \param id The id to 
	  \return The variable value.
	  \warning If you use this function to get a reference to the value,
	  	take care because the value can change asynchronously.
	*/
	const T &getValue(nodeId id, unsigned int &len) throw(Exception) {
		if (id == nobodyId || id == this->rb->getGlass()->getNodeId()) {
			return data;
		}
		else {
			chain		c;
			unsigned int	used;
			T item;

			if (!_getValue(c, id)) {
				throw Exception("Could not get remote value.");
			}
			if (!unpackf(c, 0, item, used, 1, NULL)) {
				throw Exception("Could not unpack remote value.");
			}
			len = used;
			return item;
		}
	}

	const T &getValue(nodeId id = nobodyId) throw(Exception) {
		unsigned int len;
		return this->getValue(id, len);
	}

	/*!
	  Value operator.
	  \return A reference to this instance.
	*/
	T &operator*(void) const {
		return data;
	}

	const T &getData(void) const {
		return data;
	}
};


/**
 * Specialization for arrays
 */
template<typename T,
	GlassSerializer(*packf, T),
	GlassUnserializer(*unpackf, T) >
class Remote<const T *, packf, unpackf> {
protected:
	bool		free;

	virtual bool unpack(chain &c, unsigned int &used) {
		unsigned int 	initial = 0;
		bool ret = unpackf(c, initial, this->data, used, 1, (void *)this->free);
		this->len = used - initial;
#ifdef REALLY_HARD_DEBUG
		int i = 0;
		std::cout << "Unpacking " << used << ": ";
		for (T *p(this->data); i < used; p++, i++) {
			std::cout << std::hex << (int)*p << "-";
		}
		std::cout << std::dec << std::endl;
#endif
		return ret;
	}

public:
	Remote(const char *name, T *data = NULL, unsigned int len = 1)
		throw(Exception) : Remote<T *, packf, unpackf>(name, data, len) {
		std::cerr << "pointer spec " << std::endl;
	}

	const Remote &operator=(const T &b) {
		throw Exception("Remote for array does not support assignment by operator");
	}

	const Remote &operator=(const Remote &b) {
		throw Exception("Remote for array does not support assignment by operator");
	}



	/*!
	  Assignment operator.
	  \param b The value to assign.
	  \param id The node to assign to.
	  \return
	*/
	bool setValue(nodeId id, const T *&b, unsigned int len = 1, bool copy = false, bool free = false) {
		if (id == nobodyId || id == this->rb->getGlass()->getNodeId()) {
			T *newdata;

			// release old data
			if (this->free == true && this->data) {
				delete []this->data;
			}

			// check if we need a deep copy
			if (copy == true) {
				this->free = true;
				newdata = new T[len];
				memcpy(newdata, b, len*sizeof(T));
			}
			else {
				this->free = free;
				newdata = b;
			}

			this->data = newdata;
			return true;
		}
		else {
			chain c;
			packf(c, b);
			return this->_setValue(c, id);
		}
	}
};


}//namespace



#endif // _GLASS_PLUGIN_REMOTE_VAR_HH_
