/*  Glass - a distributed computing library
    Copyright (C) 2003-2009 Bruno Barberi Gnecco
    Copyright (C) 2009-2010 Corollarium Technologies


    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef _GLASS_SERIAL_HH_
#define _GLASS_SERIAL_HH_

#include "types.h"
#include "chain.h"
#include "exception.h"

namespace libglass {

/*!
  Object serialization abstract base class.
  
  Glass has to serialize objects to send them to other nodes. If you derive
  your class SerializableObject, plugins that use serialization will work
  easily with your class, and you'll have better control of the process.

  Your class must also declare a static RegisterObject. Here's how your
  class should look like:
  
  \code

class YourClass : public SerializableObject {
private:
	// your data goes here
	static RegisterObject ro;

public:
	// your constructor and public methods.
	
	const string *name(void) const {
		static const string name = string("YourClass");
		return &name;
	}

	SerializableObject *create() const throw(Exception) {
		return new YourClass();
	}

	SerializableObject *create(const chain &c) const throw(Exception) {
		YourClass *t = new YourClass();
		if (t->unpack(c) == false) {
			throw Exception("Error unserializing");
		}
		return t;
	}

	bool pack(chain &c) {
		// your serialization method
		return true;
	}

	bool unpack(const chain &c) {
		// your serialization method
		return true;
	}
};

RegisterObject YourClass::ro = RegisterObject(new YourClass());

  \endcode


  \sa RegisterObject.
*/
class SerializableObject {
public:
	SerializableObject() { }

	/*!
	  Virtual destructor.
	*/
	virtual ~SerializableObject() { }
	
	/*!
	  A portable RTTI.

	  \return A unique name; usually the name of your class.
	*/
	virtual const string name(void) const = 0;

	/*!
	  Virtual constructor.

	  \return A new object of your class.
	*/
	virtual SerializableObject *create() const throw(Exception) = 0;

	/*!
	  \overload
	  \param s A serialized object
	*/
	virtual SerializableObject *create(const chain &s) const 
			throw(Exception) = 0;

	/*!
	  Serialization function.

	  \return The object, serialized to a string.
	*/
	virtual bool pack(chain &c) = 0;

	/*!
	  Unserialization function.

	  \param c The serialized object.
	  \retval true if unserialization was successful.
	  \retval false otherwise.
	*/
	virtual bool unpack(const chain &c) = 0;
};

/*!
  This is a simple class that registers classes that derive from 
  SerializableObject during static initialization. You must declare in your
  derived class a
  
  \code static RegisterObject ro; \endcode
  
  and initialize it with

  \code RegisterObject YourClass::ro = RegisterObject(new YourClass()); \endcode
  
  \sa SerializableObject.
*/
class RegisterObject {
public:
	RegisterObject(SerializableObject *o);
};


}

#endif // _GLASS_SERIAL_HH_
