/*  Glass - a distributed computing library
    Copyright (C) 2003-2009 Bruno Barberi Gnecco
    Copyright (C) 2009-2010 Corollarium Technologies


    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef _GLASS_PLUGIN_SHARED_HH_
#define _GLASS_PLUGIN_SHARED_HH_

#include "exception.h"
#include "packager.h"
#include "types.h"
#include "plugin.h"
#include "memorymanager.h"
#include <boost/thread.hpp>
#include <map>
#include <set>

namespace libglass {

class _Shared;

/*!
  Shared variables plugin.

  Provides synchronous shared variables among the nodes.

  Format of Shared Plugin packets:

  \li Client->server: [action (1 byte)][variable name (ended by \\0)][data]
  \li Server->client: [action (1 byte)][variable name (ended by \\0)][data]

  where valid action commands are described below.

  These are valid actions in client->server packets:

  \li 'C' = created a variable.
  \li 'U' = updated a variable.
  \li 'D' = deleted a variable.
  \li 'S' = synchronized a variable.

  These are valid actions in server->client packets:

  \li 'C' = changed variable.

  \sa _Shared, Shared.
*/
class SharedBase : public PluginBase {
	PLUGIN(SharedBase);
private:
	std::map<const string, _Shared *> vars; /**< the list of variables. */
protected:
	friend class _Shared;

	/*!
	  Registers a shared variable in the list.
	  \param v A pointer to the variable.
	  \retval true If successful.
	  \retval false Otherwise.
	*/
	bool registerShared(_Shared *v);

	/*!
	  Unregisters a shared variable, taking it off the list.
	  \param v A pointer to the variable.
	  \retval true If successful.
	  \retval false Otherwise.
	*/
	bool unregisterShared(_Shared *v);

	/*!
	  Updates the shared variable with the current local value.
	  \param s The shared variable.
	  \retval true If successful.
	  \retval false Otherwise.
	*/
	bool sendUpdate(_Shared *s);

	/*!
	  Synchronizes the value of a given variable with the most current
	  value.
	  \param s The shared variable.
	  \retval true If successful.
	  \retval false Otherwise.
	*/
	bool getUpdate(_Shared *s);

	/*!
	  Synchronizes the value of all variables.

	  \retval true If successful.
	  \retval false Otherwise.
	*/
	bool sendUpdateAll(void);


// *********************************************************************
// Destructor should be PROTECTED to ensure singleton behaviour
// However, Microsoft compilers prior to VC v7.0 .NET are buggy and
// generate the following error
//   error C2248: 'AliasBase::~AliasBase' : cannot access protected member declared in class 'libglass::AliasBase'
// The only workaround is to make the destructor public - which kind
// of destroys the whole reason for implementing the singleton class
#if defined(_MSC_VER) && (_MSC_VER < 1300)  // earlier than .NET compiler (VC 7.0)
public:
#else
protected:
#endif
// *********************************************************************

	/*!
	  Destructor.
	*/
	virtual ~SharedBase();


public:

	/**
	 * Returns the instance of _Shared with a given name or NULL if it does
	 * not exist.
	 * @param name
	 * @return
	 */
	_Shared *getByName(const string &name) const;
};

/*!
   Abstract base class for shared variables.

   \todo Prefetch new values, presend updates.   
*/
class LIBGLASS_API _Shared : public PluginInterface {
public:
	typedef enum { NEVER, NEXT_ONLY, CHANGED, FORCED } UpdateMode;
		
protected:
	friend class SharedBase;
	friend class SharedPacket;

	SharedBase		*sb;		/**< our owner plugin. */
	string			name;		/**< the variable name. */
	
	/* Ideally these should be separated in unions, but union members cannot
	   have constructors. */
	/* client: */
	bool			changed;	/**< have we changed since the last update? */
	UpdateMode		updatemode;	/**< our update mode. Default is always. */
	boost::condition_variable	trigger;	/**< to guarantee that an update arrived */

	/* server: */
	nodeId			lastchanged;	/**< who was the last to change this variable? */
	std::set<nodeId>	owners;		/**< owner list */
	chain			currentdata;	/**< serialized data */
	chain			nextdata;	/**< value for next update */

	/*!
	  Constructor.
	  \param name The variable name.
	  \throws Exception.
	*/
	explicit _Shared(const string &name, propagationMode mode = Local) throw(Exception);

	/*!
	  Destructor.
	*/
	virtual ~_Shared();

	/*!
	  Pack function. Serializes the variable to a string.
	  \return The variable, serialized to a string.
	*/
	virtual chain pack(void) { return chain(); }

	/*!
	  Unpack function. Unserializes the string to a variable.
	  \param s The string.
	  \param used The number of bytes used by the unserialization.
	  \return
	*/
	virtual bool unpack(chain &s, unsigned int &used) { return false; }
public:

	void setUpdateMode(UpdateMode updatemode) {
		this->updatemode = updatemode;
	}

	UpdateMode getUpdateMode(void) const {
		return this->updatemode;
	}
	/*!
	  Has the value changed since last update? 
	  \retval True if changed.
	  \retval False otherwise.
	*/
	bool hasChanged(void) const {
		return changed;
	}

	bool sendUpdate(void);

	/*!
	  Synchronizes this variable.

	  \retval True If successful.
	  \retval false Otherwise.
	*/
	bool getUpdate(void);

	/*!
	  Synchronizes all variables.

	  All updates are sent as one, saving bandwidth and cutting the delay.

	  \warning \b IMPORTANT! Since this method is static, it guesses that
	  the variable you want to synchronize are those of the \e current
	  Glass class. You only have to care about this if you instantiate 
	  more than one Glass derived class in the same application, which
	  you are almost certainly not doing. If you don't know for sure,
	  you are not doing it; go worry about something else.

	  \sa Glass.
	  \retval True If successful.
	  \retval false Otherwise.
	*/
	static bool sendUpdateAll(void);
};

/*!
  Shared variables base template.
  \sa _Shared.
  \param T The type to share.
  \param packf The serializer function.
  \param unpackf The unserializer function.
  \param Memory
*/
template<typename T,
	GlassSerializer(*packf, T) = toString<T>,
	GlassUnserializer(*unpackf, T) = toString<T>,
	class Memory = DefaultMemoryManager>
class Shared : public _Shared {
protected:
	/// Tests if Memory inherits fromMemoryManager
	typename boost::enable_if<boost::is_base_of<MemoryManager, Memory> >::type testMemory() {}

	boost::shared_ptr<MemoryManager> mm; /// memory manager
	T *			data; /// data
	unsigned int len; /// length argument for pack

	virtual chain pack(void) {
		chain c;
		packf(c, *data, len, NULL); // Provide default values for args 3 and 4 since required by MSVC
#ifdef REALLY_HARD_DEBUG
		std::cerr << "Packing " << len << ": " ;
		int i = 0;
		for (T p(this->data); i < len; p++, i++) {
			std::cerr << std::hex << *p << "-";
		}
		std::cerr << std::endl;
#endif
		return c;
	}

	virtual bool unpack(chain &c, unsigned int &used) {
		unsigned int 	initial = 0;
		bool ret = unpackf(c, initial, *data, used, 1, NULL);
		this->len = used - initial;
		return ret;
	}

public:

	/**
	 * Factory works better than explicit constructors, since it returns a valid
	 * variable if it has been created before, which the constructor will throw an
	 * exception saying that the variable already exists.
	 *
	 * @param name The variable name.
	 * @param data The variable initial data.
	 * @param len If it is an array, its length. Otherwise 1.
	 * @param mode Controls the propagation of this variable.
	 * @return A pointer to the existing variable if it exists,
	 */
	static Shared *factory(const char *name, const T &data = T(), unsigned int len = 1, propagationMode mode = Local) {
		const PluginManager *pm = PluginInterface::getPluginManager();
		if (!pm) {
			return NULL;
		}
		SharedBase *sb(dynamic_cast<SharedBase *>(pm->getPlugin("SharedBase")));
		_Shared *s = sb->getByName(name);
		if (s) {
			return dynamic_cast<Shared *>(s);
		}
		return new Shared(name, data, len, mode);
	}

	static Shared *factory(const string &name, const T &data = T(), unsigned int len = 1, propagationMode mode = Local) {
		return Shared::factory(name.c_str(), data, len, mode);
	}

	/*!
	  Constructor.
	  \param name The variable name.
	  \param data The variable data.
	*/
	explicit Shared(const char *name, const T &data = T(), unsigned int len = 1, propagationMode mode = Local)
			throw(Exception) : _Shared(name, mode), mm(new Memory()), len(len) {
		this->data = new(*mm, name) T(data);
	}

	explicit Shared(const std::string name, T data = T(), unsigned int len = 1, propagationMode mode = Local)
			throw(Exception) : _Shared(name, mode), mm(new Memory()), len(len) {
		this->data = new(*mm, name) T(data);
	}

	virtual ~Shared() {
		this->data->~T(); // placement new requires explicit destructor call
		mm->deallocate(this->data);
	}

	/*!
	  Assignment operator.
	  \param b The value to assign to this.
	  \return A reference to this variable value.
	*/
/*	const T &operator=(const T &b) {
		data = b;
		changed = true;
		return this->data;
	} */

	const Shared &operator=(const T &b) {
		*data = b;
		changed = true;
		return *this;
	}

	/*!
	  Assignment operator.
	  \param b The Shared to be assign to this.
	  \return A reference to this instance.
	*/
	const Shared &operator=(const Shared &b) {
		*data = *b.data;
		changed = true;
		return *this;
	}

	/*!
	  Assignment operator.
	  \param b The value to assign.
	  \param len The data length. Usually 1.
	*/
	virtual void assign(const T &b, unsigned int len = 1, bool copy = false, bool free = false) {
		*data = b;
		changed = true;
		this->len = len;
		return;
	}

	/*!
	  Returns the current variable value.

	  \return The variable value.
	*/
	const T &getData(void) const {
		return *data;
	}
	
	const unsigned int &getLen(void) const {
		return len;
	}

	/*!
	  Value operator.
	  \return The value of this variable.
	*/
	T operator*(void) const {
		return *data;
	}
};

/**
 * Specialization for arrays
 */
template<typename T,
	GlassSerializer(*packf, T),
	GlassUnserializer(*unpackf, T),
	typename Memory>
class Shared<const T *, packf, unpackf, Memory> {
protected:
	bool		free;

	virtual bool unpack(chain &c, unsigned int &used) {
		unsigned int 	initial = 0;
		bool ret = unpackf(c, initial, this->data, used, 1, (void *)this->free);
		this->len = used - initial;
#ifdef REALLY_HARD_DEBUG
		int i = 0;
		std::cout << "Unpacking " << used << ": ";
		for (T *p(this->data); i < used; p++, i++) {
			std::cout << std::hex << (int)*p << "-";
		}
		std::cout << std::dec << std::endl;
#endif
		return ret;
	}

public:
	Shared(const char *name, const T *&data = NULL, unsigned int len = 1, propagationMode mode = Local)
			throw(Exception) : Shared<T *, packf, unpackf, Memory>(name, data, len, mode) {
		// std::cerr << "pointer spec " << std::endl;
	}

	const Shared &operator=(const T &b) {
		throw Exception("Shared for array does not support assignment by operator");
	}

	const Shared &operator=(const Shared &b) {
		throw Exception("Shared for array does not support assignment by operator");
	}

	/*!
	  Assignment operator.
	  \param b The value to assign.
	*/
	void assign(const T *&b, unsigned int len, bool copy = false, bool free = false) {
		T *newdata;

		if (this->free == true && this->data) {
			delete []this->data;
		}

		if (copy == true) {
			this->free = true;
			newdata = new T[len];
			memcpy(newdata, b, len*sizeof(T));
		}
		else {
			this->free = free;
			newdata = b;
		}

		return Shared<T *, packf, unpackf>::assign(b, len);
	}

};

}//namespace


#endif // _GLASS_PLUGIN_SHARED_HH_
