/*  Glass - a distributed computing library
    Copyright (C) 2003-2009 Bruno Barberi Gnecco
    Copyright (C) 2009-2010 Corollarium Technologies


    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef _GLASS_TYPES_HH_
#define _GLASS_TYPES_HH_

#include <iostream>
#include <string>
#if !defined(_MSC_VER) 
 #include <stdint.h>
#else
 typedef unsigned int uint32_t;
#endif
#include <limits.h>
#include "msvc.h"

namespace libglass {

class chain;

/*!
  The possible relationships of two nodes.
  */
typedef enum { None = 0, /// no direct relation, but we know it exists
	Server, /// the other node is a server (higher in hierarchy)
	Client, /// the other node is a client (lower in hierarchy)
	Peer,  /// the other client is a peer (same hierarchy)
	Unregistered /// we never heard of that node.
} nodeRelation;


/*!
  The possible
  */
typedef enum { Local = 0, Global } propagationMode;

/*!
  Our string type.
  */
typedef std::string string;

}

#include "address.h"

namespace libglass {

/*!
  PluginId type.

  PluginIds are unique ids that identify each plugin. They are used, among
  other things to address packets.
  */
//typedef chain PluginId;
#define PluginId std::string

// #endif // _MSC_VER

/*!
  Node Identificator.

  These are unique ids; each node has one and only one nodeId, that is used
  to address packets internally.
*/
typedef uint32_t nodeId;

#define NODE_FORMAT std::hex << "0x" <<

/*!
  Id for "nobody".
*/
const nodeId nobodyId = 0;

/*!
  Id for "everybody" (multicast).
*/
const nodeId everybodyId = 1;

/*!
  Id for servers.
*/
const nodeId serversId = 2;

/*!
  Id for clients.
*/
const nodeId clientsId = 3;

/*!
  Id for peers.
*/
const nodeId peersId = 4;

const nodeId baseServerId = 0x10000;

/*!
  Maximum id available.
*/
const nodeId maxId = UINT_MAX;

/*!
  Converts a 4 byte string to a 32 bit number.
*/
inline unsigned int charToNum(unsigned char buf[4]) {
	return 	(((unsigned int)buf[0]) << 24) +
		(((unsigned int)buf[1]) << 16) +
		(((unsigned int)buf[2]) << 8) +
		(unsigned int)buf[3];
}

/*!
  Converts a 32 bit number to a 4 byte string.
*/
inline void numToChar(unsigned int num, unsigned char buf[4]) {
	buf[0] = (unsigned char)((num >> 24) & 0xFF);
	buf[1] = (unsigned char)((num >> 16) & 0xFF);
	buf[2] = (unsigned char)((num >> 8) & 0xFF);
	buf[3] = (unsigned char)(num & 0xFF);
}

inline void numToChar(unsigned int num, char buf[4]) {
	buf[0] = (unsigned char)((num >> 24) & 0xFF);
	buf[1] = (unsigned char)((num >> 16) & 0xFF);
	buf[2] = (unsigned char)((num >> 8) & 0xFF);
	buf[3] = (unsigned char)(num & 0xFF);
}

/*!
  Converts a 16 bit number to a 4 byte string.
*/
inline void shortToChar(unsigned short num, unsigned char buf[2]) {
	buf[0] = (unsigned char)((num >> 8) & 0xFF);
	buf[1] = (unsigned char)(num & 0xFF);
}

inline unsigned short charToShort(unsigned char buf[2]) {
	return (((unsigned int)buf[0]) << 8) +
		(unsigned int)buf[1];
}

/*!
  Provides a comparison operator for pointers to strings.
  \sa ConstStringPointerCompare.
 */
class StringPointerCompare {
public:
	/*!
	  Compares two strings that are passed as pointers.
	  \param x Pointer to the first string.
	  \param y Pointer to the second string.
	  \return The value of (*x < *y)
	 */
	int operator() (const string *x, const string *y) const {
		if (!x) {
			if (!y) {
				return false;
			}
			return true;
		}
		return *x < *y;
	}
};
/*!
  Provides a comparison operator for pointers to const strings.
  \sa StringPointerCompare.
 */
class ConstStringPointerCompare {
public:
	/*!
	  Compares two strings that are passed as pointers.
	  \param x Pointer to the first string.
	  \param y Pointer to the second string.
	  \return The value of (*x < *y)
	 */
	int operator() (const string *const x, const string *const y) const {
		return *x < *y;
	}
};

/*std::ostream& operator<<(std::ostream &o, PluginId &n) {
	chain c = n;
	o << c;
	return o;
} */

} //namespace

#endif // _GLASS_TYPES_HH_
