/*  Glass - a distributed computing library
    Copyright (C) 2003-2009 Bruno Barberi Gnecco
    Copyright (C) 2009-2010 Corollarium Technologies


    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "packet.h"
#include "alias.h"
#include "log.h"
#include "_glass.h"

using namespace libglass;

namespace libglass {

/*!
  Format of Alias Plugin packets (client->server)

  \li [GLASS_CREATE][name\0] = Created
  \li [GLASS_DELETE][name\0] = Deleted
  \li [GLASS_ASSOCIATE][name\0][targetid][alias] = Associated

  server->client packets:

  \li [GLASS_ASSOCIATE][name\0][id] = set alias of this node as name.

  \sa Alias.
*/
class AliasPacket : public Packet {
public:
	typedef enum {
		GLASS_CREATE_LOCAL = 0, GLASS_CREATE_GLOBAL, GLASS_DELETE, GLASS_ASSOCIATE
	} AliasPacketType;

private:
	AliasPacketType		type;
	chain			aliasname;
	nodeId			targetid;
	char			iddata[4];
	string			targetname;

public:
	explicit AliasPacket(Packet &p) throw(Exception) : Packet(p.getGlass()) {
		unsigned int end, end2;
		chain		c;
		p.getData(c);

		type = (AliasPacketType)c[0];
		end = c.find('\0', 1);
		aliasname = c.substr(1, end-1);

		switch (type) {
		  case GLASS_CREATE_LOCAL:
		  case GLASS_CREATE_GLOBAL:
		  	// nothing else
			break;
		  case GLASS_DELETE:
		  	// nothing else
			break;
		  case GLASS_ASSOCIATE:
			targetid = (nodeId)c.toInt(end+1);
			end2 = c.size()-1;
			targetname = c.toString(end+5, chain::npos);
			LOG(outlog(), GLASS_DEBUG, CodeDebug << "Got associate " << targetid << " to " << targetname);
			LOG(outlog(), GLASS_DEBUG, CodeDebug << "end+5=" << end+5 << "  end2=" << end2);
			for (unsigned int i = end+5; i < end2; i++) {
				LOG(outlog(), GLASS_DEBUG, CodeDebug << "name[" << i << "]=" << c[i]);
			}
			break;
		  default:
		  	throw Exception();
		  	break;
		}
		LOG(outlog(), GLASS_DEBUG, CodeDebug << "Got type " << type);
	}

	AliasPacket(AliasPacketType type, _Alias &alias) : Packet(alias.ab->getGlass()),
		type(type), aliasname(alias.name) {
	}

	virtual ~AliasPacket() {
	}

	AliasPacketType getType(void) const {
		return this->type;
	}

	const chain &getAliasName(void) const {
		return this->aliasname;
	}
	nodeId getTargetId(void) const {
		return this->targetid;
	}
	const string &getTargetName(void) const {
		return this->targetname;
	}

	bool setAssociation(nodeId id, const char *name) {
		if (!name) {
			return false;
		}

		this->targetid = id;
		// todo: we are supposing that name will still be valid when
		// encode() is called, which is true by now
		this->targetname.assign(name);
		return true;
	}

	bool setAssociation(nodeId id, string &name) {
		this->targetid = id;
		// todo: we are supposing that name will still be valid when
		// encode() is called, which is true by now
		this->targetname.assign(name);
		return true;
	}

	chain encode(unsigned int *len = NULL) {
		LOG(outlog(), GLASS_DEBUG, "Message: " << type << " with " << aliasname);
		setData((char *)&type, 1);
		appendData(aliasname);
		appendData("\0", 1);
		if (type == GLASS_ASSOCIATE) {
			numToChar(targetid, iddata);
			appendData(iddata, 4);
			appendData(targetname);
		}
		return Packet::encode(len);
	}

	std::ostream& print(std::ostream &o) {
		Packet::print(o);
		o << " Alias type " << type;
		if (type == GLASS_ASSOCIATE) {
			o << " to " << targetid << " name " << targetname;
		}
		return o;
	}
};

} // namespace


// PLUGIN_CONSTRUCTOR macro is defined in plugin.h
// The expanded code ensures that the plugin automatically registers itself
// with the PluginManager object during runtime initialisation. i.e. before
// calling main()
// This registration is achieved by the following operations:
//   Runtime initialisation of class static member AliasBase::registered
//   Runtime initialisation of class static member AliasBase::singleton
//   Definition of default class constuctor
PLUGIN_CONSTRUCTOR(AliasBase) {
}

// Define the destructor
AliasBase::~AliasBase() {
	LOG(outlog(), GLASS_DEBUG, CodeDebug);
	std::map<const string, _Alias *>::iterator it;

	it = aliases.begin();
	while (it != aliases.end()) {
		// DON'T delete it->first, it doesn't belong to us.
		delete it->second;
		it++;
	}

	aliases.clear();
}

void AliasBase::synchronize(nodeId id, nodeRelation r) {
	// TODO
	return;
}

bool AliasBase::processPacket(Packet &p) {
	LOG(outlog(), GLASS_ERROR, "XX AliasBase");
	nodeRelation			senderType = getNodeRelation(p.getSenderNode());
	chain				c;
	nodeId 				nid;
	_Alias				*alias;
	propagationMode		mode;
	std::set<nodeId>::iterator	set_iter;
	std::map<const string, _Alias *>::iterator it;

	// TODO: change getSenderNode to originalSender when it is implemented!
	try {
		AliasPacket	ap(p);

		string aliasname(ap.getAliasName().toString());
		it = aliases.find(aliasname);
		LOG(outlog(), GLASS_DEBUG, CodeDebug << "Message " << ap.getType() << " with " << aliasname);

		if (senderType == Client || senderType == Peer) {
			switch (ap.getType()) {
			  case AliasPacket::GLASS_CREATE_LOCAL: /* created */
			  case AliasPacket::GLASS_CREATE_GLOBAL: /* created */

				  if (ap.getType() == AliasPacket::GLASS_CREATE_LOCAL) {
					LOG(outlog(), GLASS_INFO, NODE_FORMAT p.getSenderNode() << " created new local alias " << aliasname);
					mode = Local;
				}
				else {
					LOG(outlog(), GLASS_INFO, NODE_FORMAT p.getSenderNode() << " created new global alias " << aliasname);
					mode = Global;
				}

				/* created an alias */
				if (it != aliases.end()) {
					// alias already exists.
					LOG(outlog(), GLASS_DEBUG, CodeDebug << "New owner of " << aliasname);
					alias = it->second;
				}
				else {
					try {
						alias = new _Alias(aliasname);
					}
					catch (Exception e) {
						LOG(outlog(), GLASS_ERROR, e.getMessage());
						return false;
					}
					LOG(outlog(), GLASS_DEBUG, CodeDebug << "Created new alias " << aliasname);
				}
				LOG(outlog(), GLASS_DEBUG, CodeDebug << "Inserting node " << NODE_FORMAT p.getSenderNode());
				alias->owners.insert(p.getSenderNode());

				if (mode == Global && senderType == Client) {
					// we are a peer who received the message from a client
					// propagate it to other peers.
					LOG(outlog(), GLASS_INFO, "Notifying peers of global alias " << aliasname);
					 // updated sender, because peers only see peers. Otherwise we'd keep the original sender id.
					p.setSenderNode(this->getGlass()->getNodeId());
					p.setTargetNodes(peersId); // send to peers
					p.setBlackMask(p.getSenderNode()); // but not to ourselves
					sendPacket(p);
				}

				/* Check for an existing association.
				todo: it would be neater to call associate(), but how to p.setTargetNode()?*/
				{
					std::map<nodeId, string>::iterator map_iter;

					map_iter = alias->associations.find(p.getSenderNode());
					LOG(outlog(), GLASS_DEBUG, CodeDebug << "Association for node " << NODE_FORMAT p.getSenderNode() <<
						(map_iter != alias->associations.end() ?  " found" : " not found"));
					if (map_iter != alias->associations.end())  {
						AliasPacket ap(AliasPacket::GLASS_ASSOCIATE, *alias);
						ap.setAssociation(p.getSenderNode(), map_iter->second);
						ap.setTargetNodes(p.getSenderNode());
						sendPacket(ap);
					}
				}
				break;
			  case AliasPacket::GLASS_DELETE: /* deleted */
				if (it == aliases.end()) {
					LOG(outlog(), GLASS_WARNING, CodeDebug << "Inexistent alias " << aliasname << " deleted.");
					break;
				}
				alias = it->second;
				set_iter = alias->owners.find(p.getSenderNode());
				if (set_iter == alias->owners.end()) {
					LOG(outlog(), GLASS_WARNING, CodeDebug << "Node deleted alias that it doesn't have.");
					break;
				}
				alias->owners.erase(set_iter);
				LOG(outlog(), GLASS_DEBUG, CodeDebug << "Node " << NODE_FORMAT p.getSenderNode() << " removed variable " << aliasname);
				if (alias->owners.empty()) {
					LOG(outlog(), GLASS_DEBUG, CodeDebug << "Nobody has " << aliasname << " anymore. Deleted.");

					if (alias->getMode() == Global) {
						// we are a peer who received the message from a client
						// propagate it to other peers.
						LOG(outlog(), GLASS_INFO, "Notifying peers of delete " << aliasname);
						 // updated sender, because peers only see peers. Otherwise we'd keep the original sender id.
						p.setSenderNode(this->getGlass()->getNodeId());
						p.setTargetNodes(peersId); // send to peers
						p.setBlackMask(p.getSenderNode()); // but not to ourselves
						sendPacket(p);
					}
					this->aliases.erase(it);
					delete alias;
				}
				break;
			  case AliasPacket::GLASS_ASSOCIATE: /* associated */
				if (it == aliases.end()) {
					LOG(outlog(), GLASS_WARNING, CodeDebug << "Inexistent alias " << aliasname << " associated.");
					break;
				}
				alias = it->second;
				nid = ap.getTargetId();
				if (nid == everybodyId) {
	//				LOG(outlog(), GLASS_DEBUG, "Associating from node " << NODE_FORMAT p.getSenderNode() << " id " << nid << ": " << ap.getTargetName());
					for (set_iter = alias->owners.begin(); set_iter != alias->owners.end(); set_iter++) {
						ap.setTargetNodes(*set_iter);
						sendPacket(ap);
					}
					alias->associations[nid] = ap.getTargetName();
				}
				else {
					set_iter = alias->owners.find(nid);
					alias->associations[nid] = ap.getTargetName();
					LOG(outlog(), GLASS_DEBUG, CodeDebug << "Associating " << nid << " with " << alias->associations[nid]);
					if (set_iter != alias->owners.end()) {
						LOG(outlog(), GLASS_DEBUG, CodeDebug << "Associated node " << nid << " exists, updating.");
						ap.setTargetNodes(nid);
						if (sendPacket(ap) == false) {
							return false;
						}
					}
				}

				if (alias->getMode() == Global && senderType == Client) { // propagate to every peer we know.
					LOG(outlog(), GLASS_INFO, "Propagating  " << aliasname << " to peers");
					std::set<nodeId> targets; // copy
					for (set_iter = alias->owners.begin(); set_iter != alias->owners.end(); set_iter++) {
						if (this->getGlass()->getNodeRelation(*set_iter) == Peer) {
							targets.insert(*set_iter);
						}
					}
					p.setTargetNodes(targets);
					p.setBlackMask(p.getSenderNode());
					sendPacket(p);
				}

				break;
			  default:
				return false;
			}
		}
		else if (senderType == Server || senderType == None) {
			switch (ap.getType()) {
			  case AliasPacket::GLASS_ASSOCIATE: /* associated */
				if (it == aliases.end()) {
					LOG(outlog(), GLASS_WARNING, CodeDebug << "Inexistent alias " << aliasname << " associated.");
					break;
				}
				alias = it->second;
				alias->_associate(ap.getTargetName());
				LOG(outlog(), GLASS_DEBUG, CodeDebug << "Associated " << alias->name << " with ");
				break;
			  default:
				return false;
			}
		}
		else {
			LOG(outlog(), GLASS_ERROR, "Invalid senderType in AliasBase.");
			return false;
		}
	}
	catch (Exception e) {
		LOG(outlog(), GLASS_ERROR, CodeDebug << e.getMessage());
		return false;
	}

	return true;
}

void AliasBase::unregisterNode(nodeId id) {
	nodeRelation			relation = getNodeRelation(id);
	_Alias				*alias;

	if (relation == Server) {
		; // server should never unregister
	}
	else if (relation == Client || relation == Peer) {
		std::map<const string, _Alias *,
			StringPointerCompare>::iterator	it;
		std::set<nodeId>::iterator set_iter;
		for (it = aliases.begin(); it != aliases.end(); it++) {
			alias = it->second;
			set_iter = alias->owners.find(id);
			if (set_iter != alias->owners.end()) {
				alias->owners.erase(set_iter);
				if (alias->owners.empty()) {
					LOG(outlog(), GLASS_DEBUG, CodeDebug << "Unregister. Nobody has " << alias->name << " anymore. Deleted.");
					delete alias;
				}
			}
		}
	}
}

bool AliasBase::registerAlias(_Alias *a) {
	LOG(outlog(), GLASS_DEBUG, CodeDebug);
	if (!a) {
		return false;
	}

	std::map<string, _Alias *>::iterator it;
	it = aliases.find(a->name);
	if (it != aliases.end()) {
		return false;
	}
	aliases[a->name] = a;
	return true;
}

bool AliasBase::unregisterAlias(_Alias *a) {
	LOG(outlog(), GLASS_DEBUG, CodeDebug);
	if (!a) {
		return false;
	}

	std::map<const string, _Alias *>::iterator it;
	it = aliases.find(a->name);
	if (it == aliases.end()) {
		return false;
	}
	aliases.erase(it);
	return true;
}

/*
 * _Alias
 */
_Alias::_Alias(const char *aliasname, propagationMode mode) throw(Exception) : PluginInterface(mode) {
	if (!aliasname) {
		throw Exception("NULL alias name");
	}
	const string n(aliasname);
	init(n);
}

_Alias::_Alias(const string &aliasname, propagationMode mode) throw(Exception) : PluginInterface(mode) {
	init(aliasname);
}

void _Alias::init(const string &name) throw(Exception) {
	this->name = name;
	ab = dynamic_cast<AliasBase *>(getPluginBase("AliasBase"));
	if (!ab) {
		throw Exception("No AliasBase class.");
	}

	if (ab->registerAlias(this) == false) {
		throw Exception("Error registering variable.");
	}

	AliasPacket::AliasPacketType type;
	if (this->getMode() == Local)
		type = AliasPacket::GLASS_CREATE_LOCAL;
	else
		type = AliasPacket::GLASS_CREATE_GLOBAL;
	AliasPacket		p(type, *this);
//	p.setTargetNodes(peersId);
	p.setTargetNodes(serversId);
	if (ab->sendPacket(p) == false) {
		ab->unregisterAlias(this);
		throw Exception("Error sending packet.");
	}
}

_Alias::~_Alias() {
	LOG(outlog(), GLASS_DEBUG, CodeDebug);

	AliasPacket		p(AliasPacket::GLASS_DELETE, *this);

	// todo: who to send to? the owners group?
//	p.setTargetNodes(peersId);
	p.setTargetNodes(serversId);

	if (ab->sendPacket(p) == false) {
		// we can't do much here, can we?
	}
	ab->unregisterAlias(this);
	associations.clear();
}

string _Alias::getName(void) const {
	return this->name;
}

bool _Alias::_associate(const string &association) {
	associated = true;
	current_association = association;
	LOG(outlog(), GLASS_DEBUG, CodeDebug << "Alias " << this->name << " changed to "
			<< current_association);
	return true;
}

bool _Alias::associate(nodeId id, string &name)  {
	return associate(id, name.c_str());
}

bool _Alias::associate(nodeId id, const char *name) {
	if (!name)
		return false;

	AliasPacket			p(AliasPacket::GLASS_ASSOCIATE, *this);
	p.setAssociation(id, name);
	return ab->sendPacket(p);
}

bool _Alias::associate(std::map<nodeId, string> &map) {
	Packet				p(this->ab->getGlass());

#ifdef TODO
	std::map<nodeId, string>::iterator	it;
	p.setData('A');
	for (it = map.begin; it != map.end; it++) {
		p.appendData(it.first);
		p.appendData(it.second);
	}
#endif
	return ab->sendPacket(p);
}

#ifdef TODO
void _Alias::associate(string (*func)(nodeId)) {
	this->func = func;
}
#endif

