/*  Glass - a distributed computing library
    Copyright (C) 2003-2009 Bruno Barberi Gnecco
    Copyright (C) 2009-2010 Corollarium Technologies


    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <iostream>
#include "barrier.h"
#include "packet.h"
#include "log.h"
#include "_glass.h"

using namespace libglass;

/*!
  Format of Barrier Plugin packets:

  \li Client->server: [action (1 byte)][barrier number (4 bytes)][group 4 bytes] 
  \li Server->client: [action (1 byte)][barrier number (4 bytes)][group 4 bytes] 

  where valid action commands are described below.

  These are valid actions in client->server packets:

  \li GLASS_CREATE = created a barrier, add me to the list;
  \li GLASS_SYNC = client called the sync();
  \li GLASS_DELETE = destroyed the barrier, take me off the list;
  \li GLASS_ENABLE = this barrier is disabled.
  \li GLASS_DISABLE = this barrier is enabled.

  These are valid actions in server->client packets:

  \li GLASS_GETOUT = you can get out of sync() now.
*/
namespace libglass {

#define BARRIER_PACKET_SIZE	10

class BarrierPacket : public Packet {
public: 
	typedef enum {
		GLASS_CREATE_LOCAL = 0, GLASS_CREATE_GLOBAL, GLASS_SYNC, GLASS_DELETE,
				GLASS_GETOUT, GLASS_DISABLE, GLASS_ENABLE
	} BarrierPacketType;

private:
	BarrierPacketType	type;
	unsigned int		number, group;
	bool			active;
	unsigned char		data[BARRIER_PACKET_SIZE];

public:
	BarrierPacket(Packet &p) throw(Exception) : Packet(p.getGlass()) {
		chain		c;
		p.getData(c);

		type = (BarrierPacketType)c[0];
		number = c.toInt(1);
		group = c.toInt(5);
		active = (c[9]) ? true : false;
	}

	BarrierPacket(BarrierPacketType type, Barrier &b) : Packet(b.bb->getGlass()),
		type(type), number(b.getNumber()), group(b.getGroup()),
		active(false) {
		;
	}

	BarrierPacket(Glass *g, BarrierPacketType type, unsigned int number,
		unsigned int group) : Packet(g), type(type), number(number), group(group),
		active(false) {
		;
	}

	virtual ~BarrierPacket() {
	}

	void setActive(bool active) {
		this->active = active;
	}

	BarrierPacketType getType(void) const {
		return this->type;
	}
	unsigned int getNumber(void) const {
		return this->number;
	} 
	unsigned int getGroup(void) const {
		return this->group;
	}
	bool getActive(void) const {
		return this->active;
	}

	chain encode(unsigned int *len = NULL) {
		data[0] = (unsigned char)type;
		numToChar(number, data+1);
		numToChar(group, data+5);
		data[9] = (unsigned char)active;
		this->setData(data, BARRIER_PACKET_SIZE);
		return Packet::encode(len);
	}

	std::ostream& print(std::ostream &o) {
		Packet::print(o);
		o << "[Barriertype = " << type << "; number=" << number << "; group=" <<
			group << "] ";
		return o;
	}
};

}


// PLUGIN_CONSTRUCTOR macro is defined in plugin.h
// The expanded code ensures that the plugin automatically registers itself
// with the PluginManager object during runtime initialisation. i.e. before
// calling main()
// This registration is achieved by the following operations:
//   Runtime initialisation of class static member BarrierBase::registered
//   Runtime initialisation of class static member BarrierBase::singleton 
//   Definition of default class constuctor
PLUGIN_CONSTRUCTOR(BarrierBase) {
}

BarrierBase::~BarrierBase() {
	std::map<unsigned int, std::map<unsigned int, Barrier *> >::
			iterator groupit;
	std::map<unsigned int, Barrier *>::iterator barrierit;
			groupit = clientmap.begin();

	LOG(outlog(), GLASS_DEBUG, CodeDebug);
	while (groupit != clientmap.end()) {
		barrierit = groupit->second.begin();
		while (barrierit != groupit->second.end()) {
			Barrier *b = barrierit->second;
			LOG(outlog(), GLASS_DEBUG, "Deleting " << barrierit->first);
			barrierit++;
			delete b;
		}
		groupit->second.clear();
		groupit++;
	}
	clientmap.clear();
	servermap.clear();
}

const char *BarrierBase::state(BarrierState s) {
	switch (s) {
		case Active:
			return "Active";
		case Inactive:
			return "Inactive";
		case Locked:
			return "Locked";
		case Disabled:
			return "Disabled";
		}
	return NULL;
}	

// Return values:
// false: An error occured during a sendPacket()
// true: Either some nodes were not locked, or if all were locked and the packet was successfully sent
bool BarrierBase::checkUpdate(unsigned int barriergroup,  unsigned int barriernum) {
	/* now check if all nodes that own barriernum are blocked. If so, issue
	   a continue packet. */
   
	// Get barrier map - the map is specific to the barrier group and barrier number pair
	// and has one entry for each node that registered the barrier.

	if (servermap.find(barriergroup) == servermap.end()) {
		return false;
	}
	if (servermap[barriergroup].find(barriernum) == servermap[barriergroup].end()) {
		return false;
	}

	std::map<nodeId, struct _BarrierData>	&bmap = servermap[barriergroup][barriernum];
	std::map<nodeId, struct _BarrierData>::iterator it;
	bool 				retval = true;

	/* possible cases:
	   - empty bmap: erase barrier from the barrier group - if resulting barrier group is empty then erase barrier group.
	   - bmap with at least one node with barrier in active state: return.
	   - bmap with all locked: issue a sync "G" packet */

	// Are there any nodes currently registered as having an interest in the barrier?
	if (bmap.empty()) {
		// No nodes are interested in the barrier so erase it from the barrier group
		servermap[barriergroup].erase(barriernum);
		// Are there any barriers left in the barrier group?
		if (servermap[barriergroup].empty()) {
			// The barrier group is empty so lets erase it
			servermap.erase(barriergroup);
			// TODO: notify peers
		}
		return true;
	}

	// Iterate over all of nodes in the barrier map
	// Each entry in the map corresponds to a node that registered the barrier
	// If any one of the barrier node states is active then just return (without sending "G" packet)
	for (it = bmap.begin(); it != bmap.end(); it++) {
		// Log node barrier states
		LOG(outlog(), GLASS_DEBUG, "Checkupdate " << barriernum << ":" << barriergroup <<  " node [" 
				<< it->first << "] is " << state(it->second.state));
		// Test barrier state on current node to see if it is in the active state
		if (it->second.state == Active) {
			LOG(outlog(), GLASS_DEBUG, "Checkupdate " << barriernum << ":" << barriergroup <<  " found active at node [" << it->first << "]");
			return true;
		}
	}

	// --- We will not reach here if any of the barrier node states are active ---

	// If we get here, then we will send the sync "G" (go) packet to any node that
	// registered the barrier AND is in a locked state

	// Create barrier packet
	BarrierPacket			p(this->getGlass(), BarrierPacket::GLASS_GETOUT, barriernum, barriergroup);

	// Iterate over all nodes in the barrier map
	// Each entry in the map corresponds to a node that registered the barrier
	for (it = bmap.begin();	it != bmap.end(); it++) {

		// If barrier is on current node is locked...
		if (it->second.state == Locked) {

			// If barrier on current node is locked but node is not active...
			if (activemap[it->first] == false) {

				// We must activate the node

				// Do the simple step first - modify the activemap
				activemap[it->first] = true;

				// Now this next bit is a bit tricky - we have to iterate over all the barriers in 
				// the barrier group and find those map entries for the node state corresponding to
				// the inactive node detected above.
				// Once found, we test barrier node state and if it is inactive then we change it to active.
				std::map<unsigned int, std::map<nodeId, _BarrierData> >	&bmap = servermap[barriergroup];
				std::map<unsigned int, std::map<nodeId, _BarrierData> >::iterator barrierit;
				std::map<nodeId, _BarrierData> *nodemap;
				std::map<nodeId, _BarrierData>::iterator nodeit;

				/* Iterate over all barriers in the same barrier group (as the locked barrier on the inactive node detected above) */
				for (barrierit = bmap.begin(); barrierit != bmap.end(); barrierit++) {

					// Get map of all nodes that have registered the current barrier
					nodemap = &barrierit->second;
					// Find map entry for the node state corresponding to the inactive node detected above
					nodeit = nodemap->find(it->first);
					// If we didn't find the node in the map then skip
					if (nodeit == nodemap->end()) {
						continue;
					}
					// If the barrier node state is inactive then set to active
					if (nodeit->second.state == Inactive) {
						LOG(outlog(), GLASS_DEBUG, "Setting node [" << it->first << "] active in barrier " << barrierit->first << ":" << barriergroup);
						nodeit->second.state = Active;
					}
				} // next barrier in the same barrier group
			} // end if

			// Target for the "G" packet is the current node
			p.setTargetNodes(it->first);
			LOG(outlog(), GLASS_DEBUG, "Sending sync \"G\" packet " << barriernum << ":" << barriergroup << " to node [" << it->first << "]");
			if (sendPacket(p) == false) {
				retval = false;
				/* don't return until everybody gets a packet! */
			}
			// Modify servermap to indicate that the barrier is in active state
			it->second.state = Active;
		} // end if
	
	} // next node that registered the barrier
	return retval;
}




bool BarrierBase::processPacket(Packet &p) {
	Barrier				*b;
	nodeId				sender = p.getSenderNode();
	nodeRelation			senderType = getNodeRelation(sender);
	unsigned int 			barriernum, barriergroup;

	LOG(outlog(), GLASS_DEBUG, CodeDebug);

	// Parse barrier packet
	BarrierPacket			bp(p);

	barriergroup = bp.getGroup();
	barriernum = bp.getNumber();
	
	if (senderType == Server || senderType == None) {
		LOG(outlog(), GLASS_DEBUG, "Received \"" << bp.getType() << "\" packet from server");
		LOG(outlog(), GLASS_DEBUG, "BarrierNum = " << bp.getNumber() << "; BarrierGroup = " << bp.getGroup());

		std::map<unsigned int, std::map<unsigned int, Barrier *> >
				::iterator 		groupit;
		std::map<unsigned int, Barrier *>	*bmap;

		switch (bp.getType()) {
		  case BarrierPacket::GLASS_GETOUT: /* go on. */
			groupit = clientmap.find(barriergroup);
			if (groupit == clientmap.end()) {
			LOG(outlog(), GLASS_DEBUG, "group not found" << barriergroup);
				// not found.
				return false;
			}
			bmap = &groupit->second;
			if (bmap->find(barriernum) == bmap->end()) {
			LOG(outlog(), GLASS_DEBUG, "bnum not found" << barriernum);
				return false;
			}
			LOG(outlog(), GLASS_DEBUG, "Posting trigger " << barriernum);
			b = (*bmap)[barriernum];

			b->syncdata.mutex.lock();
			{
				boost::lock_guard<boost::mutex> lock (b->syncdata.t_mut);
				b->syncdata.lock = false;
			}
			b->syncdata.t.notify_all();
			b->syncdata.mutex.unlock();
			break;
		  default:
			LOG(outlog(), GLASS_ERROR, "Invalid packet in BarrierBase.");
		  	return false;
		}
	}
	else if (senderType == Client || senderType == Peer) {

//		LOG(outlog(), GLASS_DEBUG, "Received \"" << *bp.getData() << "\" packet from client");
		LOG(outlog(), GLASS_DEBUG, "BarrierNum = " << bp.getNumber() << "; BarrierGroup = " << bp.getGroup());

		std::map<unsigned int, std::map<nodeId, struct _BarrierData> > *barriermap;
		std::map<unsigned int, std::map<unsigned int, 
				std::map<nodeId, struct _BarrierData> > >::iterator
							groupit;
		std::map<unsigned int, std::map<nodeId, struct _BarrierData> >::iterator barrierit;
		std::map<nodeId, struct _BarrierData> *bmap;
		std::map<nodeId, struct _BarrierData>::iterator nodeit;

		// check if group exists. If not create it. Then get the barrier data (barriermap)
		groupit = servermap.find(barriergroup);
		if (groupit == servermap.end()) {
			if (bp.getType() != BarrierPacket::GLASS_CREATE_LOCAL && bp.getType() != BarrierPacket::GLASS_CREATE_GLOBAL) {
				LOG(outlog(), GLASS_WARNING, "Group not found: " << barriergroup);
				return false;
			}
			else {
				barriermap = &servermap[barriergroup];
			}
		}
		else {
			barriermap = &groupit->second;
		}

		barrierit = barriermap->find(barriernum);
		if (barrierit == barriermap->end()) {
			if (bp.getType() != BarrierPacket::GLASS_CREATE_LOCAL && bp.getType() != BarrierPacket::GLASS_CREATE_GLOBAL) {
				LOG(outlog(), GLASS_WARNING, "Barrier not found: " << barriernum);
				return false;
			}
			else {
				bmap = &(*barriermap)[barriernum];
			}
		}
		else {
			bmap = &barrierit->second;
		}

		// ok, so we have the data at hand now. Process it.
		switch (bp.getType()) {
		 case BarrierPacket::GLASS_CREATE_LOCAL: /* created a barrier */
		 case BarrierPacket::GLASS_CREATE_GLOBAL:
			if (bmap->find(sender) != bmap->end()) {
				LOG(outlog(), GLASS_WARNING, "Sender not in bmap");
				return false;
			}
			if (activemap.find(sender) == activemap.end()) {
				/* avoid deadlock system */
				activemap[sender] = false;
			}
			// state
			if (!bp.getActive() && activemap[sender] == false) {
				LOG(outlog(), GLASS_DEBUG, "Creating inactive barrier " << barriernum << ":" << barriergroup);

				(*bmap)[sender].state = Inactive;
			}
			else { /* create it as active */
				LOG(outlog(), GLASS_DEBUG, "Creating active barrier " << barriernum << ":" << barriergroup);
				(*bmap)[sender].state = Active;
			}

			// mode
			if (bp.getType() == BarrierPacket::GLASS_CREATE_GLOBAL) {
				(*bmap)[sender].mode = Global;

				// propagate to peers
				if (senderType == Client) {
					p.setTargetNodes(peersId);
					sendPacket(p);
				}
			}
			else {
				(*bmap)[sender].mode = Local;
			}
			break;
		 case BarrierPacket::GLASS_SYNC: /* sync() called */
			/* set the flag. */
			if (bmap->find(sender) == bmap->end()) {
				LOG(outlog(), GLASS_WARNING, "Invalid sync " << barriernum << ":" << barriergroup << " from " << sender);
				// the node has not registered the barrier.
				return false;
			}
			(*bmap)[sender].state = Locked;

			LOG(outlog(), GLASS_DEBUG, "Got a sync " << barriernum << ":" << barriergroup << " from " << sender);
			checkUpdate(barriergroup, barriernum);
			break;
		 case BarrierPacket::GLASS_DELETE: /* deleted the barrier */
			/* take the node out of the list. */
			LOG(outlog(), GLASS_DEBUG, "Got a delete " << barriernum << ":" << barriergroup << " from " << sender);

			nodeit = bmap->find(sender);
			if (nodeit == bmap->end()) {
				// The node is deleting a variable that was not created.
				LOG(outlog(), GLASS_WARNING, "Invalid: node does not own barrier.");
				return false;
			}
			bmap->erase(nodeit);

			/* now check if all nodes are blocked. If so, issue a
			   continue packet. */
			checkUpdate(barriergroup, barriernum);
			break;
		  
		  default:
			LOG(outlog(), GLASS_ERROR, "Invalid packet in BarrierBase.");
			break;
		}
	}
	else {
		LOG(outlog(), GLASS_ERROR, "Invalid senderType in BarrierBase.");
		return false;
	}
	return true;
}

void BarrierBase::unregisterNode(nodeId id) {
	nodeRelation			relation = getNodeRelation(id);
	
	if (relation == Server) {
		; // client should never unregister
	}
	else if (relation == Client || relation == Peer) {
		std::map<unsigned int, std::map<unsigned int, 
				std::map<nodeId, struct _BarrierData> > >::iterator
							groupit;
		std::map<unsigned int, std::map<nodeId, struct _BarrierData> >::
				iterator 		nodeit;
		std::map<nodeId, struct _BarrierData>::iterator mapit;

		for (groupit = servermap.begin(); groupit != servermap.end(); groupit++) {
			for (nodeit = groupit->second.begin(); nodeit != 
					groupit->second.end(); nodeit++) {
				std::map<nodeId, struct _BarrierData> &bmap = nodeit->second;

				if ((mapit = bmap.find(id)) != bmap.end()) {
					bmap.erase(mapit);
					checkUpdate(groupit->first, nodeit->first);
				}
			}
		}
		activemap.erase(id);
	}
}

bool BarrierBase::registerBarrier(Barrier *b) {
	LOG(outlog(), GLASS_DEBUG, CodeDebug);
	std::map<unsigned int, Barrier *>::iterator	it;
	std::map<unsigned int, Barrier *>	&bmap = clientmap[b->getGroup()];

	it = bmap.find(b->getNumber());
	if (it == bmap.end()) { 
		bmap[b->getNumber()] = b;
		return true;
	} 
	return false;
}

bool BarrierBase::unregisterBarrier(Barrier *b) {
	std::map<unsigned int, Barrier *>::iterator	it;
	std::map<unsigned int, Barrier *>	&bmap = clientmap[b->getGroup()];

	it = bmap.find(b->getNumber());
	if (it != bmap.end()) { 
		bmap.erase(it);
		return true;
	}
	return false;
}

Barrier::Barrier(unsigned int num, unsigned int group /* = 0 */,
		bool active /* = false */, propagationMode mode /*= Local*/) throw(Exception) :
		PluginInterface(mode), number(num), group(group) {

	bb = dynamic_cast<BarrierBase *>(this->getPluginBase("BarrierBase"));
	if (!bb) {
		throw Exception("No Barrier Base class.");
	}

	LOG(outlog(), GLASS_DEBUG, CodeDebug << ": " << (void *)bb);

	if (bb->registerBarrier(this) == false) {
		throw Exception("Could not register Barrier (it probably already exists).");
	}

	/* notify the server */
	BarrierPacket::BarrierPacketType bpt;
	if (mode == Local)
		bpt = BarrierPacket::GLASS_CREATE_LOCAL;
	else
		bpt = BarrierPacket::GLASS_CREATE_GLOBAL;
	BarrierPacket			p(bpt, *this);
	p.setActive(active);

//	p.setTargetNodes(peersId);
	p.setTargetNodes(serversId);
	if (bb->sendPacket(p) == false) {
		bb->unregisterBarrier(this);
		throw Exception("Error creating barrier.");
	}
}
#if 0 // AMBIGUOUS
Barrier::Barrier(unsigned int num,
		propagationMode mode /*= Local*/) throw(Exception) :
		PluginInterface(mode), number(num), group(0) {

	LOG(outlog(), GLASS_ERROR, CodeDebug << ": " << (void *)bb);
	bool active = false;
	bb = dynamic_cast<BarrierBase *>(this->getPluginBase("BarrierBase"));
	if (!bb) {
		throw Exception("No Barrier Base class.");
	}

	LOG(outlog(), GLASS_DEBUG, CodeDebug << ": " << (void *)bb);

	if (bb->registerBarrier(this) == false) {
		throw Exception("Could not register Barrier (it probably already exists).");
	}

	/* notify the server */
	BarrierPacket::BarrierPacketType bpt;
	if (mode == Local)
		bpt = BarrierPacket::GLASS_CREATE_LOCAL;
	else
		bpt = BarrierPacket::GLASS_CREATE_GLOBAL;
	BarrierPacket			p(bpt, *this);
	p.setActive(active);

//	p.setTargetNodes(peersId);
	p.setTargetNodes(serversId);
	if (bb->sendPacket(p) == false) {
		bb->unregisterBarrier(this);
		throw Exception("Error creating barrier.");
	}
}
#endif

Barrier::~Barrier()  {
	LOG(outlog(), GLASS_DEBUG, CodeDebug << ": " << number << ": " << group);
	if (bb->unregisterBarrier(this) == false) 
		return;

	// let the server know we're no longer interested
	BarrierPacket			p(BarrierPacket::GLASS_DELETE, *this);

//	p.setTargetNodes(peersId);
	p.setTargetNodes(serversId);
	if (bb->sendPacket(p) == false) {
		// we can't do much here, can we? 
		LOG(outlog(), GLASS_ERROR, CodeDebug << ": " << "Deleting " << number << ": " << group);
	}
}

void BarrierBase::synchronize(nodeId id, nodeRelation r) {
	// TODO
	return;
}


bool Barrier::sync(void) {
	/* thread safety code. Ensures that n threads syncing the same
	   variable will be properly synchronized. */
	syncdata.mutex.lock();
		if (syncdata.lock == true) {
			boost::unique_lock<boost::mutex> lock (syncdata.t_mut);
			while (syncdata.lock) {
				syncdata.t.wait(lock);
			}
			LOG(outlog(), GLASS_DEBUG, "Waiting locked trigger " << number);
			return syncdata.retval;
		}
		else {
			syncdata.lock = true;
		}
	syncdata.mutex.unlock();

	/* send the sync packet */
	BarrierPacket		p(BarrierPacket::GLASS_SYNC, *this);
	p.setTargetNodes(serversId);

	// test return
	if ((syncdata.retval = bb->sendPacket(p)) == false) {
		syncdata.mutex.lock();
		{
			boost::lock_guard<boost::mutex> lock (syncdata.t_mut);
			syncdata.lock = false;
		}
		syncdata.t.notify_all();
		syncdata.mutex.unlock();
		return syncdata.retval;
	}

	LOG(outlog(), GLASS_DEBUG, "Waiting trigger " << number);
	{
		boost::unique_lock<boost::mutex> lock (syncdata.t_mut);
		while (syncdata.lock) {
			syncdata.t.wait(lock);
		}
	}


	return syncdata.retval;
}

bool Barrier::disable(void) {
	// todo
	return false;
}

bool Barrier::enable(void) {
	// todo
	return false;
}
