/*  Glass - a distributed computing library
    Copyright (C) 2003-2009 Bruno Barberi Gnecco
    Copyright (C) 2009-2010 Corollarium Technologies


    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <iostream>
#include <fstream>
#include "chain.h"

#if defined(_MSC_VER) && (_MSC_VER < 1300)  // earlier than .NET compiler (VC 7.0)
	/* Push the current warning state so that we can restore later */
#  pragma warning( push )
	/* Disable warning message 4290: C++ Exception Specification ignored */
#  pragma warning( disable: 4290 )
#endif

using namespace libglass;


record::record() : r(new _record), offset(0), roffset(0) {
	r->references = 0;
	r->size = 0;
	r->allocated = false;
	r->data = NULL;
}

record::record(const char *p, unsigned int size, bool allocated,
		const char *allocateddata) : r(new _record) {
	r->data = (allocateddata == NULL ? p : allocateddata);
	r->size = size;
	r->allocated = allocated;
	r->references = 1;
	offset = (allocateddata == NULL ? 0 : (int)(p-allocateddata));
	roffset = 0;
}

record::record(const record *r, int offset, int roffset) : r(r->r) {
	this->r->references++;
	this->offset = offset + r->offset;
	this->roffset = roffset + r->roffset;
}

record::record(const record &r, int offset, int roffset) : r(r.r) {
	this->r->references++;
	this->offset = offset + r.offset;
	this->roffset = roffset + r.roffset;
}

record::~record() {
	r->references--;
	if (r->references == 0) {
		if (r->allocated && r->data) {
			// Cast away the const property before deleting
			char *d = const_cast<char*>(r->data);
			// std::cerr << "Garbage collector: deleting " << (void*)d << std::endl;
			delete[] d;
		}
		delete r;
	}
}

record &record::operator=(const record &s) {
	r = s.r;
	r->references++;
	offset = s.offset;
	return *this;
}

const char *record::getData(void) const {
	return r->data+offset;
}

unsigned int record::getLength(void) const {
	return (r->size-offset-roffset > 0 ? r->size-offset-roffset : 0);
}

void record::print() const {
	std::cerr << "offset: " << offset << "; roffset: " << roffset;
	if (!r) {
		std::cerr << ". NULL r!" << std::endl;
		return;
	}
	std::cerr << ". r: " << (void *) r << "; data: " << (void *)r->data <<
		"; size: " << r->size << "; alloc: " <<
		(r->allocated == true ? "true" : "false")
		<< "; refs: " << r->references << std::endl;
}


/*
 * Chain
 */

chain::chain() {
}

chain::chain(const char *p, unsigned int s/* = 0 */) {
	append(p, s);
}

chain::chain(std::string &s) {
	append(s);
}

chain::~chain() {
	clear();
}

void chain::clear(void) {
	pieces.clear();
}

unsigned int chain::size(void) const {
	std::list<record>::const_iterator it;
	unsigned int l = 0;

	for (it = pieces.begin(); it != pieces.end(); it++) {
		l += it->getLength();
	}
	return l;
}

unsigned int chain::length(void) const {
	return this->size();
}

const char *chain::getPosition(unsigned int position) const throw(Exception) {
	std::list<record>::const_iterator it;
	for (it = pieces.begin(); it != pieces.end(); it++) {
		if (it->getLength() <= position) {
			position -= it->getLength();
		}
		else {
			return (char *)&it->getData()[position];
		}
	}
	// if we reach here, position > size.
	throw Exception();
}

char &chain::operator[](unsigned int position) throw(Exception) {
	return *const_cast<char *>(getPosition(position));
}

const chain &chain::operator=(const chain &s) {
	return assign(s);
}

const chain &chain::assign(const chain &s) {
	this->clear();
	this->pieces = s.pieces;

	return *this;
}

const chain &chain::assign(char *p, unsigned int s /* = 0 */,
		bool free /* = false */, char *allocateddata /* = NULL */) {
	this->clear();
	this->append(p, s, free, allocateddata);
	return *this;
}

const chain &chain::assign(std::string &s) {
	this->clear();
	this->append(s);
	return *this;
}

const bool chain::operator<(const chain &c) const {
	unsigned int endc1, endc2;

	endc1 = this->size();
	endc2 = c.size();
	for (unsigned int i = 0; i < endc1 && i < endc2; i++) {
		if (*(this->getPosition(i)) < *(c.getPosition(i))) return true;
	}
	return false;
}

const chain &chain::operator+=(const chain &s) {
	this->append(s);
	return *this;
}

const chain &chain::operator+=(char *p) {
	this->append(p, 0);
	return *this;
}

const chain &chain::operator+=(std::string &s) {
	this->append(s);
	return *this;
}

#ifdef TODO
const chain chain::operator+(const chain &s) const {
	chain r;
	// todo
	return r;
}
#endif

bool chain::operator==(const chain &s) const {
	unsigned int thispos, spos;
	bool last = false;
	std::list<record>::const_iterator thisit, sit;

	if (s.size() != this->size()) {
		return false;
	}
	// sizes are equal, compare them directly

	thispos = spos = 0;
	thisit = this->pieces.begin();
	sit = s.pieces.begin();
	if (thisit == this->pieces.end() && sit == s.pieces.end()) {
		return true;
	}
	else if (thisit == this->pieces.end() || sit == s.pieces.end()) {
		return false;
	}
	while (true) {
		if (thispos >= thisit->getLength()) {
			thisit++;
			if (thisit == this->pieces.end()) {
				last = true;
			}
			else {
				thispos = 0;
				continue;
			}
		}
		if (spos >= sit->getLength()) {
			sit++;
			if (sit == s.pieces.end()) {
				last = true;
			}
			else {
				spos = 0;
				continue;
			}
		}
		if (last == true) {
			if (sit == s.pieces.end() &&
				thisit == this->pieces.end()) {
				return true;
			}
			else {
				return false;
			}
		}

		if (thisit->getData()[thispos++] != sit->getData()[spos++]) {
			return false;
		}
	}
	return false;
}

bool chain::operator!=(const chain &s) const {
	return !(s == *this);
}

void chain::append(const chain &s) {
	std::list<record>::const_iterator it;

	for (it = s.pieces.begin(); it != s.pieces.end(); it++) {
		record r(*it);
		pieces.push_back(r);
	}
}

void chain::append(const char *p, unsigned int s /* = 0 */, bool free /* = false */,
		char *allocatedbegin /* = NULL */) {
	if (s == 0) {
		s = strlen(p) + 1;
	}
	record r(p, s, free, allocatedbegin);
	pieces.push_back(r);
}

void chain::append(std::string &s) {
	append((char *)s.data(), s.length());
}

void chain::appendcopy(chain &s) {
	std::list<record>::iterator it;

	for (it = s.pieces.begin(); it != s.pieces.end(); it++) {
		char *data = new char[it->getLength()];
		memcpy(data, it->getData(), it->getLength());
		record r(data, it->getLength(), true);
		pieces.push_back(r);
	}
}

void chain::appendcopy(const char *p, unsigned int s /* = 0 */) {
	if (s == 0) {
		s = strlen(p) + 1;
	}
	char *data = new char[s];
	memcpy(data, p, s);
	record r(data, s, true, data);
	pieces.push_back(r);
}

void chain::appendcopy(std::string &s) {
	appendcopy((char *)s.data(), s.length()+1);
}

#ifdef DEPRECATED
chain *chain::concatenate(int begin, int end) const {
	std::list<record>::const_iterator it;
	int i = 0;

	if (begin < 0 || (end < begin && end != -1)) {
		return NULL;
	}

	/* get start record */
	for (it = pieces.begin(); it != pieces.end(); i++, it++) {
		if (i == begin)
			break;
	}
	if (it == pieces.end()) {
		return NULL;
	}

	chain *s = new chain();
	for (; (end == -1 ? true : i <= end) && it != pieces.end(); i++, it++) {
		s->appendcopy(it->getData(), it->getLength());
	}
	return s;
}
#endif

char *chain::toArray(unsigned int begin, unsigned int end) const throw(Exception) {
	std::list<record>::const_iterator it;
	unsigned int		total = 0;
	unsigned int		initial = begin;
	char			*buf, *p;

	if (end < begin) {
		throw Exception("Invalid arguments: out of bounds.");
	}

	/* find the start */
	for (it = pieces.begin(); it != pieces.end(); it++) {
		int i = it->getLength();
		i++;
		if (it->getLength() <= initial) {
			initial -= it->getLength();
		}
		else {
			break;
		}
	}

	if (it == pieces.end()) { /* OOB */
		if (initial == 0) {
			// empty
			return NULL;
		}
		throw Exception("Invalid arguments: out of bounds.");
	}


	if (this->size() < end) { /* end OOB, ok */
		end = npos;
	}

	buf = new char[end-begin+1];
	p = buf;
	for (; it != pieces.end(); it++) {
		if (end == npos || total + (it->getLength() - initial) <=
				(end - begin + 1)) {
			memcpy(p, it->getData() + initial, it->getLength()-initial);
			p += it->getLength()-initial;
			total += it->getLength() - initial;
			initial = 0;
			if (total == (end - begin + 1)) {
				break;
			}
		}
		else {
			memcpy(p, it->getData() + initial,
					((end-begin+1)-total));
			break;
		}
	}
	return buf;
}

std::string chain::toString(unsigned int begin, unsigned int end) const
		 throw(Exception) {
	std::string 		s;
	std::list<record>::const_iterator it;
	int			total = 0;
	unsigned int		initial = begin;

	if (end != npos && end < begin) {
		throw Exception("Invalid arguments: out of bounds.");
	}

	/* find the start */
	for (it = pieces.begin(); it != pieces.end(); it++) {
		if (it->getLength() <= initial) {
			initial -= it->getLength();
		}
		else {
			break;
		}
	}

	if (it == pieces.end()) { /* OOB */
		throw Exception("Invalid arguments: out of bounds.");
	}

	if (this->size() < end) { /* end OOB, ok */
		end = npos;
	}

	for (; it != pieces.end(); it++) {
		if (end == npos || total + (it->getLength() - initial) <=
				(end - begin + 1)) {
			s.append(it->getData() + initial, it->getLength()-initial);
			total += it->getLength() - initial;
			initial = 0;
			if (total == (signed)(end - begin + 1)) {
				break;
			}
		}
		else {
			s.append(it->getData() + initial,
				(((signed)(end-begin+1))-total));
			break;
		}
	}
	return s;
}

std::string chain::toStringNull(unsigned int begin, unsigned int end) const throw(Exception) {
	unsigned int last = this->find('\0', begin) - 1;
	if (last > end) last = end;
	std::string s;
	try {
		return this->toString(begin, last);
	}
	catch (Exception e) {
		; // do nothing. This is expected.
	}
	return s;
}


int chain::toInt(unsigned int begin) const throw(Exception) {
	char c[4];
	substrcopy(c, begin, begin+3);
	return (int)charToNum((unsigned char *)c);
}

chain chain::substr(unsigned int begin, unsigned int end) const
		throw(Exception) {
	chain 			s;
	int			total = 0;
	unsigned int		initial = begin;
	std::list<record>::const_iterator it;

	if (end != npos && end < begin) {
		throw Exception("Invalid arguments: out of bounds.");
	}

	/* find the start */
	for (it = pieces.begin(); it != pieces.end(); it++) {
		if (it->getLength() <= initial) {
			initial -= it->getLength();
		}
		else {
			break;
		}
	}

	if (it == pieces.end()) { /* OOB */
		throw Exception("Invalid arguments: out of bounds.");
	}

	if (this->size() < end) { /* OOB, ok */
		end = npos;
	}

	for (; it != pieces.end(); it++) {
		if (end == npos || total + (it->getLength() - initial) <=
				(end - begin + 1)) {
			record r(*it, initial);
			s.pieces.push_back(r);
			total += it->getLength() - initial;
			initial = 0;
			if (total == (signed)(end - begin + 1)) {
				break;
			}
		}
		else {
			record r(*it, initial, it->getLength() - (int)initial -
					((end-begin+1)-total));
			s.pieces.push_back(r);
			break;
		}
	}
	return s;
}

bool chain::substrcopy(char *dest, unsigned int begin, unsigned int end) const
		throw(Exception) {
	if (!dest) {
		return false;
	}

	std::list<record>::const_iterator it;
	chain c(this->substr(begin, end));

	for (it = c.pieces.begin(); it != c.pieces.end(); it++) {
		memcpy(dest, it->getData(), it->getLength());
		dest += it->getLength();
	}

	return true;
}

chain chain::duplicate(void) {
	chain			c;

	c.appendcopy(*this);

	return c;
}

unsigned int chain::find(char c, unsigned int initial /* = 0 */) const {
	std::list<record>::const_iterator it;
	unsigned int pos = initial;
	char *p;

	if (pos != 0) {
		for (it = pieces.begin(); it != pieces.end(); it++) {
			if (it->getLength() <= initial) {
				initial -= it->getLength();
			}
			else if ((p = strchr(const_cast<char *>(it->getData())+initial, c))) {
				pos += p - (it->getData()+initial);
				return pos;
			}
			else {
				pos += it->getLength()-initial;
				it++;
				break;
			}
		}
	}
	else {
		it = pieces.begin();
	}

	for (; it != pieces.end(); it++) {
		if ((p = strchr(const_cast<char *>(it->getData()), c))) {
			pos += p - it->getData();
			return pos;
		}
		else {
			pos += it->getLength();
		}
	}
	return npos;
}

// iterator
chain_iterator chain::begin(void) {
	chain_iterator it;
	it.setiterator(pieces.begin());
	return it;
}

chain_iterator chain::end(void) {
	chain_iterator it;
	it.setiterator(pieces.end());
	return it;
}

void chain::erase(chain_iterator it) { //x
#ifdef OLD
	record r = *(it.it);
	r.r->references--;
	if (r.r->references == 0 && r.r->allocated) {
		delete r.r->data;
	}
	pieces.erase(it.it);
#endif
}

void chain::debug(void) const {
	std::list<record>::const_iterator it;

	for (it = pieces.begin(); it != pieces.end(); it++) {
		it->print();
	}
}

namespace libglass {

std::ostream& operator<<(std::ostream &o, chain &p) {
	std::list<record>::iterator it;

	for (it = p.pieces.begin(); it != p.pieces.end(); it++) {
		o.write(it->getData(), it->getLength());
	}
	return o;
}

} //namespace

#if defined(_MSC_VER) && (_MSC_VER < 1300)  // earlier than .NET compiler (VC 7.0)
	/* Restore compiler warning state to previous state */
#  pragma warning( pop )
#endif
