/*  Glass - a distributed computing library
    Copyright (C) 2003-2009 Bruno Barberi Gnecco
    Copyright (C) 2009-2010 Corollarium Technologies


    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <iostream>
#include <boost/thread.hpp>
#include "packetclientserver.h"
#include "glass.h"
#include "glassclient.h"
#include "protocol.h"
#include "log.h"
#include "_glass.h"

/*!
 Use the following macro to define behaviour in GlassClient::~GlassClient
 Set to 1 to wait for worker thread to terminate normally
 Set to 0 to kill the worker thread abruptly.
*/
#define GLASS_CLIENT_WAIT_FOR_WORKER_THREAD 1

using namespace libglass;

GlassClient::GlassClient(Protocol *p, const std::string &server, const std::string &nickname,
		bool tryuntilconnect) throw(Exception) :
		Glass(p),
		GlassThread(false),
		initialized(false),
		initsuccess(false),
		quitting(false),
		tryuntilconnect(tryuntilconnect),
		ismaster(false) {
	LOG(outlog(), GLASS_DEBUG, CodeDebug);

	this->nickname.assign(nickname);
	this->serveraddr.host.assign(server);
	this->serveraddr.relation = Server;

	/* lock. We shan't return while we have not either connected to the
	   server or died trying. */
	this->start(); // This call is asynchronous and returns immediately. It will
	         // cause GlassClient::execute() to run in a new worker thread.

	// Wait for the worker thread to indicate that it has
    // successfully connected or failed to connect.
    // The initialized member tells us which of these two
    // situations occured.

	LOG(outlog(), GLASS_INFO, "Waiting for client init, thread " << this->t.get_id());
	while (initialized == false) { /* execute() method should set it true if all's ok */
		boost::posix_time::time_duration td = boost::posix_time::milliseconds(50);
		boost::unique_lock<boost::mutex> lock(mut);
		this->cond.timed_wait(lock, td);

		// TODO: if it takes too long....
	}
	LOG(outlog(), GLASS_INFO, "Condition posted.");
	if (initsuccess == false) {
		throw Exception("Could not connect");
	}
	LOG(outlog(), GLASS_DEBUG, "Successfully constructed GlassClient");

	this->setProtocolOwner(this);
}

GlassClient::~GlassClient() {
	LOG(outlog(), GLASS_DEBUG, CodeDebug);

	lock(mutexdata);

	quitting = true;

	// Create bye packet
	ClientServerPacket p(this, ClientServerPacket::BYE);
	/* Send packet to let the server know we are quitting */
	sendPacket(p);

#if GLASS_CLIENT_WAIT_FOR_WORKER_THREAD
	LOG(outlog(), GLASS_DEBUG, "Waiting for client worker thread to terminate");
	// TODO cond.wait(); // Wait for worker thread to terminate before
										 // continuing
	LOG(outlog(), GLASS_DEBUG, "Client worker thread has terminated normally");
#else
	LOG(outlog(), GLASS_DEBUG, "Killing client worker thread");
	this->join();
	LOG(outlog(), GLASS_DEBUG, "Client worker thread has been killed");
#endif // GLASS_CLIENT_WAIT_FOR_WORKER_THREAD

	/* IMPORTANT: clean IS A VIRTUAL DESTRUCTOR. Do not delete any fields
	or do anything that will compromise the state of the class before
	calling it. See the doxygen documentation for more information. */
	Glass::clean();

	/* now you can free anything you want */
}

// Overrides the pure virtual method in base class Thread
// This method is called either when the thread terminates normally or when an
// exception is raised within the thread's context. 
void GlassClient::cleanup(void) {
	LOG(outlog(), GLASS_DEBUG, CodeDebug);
}

bool GlassClient::sendPacket(Packet &p) {
	LOG(outlog(), GLASS_DEBUG, CodeDebug);

	// right now, we can only send packets directly to the server.
	p.setTargetAddresses(IdToAddress(this->serverid));
	p.setSenderNode(Glass::id);

	return protocol->send(p);
}

bool GlassClient::rootPacket(Packet &p) {
	ClientServerPacket		sp(p);

	if (quitting) {
		return true;
	}

	LOG(outlog(), GLASS_NOTICE, "Client ROOTPACKET: " << sp.getType());
	switch (sp.getType()) {
	  case ClientServerPacket::ID:
		// This is acknowledement packet sent by the server in response to
		// the hello packet the client send immediately after connection.
		/* our id arrived, now we have a name. */
	  {
		Glass::id = sp.getNodeId();
		this->unregisterNode(0);
		nodeAddress addr(p.getSenderNodeAddress());
		this->registerNode(addr, p.getSenderNode(), Server, true);
		LOG(outlog(), GLASS_INFO, "We have an id:" << NODE_FORMAT this->Glass::id);
	  }
		break;

	  case ClientServerPacket::QUIT_ALL:
		// Quit packet
		LOG(outlog(), GLASS_INFO, "Got a quit from " << NODE_FORMAT p.getSenderNode());
		exit(1);
		break;

	  case ClientServerPacket::BYE:
		// Bye packet
		/* the show is over */
// TODO!!! how do we quit gracefully?
		this->~GlassClient(); // problem: dyn X static.
		break;

	  case ClientServerPacket::NODE_LIST:
		LOG(outlog(), GLASS_INFO, "Got NODE_LIST size " << sp.getNodeList().size());
		/* how crowded is the party? */
		if (sp.getMasterId() == this->getNodeId()) {
			LOG(outlog(), GLASS_INFO, "We've been told we are master.");
			this->ismaster = true;
		}
		else {
			this->ismaster = false;
		}
		{
			lock(mutexdata);

			// clean everything but the server
#ifdef BIMAP
			for (AddressIdMap::map_by<Id>::iterator it(addridmap.right.begin()); it != addridmap.right.end(); ) {
				if (it->get<Address>().relation != Server) {
					nodeId id(it->get<Id>());
					LOG(outlog(), GLASS_DEBUG, "erasing " << id << " with relation " << it->get<Address>().relation);
#endif
			for (AddressIdMap::iterator it(addridmap.begin()); it != addridmap.end(); ) {
				if (it->second.relation != Server) {
					nodeId id(it->first);
					LOG(outlog(), GLASS_DEBUG, "erasing " << id << " with relation " << it->second.relation);
					it++;
					this->unregisterNode(id);
				}
				else {
					it++;
				}
			}

			const std::map<nodeId, nodeNickName> &nodelist(sp.getNodeList());
			LOG(outlog(), GLASS_DEBUG, "Inserting " << nodelist.size() << ", current:" << addridmap.size());
			for (std::map<nodeId, nodeNickName>::const_iterator it(nodelist.begin()); it != nodelist.end(); it++) {
				nodeAddress a;
				a.nickname = it->second;
				LOG(outlog(), GLASS_DEBUG, "Inserting " << it->first << " addr " << a);
				this->registerNode(a, it->first, None, false);
			}
		}
		LOG(outlog(), GLASS_DEBUG, "Got NODE_LIST. Master " << NODE_FORMAT sp.getNodeId()
			<< " this->id=" << NODE_FORMAT getNodeId());
		break;

	  default:
		LOG(outlog(), GLASS_ERROR, "Invalid root packet.");
		return false;
	}
	return true;
}

nodeId GlassClient::registerNode(nodeAddress &addr, nodeId id, nodeRelation r, bool force) {
	LOG(outlog(), GLASS_DEBUG, "Register node GlassClient " << addr<< "," << id);

	if (r == Server) {
		this->serverid = id;
	}
	addr.setRelation(r);

	if (this->setNode(addr, id, r, force) == false) {
		LOG(outlog(), GLASS_DEBUG, "Register node GlassClient " << addr<< "," << id);
		return nobodyId;
	}

	return this->serverid;
}

// This method is overrides the pure virtual method in base class and is
// executed in a new thread when the start() method is called. Note that each thread object
// can be run only once.
void GlassClient::execute(void) {
	int				ret;

	LOG(outlog(), GLASS_DEBUG, CodeDebug);

	/* let us not forget, my fellows, that GlassClient constructor waits
	   for us to  return. So, cond.notify() as soon as thou canst, and even if
		 thou  dieth, dost that before thy heart beats no more. */

	int tries = 0;
	while (true) {
		this->serveraddr = protocol->connect(this->serveraddr);
		if (this->serveraddr.status != PROTOCOL_OK && this->tryuntilconnect == false) {
			LOG(outlog(), GLASS_ERROR, "Could not connect to server.");
			initsuccess = false;
			initialized = true;
			this->cond.notify_all(); // Signal GlassClient constructor so that it can continue
			return; // Failed to connect
		}
		else if (this->serveraddr.status == PROTOCOL_OK) {
			break;
		}
		tries++;
		std::cerr<< "Could not connect to server, retrying " << tries;
		LOG(outlog(), GLASS_INFO, "Could not connect to server, retrying " << tries);
		SLEEP(1);
	}


	this->registerNode(this->serveraddr, nobodyId, Server, true);

	LOG(outlog(), GLASS_DEBUG, "register node");

	/* handshaking */
	{
		ClientServerPacket	cp(this, ClientServerPacket::HELLO);
		cp.setNickName(this->nickname);
		cp.setTargetNodes(this->serverid);
		// Send "H" hello packet to server - we expect to receive an "I" id packet
		// in acknowledgement
		if (sendPacket(cp) == false) {
			LOG(outlog(), GLASS_ERROR, "Sending \"H\" hello packet.");
			initsuccess = false;
			initialized = true;
			this->cond.notify_all();
			return;
		}
	}

	/* wait for acknowledgement */
	LOG(outlog(), GLASS_INFO, "Waiting for \"I\" packet acknowledgement.");
	int packetproc = 0;
	do {
		Packet				p(this);
		int ret = protocol->recv(p);
		if (ret != PROTOCOL_OK && ret != PROTOCOL_TIMEOUT && ret != PROTOCOL_UNREGISTERED_NODE) {
			LOG(outlog(), GLASS_ERROR, "Receiving packet: " << ret);
			initsuccess = false;
			initialized = true;
			this->cond.notify_all();
			return;
		}
		else if (ret == PROTOCOL_OK || ret == PROTOCOL_UNREGISTERED_NODE) { // success
			LOG(outlog(), GLASS_DEBUG, "got something.");
			if (this->processPacket(p, true) == true) {
				packetproc++;
			}
			else {
				LOG(outlog(), GLASS_ERROR, "Got a false on init processpacket");
			}
		}
		else { // timeout
			; // do nothing
		}
	} while (packetproc < 2);

	LOG(outlog(), GLASS_NOTICE, "Connected to server with success.");

	initialized = true; // Initialization successful
	initsuccess = true;
	this->cond.notify_all(); // Signal GlassClient constructor so that it can continue

	LOG(outlog(), GLASS_INFO, "Posted notification");

	/* process packet: */
	while(1) {
		Packet				p(this);
		ret = protocol->recv(p);

		switch (ret) {
		case PROTOCOL_NOT_INITIALIZED:
			LOG(outlog(), GLASS_ERROR, "Fatal protocol error: " << ret);
			return;

		case PROTOCOL_UNREGISTERED_NODE:
			LOG(outlog(), GLASS_ERROR, "Got message from unregistered node: " << p.getSenderNode());
			break;

		case PROTOCOL_OK: // as expected
			if (processPacket(p) == false) {
				// probably a misconstrued packet, ignore.
				LOG(outlog(), GLASS_WARNING, "Error processing packet, ignoring.");
			}
			break;
		case PROTOCOL_CONNECTION_CLOSED:
			// ICOF!
			if (!quitting) { 
				/* if we are not quitting, a lost connection is
				   unexpected */
				LOG(outlog(), GLASS_ERROR, "Lost connection.");
				// TODO: reconnect
			}
			return;
		default:
			// ret >=1
			// non-fatal, ignore
			break;
		}
		if (quitting == true) {
			return;
		}
	}
}

unsigned int GlassClient::getTotalNodes(void) {
	lock lock(mutexdata);
	return addridmap.size();
}

std::map<nodeId, nodeNickName> GlassClient::getClientList(void) const {
	lock lock(mutexdata);
	std::map<nodeId, nodeNickName> clientlist;
	AddressIdMap::const_iterator it;
	for (it = addridmap.begin(); it != addridmap.end(); it++) {
		clientlist[it->first] = it->second.nickname;
	}
	return clientlist;
}

bool GlassClient::isMaster(void) {
	return ismaster;
}

void GlassClient::killAll(void) {
	ClientServerPacket		p(this, ClientServerPacket::QUIT_ALL);
	sendPacket(p);

	SLEEP(5);
}
