/*  Glass - a distributed computing library
    Copyright (C) 2003-2009 Bruno Barberi Gnecco
    Copyright (C) 2009-2010 Corollarium Technologies


    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "event.h"
#include "packet.h"
#include "_glass.h"
#include "serial.h"
#include "objectmanager.h"

using namespace libglass;

namespace libglass {

/*!
 Format of Event Plugin packets:

  \li [action (1 byte)][name\0][data]

  These are valid actions in client->server packets:

  \li GLASS_CREATE(_LOCAL/_GLOBAL) = created an event, add me to the list;
  \li GLASS_DELETE = destroyed the event, take me off the list;
  \li GLASS_EVENT = queued an event.

  These are valid actions in server->client packets:

  \li GLASS_EVENT = queued an event.
*/
class EventPacket : public Packet {
public:
	typedef enum {
		GLASS_CREATE_LOCAL = 0, GLASS_CREATE_GLOBAL, GLASS_DELETE, GLASS_EVENT
	} EventPacketType;

private:
	EventPacketType		type;
	string			eventname;
	chain			eventdata;

public:
	EventPacket(Packet &p) throw(Exception) : Packet(p.getGlass()) {
		chain		c;
		unsigned int	end;
		p.getData(c);

		type = (EventPacketType)c[0];
		end = c.find('\0', 1);
		eventname = c.toString(1, end-1);
		if (type == GLASS_EVENT) {
			eventdata = c.substr(end+1, chain::npos);
		}
	}

	EventPacket(EventPacketType type, _Event &event) : Packet(event.eb->getGlass()),
		type(type), eventname(event.name)  {
		// todo
	}

	virtual ~EventPacket() {
	}

	EventPacketType getType(void) const {
		return this->type;
	}

	string &getEventName(void) {
		return eventname;
	}

	void setEventData(chain &c) {
		this->eventdata = c;
	}

	chain &getEventData(void) {
		return eventdata;
	}

	chain encode(unsigned int *len = NULL) {
		setData((char *)&type, 1);
		appendData(eventname);
		appendData("\0", 1);

		if (type == GLASS_EVENT) {
			appendData(eventdata);
		}
		return Packet::encode(len);
	}

	std::ostream& print(std::ostream &o) {
		Packet::print(o);
		return o;
	}
};

} // namespace


// PLUGIN_CONSTRUCTOR macro is defined in plugin.h
// The expanded code ensures that the plugin automatically registers itself
// with the PluginManager object during runtime initialisation. i.e. before
// calling main()
// This registration is achieved by the following operations:
//   Runtime initialisation of class static member EventBase::registered
//   Runtime initialisation of class static member EventBase::singleton
//   Definition of default class constuctor
PLUGIN_CONSTRUCTOR(EventBase) {
	LOG(outlog(), GLASS_DEBUG, CodeDebug);
}

EventBase::~EventBase() {
	LOG(outlog(), GLASS_DEBUG, CodeDebug);
	std::map<const string, _Event *>::iterator it;

	it = events.begin();
	while (it != events.end()) {
		// DON'T delete it->first, it doesn't belong to us.
		delete it->second;
		it++;
	}

	events.clear();
}

void EventBase::synchronize(nodeId id, nodeRelation r) {
	if (r == Peer) {
		// TODO
	}
	else {
		; // nothing
	}
	return;
}

bool EventBase::processPacket(Packet &p) {
	nodeRelation			senderType = getNodeRelation(p.getSenderNode());
	std::set<nodeId>::iterator	set_iter;
	nodeId				sender = p.getSenderNode();
	std::map<const string, _Event *,
			StringPointerCompare>::iterator events_it;
	std::map<string, _EventBaseData >::iterator owners_it;

	// Parse event packet
	try {
		EventPacket ep(p);
		string				&eventname = ep.getEventName();

		if (senderType == Client || senderType == Peer) {
			switch (ep.getType()) {
			  case EventPacket::GLASS_CREATE_LOCAL: /* created  */
				LOG(outlog(), GLASS_INFO, NODE_FORMAT sender << " created new local event " << eventname);
				this->eventdata[eventname].owners.insert(sender);
				break;

			  case EventPacket::GLASS_CREATE_GLOBAL: /* created  */
					LOG(outlog(), GLASS_INFO, NODE_FORMAT sender << " created new global event " << eventname);
				this->eventdata[eventname].owners.insert(sender);
				if (senderType == Client) {
					// we are a peer who received the message from a client
					// propagate it to other peers.
					LOG(outlog(), GLASS_INFO, "Notifying peers of created global event " << eventname);
					 // updated sender, because peers only see peers. Otherwise we'd keep the original sender id.
					p.setReliable(true);
					p.setSenderNode(this->getGlass()->getNodeId());
					p.setTargetNodes(peersId); // send to peers
					p.setBlackMask(p.getSenderNode()); // but not to ourselves
					sendPacket(p);
				}
				break;

			  case EventPacket::GLASS_DELETE: /* event deleted */
				owners_it = this->eventdata.find(eventname);
				if (owners_it == this->eventdata.end()) {
					LOG(outlog(), GLASS_WARNING, "Inexistent event " << eventname << " deleted.");
					break;
				}
				set_iter = owners_it->second.owners.find(sender);
				if (set_iter == owners_it->second.owners.end()) {
					LOG(outlog(), GLASS_WARNING, "Node deleted event that it doesn't have.");
					break;
				}
				owners_it->second.owners.erase(set_iter);
				LOG(outlog(), GLASS_INFO, "Node " << sender << " removed event " << eventname);
				if (owners_it->second.owners.empty()) {
					LOG(outlog(), GLASS_DEBUG, "Nobody has " << eventname << " anymore. Deleted.");
					if (owners_it->second.mode == Global) {
						// we are a peer who received the message from a client
						// propagate it to other peers.
						LOG(outlog(), GLASS_INFO, "Notifying peers of delete event " << eventname);

						// updated sender, because peers only see peers. Otherwise we'd keep the original sender id.
						p.setReliable(true);
						p.setSenderNode(this->getGlass()->getNodeId());
						p.setTargetNodes(peersId); // send to peers
						p.setBlackMask(p.getSenderNode()); // but not to ourselves
						sendPacket(p);
					}

					this->eventdata.erase(owners_it);
				}
				break;

			  case EventPacket::GLASS_EVENT: /* event enqueued */
				LOG(outlog(), GLASS_INFO, "Got event " << eventname);
				owners_it = this->eventdata.find(eventname);
				if (owners_it == this->eventdata.end()) {
					LOG(outlog(), GLASS_WARNING, "Inexistent event " << eventname << " enqueued.");
					break;
				}

				p.setReliable(false);
				p.setBlackMask(p.getSenderNode());
				p.addWhiteMask(this->eventdata[eventname].owners);
				sendPacket(p);
				break;

			  default:
				LOG(outlog(), GLASS_ERROR, "Invalid packet in EventBase.");
				break;
			}
		}
		else if (senderType == Server || senderType == None) { // indirect
			switch (ep.getType()) {
			  case EventPacket::GLASS_EVENT: /* event enqueued. */
				LOG(outlog(), GLASS_INFO, "Got event (server) " << eventname);
				events_lock.lock();
				if (events.empty()) {
					LOG(outlog(), GLASS_ERROR, "No events!");
					events_lock.unlock();
					return false;
				}
				events_it = events.find(eventname);
				if (events_it == events.end()) {
					LOG(outlog(), GLASS_ERROR, "Event not found!");
					events_lock.unlock();
					return false;
				}

				LOG(outlog(), GLASS_INFO, "Pushing event " << eventname);
				if (!events_it->second->pushEvent(ep.getEventData(), ep.getSenderNode())) {
					LOG(outlog(), GLASS_ERROR, "Error pushing event");
				}

				events_lock.unlock();
				break;
			  default:
				LOG(outlog(), GLASS_ERROR, "Invalid packet in EventBase.");
			  	return false;
			}
		}
		else {
			LOG(outlog(), GLASS_ERROR, "Invalid senderType in EventBase.");
			return false;
		}
	}
	catch (Exception e) {
		LOG(outlog(), GLASS_ERROR, "Decoding packet in event:" << e.getMessage());
		return false;
	}
	return true;
}

void EventBase::unregisterNode(nodeId id) {
	std::map<string, _EventBaseData >::iterator it;

	for (it = this->eventdata.begin(); it != this->eventdata.end(); it++) {
		struct _EventBaseData &ed(it->second);
		ed.owners.erase(id);
		if (ed.owners.empty()) {
			this->eventdata.erase(it);
		}
	}
}

bool EventBase::registerEvent(_Event *e) {
	if (!e) {
		return false;
	}

	events_lock.lock();
	std::map<const string, _Event *>::iterator it;
	it = events.find(e->name);
	if (it != events.end()) {
		events_lock.unlock();
		return false;
	}
	events[e->name] = e;
	events_lock.unlock();
	return true;
}

bool EventBase::unregisterEvent(_Event *e) {
	if (!e) {
		return false;
	}

	events_lock.lock();

	std::map<const string, _Event *>::iterator it;
	it = events.find(e->name);
	if (it == events.end()) {
		events_lock.unlock();
		return false;
	}
	events.erase(it);
	events_lock.unlock();
	return true;
}

_Event *EventBase::getByName(const string &name) const {
	// TODO: possible race condition with un|registerEvent/getByName/un. Mutex this.
	std::map<const string, _Event *>::const_iterator it;
	it = events.find(name);
	if (it == events.end()) {
		return NULL;
	}
	return it->second;
}


/*
 * Event
 */

_Event::_Event(const char *name, propagationMode mode) throw(Exception) : PluginInterface(mode) {
	if (!name) {
		throw Exception("NULL event name");
	}
	this->name = string(name);

	eb = dynamic_cast<EventBase *>(getPluginBase("EventBase"));
	if (!eb) {
		throw Exception("No Event Base class.");
	}

	if (eb->registerEvent(this) == false) {
		throw Exception("Could not register Event (it probably already exists).");
	}

	/* notify the server */
	EventPacket::EventPacketType ept;
	if (mode == Local)
		ept = EventPacket::GLASS_CREATE_LOCAL;
	else
		ept = EventPacket::GLASS_CREATE_GLOBAL;
	EventPacket		ep(ept, *this);
//	ep.setTargetNode(peersId);
	ep.setTargetNodes(serversId);
	if (eb->sendPacket(ep) == false) {
		eb->unregisterEvent(this);
		throw Exception("Error sending packet.");
	}
}

_Event::~_Event() {
	LOG(outlog(), GLASS_DEBUG, CodeDebug);

	EventPacket		ep(EventPacket::GLASS_DELETE, *this);

//	ep.setTargetNodes(peersId);
	ep.setTargetNodes(serversId);
	if (eb->sendPacket(ep) == false) {
		// we can't do much here, can we?
	}
	eb->unregisterEvent(this);
}

bool _Event::_enqueueEvent(chain &c, nodeId target) {
	EventPacket		ep(EventPacket::GLASS_EVENT, *this);
	LOG(outlog(), GLASS_DEBUG, CodeDebug << "Name: " << name << " to " << NODE_FORMAT(target));

	ep.setReliable(false);
	ep.setEventData(c);
	ep.setTargetNodes(target);
	return eb->sendPacket(ep);
}

bool _Event::_enqueueEvent(chain &c, std::vector<nodeId> &targets) {
	EventPacket		ep(EventPacket::GLASS_EVENT, *this);
	LOG(outlog(), GLASS_DEBUG, CodeDebug << "Name: " << name);
	ep.setReliable(false);
	ep.setEventData(c);
	ep.setTargetNodes(targets);
	return eb->sendPacket(ep);
}

EventSerializable::EventSerializable(const char *name,
		void (*func)(SerializableObject *o) /* = NULL */,
		propagationMode mode)
		throw(Exception) : _Event(name, mode), handler(func) {

}

EventSerializable::~EventSerializable() {

}

bool EventSerializable::pushEvent(chain &c, nodeId sender){
	SerializableObjectManager &som = SerializableObjectManager::getHandle();
	const char *buf = c.begin().data() + 1;

	// get name
	string objname = string(buf);
	LOG(outlog(), GLASS_DEBUG, "Pushing serialized event " << objname);
	SerializableObject *o = som.newInstance(objname);
	if (o) {
		// get data
		LOG(outlog(), GLASS_DEBUG, "name.length() = " << objname.length());
		objname.assign(buf + objname.length()+1, c.length()-(objname.length()+1)-1);
		LOG(outlog(), GLASS_DEBUG, c.length() << "name.length() = " << objname.length());
		o->unpack(chain(objname));
		q.push(o);
		return true;
	}
	LOG(outlog(), GLASS_ERROR, "Could not find ");
	return false;
}

void EventSerializable::clear(void) {
	// todo
}

bool EventSerializable::enqueueEvent(SerializableObject &o, nodeId target) {
	chain c, c2;
	c.append(o.name().data(), o.name().size());
	o.pack(c2);
	c.append(c2);
	return _enqueueEvent(c, target);
}

#ifdef TODO
bool EventSerializable::enqueueEvent(SerializableObject &o,
		std::vector<nodeId> &targets) {
	return false;
}
#endif

SerializableObject *EventSerializable::getEvent(void) throw(EmptyException) {
	if (q.empty()) {
		throw EmptyException();
	}
	SerializableObject *o = q.front();
	q.pop();
	return o;
}

bool EventSerializable::isEmpty(void) {
	return q.empty();
}

bool EventSerializable::setHandler(void (*func)(SerializableObject *o)) {
	this->handler = func;

	// todo: for each event in queue, func(it)
	return true;
}

unsigned int EventSerializable::queueSize(void) {
	return q.size();
}

