/*  Glass - a distributed computing library
    Copyright (C) 2003-2009 Bruno Barberi Gnecco
    Copyright (C) 2009-2010 Corollarium Technologies


    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "_glass.h"
#include "plugin.h"
#include "glassmanager.h"
#include "packet.h"
#include "protocol.h"
#include "types.h"

#include <sstream>
#include <fstream>
#include <vector>
#include <map>
#include <stdlib.h>

using namespace libglass;

Glass::Glass(Protocol *p) throw(Exception) :
		protocol(p),
		id(nobodyId),
		pm(NULL),
		pt_id(boost::this_thread::get_id()) {

	LOG(outlog(), GLASS_DEBUG, CodeDebug);
	if (!p) {
		throw Exception("Null protocol");
	}

	if (GlassManager::getHandle().registerGlass(this) == false) {
		throw Exception("Could not register Glass.");
	}

	this->pm = PluginManagerBase::getHandle().createPluginManager(this);
	if (!this->pm) {
		throw Exception("Could not create plugin manager.");
	}
	
	this->setProtocolOwner(this);
}

Glass::~Glass() {
	/* clean() is a "virtual destructor", called before ~Glass() by the
	derived class (i.e. GlassServer, GlassClient, etc). Do whatever needs
	access to virtual functions there. */
	LOG(outlog(), GLASS_DEBUG, "Deleting protocol.");
	delete protocol;
	// LOG(outlog(), GLASS_DEBUG, "Got out.");
	GlassManager::getHandle().unregisterGlass(this);
}

/* clean() method is a "virtual destructor", called before ~Glass() by the
	derived class (i.e. GlassServer, GlassClient, etc). Do whatever needs
	access to virtual functions there. */
void Glass::clean(void) {
	LOG(outlog(), GLASS_DEBUG, "Cleaning up plugins.");

	// TODO - THIS SHOULD BE AN ATOMIC OPERATION PROTECTED BY
	// Copy pointer to plugin manager
	PluginManager	*pm_temp = pm;
	// Prevent other methods from accessing the plugin manager
	pm = NULL;
	// Now delete the plugin manager
	delete pm_temp;
	pm_temp = NULL; // To prevent misuse of pointer
	// END OF ATOMIC OPERATION
}

void Glass::setProtocolOwner(const Glass *g) {
	this->protocol->setOwner(g);
}

bool Glass::processPacket(Packet &p, bool rootOnly) {
	if (pm == NULL) {
		LOG(outlog(), GLASS_ERROR, "NULL plugin manager.");
		return false;
	}
	return pm->processPacket(p, rootOnly);
}

bool Glass::sendPacket(Packet &p) {
	LOG(outlog(), GLASS_DEBUG, CodeDebug);

	const std::set<nodeId> &destId(p.getTargetNodes());
	const std::set<nodeId> &blackId(p.getBlackMask());
	const std::set<nodeId> &whiteId(p.getWhiteMask());
	std::set<nodeAddress> &destaddr(p.getTargetAdresses());
	destaddr.clear();

	// check if we are broadcasting
	if (destId.find(everybodyId) != destId.end()) {
		AddressIdMap::const_iterator it;
		for (it = addridmap.begin(); it != addridmap.end(); it++) {
			LOG(outlog(), GLASS_DEBUG, CodeDebug << ": " << "EverybodyId: " << NODE_FORMAT(it->first) << "=>" << it->second);

			/*
			 * if there is a white list, only send if node is in white list and not black listed
			 * otherwise, send if node is not black listed
			 */
			if (it->second.isValidFD() &&
				it->second.isDirectRelation() &&
					blackId.find(it->first) == blackId.end() &&
					(whiteId.size() == 0 || whiteId.find(it->first) != whiteId.end())) {
				LOG(outlog(), GLASS_DEBUG, "Adding to packetlist " << it->second);
				destaddr.insert(it->second);
			}
			else {
				LOG(outlog(), GLASS_DEBUG, it->first << "no add " << it->second.isValidFD() << "," <<
						(it->second.isDirectRelation()) << "," <<
						(blackId.find(it->first) == blackId.end()) << "," <<
						(whiteId.size() == 0) << "," <<
						(whiteId.find(it->first) != whiteId.end()));
			}
		}
	}
	else { // no broadcast, individuals
		std::set<nodeId>::iterator 	it;
		for (it = destId.begin(); it != destId.end(); it++) {
			LOG(outlog(), GLASS_DEBUG, CodeDebug << "In list to send: id " << NODE_FORMAT(*it) << ", addr " << IdToAddress(*it));

			if (clientsId == *it) { // send to all clients.
				LOG(outlog(), GLASS_DEBUG, CodeDebug << ": " << "to ClientsId.");
				AddressIdMap::const_iterator it;
				for (it = addridmap.begin(); it != addridmap.end(); it++) {
					if (it->second.relation == Client) {
						LOG(outlog(), GLASS_DEBUG, CodeDebug << ": " << "ClientsId: " << it->first << ": " << it->second);
						// if in white list and not in black list
						if (it->second.isValidFD() &&
							it->second.isDirectRelation() &&
							blackId.find(it->first) == blackId.end() && (whiteId.size() == 0 || whiteId.find(it->first) != whiteId.end())) {
							destaddr.insert(it->second);
						}
					}
				}
			}
			else if (peersId == *it) { // send to all peers.
				LOG(outlog(), GLASS_DEBUG, CodeDebug << ": " << "to peersId.");
				AddressIdMap::const_iterator it;
				for (it = addridmap.begin(); it != addridmap.end(); it++) {
					if (it->second.relation == Peer) {
						LOG(outlog(), GLASS_DEBUG, CodeDebug << ": " << "peersId: " << it->first << ": " << it->second);
						if (it->second.isValidFD() &&
							it->second.isDirectRelation() &&
							blackId.find(it->first) == blackId.end() && (whiteId.size() == 0 || whiteId.find(it->first) != whiteId.end())) {
							destaddr.insert(IdToAddress(it->first));
						}
					}
				}
			}
			else if (serversId == *it) {
				LOG(outlog(), GLASS_DEBUG, CodeDebug << ": " << "to serversId.");
				AddressIdMap::const_iterator it;
				for (it = addridmap.begin(); it != addridmap.end(); it++) {
					if (it->second.relation == Server) {
						LOG(outlog(), GLASS_DEBUG, CodeDebug << ": " << "serversId: " << it->first << ": " << it->second);
						if (it->second.isValidFD() &&
							it->second.isDirectRelation() &&
							blackId.find(it->first) == blackId.end() && (whiteId.size() == 0 || whiteId.find(it->first) != whiteId.end())) {
							destaddr.insert(IdToAddress(it->first));
						}
					}
				}
			}
			else {
				const nodeAddress &addr(this->IdToAddress(*it));
				if (addr.isValidFD() && addr.isDirectRelation() &&
						blackId.find(*it) == blackId.end() && (whiteId.size() == 0 || whiteId.find(*it) != whiteId.end())) {

					LOG(outlog(), GLASS_DEBUG, CodeDebug << "Sending: id " << NODE_FORMAT(*it) << ", addr " << addr);
					if (addr == nobodyAddress) {
						LOG(outlog(), GLASS_ERROR, CodeDebug << "Unknown id/addr: " << NODE_FORMAT(*it) << ", addr " << addr);
						return false;
					}
					else {
						destaddr.insert(addr);
					}
				}
				else {
					LOG(outlog(), GLASS_DEBUG, CodeDebug << "Blacklisting: id " << NODE_FORMAT(*it) << ", addr " << IdToAddress(*it));
				}
			}
		}
	}

#ifdef DETAILED_DEBUG
	LOG(outlog(), GLASS_DEBUG, "======================================");
	for (std::set<nodeId>::const_iterator it(destId.begin()); it != destId.end(); it++) {
		LOG(outlog(), GLASS_DEBUG, *it);
	}
	LOG(outlog(), GLASS_DEBUG, "--------------------------------------");
	for (std::set<nodeAddress>::const_iterator it(destaddr.begin()); it != destaddr.end(); it++) {
		LOG(outlog(), GLASS_DEBUG, *it);
	}

	LOG(outlog(), GLASS_DEBUG, "size"  << p.getTargetAdresses().size());
	LOG(outlog(), GLASS_DEBUG, "======================================");
#endif

	/* if this server created the packet, set the id. Otherwise we are
	 * just forwarding it. */
	if (p.getSenderNode() == nobodyId) {
		p.setSenderNode(id);
	}
	return protocol->send(p);
}


PluginManager *Glass::getPluginManager(void) const { 
	return pm; 
}

PluginBase *Glass::getPlugin(PluginId id) const { 
	return pm->getPlugin(id); 
}

const nodeAddress &Glass::IdToAddress(nodeId id) const {
	lock lock(mutexdata);
	AddressIdMap::const_iterator it;

	it = addridmap.find(id);
	if (it == addridmap.end()) {
		return nobodyAddress;
	}

	return it->second;
}

const nodeAddress Glass::FDToAddress(nodeFD fd) const {
	lock lock(mutexdata);
	AddressIdMap::const_iterator it;

	for (it = addridmap.begin(); it != addridmap.end(); it++) {
		if (it->second.fd == fd) {
			return it->second;
		}
	}
	nodeAddress addr(nobodyAddress);
	addr.fd = fd;
	return addr;
}

nodeId Glass::AddressToId(const nodeAddress &addr) const {
	lock lock(mutexdata);
	AddressIdMap::const_iterator it;

	for (it = addridmap.begin(); it != addridmap.end(); it++) {
		if (it->second == addr) {
			return it->first;
		}
	}
	return nobodyId;
}

nodeId Glass::NickNameToId(const nodeNickName &nick) const {
	lock lock(mutexdata);
	AddressIdMap::const_iterator it;

	for (it = addridmap.begin(); it != addridmap.end(); it++) {
		if (it->second.nickname == nick) {
			return it->first;
		}
	}
	return everybodyId;
}



// Register node to associate it with the glass object
// For example, a GlassServer will register client nodes as they connect
// Registration involves the following steps:
//   Check to see if the node is already registered
//   Add the nodeID to the addridmap map
//   Add the nodeID from the idtorelation map
bool Glass::setNode(nodeAddress &addr, nodeId id, nodeRelation r, bool force /* = false */) {
	lock lock(mutexdata);
	AddressIdMap::iterator it;

	/* first, find to see if they exist */
	it = addridmap.find(id);
	if (it != addridmap.end()) {
		if (force == false) {
			LOG(outlog(), GLASS_DEBUG, "id " << id << " is already registered, skipping" );
			return false; // Node is already registered
		}
		else {
			LOG(outlog(), GLASS_DEBUG, "Replacing id " << id );
			addridmap.erase(it);
		}
	}
	
	/* Node is not registered yet, so we can safely add the node ID
	   and node file descriptor to the map containers */
	addr.relation = r;
	addridmap.insert(std::pair<nodeId, nodeAddress>(id, addr));
	LOG(outlog(), GLASS_DEBUG, CodeDebug << "Adding " << id << " <=> " << addr);

	return true;
}

bool Glass::setRelation(nodeId id, nodeRelation r) {
	lock lock(mutexdata);
	AddressIdMap::iterator idaddr;

	/* first, find to see if id exists */
	idaddr = addridmap.find(id);
	if (idaddr != addridmap.end()) {
		return false;
	}
	nodeAddress a2(idaddr->second);
	// addridmap.replace_data(idaddr, a2);
	return true;
}

// Unregister a node that has been associated with the Glass object
// For example, a GlassServer will unregister the client nodes as they disconnect
// To unregister the node we must complete the following steps:
//   Notify all plugins that the node is going to be unregistered
//   Remove the nodeID from the addridmap.right map
//   Remove the nodeFD from the addridmap.by<Address>() map
//   Remove the nodeID from the idtorelation map
bool Glass::unregisterNode(nodeId id) {
	lock lock(mutexdata);
	AddressIdMap::const_iterator idaddr;

	// Find the file desciptor ID corresponding to node ID
	idaddr = addridmap.find(id);
	// Check if we found the node ID
	if (idaddr == addridmap.end()) {
		return false; // Node ID not found
	}

	/* notify plugins */
	if (pm)
		pm->unregisterNode(id);

	/* unregister */
	addridmap.erase(id);

	return true;
}

nodeId Glass::getNodeId(void) const {
	return this->id;
}

nodeNickName Glass::getNodeNickname(void) const {
	return this->nickname;
}

nodeRelation Glass::getNodeRelation(nodeId id) const {
	lock lock(mutexdata);
	
	AddressIdMap::const_iterator addrid;
	addrid = addridmap.find(id);
	if (addrid == addridmap.end()) {
		return Unregistered;
	}
	return addrid->second.relation;
}

void Glass::setCurrent(void) {
	GlassManager::getHandle().setCurrent(this);
}

void Glass::lockCurrent(void) {
	GlassManager::getHandle().lockCurrent(this);
}

void Glass::unlockCurrent(void) {
	GlassManager::getHandle().unlockCurrent();
}

unsigned int Glass::getTotalNodes(void) {
	lock lock(mutexdata);
	return addridmap.size() + 1; // +1 because this very node is not in the list
}

#ifdef BIMAP
void Glass::printAddrmap(void) const {
	AddressIdMap::const_iterator it;
	LOG(outlog(), GLASS_DEBUG, "Id <=> Address");
	for (it = addridmap.begin(); it != addridmap.end(); it++) {
		LOG(outlog(), GLASS_DEBUG, it->first << " <=> " << it->second);
	}
}
#endif
