/*  Glass - a distributed computing library
    Copyright (C) 2003-2009 Bruno Barberi Gnecco
    Copyright (C) 2009-2010 Corollarium Technologies


    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef _GLASS_MANAGER_HH_
#define _GLASS_MANAGER_HH_

#include "_glass.h"
#include <list>
#include <boost/thread.hpp>
#include <map>
#include <algorithm>

namespace libglass {

/*!
   This class is responsible for keeping a list of all Glasses instances, and
   knowing which one is current.
*/
class GlassManager {
private:
	std::multimap<const boost::thread::id, Glass *>
				glasslist;	/**< List of Glasses. */

	Glass			*current;	/**< current Glass. */
	bool			lock;		/**< lock current. */
	boost::mutex		mutex;

	/*! 
	  Private constructor. This class is a singleton. 
	*/
	GlassManager() : current(NULL), lock(false) { 
		LOG(outlog(), GLASS_DEBUG, CodeDebug);
	}
	
	/*! 
	  Operator= overloading. This class is a singleton.
	*/
	void operator=(GlassManager &);

	/*! 
	  Avoids cloning. This class is a singleton.
	*/
	GlassManager(const GlassManager &);

protected:
	friend class Glass;

	/*! 
	  Returns a handle to this class. This class is a singleton.
	  \return A handle to this class.
	*/
	static GlassManager& getHandle(void) {
		static GlassManager *singleton = new GlassManager();
		return *singleton;
	}

	/*!
	  Registers a new Glass instance.

	  The new glass is set as current, unless some other instance locked
	  itself as current.

	  \param g The new Glass instance.
	  \retval true If the class was set as current.
	  \retval false If some other class locked itself as current. The
	    Glass instance \e is added to the list, so do not consider this
	    an error.
	*/
	bool registerGlass(Glass *g) {
		if (!g) {
			return false;
		}
		mutex.lock();

		// are we registered yet?
		std::multimap<const boost::thread::id, Glass *>::iterator it;
		for (it = glasslist.begin(); it != glasslist.end(); it++) {
			if (it->second == g) {
				mutex.unlock();
				return false;
			}
		}
		glasslist.insert(std::pair<const boost::thread::id, Glass *>(g->pt_id, g));
		if (!lock) {
			current = g;
		}
		mutex.unlock();
		return true;
	}


	/*!
	  Unregisters an existing Glass instance.

	  If the deleted glass is current, current is set to the first glass
	  of the list; if a lock was present in this case, it's removed. To
	  avoid possible logic bugs, never delete the current Glass instance.
	  Undefined behaviour may occur if no Glasses are registered.

	  \param g The Glass instance.
	*/
	void unregisterGlass(Glass *g) {
		mutex.lock();
		std::multimap<const boost::thread::id, Glass *>::iterator it;
		for (it = glasslist.begin(); it != glasslist.end(); it++) {
			if (it->second == g) {
				glasslist.erase(it);
				break;
			}
		}
		if (current == g && it != glasslist.end()) {
			lock = false;
			if (glasslist.empty()) {
				current = NULL;
			}
			else {
				current = glasslist.begin()->second;
			}
		}
		mutex.unlock();
	}

	/*!
	  Sets the given Glass as current.

	  \param g The Glass instance.
	  \retval true Successful.
	  \retval false If null argument, g is not registered or a lock is 
	  in effect.
	*/
	bool setCurrent(Glass *g) {
		mutex.lock();
		if (lock || !g) {
			mutex.unlock();
			return false;
		}

		std::multimap<const boost::thread::id, Glass *>::iterator it = glasslist.find(g->pt_id);
		if (it == glasslist.end()) {
			mutex.unlock();
			return false;
		}

		current = g;
		mutex.unlock();
		return true;
	}

	/*!
	  Returns the current Glass.

	  If no lock is set and being run a multi-threaded environment, this
	  function returns a pointer to the Glass instance associated with the
	  current thread, or the first instance in the list if no association
	  found.

	  \return The current Glass instance.
	  \sa getCurrentGlass.
	*/
	Glass *getCurrent(void) {
		Glass			*g;
		boost::mutex::scoped_lock l(mutex);

		if (lock) {
			g = current;
		}
		else {
			LOG(outlog(), GLASS_DEBUG, CodeDebug << "Accessing glasslist");
			std::multimap<const boost::thread::id, Glass *>::iterator it = glasslist.find(boost::this_thread::get_id());
			if (it == glasslist.end()) {
				g = glasslist.begin()->second;
			}
			else {
				g = it->second;
			}
		}

		return g;
	}

	/*!
	  Locks a Glass as current.

	  \param g The Glass instance.
	  \retval true Successful.
	  \retval false If null argument, g is not registered or a lock is 
	  in effect.
	*/
	bool lockCurrent(Glass *g) {
		// mutex.lock() would lock when setCurrent is called, solve it
		if (setCurrent(g) == false) {
			return false;
		}
		lock = true;
		return true;
	}

	/*!
	  Unlocks current Glass.

	  Has no effect if no lock is in effect.
	*/
	void unlockCurrent(void) {
		mutex.lock();
		lock = false;
		mutex.unlock();
	}

public:

	/*!
	  Returns the current Glass.

	  \return The current Glass instance.
	*/	
	static Glass *getCurrentGlass(void) {
		return getHandle().getCurrent();
	}
};

} //namespace

#endif
