/*  Glass - a distributed computing library
    Copyright (C) 2003-2009 Bruno Barberi Gnecco
    Copyright (C) 2009-2010 Corollarium Technologies


    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "_glass.h"
#include "interprocess.h"
#include "log.h"
#include <boost/bimap.hpp>

namespace libglass {

/**
 * Handles all shared memory. This is a singleton.
 *
 * Interprocess essentially is a wrapper for this class.
 */
class InterprocessSingleton : public MemoryManager {
private:
	/*
	 * Singleton stuff.
	 */
	InterprocessSingleton();
	void operator=(InterprocessSingleton &);
	InterprocessSingleton(const InterprocessSingleton &);

protected:
	boost::interprocess::named_mutex mutex; /// interprocess mutex
	boost::bimap<void *, std::string> shmap; /// map: memory region <-> identifier
	std::map<void *, MemData> shdata; /// map: memory region <-> memory data

public:
	/*!
	  Returns a handle to this class. This class is a singleton.
	  \return A handle to this class.
	*/
	static InterprocessSingleton& getHandle(void) {
		static InterprocessSingleton *singleton = new InterprocessSingleton();
		return *singleton;
	}

	virtual ~InterprocessSingleton();

	void* allocate(size_t nbytes, const std::string &identifier = "");

	void deallocate(void *p);

	Lock scoped_lock(void);

	Lock scoped_lock(MemoryManager &mm);
};


InterprocessLock::InterprocessLock(MemoryManager &mm, boost::interprocess::named_mutex &mutex) : _Lock(mm), mutex(mutex) {
	LOG(outlog(), GLASS_DEBUG, "InterprocessLock");
	mutex.lock();
}

InterprocessLock::~InterprocessLock() {
	LOG(outlog(), GLASS_DEBUG, "~InterprocessLock");
	try {
		mutex.unlock();
	}
	catch (...) {

	}
}

/**
 * Unlocks
 */
void InterprocessLock::unlock(void) {
	mutex.unlock();
}

InterprocessSingleton::InterprocessSingleton() :
	mutex(boost::interprocess::open_or_create, "libglass_interprocess_m") {

	LOG(outlog(), GLASS_DEBUG, "Interprocess");
}

InterprocessSingleton::~InterprocessSingleton() {
	LOG(outlog(), GLASS_DEBUG, "~Interprocess");
}


void* InterprocessSingleton::allocate(size_t nbytes, const std::string &identifier) {
	Lock l(this->scoped_lock());
	void *p; // our memory region
	bool firstalloc = false;

	LOG(outlog(), GLASS_DEBUG, "Allocating sh memory " << identifier << " with " << nbytes << " bytes");

	// have we allocated it already?
	try {
		p = this->shmap.right.at(identifier);
	}
	catch( std::out_of_range & e) {
		firstalloc = true;
	}

	if (!firstalloc) {
		// if (this->shdata[xx].length != nbytes)
		// TODO: check if nbytes is equal or has changed.
	}

	// TODO: we should test the following case:
	// 1. allocate "x"
	// 2. allocate "x" again --> it should work and return the same address, not replicating

	// TODO: we should test the following case:
	// 1. allocate
	// 2. deallocate (but another client still has memory allocated)
	// 3. allocate again

	if (firstalloc) {
		try {
			boost::interprocess::offset_t shmSize = 0;
			MemData md(new _MemData);
			md->shm = new boost::interprocess::shared_memory_object(boost::interprocess::open_or_create, identifier.c_str(), boost::interprocess::read_write);
			md->length = nbytes;

			if (firstalloc) {
				LOG(outlog(), GLASS_DEBUG, "Truncating");
				md->shm->truncate(nbytes);
				md->shm->get_size(shmSize);
			}

			md->region = new boost::interprocess::mapped_region(*md->shm, boost::interprocess::read_write);
			if (firstalloc) {
				this->shmap.insert(boost::bimap<void *, string>::value_type(md->region->get_address(), identifier));
				this->shdata[md->region] = md;
			}

			LOG(outlog(), GLASS_DEBUG, "Allocated: " << md->region->get_address() << " size " << shmSize);
			p = md->region->get_address();
		}
		catch(boost::interprocess::interprocess_exception &e){
			std::cout << e.what() << std::endl;
			return NULL;
		}
	}

	return p;
}

void InterprocessSingleton::deallocate(void *p) {
	LOG(outlog(), GLASS_DEBUG, "Deallocating sh memory");
	Lock l(this->scoped_lock());

	try {
		std::string id(this->shmap.left.at(p));
		boost::interprocess::shared_memory_object::remove(id.c_str());

		this->shmap.left.erase(p);
		this->shdata.erase(p);
	}
	catch( std::out_of_range & e ) {
		LOG(outlog(), GLASS_DEBUG, "deallocating memory with no identifier");
	}
}

Lock InterprocessSingleton::scoped_lock(void) {
	InterprocessLock *l = new InterprocessLock(*this, mutex);
	return Lock(l);
}

Interprocess::Interprocess() : MemoryManager() {
}

Interprocess::~Interprocess() {
}

void* Interprocess::allocate(size_t nbytes, const std::string &identifier) {
	return InterprocessSingleton::getHandle().allocate(nbytes, identifier);
}

void Interprocess::deallocate(void *p) {
	return InterprocessSingleton::getHandle().deallocate(p);
}

Lock Interprocess::scoped_lock() {
	return InterprocessSingleton::getHandle().scoped_lock();
}


}
