/*  Glass - a distributed computing library
    Copyright (C) 2003-2009 Bruno Barberi Gnecco
    Copyright (C) 2009-2010 Corollarium Technologies


    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef _GLASS_LOG_HH_
#define _GLASS_LOG_HH_

#include "glassconfig.h"
#include "types.h"
#include <signal.h>
#include <iostream>
#include <fstream>
#include <sstream>
#include <boost/thread.hpp>

namespace libglass {

#define BREAK_HERE raise(SIGINT) /// breakpoint for GDB. Helpful for debugging.

/*!
  Log base class.
*/
class Log {
public:
	/*!
	  Verbose levels.
	*/
	typedef enum { 
		GLASS_DEBUG,		/**< Debug messages. */
		GLASS_INFO,			/**< Informational messages */
		GLASS_NOTICE,		/**< normal but significant condition. */
		GLASS_WARNING, 		/**< Warning messages. */
		GLASS_ERROR,		/**< Error messages. */
		GLASS_CRITICAL,		/**< Critical conditions */
		GLASS_ALERT,		/**< Action must be taken immediately */
		GLASS_EMERGENCY		/**< System is unusable */
	} verboseLevel;

public:
	boost::mutex		mutex;

protected:
	verboseLevel		level;	/**< The minimum verboseLevel to print. */
	std::ostream		*o;	/**< The output ostream. */

	/*!
	  Converts a verboseLevel to a string.
	  \param l The level.
	  \return The name of that level.
	*/
	inline const char *strlevel(verboseLevel l) const {
		switch (l) {
		  case GLASS_DEBUG:
			return (const char *)"Debug: ";
		  case GLASS_INFO:
			return (const char *)"Info: ";
		  case GLASS_NOTICE:
			return (const char *)"Notice: ";
		  case GLASS_WARNING:
			return (const char *)"WARNING: ";
		  case GLASS_ERROR:
			return (const char *)"ERROR: ";
		  case GLASS_CRITICAL:
			return (const char *)"CRITICAL: ";
		  case GLASS_ALERT:
			return (const char *)"ALERT: ";
		  case GLASS_EMERGENCY:
			return (const char *)"EMERGENCY: ";
		}
		// should never happen
		return NULL;
	}

public:
	Log();
	Log(std::ostream &o);
	/*!
	   Logs to filename. If fifo == true, opens filename as a fifo.
	   \param filename The file name.
	   \param fifo If true, opens the file as a fifo.
	*/
	Log(const char *filename, bool fifo = false);
	Log(const char *hostname, int port);
	virtual ~Log();

	/*!
	   Sets verbose level. Messages with level lower than this will not
	   be printed.
	   \param l Verbose level
	*/
	inline void setLevel(verboseLevel l) { this->level = l; }
	/*!
	   Gets current verbose level.
	   \return The current verbose level
	*/
	inline verboseLevel getLevel(void) const { return level; }
	/*!
	   Compares a given verbose level to the current one and returns true
	   if it is at least as important.
	   \param l Verbose level to compare.
	   \retval true 
	   \retval false
	*/
	inline bool isInLevel(verboseLevel l) { return l >= this->level; }

	/*!
	   Prints an information header, containing date/time, message level,etc.
	   \param l Verbose level
	*/
	virtual void header(verboseLevel l);

	std::string getHeader(verboseLevel l) const;

	/*!
	   Overloads << operator.
	   \param str The std::string.
	   \return A reference for this log.
	*/
 	virtual Log & operator<< (const std::string &str);
 
	/*!
	   Overloads << operator.
	   \param str The character string, ended by \\0.
	   \return A reference for this log.
	*/
 	virtual Log & operator<< (const char *str);
 
	/*!
	   \return 
	*/
	bool lock(void);

	/*!
	   \return 
	*/
	bool unlock(void);
};

/*!
   Default Log instance for Glass output.
   \return A reference to the Log instance.
*/
extern Log &outlog(void);

/*!
   Extends Log for packets. Used by Protocol classes.
*/
class PacketLog : public Log {
public:
	PacketLog();
	PacketLog(std::ostream &o);
	PacketLog(const char *filename, bool fifo = false);
	PacketLog(const char *hostname, int port);
	virtual void header(verboseLevel l);
};

/*!
   A string with the name of the current function and useful extra information.
*/
#define CodeDebug	FUNCTION_NAME << ":" << __LINE__ << " "

/*!
   \def LOG(l, level, message)
   \param l The Log class to use. 
   \param level The message level.
   \param message The message itself. This may be any expression that is
    std::stringstream can process.
*/
#ifdef ENABLE_GLASS_LOG
 #define LOG(l, level, message) \
 	{ \
		boost::mutex::scoped_lock(l.mutex); \
		if (l.isInLevel(libglass::Log::level)) { \
			std::stringstream ___s; \
			___s << message; \
			fprintf(stderr, "%s%s\n", l.getHeader(libglass::Log::level).c_str(), ___s.str().c_str()); \
		} \
 	}
//          l << ___s.str();
#else
 #define LOG(l, level, message)
#endif

}

#endif
