/*  Glass - a distributed computing library
    Copyright (C) 2003-2009 Bruno Barberi Gnecco
    Copyright (C) 2009-2010 Corollarium Technologies


    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "_glass.h"

using namespace libglass;

namespace libglass {

template<>
bool fromString<std::string>(const chain &c, unsigned int offset, 
		std::string &data, unsigned int &used, unsigned int nmemb, 
		void *extra) {
	try {
		data = c.toString(offset);
		used = c.size() - offset;
	}
	catch (Exception e) {
		return false;
	}
	return true;
}

template<>
bool rawPack<std::string>(chain &c, const std::string & s, unsigned int nmemb, 
		void *extra) {
	c.append(s.data(), s.size());
	return true;
}
template<>
bool rawUnpack<std::string>(const chain &c, unsigned int offset, 
		std::string &data, unsigned int &used, unsigned int nmemb, 
		void *extra) {
	try {
		data = c.toString(offset);
		used = c.size() - offset;
	}
	catch (Exception e) {
		return false;
	}
	return true;
}

template<>
bool rawPack<char *>(chain &c, char* const &t, unsigned int nmemb, void *extra) {
	c.append(t, nmemb);
	return true;
}
template<>
bool rawUnpack<char *>(const chain &c, unsigned int offset,
		char * &data, unsigned int &used, unsigned int nmemb,
		void *extra) {
	try {
		LOG(outlog(), GLASS_NOTICE, "Deleting in rawUnpack: " << (void *)data);
		if (data && extra) { // delete
			delete[] data;
			data = NULL;
		}
	}
	catch (Exception e) {
		; // nothin
	}
	try {
		used = c.size() - offset;
		data = c.toArray(offset, used);
	}
	catch (Exception e) {
		return false;
	}
	return true;
}

	
#ifdef DEPRECATED
/*
 * string
 */
template<>
bool fromString<string>(const string &s, string &data, int &used) {
	data = s;
	used = s.length();
	return true;
} 

template<>
bool fromString<string>(const char *s, int length, string &data, int &used) {
	return false;
}

template<>
string toString<string>(const string& t) {
	return t;
} 

template<>
int toString<string>(const string& t, char *&buffer) {
	return 0;
} 

/*
 * short
 */
template<>
bool fromString<short>(const string &s, short &data, int &used) {
// todo: test length
	used = 2;
	data = (short)(s[1] & 0xFF) + ((s[0] << 8) & 0xFF00);
	return true;
} 

template<>
bool fromString<short>(const char *s, int length, short &data, int &used) {
	if (length < 2) {
		return false;
	}
	used = 2;
	data = (short)(s[1] & 0xFF) + ((s[0] << 8) & 0xFF00);
	return true;
}

template<>
string toString<short>(const short& t) {
	char			buf[2];

	buf[1] = (char)(t) & 0xFF;
	buf[0] = (char)((t) >> 8) & 0xFF;
	return string(buf, 2);
} 

template<>
int toString<short>(const short& t, char *&buffer) {
	buffer[1] = (char)(t) & 0xFF;
	buffer[0] = (char)((t) >> 8) & 0xFF;
	return 2;
} 

/*
 * int
 */

template<>
bool fromString<int>(const string &s, int &data, int &used) {
// todo: test length
	used = 4;
	data = (int)(s[3] & 0xFF) + ((s[2] << 8) & 0xFF00) +
		((s[1] << 16) & 0xFF0000) + ((s[0] << 24) & 0xFF000000);
	return true;
} 

template<>
bool fromString<int>(const char *s, int length, int &data, int &used) {
	if (length < 4) {
		return false;
	}
	used = 4;
	data = (int)(s[3] & 0xFF) + ((s[2] << 8) & 0xFF00) +
		((s[1] << 16) & 0xFF0000) + ((s[0] << 24) & 0xFF000000);
	return true;
}

template<>
string toString<int>(const int& t) {
	char			buf[4];

	buf[3] = (char)(t) & 0xFF;
	buf[2] = (char)((t) >> 8) & 0xFF;
	buf[1] = (char)((t) >> 16) & 0xFF;
	buf[0] = (char)((t) >> 24) & 0xFF;
	return string(buf, 4);
} 

template<>
int toString<int>(const int& t, char *&buffer) {
	buffer[3] = (char)(t) & 0xFF;
	buffer[2] = (char)((t) >> 8) & 0xFF;
	buffer[1] = (char)((t) >> 16) & 0xFF;
	buffer[0] = (char)((t) >> 24) & 0xFF;
	return 4;
} 

/*! 
  Float digit encoder.
  \param c The digit.
  \return The encoded digit.
  \sa _digit_decode().
*/
static char _digit_encode(char c) {
	if (c >= '0' && c <= '9')
		return c - '0';
	else if (c == '-')
		return 0xA;
	else if (c == '+')
		return 0xB;
	else if (c == 'e' || c == 'E')
		return 0xC;
	else if (c == '.')
		return 0xD;
	else if (c == ' ')
		return 0xE;
	else {			/* should NEVER happen */
		return 0xF;
	}
}

/*! 
  Float digit decoder.
  \param c The encoded digit.
  \return The digit.
  \sa _digit_decode().
*/
static char _digit_decode(char d) {
	if (d >= 0 && d <= 9)
		return d + '0';
	else if (d == 0xA)
		return '-';
	else if (d == 0xB)
		return '+';
	else if (d == 0xC)
		return 'e';
	else if (d == 0xD)
		return '.';
	else if (d == 0xE)
		return ' ';
	else {			/* should NEVER happen */
		return ' ';
	}
}

template<>
bool fromString<float>(const char *s, int length, float &data, int &used) {
	char			temp[13];
	int 			i;

	if (length < 13) {
		// todo: error
		return false;
	}

	for (i = 0; i < 6; i++) {
		temp[i << 1] = _digit_decode(s[i] & 0xF);
		temp[(i << 1) + 1] = _digit_decode((s[i] >> 4) & 0xF);
	}
	temp[12] = '\0';
	used = 6;
	data = (float)strtod(temp, NULL);
	return true;
}

template<>
bool fromString<float>(const string& s, float &data, int &used) {
	char 			temp[13];
	int 			i;

	for (i = 0; i < 6; i++) {
		temp[i << 1] = _digit_decode(s[i] & 0xF);
		temp[(i << 1) + 1] = _digit_decode((s[i] >> 4) & 0xF);
	}
	temp[12] = '\0';
	used = 6;
	data = (float)strtod(temp, NULL);
	return true;
}

template<>
string toString<float>(const float& t) {
	string				s("000000");
	char				temp[13];
	int				i;

#ifndef WIN32
	snprintf(temp, 13, "%#.5e", t);
#else
	_snprintf(temp, 13, "%#.5e", t);
#endif

	for (i = 0; i < 6; i++) {
		s[i] = (_digit_encode(temp[i << 1])) +
			(_digit_encode(temp[(i << 1) + 1]) << 4);
	}
//	used = 6;
	return s;
}

template<>
int toString<float>(const float& t, char *&buffer) {
	char				temp[13];
	int				i;

	// todo: length of buffer < 6 : return
#ifndef WIN32
	snprintf(temp, 13, "%#.5e", t);
#else
	_snprintf(temp, 13, "%#.5e", t);
#endif

	for (i = 0; i < 6; i++) {
		buffer[i] = (_digit_encode(temp[i << 1])) +
			(_digit_encode(temp[(i << 1) + 1]) << 4);
	}
	
	return 6;
}


#ifdef HAVE_IEEE754_H

char *float754_pack(void *data, int *n) {
	char *buf;
	union ieee754_float f;

	f.f = *(float *)data;
	buf = (char *)malloc(4);
	if (!buf) {
		*n = -1;
		return NULL;
	}
	buf[0] = (f.ieee.negative << 7) | ((f.ieee.exponent >> 1) & 0x7F);
	buf[1] = ((f.ieee.exponent & 0x01) << 7) | ((f.ieee.
						     mantissa >> 16) & 0x7F);
	buf[2] = (f.ieee.mantissa >> 8) & 0xFF;
	buf[3] = (f.ieee.mantissa) & 0xFF;
	*n = 4;
	return buf;
}
void *float754_unpack(char *data, int *n) {
	union ieee754_float f;
	float *g;
	g = (float *)malloc(sizeof(float));
	if (!g)
		return NULL;
	f.ieee.negative = data[0] >> 7;
	f.ieee.exponent = ((data[0] << 1) & 0xFE) + ((data[1] >> 7) & 0x01);
	f.ieee.mantissa = (data[3] & 0xFF) + ((data[2] << 8) & 0xFF00) +
		(((data[1] & 0x7F) << 16) & 0xFF0000);
	*n = 4;
	*g = f.f;
	return g;
}
#endif


#ifdef LATER

/* unsigned_char */
char *unsigned_char_pack(void *data, int *n) {
	char *buf;
	buf = malloc(4);
	if (!buf || !data) {
		*n = -1;
		return NULL;
	}
	memset(buf, 0, 4);
	buf[3] = *(unsigned char *)data;
	*n = 4;
	return buf;
}
void *unsigned_char_unpack(char *data, int *n) {
	unsigned char *c;
	c = (unsigned char *)malloc(1);
	if (!c)
		return NULL;
	*c = data[3];
	*n = 4;
	return (void *)c;
}
void unsigned_char_free(void *data) {
	dice_free(data);
}

/* signed_char */
char *signed_char_pack(void *data, int *n) {
	return unsigned_char_pack(data, n);
}
void *signed_char_unpack(char *data, int *n) {
	return unsigned_char_unpack(data, n);
}
void signed_char_free(void *data) {
	dice_free(data);
}

/* unsigned_short */
char *unsigned_short_pack(void *data, int *n) {
	char *buf;
	unsigned short *s;
	buf = malloc(4);
	if (!buf || !data) {
		*n = -1;
		return NULL;
	}
	memset(buf, 0, 4);
	s = (unsigned short *)data;
	buf[3] = (char)(*s) & 0xFF;
	buf[2] = (char)((*s) >> 8) & 0xFF;
	*n = 4;
	return buf;
}
void *unsigned_short_unpack(char *data, int *n) {
	unsigned short *s;
	s = (unsigned short *)malloc(sizeof(unsigned short));
	if (!s)
		return NULL;
	*s = (unsigned short)(data[3] & 0xFF) + ((data[2] << 8) & 0xFF00);
	*n = 4;
	return (void *)s;
}
void unsigned_short_free(void *data) {
	dice_free(data);
}

/* signed_short */
char *signed_short_pack(void *data, int *n) {
	return unsigned_short_pack(data, n);
}
void *signed_short_unpack(char *data, int *n) {
	return unsigned_short_unpack(data, n);
}
void signed_short_free(void *data) {
	dice_free(data);
}

/* unsigned long */
char *unsigned_long_pack(void *data, int *n) {
	char *buf;
	unsigned long *l;
	l = (unsigned long *)data;
	buf = malloc(4);
	if (!buf || !data) {
		*n = -1;
		return NULL;
	}
	buf[3] = (char)(*l) & 0xFF;
	buf[2] = (char)((*l) >> 8) & 0xFF;
	buf[1] = (char)((*l) >> 16) & 0xFF;
	buf[0] = (char)((*l) >> 24) & 0xFF;
	*n = 4;
	return buf;
}
void *unsigned_long_unpack(char *data, int *n) {
	unsigned long *l;
	l = (unsigned long *)malloc(sizeof(unsigned long));
	if (!l)
		return NULL;
	*l = (unsigned long)(data[3] & 0xFF) + ((data[2] << 8) & 0xFF00) +
		((data[1] << 16) & 0xFF0000) + ((data[0] << 24) & 0xFF000000);
	*n = 4;

	return (void *)l;
}
void unsigned_long_free(void *data) {
	dice_free(data);
}

/* signed_long */
char *signed_long_pack(void *data, int *n) {
	return unsigned_long_pack(data, n);
}
void *signed_long_unpack(char *data, int *n) {
	return unsigned_long_unpack(data, n);
}
void signed_long_free(void *data) {
	dice_free(data);
}

/* double, transfer using printf */
char *double_pack(void *data, int *n) {
	char *buf;
	char temp[23];
	double f;
	int i;

	f = *(double *)data;
	buf = (char *)malloc(12);
	if (!buf) {
		*n = -1;
		return NULL;
	}
#ifndef WIN32
	snprintf(buf, 23, "%#.15e", f);
#else
	_snprintf(buf, 23, "%#.15e", f);
#endif
	for (i = 0; i < 12; i++)
		buf[i] = (_digit_encode(temp[i << 1])) +
			(_digit_encode(temp[(i << 1) + 1]) << 4);
	*n = 12;
	return buf;
}
void *double_unpack(char *data, int *n) {
	double *g;
	char temp[23];
	int i;

	g = (double *)malloc(sizeof(double));
	if (!g)
		return NULL;
	for (i = 0; i < 6; i++) {
		temp[i << 1] = _digit_decode(data[i] & 0xF);
		temp[(i << 1) + 1] = _digit_decode((data[i] >> 4) & 0xF);
	}
	temp[22] = '\0';
	*n = 12;
	*g = (double)strtod(data, NULL);
	return g;
}
void double_free(void *data) {
	dice_free(data);
}

#ifdef HAVE_IEEE754_H
/* float as defined in ieee754. Needs ieee754.h. */
char *float754_pack(void *data, int *n) {
	char *buf;
	union ieee754_float f;

	f.f = *(float *)data;
	buf = (char *)malloc(4);
	if (!buf) {
		*n = -1;
		return NULL;
	}
	buf[0] = (f.ieee.negative << 7) | ((f.ieee.exponent >> 1) & 0x7F);
	buf[1] = ((f.ieee.exponent & 0x01) << 7) | ((f.ieee.
						     mantissa >> 16) & 0x7F);
	buf[2] = (f.ieee.mantissa >> 8) & 0xFF;
	buf[3] = (f.ieee.mantissa) & 0xFF;
	*n = 4;
	return buf;
}
void *float754_unpack(char *data, int *n) {
	union ieee754_float f;
	float *g;
	g = (float *)malloc(sizeof(float));
	if (!g)
		return NULL;
	f.ieee.negative = data[0] >> 7;
	f.ieee.exponent = ((data[0] << 1) & 0xFE) + ((data[1] >> 7) & 0x01);
	f.ieee.mantissa = (data[3] & 0xFF) + ((data[2] << 8) & 0xFF00) +
		(((data[1] & 0x7F) << 16) & 0xFF0000);
	*n = 4;
	*g = f.f;
	return g;
}
void float754_free(void *data) {
	dice_free(data);
}

/* double as defined in ieee754. Needs ieee754.h. */
char *double754_pack(void *data, int *n) {
	char *buf;
	union ieee754_double d;

	d.d = *(double *)data;
	buf = (char *)malloc(8);
	if (!buf) {
		*n = -1;
		return NULL;
	}
	buf[0] = (d.ieee.negative << 7) + ((d.ieee.exponent >> 4) & 0x7F);
	buf[1] = (((d.ieee.exponent) << 4) & 0xF0) +
		((d.ieee.mantissa0 >> 16) & 0x0F);
	buf[2] = (d.ieee.mantissa0 >> 8) & 0xFF;
	buf[3] = (d.ieee.mantissa0) & 0xFF;
	buf[4] = (d.ieee.mantissa1 >> 24) & 0xFF;
	buf[5] = (d.ieee.mantissa1 >> 16) & 0xFF;
	buf[6] = (d.ieee.mantissa1 >> 8) & 0xFF;
	buf[7] = (d.ieee.mantissa1) & 0xFF;
	*n = 8;
	return buf;
}
void *double754_unpack(char *data, int *n) {
	union ieee754_double d;
	double *e;
	e = (double *)malloc(sizeof(double));
	if (!e) {
		return NULL;
	}
	d.ieee.negative = data[0] >> 7;
	d.ieee.exponent = ((data[0] << 4) & 0x7F0) + ((data[1] >> 4) & 0x0F);
	d.ieee.mantissa0 = (data[3] & 0xFF) + ((data[2] << 8) & 0xFF00)
		+ (((data[1] & 0x0F) << 16) & 0xFF0000);
	d.ieee.mantissa1 = (data[7] & 0xFF) + ((data[6] << 8) & 0xFF00)
		+ ((data[5] << 16) & 0xFF0000) +
		((data[4] << 24) & 0xFF000000);
	*n = 8;
	*e = d.d;
	return e;
}
void double754_free(void *data) {
	dice_free(data);
}
#endif

/* string */
char *string_pack(void *data, int *n) {
	char *buf;
	int length, total;

	length = strlen((char *)data);
	total = length + 4 - (length % 4 == 0 ? 4 : length % 4) + 4;	/* make it mod4=0 to conform with RFC1014 */
	buf = (char *)malloc(total);
	if (!buf)
		dice_printf(NULL, "error");
	memset(buf, 0, total);
	buf[3] = (char)(length) & 0xFF;
	buf[2] = (char)((length) >> 8) & 0xFF;
	buf[1] = (char)((length) >> 16) & 0xFF;
	buf[0] = (char)((length) >> 24) & 0xFF;
	memcpy(buf + 4, data, length);
	*n = total;
	return buf;
}
void *string_unpack(char *data, int *n) {
	char *buf;
	int length;
	length = (int)(data[3] & 0xFF) + (((int)data[2] << 8) & 0xFF00) +
		(((int)data[1] << 16) & 0xFF0000) +
		(((int)data[0] << 24) & 0xFF000000);

	buf = malloc(length + 1);
	if (!buf) {
		dice_printf("awer", "%s", strerror(errno));
	}
	memcpy(buf, data + 4, length);
	buf[length] = '\0';
	*n = length + 4 - (length % 4 == 0 ? 4 : length % 4) + 4;
	return (void *)buf;
}
void string_free(void *data) {
	dice_free(data);
}

#endif

#endif //DEPRECATED
}
