/*  Glass - a distributed computing library
    Copyright (C) 2003-2009 Bruno Barberi Gnecco
    Copyright (C) 2009-2010 Corollarium Technologies


    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "packet.h"
#include "types.h"
#include "_glass.h"
#ifndef WIN32
#include <sys/time.h>
#endif

using namespace libglass;


Packet::Packet(Glass *g) : from(nobodyId), data(), pid(), reliable(true), maximumDataLength(0), timeout(0), g(g) {
}

Packet::~Packet() {
}

Glass *Packet::getGlass(void) const {
	return this->g;
}
void Packet::setData(unsigned char *data, unsigned int length, bool free, unsigned char *allocatedbegin) {
	if (!data)
		return;
	this->data.assign((char *)data, length, free, (char *)allocatedbegin);
}

void Packet::setData(const char *data, unsigned int length, bool free, unsigned char *allocatedbegin) {
	if (!data)
		return;
	this->data.assign((char *)data, length, free, (char *)allocatedbegin);
}

void Packet::setData(chain &s) {
	this->data.assign(s);
}

void Packet::appendData(unsigned char *data, unsigned int length, bool free, unsigned char *allocatedbegin) {
	if (!data)
		return;
	this->data.append((char *)data, length, free, (char *)allocatedbegin);
}

void Packet::appendData(const char *data, unsigned int length, bool free, char *allocatedbegin) {
	if (!data)
		return;
	this->data.append((char *)data, length, free, allocatedbegin);
}

void Packet::appendData(chain &s) {
	this->data.append(s);
}

void Packet::appendData(string &s) {
	this->data.append(s);
}

unsigned int Packet::getDataLength(void) const {
	return data.length();
}

chain &Packet::getData(void) {
	return data;
}

void Packet::getData(chain &s) {
	s = data;
}

#ifdef DEPRECATED
unsigned char *Packet::getData(void) const {
	return (unsigned char *)data.toArray();
}
#endif

void Packet::setPluginId(PluginId id) {
// todo: This may have to be changed to fit the future pluginID format.
	this->pid = id;
}

PluginId Packet::getPluginId(void) const {
	return pid;
}

void Packet::setTargetNodes(nodeId id) {
	targetIds.clear();
	targetIds.insert(id);
}

void Packet::setTargetNodes(std::set<nodeId> &id_set) {
	targetIds = id_set;
}

void Packet::setTargetNodes(std::vector<nodeId> &id_set) {
	std::vector<nodeId>::iterator	it;
	for (it = id_set.begin(); it != id_set.end(); it++) {
		targetIds.insert(*it);
	}
}

void Packet::addWhiteMask(nodeId id) {
	whiteIds.insert(id);
}

void Packet::addWhiteMask(std::set<nodeId> &id_set) {
	whiteIds.insert(id_set.begin(), id_set.end());
}

void Packet::setWhiteMask(nodeId id) {
	whiteIds.clear();
	whiteIds.insert(id);
}

void Packet::setWhiteMask(std::set<nodeId> &id_set) {
	whiteIds = id_set;
}


void Packet::addBlackMask(nodeId id) {
	blackIds.insert(id);
}

void Packet::addBlackMask(std::set<nodeId> &id_set) {
	blackIds.insert(id_set.begin(), id_set.end());
}

void Packet::setBlackMask(nodeId id) {
	blackIds.clear();
	blackIds.insert(id);
}

void Packet::setBlackMask(std::set<nodeId> &id_set) {
	blackIds = id_set;
}


void Packet::addTargetNodes(nodeId id) {
	targetIds.insert(id);
}

void Packet::addTargetNodes(std::set<nodeId> &id_set) {
	targetIds.insert(id_set.begin(), id_set.end());
}

std::set<nodeId> & Packet::getTargetNodes(void) {
	return this->targetIds;
}

std::set<nodeId> & Packet::getBlackMask(void) {
	return this->blackIds;
}

std::set<nodeId> & Packet::getWhiteMask(void) {
	return this->whiteIds;
}


void Packet::clearTargetNodes(void) {
	targetIds.clear();
	targetAddresses.clear();
	this->clearBlackMask();
	this->clearWhiteMask();
}

void Packet::clearBlackMask(void) {
	blackIds.clear();
}

void Packet::clearWhiteMask(void) {
	whiteIds.clear();
}

void Packet::setSenderNode(nodeId id) {
	this->from = id;
}

nodeId Packet::getSenderNode(void) const {
	return from;
}

void Packet::setSenderNodeAddress(const nodeAddress &addr) {
	this->fromaddress = addr;
}

const nodeAddress Packet::getSenderNodeAddress(void) const {
	return fromaddress;
}

bool Packet::isReliable(void) const {
	return this->reliable;
}

void Packet::setReliable(bool reliable) {
	this->reliable = reliable;
}

void Packet::setTargetAddresses(const nodeAddress &addr) {
	targetAddresses.clear();
	targetAddresses.insert(addr);
}

void Packet::setTargetAddresses(std::set<nodeAddress> &addrs) {
	targetAddresses = addrs;
}

std::set<nodeAddress> &Packet::getTargetAdresses(void) {
	return this->targetAddresses;
}

chain Packet::encode(unsigned int *len) {
	unsigned char		*w;
	chain			m;
	unsigned int 		length, headerlen, i;

	/* our packet structure is the following:
	   [header length - 2 bytes]
	   [data length - 4 bytes]
	   [sender id - 4 bytes]
	   [target ids - ended by nobodyId]
	   [pluginid - ended by '\0']
	   [data] */

	// remove any 'nobodyId' that may be present.
	for (std::set<nodeId>::const_iterator it(targetIds.begin()); it != targetIds.end();) {
		if (*it == nobodyId) {
			targetIds.erase(it++);
		}
		else {
			++it;
		}
	}
	// make header.
	headerlen = 2 + 4 + 4 + (targetIds.size()+1)*sizeof(nodeId);

	w = new unsigned char[headerlen];	// we don't need to copy the pluginid
	memset(w, 0, headerlen);
	numToChar(from, w+6);			// convert sender id

	// add targets
	i = 10;
	for (std::set<nodeId>::iterator it(targetIds.begin());
			it != targetIds.end(); it++, i += sizeof(nodeId)) {
		if (*it == nobodyId) { // skip or we'll get into trouble
			continue;
		}
		numToChar(*it, w+i);
	}
	numToChar(nobodyId, w+i);
	m.append((char*)w, headerlen, true);
	m.append(pid);
	m.append("\0");

	// header length
	headerlen += pid.length() + 1;
	shortToChar(headerlen, w);

	// append data
	m.append(data);

	/* total length */
	length = m.length();
	numToChar(length, w+2);

	if (len) {
		*len = length;
	}
	return m;
}

bool Packet::decode(unsigned char *buffer, unsigned int length)  {
	chain c;
	c.append((char *)buffer, length, true);
	return decode(c);
}

bool Packet::decode(unsigned char *header, unsigned int headerlen,
		unsigned char *databuf, unsigned int datalen)  {
	unsigned int 		i, l = 0, length = 0;

	/* check for valid parameters */
	if ((length = (headerlen + datalen)) <= 18) {
		LOG(outlog(), GLASS_ERROR, CodeDebug << "Parameter: " << length);
		return false;
	}

	/* get length and check that it's indeed correct */
	l = charToNum(header + 2);
	if (l != length) {
		LOG(outlog(), GLASS_ERROR, CodeDebug << "Length mismatch: " << length << " != " << l);
		return false;
	}

	/* get "from" */
	from = charToNum(header + 6);

	/* get "to" */
	try {
		nodeId			id;

		i = 10;
		id = (nodeId)charToNum(header + i);
		targetIds.clear();
		while (id != nobodyId) {
			targetIds.insert(id);
			i += 4;
			id = (nodeId)charToNum(header + i);
		}
		i += 4;

		/* get PluginId */
		this->pid = PluginId((char *)header + i, headerlen - i - 1);

		/* last but not least, data */
		this->data.assign((char *)databuf, datalen, true);
	}
	catch (Exception e) {
		LOG(outlog(), GLASS_ERROR, CodeDebug << "Caught exception parsing packet");
		return false;
	}

	return true;
}

bool Packet::decode(chain &buffer) {
	unsigned int 		i, l = 0, length = 0;

	/* this code is duplicated from decode(char, int, char, int).
	 * I don't like to duplicate code, but there's a significant
	 * simplification in memory management.
	 */

	/* check for valid parameters */
	if ((length = buffer.size()) <= 18) {
		LOG(outlog(), GLASS_ERROR, CodeDebug << "Parameter: " << length);
		return false;
	}

	/* get length and check that it's indeed correct */
	l = buffer.toInt(2);
	if (l != length) {
		LOG(outlog(), GLASS_ERROR, CodeDebug << length << "!=" << l);
		return false;
	}

	/* get "from" */
	from = buffer.toInt(6);

	/* get "to" */
	try {
		nodeId			id;
		i = 10;
		id = (nodeId)buffer.toInt(i);
		targetIds.clear();
		while (id != nobodyId) {
			targetIds.insert(id);
			i += 4;
			id = (nodeId)buffer.toInt(i);
		}
		i += 4;

		/* get PluginId */
		l = buffer.find('\0', i);
		pid = buffer.toString(i, l-1);

		/* last but not least, data */
		data = buffer.substr(l+1, chain::npos);
	}
	catch (Exception e) {
		LOG(outlog(), GLASS_ERROR, CodeDebug << "Caught exception parsing packet");
		return false;
	}

	return true;
}


std::ostream& Packet::print(std::ostream &o) {
	double timestamp = 0;
#ifdef WIN32

#else
	struct timeval detail_time;
	gettimeofday(&detail_time,NULL);
	timestamp = detail_time.tv_sec + ((double)detail_time.tv_usec/1000000.0);
#endif
	o << "Packet: [PID=" << this->pid
		<< "; Time=" << timestamp
		<< "; DATA=\"" << (((unsigned int)(this->data[0])) & 0xFF)
		<< "\"; SenderId=" << this->getSenderNode()
		<< "; TargetId = ";

	// Duplicate the target node ID set
	std::set<nodeId> 	&dest = this->getTargetNodes();

	if (dest.empty())
		o << "NULL";
	else {
		// Set iterator
		std::set<nodeId>::iterator it;
		// Iterate over all targets IDs in the set
		it = dest.begin();
		while (it != dest.end()) {
			// Copy current target node ID
			nodeId targetid = *it;
			// Output target node ID
			o << targetid;
			// Advance iterator
			it++;
			// Add comma seperator
			if (it != dest.end())
				o << ",";
		} // while
	} // if

	o << "]";

	return o;
}


namespace libglass {

std::ostream& operator<<(std::ostream &o, Packet &p) {
	return p.print(o);
}

}
