/*  Glass - a distributed computing library
    Copyright (C) 2003-2009 Bruno Barberi Gnecco
    Copyright (C) 2009-2010 Corollarium Technologies


    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "_glass.h"
#include "packetclientserver.h"
#include <map>

#define MAX_NICKNAME_LEN 256

using namespace libglass;

ClientServerPacket::ClientServerPacket(Packet &p, bool autostart) throw(Exception) : Packet(p.getGlass()) {
	if (autostart && !this->parse(p)) {
		throw Exception("Inexistent message type");
	}
}

ClientServerPacket::ClientServerPacket(Glass *g, int t) throw(Exception) :
		Packet(g), type(t), id(nobodyId), master(nobodyId) {
}

ClientServerPacket::ClientServerPacket(Glass *g) throw(Exception) : Packet(g) {
}

ClientServerPacket::~ClientServerPacket() {
}

bool ClientServerPacket::parse(Packet &p) {
	chain		c;
	unsigned int 	number = 0, i, j;

	LOG(outlog(), GLASS_DEBUG, CodeDebug);

	p.getData(c);

	type = (ClientServerPacketType)c[0];
	switch(type) {
	  case ID:
		this->id = c.toInt(1);
		break;
	  case HELLO:
		if (c.length() > 2)
			this->nickname = c.toString(1, c.length()-1);
		break;
	  case SET_MASTER:
		break;
	  case QUIT_ALL:
		; // nothing
		break;
	  case BYE:
		; // nothing
		break;
	  case NODE_LIST:
		number = c.toInt(1);
		this->master = c.toInt(5);
		LOG(outlog(), GLASS_DEBUG, "master: " << master << "; size= " << number);

		for (i = 0, j = 9; i < number; i++, j += sizeof(nodeId) + MAX_NICKNAME_LEN) {
			nodeId id = c.toInt(j);
			this->nodelist[id] = c.toStringNull(j + sizeof(nodeId), j + sizeof(nodeId) + MAX_NICKNAME_LEN);
		}
		break;
	  default:
		return false;
	}
	return true;
}

ClientServerPacket::ClientServerPacketType ClientServerPacket::getType(void) const {
	return static_cast<ClientServerPacketType>(this->type);
}

nodeId ClientServerPacket::getNodeId(void) const {
	return this->id;
}

void ClientServerPacket::setNodeId(nodeId id) {
	this->id = id;
}

const std::map<nodeId, nodeNickName> &ClientServerPacket::getNodeList(void) const {
	return nodelist;
}

void ClientServerPacket::setNodeList(nodeId master, const std::map<nodeId, nodeNickName> &l) {
	this->master = master;
	this->nodelist = l;
}

/*
bool ClientServerPacket::setNodeList(std::map<nodeId, nodeRelation> &idtorelation) {
	return true;
}
*/
chain ClientServerPacket::encode(unsigned int *len) {
	unsigned int size = 1;

	data[0] = (unsigned char)this->type;

	switch(type) {
	  case ID:
		numToChar(id, data+1);
		size += 4;
		setData(data, size);
		break;
	  case HELLO:
		setData((char *)&data, 1);
		appendData(nickname);
		break;
	  case QUIT_ALL:
		setData(data, size);
		break;
	  case BYE:
		setData(data, size);
		break;
	  case NODE_LIST:
		{
			LOG(outlog(), GLASS_DEBUG, "node list");

			unsigned int listsize = nodelist.size();
			numToChar(listsize, data+1);
			size += 4;
			setData(data, size);

			// master			
			numToChar(master, data+5);
			size += 4;
			setData(data, size);
			
			LOG(outlog(), GLASS_DEBUG, "master: " << id << "; size= " << listsize);

			chain c;
			char *nodedata, *t;
			// TODO: static is ugly, use dynamic hostname buffer size
			unsigned int datasize = listsize*(sizeof(nodeId) + MAX_NICKNAME_LEN);
			nodedata = new char[datasize];
			memset(nodedata, 0, datasize);

			LOG(outlog(), GLASS_INFO, "Sending NODE_LIST. size=" << nodelist.size());

			t = nodedata;
			for (std::map<nodeId, nodeNickName>::const_iterator it(nodelist.begin()); it != nodelist.end(); it++ ) {
				LOG(outlog(), GLASS_INFO, "Adding to NODE_LIST. nodeId " << it->first << ", nickname " << it->second);

				// first, the node id
				numToChar(it->first, t);
				t += sizeof(nodeId);
				int nicklen = it->second.length();
				if (nicklen > MAX_NICKNAME_LEN) {
					nicklen = MAX_NICKNAME_LEN-1;
				}
				strncpy(t, it->second.c_str(), nicklen);
				t += MAX_NICKNAME_LEN;
			}
			this->appendData(nodedata, datasize, true);
			appendData(c);
		}
		break;
	  default:
	  	// throw Exception("Inexistent message type");
		break;
	}
	this->setPluginId(rootBaseId);
	return Packet::encode(len);
}

std::ostream& ClientServerPacket::print(std::ostream &o) {
	Packet::print(o);
	o << "type: " << type;
	return o;
}
