/*  Glass - a distributed computing library
    Copyright (C) 2003-2009 Bruno Barberi Gnecco
    Copyright (C) 2009-2010 Corollarium Technologies


    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "_glass.h"
#include "packetpeer.h"
#include <map>

#define MAX_HOSTNAME_LEN 256

using namespace libglass;

PeerPacket::PeerPacket(Packet &p, bool autostart) throw(Exception) : ClientServerPacket(p, false) {
	if (autostart && !this->parse(p)) {
		throw Exception("Inexistent message type");
	}
}

PeerPacket::PeerPacket(Glass *g, int t) throw(Exception) : ClientServerPacket(g, t) {
	this->type = static_cast<ClientServerPacket::ClientServerPacketType>(t);
	this->id = nobodyId;
}

PeerPacket::~PeerPacket() {
}

bool PeerPacket::parse(Packet &p) {
	LOG(outlog(), GLASS_DEBUG, CodeDebug);
	chain		c;
	unsigned int 	number = 0, i, j;

	p.getData(c);
	type = (((int)(c[0])) & 0xFF); // ahh... C++ is so pretty....

	LOG(outlog(), GLASS_DEBUG, "Got peerpacket parse type " << type);

	switch(type) {
	case PEER_HELLO:
	case PEER_START:
		{
			this->node.setPort(c.toInt(1));
			this->node.host = c.toString(5, chain::npos);
			LOG(outlog(), GLASS_INFO, "Got PEER_HELLO (" << this->node.port << "," << this->node.host);
		}
		break;
	case PEER_LIST:
		number = c.toInt(1);
		for (i = 0, j = 5; i < number; i++, j += sizeof(nodeId) + sizeof(uint32_t) + MAX_HOSTNAME_LEN) {
			nodeId id = c.toInt(j);
			unsigned int port = c.toInt(j + sizeof(nodeId));
			std::string host = c.toString(j + sizeof(nodeId) + sizeof(uint32_t));
			peerlist[id] = nodeAddress(host, port);
		}
		break;
	case PEER_CHANGE_ID:
		this->id = c.toInt(1);
		;
	default:
		return ClientServerPacket::parse(p);
	}
	return true;
}

chain PeerPacket::encode(unsigned int *len) {
	unsigned int size = 1;

	data[0] = (unsigned char)this->type;

	switch(this->type) {
	case PEER_START:
	case PEER_HELLO: {
		LOG(outlog(), GLASS_DEBUG, "Peer send packet type " << this->type);
		unsigned char *tmp = new unsigned char[512];
		tmp[0] = (unsigned char)this->type;
		numToChar(this->node.port, tmp+1);
		unsigned int l = this->node.host.length();
		strncpy((char *)tmp+5, this->node.host.c_str(), l);
		setData(tmp, 5+l, true);
		break;
	}
	case PEER_LIST:
		{
			// first, the list size
			unsigned int listsize = peerlist.size();
			numToChar(listsize, data+1);
			size += 4;
			setData(data, size);

			chain c;
			char *nodedata, *t;
			// TODO: static is ugly, use dynamic hostname buffer size
			unsigned int datasize = listsize*(sizeof(nodeId) + sizeof(uint32_t) + MAX_HOSTNAME_LEN);
			nodedata = new char[datasize];
			memset(nodedata, 0, datasize);

			LOG(outlog(), GLASS_INFO, "Sending PEER_LIST. size=" << peerlist.size());

			t = nodedata;
			for (std::map<nodeId, nodeAddress>::iterator it(this->peerlist.begin()); it != peerlist.end(); it++ ) {
				LOG(outlog(), GLASS_INFO, "Adding to PEER_LIST. nodeId " << it->first << ", port" << it->second.port << ", hostname " << it->second.host);

				// first, the node id
				numToChar(it->first, t);
				t += sizeof(nodeId);
				numToChar(it->second.port, t);
				// second, the port (4 bytes)
				t += sizeof(uint32_t);
				int hostlen = it->second.host.length();
				if (hostlen > MAX_HOSTNAME_LEN) {
					hostlen = MAX_HOSTNAME_LEN-1;
				}
				strncpy(t, it->second.host.c_str(), hostlen);
				t += MAX_HOSTNAME_LEN;
			}
			c.append(nodedata, datasize, true);
			appendData(c);
		}
		break;
	case PEER_CHANGE_ID:
		numToChar(this->id, data+1);
		setData(data, 1 + sizeof(nodeId));
		break;
	default:
		return ClientServerPacket::encode(len);
		break;
	}
	this->setPluginId(rootBaseId);
	return Packet::encode(len);
}

std::ostream& PeerPacket::print(std::ostream &o) {
	Packet::print(o);
	o << "type: " << type;
	return o;
}
