/*  Glass - a distributed computing library
    Copyright (C) 2003-2009 Bruno Barberi Gnecco
    Copyright (C) 2009-2010 Corollarium Technologies


    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "plugin.h"
#include "_glass.h"
#include "packet.h"
#include <map>

using namespace libglass;

/*
 * PluginManagerBase
 */
PluginManagerBase::PluginManagerBase() { 
	LOG(outlog(), GLASS_DEBUG, CodeDebug);
}

PluginManager *PluginManagerBase::createPluginManager(Glass *g) {
	std::map<PluginId, PluginBase *> 	*plugins =
			new std::map<PluginId, PluginBase *>;
	
	LOG(outlog(), GLASS_DEBUG, CodeDebug);

	/* the factory */
	PluginManager *pm = new PluginManager(g, plugins);

	/* get a copy of the plugin map, with fresh new instances inside */
	for (std::map<PluginId, PluginBase *>::iterator it = pluginmap.begin();
			it != pluginmap.end(); it++) {
		PluginBase *pb = it->second->instance(pm);
		if (!pb) {
			LOG(outlog(), GLASS_ERROR, "NULL plugin instantiation: " << it->first);
		}
		else {
			(*plugins)[it->first] = pb;
		}
	}

	return pm;
}

PluginManagerBase::~PluginManagerBase() {
	LOG(outlog(), GLASS_DEBUG, CodeDebug);
	pluginmap.clear();
}

bool PluginManagerBase::registerPlugin(PluginBase *base) {
	LOG(outlog(), GLASS_DEBUG, CodeDebug);

	if (!base) {
		LOG(outlog(), GLASS_ERROR, CodeDebug << ": " << "NULL base.");
		return false;
	}
	
	if (pluginmap.find(base->getPluginId()) != pluginmap.end()) {
		LOG(outlog(), GLASS_ERROR, CodeDebug << ": " << "Duplicated id.");
		return false;
	}
	{ 
		PluginId e(base->getPluginId());
		LOG(outlog(), GLASS_DEBUG, "Registering plugin " << e);
		pluginmap[e] = base;
	}

	return true;
}

PluginManagerBase& PluginManagerBase::getHandle(void) {
	static PluginManagerBase *singleton = new PluginManagerBase();

	return *singleton;
}

/*
 * Plugin Manager 
 */
PluginManager::PluginManager(Glass *g, 
		std::map<PluginId, PluginBase *> *pluginmap) :
		pluginmap(pluginmap), glass(g) {
	LOG(outlog(), GLASS_DEBUG, CodeDebug);
}

PluginManager::~PluginManager() {
	LOG(outlog(), GLASS_DEBUG, CodeDebug);

	std::map<PluginId, PluginBase *>::iterator it = pluginmap->begin();
	while (it != pluginmap->end()) {
		delete (*it++).second;
	}

	delete pluginmap;
}

bool PluginManager::processPacket(Packet &p, bool rootOnly) {
	std::map<PluginId, PluginBase *>::iterator it;

	// LOG(outlog(), GLASS_DEBUG, "PM starting to process packet:" );

	// Compare plugin ID in the packet to the list
	// of plugins that we have registered.

	if (pluginmap == NULL) {
		// this may happen during destruction of the class.
		return false;
	}
	/* now let the plugin handle it */
	it = pluginmap->find(p.getPluginId());
	if (it == pluginmap->end()) {
		LOG(outlog(), GLASS_DEBUG, "Waiting for \"I.");
		// If the plugin ID contained in the packet does not
		// match any of the registered plugins then
		// we may have a message to the glass class itself
		if (p.getPluginId() == rootBaseId) {
			LOG(outlog(), GLASS_INFO, "Redirecting packet received from node " << NODE_FORMAT p.getSenderNode() << " to root glass class for processing");
			return glass->rootPacket(p); // Let glass class process the packet
		}

		// If reached here then plugin ID in packet is not recognised
		LOG(outlog(), GLASS_ERROR, "Packet uses unregistered plugin ID \"" << p.getPluginId() << "\" - Ignoring");
		LOG(outlog(), GLASS_ERROR, "Perhaps " << rootBaseId.length() << "," << p.getPluginId().length());
		return false; // Unrecognised plugin ID
	}
	else if (rootOnly == true) {
		return false;
	}

	// If reached here then we recognise the plugin ID
	LOG(outlog(), GLASS_INFO, "Redirecting packet received from node " << NODE_FORMAT p.getSenderNode() << " to plugin " << p.getPluginId() << " for processing.");
	PluginBase *pb = it->second;
	return pb->processPacket(p); // Pass packet to plugin for processing
}

bool PluginManager::sendPacket(Packet &p) {
	return getGlass()->sendPacket(p);
}

void PluginManager::unregisterNode(nodeId id) {
	std::map<PluginId, PluginBase *>::iterator it;

	for (it = pluginmap->begin(); it != pluginmap->end(); it++) {
		(it->second)->unregisterNode(id);
	}
}

PluginBase *PluginManager::getPlugin(PluginId id) const {
	std::map<PluginId, PluginBase *>::iterator it;

	it = pluginmap->find(id);
	if (it == pluginmap->end()) {
		LOG(outlog(), GLASS_ERROR, CodeDebug << " no pluginid: " << id << "(" << id.size() << ")");
		for (it = pluginmap->begin(); it != pluginmap->end(); it++) {
			LOG(outlog(), GLASS_ERROR, CodeDebug << it->first <<"(" << it->first.size() << ")");
		}
		return NULL;
	}
	return it->second;
}

Glass *PluginManager::getGlass(void) const {
	return glass;
}
void PluginManager::printPlugins(void) {
	std::map<PluginId, PluginBase *>::iterator it;

	LOG(outlog(), GLASS_ERROR, CodeDebug << " Plugins: ");
	for (it = pluginmap->begin(); it != pluginmap->end(); it++) {
		LOG(outlog(), GLASS_ERROR, "\t" << it->first <<"(" << it->first.size() << ")");
	}
}


/*
 * PluginBase
 */

PluginBase::PluginBase(const PluginId id, bool *registered) : id(id) { 
	LOG(outlog(), GLASS_DEBUG, CodeDebug << "=====\"" << id << "\" : " << id.size());
	if (*registered == false) {
		PluginManagerBase::getHandle().registerPlugin(this);
		*registered = true;
	}
}

PluginBase::~PluginBase() { 
	LOG(outlog(), GLASS_DEBUG, CodeDebug);
}

PluginId PluginBase::getPluginId(void) const {
	return id;
}

nodeRelation PluginBase::getNodeRelation(nodeId id) const {
	if (pm_parent)
		return pm_parent->getGlass()->getNodeRelation(id);
	return Unregistered;
}

bool PluginBase::sendPacket(Packet &p) {
	p.setPluginId(this->id);
	return pm_parent->sendPacket(p);
}

Glass *PluginBase::getGlass(void) const {
	if (pm_parent)
		return pm_parent->getGlass();
	return NULL;
}

/*
 * PluginInterface
 */

PluginInterface::PluginInterface(propagationMode mode) : g(GlassManager::getCurrentGlass()), mode(mode) {
	LOG(outlog(), GLASS_DEBUG, CodeDebug << " PluginInterface()");
	if (!g) {
		throw Exception("Null glass manager in PluginInterface");
	}
}

PluginInterface::~PluginInterface() {
	;
}

PluginBase *PluginInterface::getPluginBase(PluginId id) {
	LOG(outlog(), GLASS_DEBUG, CodeDebug << this);
	const PluginManager *pm = PluginInterface::getPluginManager();
	if (!pm) {
		LOG(outlog(), GLASS_ERROR, CodeDebug << "NULL pm");
		return NULL;
	}
	return pm->getPlugin(id);
}

propagationMode PluginInterface::getMode(void) const {
	return this->mode;
}

const PluginManager *PluginInterface::getPluginManager(void) {
	Glass *gg = GlassManager::getCurrentGlass();
	LOG(outlog(), GLASS_DEBUG, CodeDebug << gg);
	if (!gg) {
		LOG(outlog(), GLASS_ERROR, CodeDebug << gg);
		return NULL;
	}
	LOG(outlog(), GLASS_DEBUG, CodeDebug << gg);
	PluginManager *pm = gg->getPluginManager();
	if (!pm) {
		return NULL;
	}
	return pm;
}

const Glass *PluginInterface::getGlass(void) const {
	return this->g;
}
