/*  Glass - a distributed computing library
    Copyright (C) 2003-2009 Bruno Barberi Gnecco
    Copyright (C) 2009-2010 Corollarium Technologies


    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "remotevar.h"
#include "log.h"
#include "packet.h"
#include "_glass.h"

using namespace libglass;


namespace libglass {

/*!
  Format of Remote Plugin packets:

  \li [action (1 byte)][variable name (ended by \\0)][data]

  where valid action commands are described below.

  These are valid actions in client->server packets:

  \li GLASS_CREATE = created a remote variable.
  \li GLASS_DELETE = deleted a variable.
  \li GLASS_SET = set a variable value on a remote instance.
  \li GLASS_GET = get a variable value on a remote instance.

  These are valid actions in server->client packets:

  \li GLASS_SET = changed variable value.
  \li GLASS_VALUE = changed variable value.
*/
class RemotePacket : public Packet {
public: 
	typedef enum  {
		GLASS_CREATE = 0, GLASS_DELETE, GLASS_SET, GLASS_GET, GLASS_VALUE
	} RemotePacketType;

private:
	RemotePacketType	type;
	string			remotename;
	chain			remotedata;
	_Remote		*r;
	char			iddata[4];
	nodeId			targetid;

public:
	RemotePacket(Packet &p) throw(Exception) : Packet(p.getGlass()), r(NULL) {
		chain		c;
		unsigned int	end;
		p.getData(c);

		type = (RemotePacketType)c[0];
		end = c.find('\0', 1);
		remotename = c.toString(1, end-1);
		if (type == GLASS_SET || type == GLASS_VALUE) {
			targetid = (nodeId)c.toInt(end+1);
			remotedata = c.substr(end+1, chain::npos);
		}
		else if (type == GLASS_GET) {
			targetid = (nodeId)c.toInt(end+1);
		}
	}

	RemotePacket(RemotePacketType type, _Remote &remote) :
		Packet(remote.rb->getGlass()), type(type), r(&remote)  {
	}

	virtual ~RemotePacket() {
	}

	RemotePacketType getType(void) const {
		return this->type;
	}

	void setType(RemotePacketType type) {
		this->type = type;
	}

	string &getRemoteName(void) {
		if (r) {
			return r->name;
		}
		else {
			return remotename;
		}
	}

	chain & getRemoteData(void) {
		return remotedata;
	}

	void setRemoteData(const chain &c) {
		this->remotedata = c;
	}

	chain encode(unsigned int *len = NULL) {
		setData((char *)&type, 1);
		if (!r) {
			LOG(outlog(), GLASS_ERROR, CodeDebug << "Encoding packet");
			// throw Exception()
			return chain();
		}
		appendData(r->name);
		appendData("\0", 1);
		if (type == GLASS_SET || type == GLASS_VALUE) {
			appendData(this->remotedata);
		}
		return Packet::encode(len);
	}

	std::ostream& print(std::ostream &o) {
		Packet::print(o);
		return o;
	}
};

} // namespace

// PLUGIN_CONSTRUCTOR macro is defined in plugin.h
// The expanded code ensures that the plugin automatically registers itself
// with the PluginManager object during runtime initialisation. i.e. before
// calling main()
// This registration is achieved by the following operations:
//   Runtime initialisation of class static member RemoteBase::registered
//   Runtime initialisation of class static member RemoteBase::singleton 
//   Definition of default class constuctor
PLUGIN_CONSTRUCTOR(RemoteBase) {
	LOG(outlog(), GLASS_DEBUG, CodeDebug);
}

RemoteBase::~RemoteBase() {
	LOG(outlog(), GLASS_DEBUG, CodeDebug);
	std::map<const string, _Remote *>::iterator it;
	
	it = vars.begin();
	while (it != vars.end()) {
		delete it->second;
		it++;
	}

	vars.clear();
}

void RemoteBase::synchronize(nodeId id, nodeRelation r) {
	// TODO
	return;
}

bool RemoteBase::processPacket(Packet &p) {
	nodeRelation			senderType = getNodeRelation(p.getSenderNode());
	unsigned int			used;
	string				varname;
	_Remote 			*r;
	std::set<nodeId>::iterator	set_iter;
	std::map<const string, _Remote *>::iterator it;

	try {
		RemotePacket		rp(p);
		string				&varname = rp.getRemoteName();
		it = vars.find(varname);

		if (senderType == Client || senderType == Peer) {
			switch (rp.getType()) {
			  case RemotePacket::GLASS_CREATE: /* created */
				if (it != vars.end()) {
					// variable already exists.
					// do nothing.
					r = it->second;
				}
				else {
					try {
						r = new _Remote(varname);
					}
					catch (Exception e) {
						LOG(outlog(), GLASS_ERROR, e.getMessage());
						return false;
					}
					LOG(outlog(), GLASS_DEBUG, "Created new remote variable " << varname);
				}
				r->owners.insert(p.getSenderNode());

				if (senderType == Client) {
					// we are a peer who received the message from a client
					// propagate it to other peers.
					LOG(outlog(), GLASS_INFO, "Notifying peers of remote " << varname);
					 // updated sender, because peers only see peers. Otherwise we'd keep the original sender id.
					p.setReliable(true);
					p.setSenderNode(this->getGlass()->getNodeId());
					p.setTargetNodes(peersId); // send to peers
					p.setBlackMask(p.getSenderNode()); // but not to ourselves
					sendPacket(p);
				}

				break;
			  case RemotePacket::GLASS_DELETE: /* deleted */
				if (it == vars.end()) {
					LOG(outlog(), GLASS_WARNING, "Inexistent variable " << varname << " deleted.");
					break;
				}
				r = it->second;
				set_iter = r->owners.find(p.getSenderNode());
				if (set_iter == r->owners.end()) {
					LOG(outlog(), GLASS_WARNING, "Node deleted variable that it doesn't have.");
					break;
				}
				r->owners.erase(set_iter);
				LOG(outlog(), GLASS_DEBUG, "Node " << NODE_FORMAT p.getSenderNode() << " removed variable " << varname);
				if (r->owners.empty()) {
					LOG(outlog(), GLASS_DEBUG, "Nobody has " << varname << " anymore. Deleted.");
					p.setSenderNode(this->getGlass()->getNodeId());
					p.setTargetNodes(peersId); // send to peers
					p.setBlackMask(p.getSenderNode()); // but not to ourselves
					sendPacket(p);
					vars.erase(it);
					delete r;
				}
				break;
			  case RemotePacket::GLASS_SET: /*  */
				LOG(outlog(), GLASS_DEBUG, "GLASS_SET from " << NODE_FORMAT p.getSenderNode());
				if (it == vars.end()) {
					LOG(outlog(), GLASS_WARNING, "Update for a variable " << varname << " that doesn't exist.");
					break;
				}
				// variable already exists, we're updating
				r = it->second;

				// r->lastchanged = p.getSenderNode();
				LOG(outlog(), GLASS_INFO, "Updated variable " << varname << NODE_FORMAT p.getSenderNode());

				p.setReliable(true); // TODO: user should set this.
				p.setBlackMask(p.getSenderNode());
				sendPacket(p);

				if (senderType == Client) { // propagate to every peer we know.
					LOG(outlog(), GLASS_INFO, "Propagating  " << varname << " to peers");
					std::set<nodeId> targets; // copy
					for (set_iter = r->owners.begin(); set_iter != r->owners.end(); set_iter++) {
						if (this->getGlass()->getNodeRelation(*set_iter) == Peer) {
							targets.insert(*set_iter);
						}
					}
					p.setTargetNodes(targets);
					p.setBlackMask(p.getSenderNode());
					sendPacket(p);
				}
				break;
			  case RemotePacket::GLASS_GET: /*  */
				  // TODO
				break;
			  case RemotePacket::GLASS_VALUE: /*  */
				  // TODO
				break;
			  default:
				LOG(outlog(), GLASS_ERROR, "Invalid packet in SharedBase.");
				return false;
			}
		}
		else if (senderType == Server || senderType == None) {
			switch (rp.getType()) {
			  case RemotePacket::GLASS_SET: /*  */
				if (it == vars.end()) {
					// variable doesn't exist. This shouldn't happen.
					LOG(outlog(), GLASS_WARNING, "Got update of a variable that I don't have.");
					break;
				}
				// TODO: mutex
				r = it->second;
				if (r->unpack(rp.getRemoteData(), used) == false) {
					LOG(outlog(), GLASS_ERROR, "Could not unpack variable " << varname);
				}
				// trigger.post()
				break;
			  case RemotePacket::GLASS_GET: /*  */
				// TODO
				break;
			  case RemotePacket::GLASS_VALUE: /*  */
				// TODO
				break;
			  default:
				LOG(outlog(), GLASS_ERROR, "Invalid packet in BarrierBase.");
				return false;
			}
		}
		else {
			return false;
		}
	}
	catch (Exception e) {
		LOG(outlog(), GLASS_ERROR, "Decoding packet in remote:" << e.getMessage());
		return false;
	}
	return true;
}

_Remote *RemoteBase::getByName(const string &name) const {
	// TODO: possible race condition with un|registerShared/getByName/un. Mutex this.
	std::map<const string, _Remote *>::const_iterator it;
	it = vars.find(name);
	if (it == vars.end()) {
		return NULL;
	}
	return it->second;
}

void RemoteBase::unregisterNode(nodeId id) {
	nodeRelation			relation = getNodeRelation(id);
	_Remote				*r;

	if (relation == Server) {
		; // server should never unregister
	}
	else if (relation == Client || relation == Peer) {
		std::map<const string, _Remote *>::iterator it;
		for (it = vars.begin(); it != vars.end(); it++) {
			r = it->second;
			r->owners.erase(id);
			if (r->owners.empty()) {
				// TODO: possible race condition with un|registerShared/getByName/un. Mutex this.
				LOG(outlog(), GLASS_DEBUG, "Nobody has " << r->name << " anymore. Deleted.");
				std::map<const string, _Remote *>::iterator it2(it++);
				vars.erase(it2);
				delete r;
			}
			else {
				it++;
			}
		}
	}
}

bool RemoteBase::registerRemote(_Remote *v) { 
	LOG(outlog(), GLASS_DEBUG, CodeDebug);
	if (!v) {
		return false;
	}

	// TODO: possible race condition with un|registerShared/getByName/un. Mutex this.

	std::map<const string, _Remote *>::iterator it;
	it = vars.find(v->name);
	if (it != vars.end()) {
		return false;
	}
	vars[v->name] = v;
	return true;
}

bool RemoteBase::unregisterRemote(_Remote *v) { 
	LOG(outlog(), GLASS_DEBUG, CodeDebug);
	if (!v) {
		return false;
	}

	std::map<const string, _Remote *>::iterator it;
	it = vars.find(v->name);
	if (it == vars.end()) {
		return false;
	}
	vars.erase(it);
	return true;
}


_Remote::_Remote(const string &name) throw(Exception) :
			PluginInterface(Global),
			name(name) {
	if (!name.length()) {
		throw Exception("NULL remote name");
	}

	rb = dynamic_cast<RemoteBase *>(getPluginBase("RemoteBase"));
	if (!rb) {
		throw Exception("No Remote Base class.");
	}

	if (rb->registerRemote(this) == false) {
		throw Exception("Could not register Remote variable (it probably already exists).");
	}

	/* notify the server */
	RemotePacket		rp(RemotePacket::GLASS_CREATE, *this);
//	rp.setTargetNode(peersId);
	rp.setTargetNodes(serversId);
	if (rb->sendPacket(rp) == false) {
		rb->unregisterRemote(this);
		throw Exception("Error sending packet.");
	}
}

_Remote::~_Remote() {
	LOG(outlog(), GLASS_DEBUG, CodeDebug);	

	RemotePacket		rp(RemotePacket::GLASS_DELETE, *this);

//	rp.setTargetNodes(peersId);
	rp.setTargetNodes(serversId);
	if (rb->sendPacket(rp) == false) {
		// we can't do much here, can we?
	}

	LOG(outlog(), GLASS_DEBUG, CodeDebug << "CALL UNREGISTERED");
	rb->unregisterRemote(this);
}

bool _Remote::_setValue(const chain &c, nodeId id) {
	RemotePacket		rp(RemotePacket::GLASS_SET, *this);

	rp.setTargetNodes(id);
	rp.setRemoteData(c);
	if (rb->sendPacket(rp) == false) {
		// we can't do much here, can we?
	}
	return true;
}

bool _Remote::_getValue(chain &c, nodeId id) {
	RemotePacket		rp(RemotePacket::GLASS_GET, *this);

	rp.setTargetNodes(id);
	if (rb->sendPacket(rp) == false) {
		// we can't do much here, can we?
	}
	return true;
}

bool _Remote::copyDataToRemote(nodeId id) { // TODO: review
#ifdef TODO
	chain			c(this->pack());
	return _setValue(c, id);
#endif
	return false;
}

bool _Remote::copyDataFromRemote(nodeId id) { // TODO: review
#ifdef TODO
	chain			c;
	unsigned int	i = 0;

	if (_getValue(c, id) == false) {
		return false;
	}

	return unpack(c, i);
#endif
	return false;
}
