/*  Glass - a distributed computing library
    Copyright (C) 2003-2009 Bruno Barberi Gnecco
    Copyright (C) 2009-2010 Corollarium Technologies


    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <iostream>
#include <vector>
#include <sstream>
#include <fstream>

#if !defined(WIN32)
#  include <unistd.h>
#endif // WIN32

#include "glassserver.h"
#include "_glass.h"
#include "protocol.h"
#include "packetclientserver.h"

using namespace libglass;

/* 
 * GlassServer::GlassServerConnection
 */

GlassServer::GlassServerConnection::GlassServerConnection(GlassServer *s)
		throw(Exception) : s(s) {
	if (!s)
		throw Exception();
}

/*!
   Overrides the pure virtual method in base class Thread */
void GlassServer::GlassServerConnection::cleanup(void) {
	// nothing to do yet
}

/* This method is overrides the pure virtual method in Thread base class and is
   executed in a new thread when the start() method is called. Note that each thread object
   can be run only once. */
void GlassServer::GlassServerConnection::execute(void) {
	LOG(outlog(), GLASS_DEBUG, CodeDebug);
		
	/* get connections from new clients. */
	while (1) {
		LOG(outlog(), GLASS_DEBUG, "Waiting for new connection.");
		nodeAddress addr(s->getProtocol()->accept());
		if (addr.fd < 0) {
			LOG(outlog(), GLASS_ERROR, "Error in protocol: " << addr);
			return;
		}
		LOG(outlog(), GLASS_DEBUG, "Finished new connection " << addr.fd);
	}
}



/* 
 * GlassServer
 */

GlassServer::GlassServer(Protocol *p, bool autostart) throw(Exception) : Glass(p),
		master(nobodyId) {
	LOG(outlog(), GLASS_DEBUG, CodeDebug);

	/* open a port and listen to it */
	if (protocol->listen() != 0) {
		LOG(outlog(), GLASS_ERROR, "Error listening. The port is probably being used by another process.");
		exit(1);
	}

	/* since this is a simple client->server architecture, the server
	  can have its id hardcoded */
	id = baseServerId;

	/* this opens a new thread, which handles new connections. If it
	  throws an exception, let the user handle it. */
	sc = new GlassServerConnection(this);
	sc->start();

	if (autostart) {
		/* this is our packet read/process loop */
		this->start();

		LOG(outlog(), GLASS_ERROR, "Infinite loop ended!");
		exit(1);
	}
}

GlassServer::~GlassServer() {
	/* IMPORTANT: clean IS A VIRTUAL DESTRUCTOR. Do not delete any fields
	or do anything that will compromise the state of the class before
	calling it. See the doxygen documentation for more information. */
	clean();
}

int GlassServer::start(void) {
	Packet			p(this);
	int			ret;

	LOG(outlog(), GLASS_DEBUG, CodeDebug);

	/* loop forever, processing packets */
	while (1) {
		p.clearTargetNodes();
		p.setSenderNode(everybodyId);

		ret = protocol->recv(p);
		if (ret == PROTOCOL_NOT_INITIALIZED) {
			LOG(outlog(), GLASS_ERROR, CodeDebug << ": " << "Error in protocol receive method.");
			continue;
		}
		else if (ret == PROTOCOL_CONNECTION_CLOSED) {
			LOG(outlog(), GLASS_DEBUG, CodeDebug << ": " << "Closed connection.");
			continue;
		}
		else if (ret == PROTOCOL_TIMEOUT) {
			LOG(outlog(), GLASS_DEBUG, CodeDebug << ": " << "Timeout.");
			/* time out. This is a normal behaviour. It's used to
			   ensure that the node list is updated every now 
			   and then after a new connection is made on the
			   other thread (GlassServerConnection). */
			continue;
		}
		else if (ret == PROTOCOL_INVALID_PACKET) {
			LOG(outlog(), GLASS_DEBUG, CodeDebug << ": " << "Invalid packet.");
			/* invalid packet, ignore and hope everything will 
			be fine */
			continue;
		}
		LOG(outlog(), GLASS_DEBUG, CodeDebug << "Got packet");
		processPacket(p);
	}
	return 0;
}

void GlassServer::stop(void) {
	LOG(outlog(), GLASS_DEBUG, CodeDebug);
//	delete protocol();
	// TODO!
}

Protocol *GlassServer::getProtocol(void) const {
	return protocol;
}

bool GlassServer::rootPacket(Packet &p) {
	ClientServerPacket		sp(p);

	LOG(outlog(), GLASS_NOTICE, "Server ROOTPACKET: " << sp.getType() << " from " << NODE_FORMAT p.getSenderNode() << "," << p.getSenderNodeAddress());
	switch (sp.getTypeAbstract()) {
	  case ClientServerPacket::HELLO: /* hello */
	  	{
			LOG(outlog(), GLASS_INFO, "Got HELLO");
			nodeAddress addr(p.getSenderNodeAddress());
			addr.nickname = sp.getNickName();
			nodeId newid = this->GlassServer::registerNode(addr, p.getSenderNode(), Client, true);

			if (this->master == nobodyId) {
				this->master = newid;
			}

			ClientServerPacket	response(this, ClientServerPacket::ID);
			response.setNodeId(newid);
			// Protocol class already registered this node's FD/Id
			response.setTargetNodes(newid);
			if (sendPacket(response) == false) {
				// todo
			}

			// Notify other nodes
			updateClients();
		}
		break;

	  case ClientServerPacket::BYE: /* it's leaving. */
		LOG(outlog(), GLASS_DEBUG, CodeDebug << ": " << "Got a bye bye from " << NODE_FORMAT p.getSenderNode());
		{
			nodeAddress addr(nobodyAddress);

			// Get sender nodeFD from nodeID before we unregister the node (and erase from map)
			addr = this->IdToAddress(p.getSenderNode());
			// Check that we found the nodeID
			if (addr == nobodyAddress) {
				// Failed to convert nodeID to nodeFD
				LOG(outlog(), GLASS_ERROR, CodeDebug << ": " << "Failed to find nodeID " << NODE_FORMAT p.getSenderNode());
			}

			// Unregister the node...
			//   Notify all plugins that the node is going to be unregistered
			//   Remove the nodeID from the idtofdmap map
			//   Remove the nodeFD from the fdtoidmap map
			//   Remove the nodeID from the idtorelation map   
			this->unregisterNode(p.getSenderNode());

			// Disconnect client (only if we have a valid file descriptor)
			if (addr.isValidFD())
				protocol->disconnect(addr);

			// Notify other nodes that it left
			updateClients();
		}
		break;

	  case ClientServerPacket::QUIT_ALL:
		{
			/* everybody out, the party is over. */
			LOG(outlog(), GLASS_DEBUG, CodeDebug << " Got a killall from " << NODE_FORMAT p.getSenderNode());
			ClientServerPacket	response(this, ClientServerPacket::QUIT_ALL);
			response.setTargetNodes(everybodyId);
			if (sendPacket(response) == false) {
				// tough luck
			}
			exit(0);
		}
		break;
		
	  default:
		LOG(outlog(), GLASS_ERROR, "Invalid server root packet." << sp.getTypeAbstract());
		return false;
	}
	return true;
}

bool GlassServer::updateClients(void) { 
	std::map<nodeId, nodeNickName>	newnodelist;
	ClientServerPacket		response(this, ClientServerPacket::NODE_LIST);

	LOG(outlog(), GLASS_INFO, "Updating clients, total nodes: " << this->getTotalNodes());
	AddressIdMap::const_iterator it;
	for (it = addridmap.begin(); it != addridmap.end(); it++) {
		if (it->second.relation == Client) {
			LOG(outlog(), GLASS_DEBUG, "Node: " << it->first);
			nodeId nid = it->first;
			newnodelist[nid] = it->second.nickname;
		}
	}
	response.setTargetNodes(clientsId);
	response.setNodeList(master, newnodelist);
	return this->sendPacket(response);
}

// Add node with our register of connected clients
nodeId GlassServer::registerNode(nodeAddress &addr, nodeId id, nodeRelation r, bool force) {
	nodeId				newid;

	LOG(outlog(), GLASS_INFO, CodeDebug);

	/* generates new id */
	if (getTotalNodes() == 0) { // empty list, get first one
		newid = this->id+1;
	}
	else {
		newid = this->id;
		// find maximum client id
		AddressIdMap::const_iterator it;
		for (it = addridmap.begin(); it != addridmap.end(); it++) {
			if (it->second.relation == Client && it->first > newid) {
				newid = it->first;
			}
		}
		newid++; // next after max
		if (newid >= this->id + 0xFFFF) { // full
			// BUG: TODO: if this->id = 0xFFFF and full, this will fail
			return nobodyId;
		}
	}

	/* Let base class do it work...
	 * - Check to see if the node is already registered
	 * - Add the nodeID from the idtofdmap map
	 * - Add the nodeFD from the fdtoidmap map
	 * - Add the nodeID from the idtorelation map
	 */
	if (this->setNode(addr, newid, Client, true) == false) {
		return nobodyId;
	}

	// Increment count of connected clients
	LOG(outlog(), GLASS_DEBUG, CodeDebug << ": " << "New node " << newid << "," << addr << " : " << this->getTotalNodes() << " nodes registered.");

	if (master == nobodyId) {
		master = newid;
	}
	return newid;
}

// Remove node from our register of connected clients
bool GlassServer::unregisterNode(nodeId id) {
	/* Let the base class do its work...
	 * - Inform all plugins that the node is being unregistered
	 * - Notify all plugins that the node is going to be unregistered
	 * - Remove the nodeID from the idtofdmap map
	 * - Remove the nodeFD from the fdtoidmap map
	 * - Remove the nodeID from the idtorelation map
	 */
	if (Glass::unregisterNode(id) == false) {
		return false; // Failed
	}

	// do we need to assign a new master?
	if (master == id) {
		// yes, get first in list
		AddressIdMap::const_iterator it(addridmap.begin());
		if (it == addridmap.end()) {
			master = nobodyId;
		}
		else {
			master = it->first;
		}

		LOG(outlog(), GLASS_DEBUG, "Old master " << id << " quit, new one is " << master);
	}

	LOG(outlog(), GLASS_DEBUG, CodeDebug << ": " << "Node quit i:" << id << ": " << this->getTotalNodes() << " nodes registered.");

	this->updateClients();
	return true; // Success
}

#ifdef DEPRECATED
// Remove node from our register of connected clients
bool GlassServer::unregisterNode(nodeAddress &addr) {
	/* Let the base class do its work...
	 * - Inform all plugins that the node is being unregistered
	 * - Notify all plugins that the node is going to be unregistered
	 * - Remove the nodeID from the idtofdmap map
	 * - Remove the nodeFD from the fdtoidmap map
	 * - Remove the nodeID from the idtorelation map
	 */
	nodeId id(AddressToId(addr));

	if (Glass::unregisterNode(addr) == false) {
		return false; // Failed
	}

	if (master == id) {
		// yes, get first in list
		AddressIdMap::map_by<Id>::const_iterator it(addridmap.right.begin());
		if (it == addridmap.right.end()) {
			master = nobodyId;
		}
		else {
			master = it->get<Id>();
		}
		LOG(outlog(), GLASS_DEBUG, "Old master " << id << " quit, new one is " << master);	
	}
	this->updateClients();

	LOG(outlog(), GLASS_DEBUG, CodeDebug << ": " << "Node quit f: " << addr << ": " << this->getTotalNodes() << " nodes registered.");
	return true; // Success
}
#endif
