/*  Glass - a distributed computing library
    Copyright (C) 2003-2009 Bruno Barberi Gnecco
    Copyright (C) 2009-2010 Corollarium Technologies


    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "_glass.h"
#include "glassserverpeer.h"
#include "packetpeer.h"
#include "protocol.h"

using namespace libglass;

GlassServerPeer::GlassServerPeer(Protocol *p, const char *peer, unsigned int peerport) throw(Exception) : GlassServer(p, false) {
	LOG(outlog(), GLASS_DEBUG, CodeDebug);

	if (peer) {
		this->seedpeer.assign(peer);
		this->seedpeerport = peerport;
	}

	// generate a new id. We'll check upon connection whether there's a clash and change it if neeeded
	this->id = this->generateId();

	LOG(outlog(), GLASS_NOTICE, "Our peer id: " << NODE_FORMAT this->id);

	/* this is our packet read/process loop */
	this->start();
}

GlassServerPeer::~GlassServerPeer() {

}

int GlassServerPeer::start(void) {
	Packet				p(this);
	bool				packetproc;

	if (this->seedpeer.size()) {
		char				*srv = strdup(this->seedpeer.c_str());

		LOG(outlog(), GLASS_INFO, "Connecting to seedpeer " << this->seedpeer << ":" << this->seedpeerport);

		// first, connect to seed peer.
		nodeAddress address(protocol->connect(srv, this->seedpeerport));
		if (address.status < 0) {
			// error. No peer found. We are the first ones, hooray.
			LOG(outlog(), GLASS_DEBUG, "Could not connect to seed peer. We should be the first ones.");
			free(srv);
		}
		else {
			free(srv);

			LOG(outlog(), GLASS_INFO, "Connected to peer");
			// register the node as a temporary nobody. We'll correct that later when
			// we get the remote peer id.
			this->registerNode(address, nobodyId, Peer, true);

			/* handshaking */
			{
				PeerPacket cp(this, PeerPacket::PEER_START);
				cp.setTargetNodes(everybodyId);
				cp.setnodeAddress(nodeAddress(this->protocol->getLocalHostname(), this->protocol->getListenPort()));

				// Send start packet to server - we expect to receive a PEER_LIST
				// in acknowledgement
				LOG(outlog(), GLASS_DEBUG, "Sending packet");
				if (sendPacket(cp) == false) {
					LOG(outlog(), GLASS_ERROR, "Sending start packet.");
					return -2;
				}
			}

			/* wait for acknowledgement */
			LOG(outlog(), GLASS_INFO, "Waiting for PEER_LIST packet acknowledgement.");
			do {
				int ret = protocol->recv(p);
				if (ret != PROTOCOL_OK && ret != PROTOCOL_TIMEOUT && ret != PROTOCOL_UNREGISTERED_NODE) {
					LOG(outlog(), GLASS_ERROR, "Receiving packet: " << ret);
					return -3;
				}

				LOG(outlog(), GLASS_DEBUG, "from " << p.getSenderNode());

				// add to peer list.
				this->registerNode(address, p.getSenderNode(), Peer, true);

				// we are expeciting PEER_LIST, though we may get a PEER_CHANGE_ID before.
				// since they will follow each other immediately, we're probably safe.
				packetproc = this->processPacket(p, true);
				if (packetproc == true) {
					// initialized = true; // Initialisation successed
				}
			} while (packetproc == false);

			LOG(outlog(), GLASS_INFO, "Starting loop.");

		}
	}

	return GlassServer::start();
}

nodeId GlassServerPeer::generateId(void) const {
#ifdef _MSC_VER
	srand(GetTickCount());
#else
	/* we need a unique id. */
	struct timeval tv;
	gettimeofday(&tv, NULL);
	srand(tv.tv_sec*1000000 + tv.tv_usec);
#endif
	nodeId id = (rand() & 0xFFFF0000);
	if (id == 0) {
		id = 0x00010000;
	}
	return id;
}

bool GlassServerPeer::rootPacket(Packet &p) {
	LOG(outlog(), GLASS_DEBUG, CodeDebug << ": from " << NODE_FORMAT p.getSenderNode());
	try {
		PeerPacket		sp(p);

		LOG(outlog(), GLASS_INFO, "Serverpeer Got packet " << sp.getTypeAbstract());

		switch (sp.getTypeAbstract()) {
		case PeerPacket::PEER_START: /* start. It's a hello with a "send me the list" */
			LOG(outlog(), GLASS_DEBUG, "Got PEER_START. ");
			if (this->peerlist.find(p.getSenderNode()) != this->peerlist.end()) {
				LOG(outlog(), GLASS_INFO, "Ops. We need to change the id. ");
				nodeId n;

				do {
					n += 0x10000;
					if (!n) n = 0x10000;
				} while (this->peerlist.find(n) != this->peerlist.end());

				// send new id
				PeerPacket response(this, PeerPacket::PEER_CHANGE_ID);
				response.setNodeId(n);

				// override, because we'll need it later
				p.setSenderNode(n);

				// and send it back
				response.setTargetNodes(p.getSenderNode());
				response.setPeerList(this->peerlist);
				if (sendPacket(response) == false) {
					// todo
				}
			}
			// and continue with the rest of processing
		case PeerPacket::PEER_HELLO: /* hello */ {
			LOG(outlog(), GLASS_INFO, "Got PEER_HELLO. ");
			// add to list

			peerlist[p.getSenderNode()] = sp.getnodeAddress();

			if (sp.getTypeAbstract() == PeerPacket::PEER_START) {
				nodeAddress addr(p.getSenderNodeAddress());
				this->registerNode(addr, p.getSenderNode(), Peer, true);

				PeerPacket response(this, PeerPacket::PEER_LIST);
				response.setTargetNodes(p.getSenderNode());
				response.setPeerList(this->peerlist);
				if (sendPacket(response) == false) {
					// todo
				}

				// Notify other nodes
				updateClients();

			}
			return true;
		}
		case PeerPacket::PEER_LIST: { /* list */
			/* how crowded is the party? */
			const std::map<nodeId, nodeAddress> newpeerlist(sp.getPeerList());
			int totalnodes = newpeerlist.size();
			LOG(outlog(), GLASS_INFO, "Got PEER_LIST. size=" << totalnodes);

			for (std::map<nodeId, nodeAddress>::const_iterator it(newpeerlist.begin()); it != newpeerlist.end(); it++) {
				nodeId n(it->first);
				nodeAddress pn(it->second);

				if (n == this->id) {// don't connect to ourselves.
					LOG(outlog(), GLASS_DEBUG, "Skipping self connect.");
					continue;
				}

				if (this->peerlist.find(n) != this->peerlist.end()) {
					LOG(outlog(), GLASS_DEBUG, "Skipping connected node. nodeId " << n<< ", port" << pn.port << ", hostname " << pn.host);
					continue;
				}

				LOG(outlog(), GLASS_INFO, "Connecting to PEER_LIST. nodeId " << n << ", port" << pn.port << ", hostname " << pn.host);

				char *tmphost = strdup(pn.host.c_str());
				nodeAddress address(protocol->connect(tmphost, pn.port));
				if (address.status < 0) {
					// error. No peer found. We are the first ones, hooray.
					LOG(outlog(), GLASS_NOTICE, "Could not connect to peer " << pn.host << ". We should be the first ones.");
				}
				else {
					// add to peer list.
					this->setNode(address, it->first, Peer);
				}
				free(tmphost);
			}
			return true;
		}
		case PeerPacket::PEER_CHANGE_ID:
			LOG(outlog(), GLASS_INFO, "Got PEER_CHANGE_ID " << sp.getNodeId());
			this->id = sp.getNodeId();
			return true;

		case PeerPacket::PEER_BYE: /* bye */
			LOG(outlog(), GLASS_INFO, "Got PEER_BYE " << sp.getNodeId());
			this->peerlist.erase(sp.getNodeId());
			this->unregisterNode(sp.getNodeId());
			return true;

		default:
			return GlassServer::rootPacket(p);
		}
		return false;
	}
	catch (Exception e) {
		LOG(outlog(), GLASS_ERROR, e.getMessage());
	}
	return false;
}

nodeId GlassServerPeer::registerNode(nodeAddress &address, nodeId id, nodeRelation r, bool force) {
	LOG(outlog(), GLASS_DEBUG, "Server peer: register node." << address << "," << id);
	if (r == Peer) {
		if (this->setNode(address, id, Peer, true) == false) {
			LOG(outlog(), GLASS_ERROR, "Setting node." << address << "," << id);
			return nobodyId;
		}
	}
	return id;
}

bool GlassServerPeer::unregisterNode(nodeId id) {
	if (this->peerlist.find(id) != this->peerlist.end()) {
		this->peerlist.erase(id);
		if (Glass::unregisterNode(id) == false) {
			return false; // Failed
		}
	}
	else {
		return GlassServer::unregisterNode(id);
	}
	return true;
}

nodeAddress GlassServerPeer::IdToAddress(nodeId id) {
	nodeAddress address(GlassServer::IdToAddress(id));

	if (address == nobodyAddress) {
		// we know how to route messages to clients of other peers. Let's check it
		address = GlassServer::IdToAddress(id & 0xFFFF0000);
	}
	return address;
}
