/*  Glass - a distributed computing library
    Copyright (C) 2003-2009 Bruno Barberi Gnecco
    Copyright (C) 2009-2010 Corollarium Technologies


    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "shared.h"
#include "log.h"
#include "packet.h"
#include "_glass.h"

using namespace libglass;

namespace libglass {

/*!
  Format of Shared Plugin packets:

  \li [action (1 byte)][variable name (ended by \\0)][data]

  where valid action commands are described below.

  These are valid actions in client->server packets:

  \li GLASS_CREATE_LOCAL/GLOBAL = created a variable.
  \li GLASS_UPDATE = updated a variable.
  \li GLASS_DELETE = deleted a variable.
  \li GLASS_SYNC = synchronized a variable.

  These are valid actions in server->client packets:

  \li GLASS_CHANGED = changed variable value.
*/
class SharedPacket : public Packet {
public:
	typedef enum {
		GLASS_CREATE_LOCAL = 0, GLASS_CREATE_GLOBAL, GLASS_UPDATE, GLASS_DELETE, GLASS_SYNC,
		GLASS_CHANGED
	} SharedPacketType;

private:
	SharedPacketType	type;
	chain			shareddata;
	_Shared			*sh;

public:
	SharedPacket(Packet &p) throw(Exception) : Packet(p.getGlass()), sh(NULL) {
		chain		c;
		unsigned int	end;
		p.getData(c);

		type = (SharedPacketType)c[0];
		end = c.find('\0', 1);
		identifier = c.toString(1, end-1);
		if (type == GLASS_UPDATE || type == GLASS_CHANGED) {
			shareddata = c.substr(end+1, chain::npos);
		}
	}

	SharedPacket(SharedPacketType type, _Shared &shared) : Packet(shared.sb->getGlass()),
		type(type), sh(&shared)  {
	}

	virtual ~SharedPacket() {
	}

	SharedPacketType getType(void) const {
		return this->type;
	}

	void setType(SharedPacketType type) {
		this->type = type;
	}

	string &getSharedName(void) {
		if (sh) {
			return sh->name;
		}
		else {
			return identifier;
		}
	}

	chain & getSharedData(void) {
		return shareddata;
	}

	chain encode(unsigned int *len = NULL) {
		setData((char *)&type, 1);
		if (!sh) {
			LOG(outlog(), GLASS_ERROR, CodeDebug << "Encoding packet");
			// throw Exception()
			return chain();
		}
		appendData(sh->name);
		appendData("\0", 1);
		if (type == GLASS_UPDATE || type == GLASS_CHANGED) {
			appendData(sh->nextdata);
		}
		return Packet::encode(len);
	}

	std::ostream& print(std::ostream &o) {
		Packet::print(o);
		return o;
	}
};

} // namespace


// PLUGIN_CONSTRUCTOR macro is defined in plugin.h
// The expanded code ensures that the plugin automatically registers itself
// with the PluginManager object during runtime initialisation. i.e. before
// calling main()
// This registration is achieved by the following operations:
//   Runtime initialisation of class static member SharedBase::registered
//   Runtime initialisation of class static member SharedBase::singleton
//   Definition of default class constuctor
PLUGIN_CONSTRUCTOR(SharedBase) {
	LOG(outlog(), GLASS_DEBUG, CodeDebug);
}

SharedBase::~SharedBase() {
	LOG(outlog(), GLASS_DEBUG, CodeDebug);
	std::map<const string, _Shared *>::iterator it;

	it = vars.begin();
	while (it != vars.end()) {
		delete it->second;
		it++;
	}

	vars.clear();
}


void SharedBase::synchronize(nodeId id, nodeRelation r) {
	// TODO
	return;
}

bool SharedBase::processPacket(Packet &p) {
	nodeRelation		senderType = getNodeRelation(p.getSenderNode());
	std::set<nodeId>::iterator	set_iter;
	std::map<const string, _Shared *>::iterator it;
	_Shared 			*sh;
	unsigned int		used;
	propagationMode		mode;

	// Parse event packet
	try {
		SharedPacket		sp(p);
		string				&varname = sp.getSharedName();
		it = vars.find(varname);

		if (senderType == Client || senderType == Peer) {
			switch (sp.getType()) {
			  case SharedPacket::GLASS_CREATE_LOCAL: /* created */
			  case SharedPacket::GLASS_CREATE_GLOBAL: /* created */
				if (sp.getType() == SharedPacket::GLASS_CREATE_LOCAL) {
					LOG(outlog(), GLASS_INFO, NODE_FORMAT p.getSenderNode() << " created new local shared " << varname);
					mode = Local;
				}
				else {
					LOG(outlog(), GLASS_INFO, NODE_FORMAT p.getSenderNode() << " created new global shared " << varname);
					mode = Global;
				}

				if (it != vars.end()) {
					// variable already exists.
					// do nothing.
					sh = it->second;
				}
				else {
					try {
						sh = new _Shared(varname, mode);
					}
					catch (Exception e) {
						LOG(outlog(), GLASS_ERROR, e.getMessage());
						return false;
					}
					LOG(outlog(), GLASS_INFO, "Created new shared variable " << varname);
				}
				sh->owners.insert(p.getSenderNode());

				if (mode == Global && senderType == Client) {
					// we are a peer who received the message from a client
					// propagate it to other peers.
					LOG(outlog(), GLASS_INFO, "Notifying peers of global shared " << varname);
					 // updated sender, because peers only see peers. Otherwise we'd keep the original sender id.
					p.setReliable(true);
					p.setSenderNode(this->getGlass()->getNodeId());
					p.setTargetNodes(peersId); // send to peers
					p.setBlackMask(p.getSenderNode()); // but not to ourselves
					sendPacket(p);
				}

				break;
			  case SharedPacket::GLASS_UPDATE: /* updated */
				LOG(outlog(), GLASS_DEBUG, "GLASS_UPDATE from " << NODE_FORMAT p.getSenderNode());
				if (it == vars.end()) {
					LOG(outlog(), GLASS_WARNING, "Update for a variable " << varname << " that doesn't exist.");
					break;
				}
				// variable already exists, we're updating
				sh = it->second;
				if (sh->lastchanged == nobodyId || sh->lastchanged >= p.getSenderNode()) {
					// data is dynallocated, so we don't have a leak problem. chain deals with free later.
					sh->nextdata = sp.getSharedData();
					sh->lastchanged = p.getSenderNode();
					LOG(outlog(), GLASS_INFO, "Updated variable " << varname << NODE_FORMAT p.getSenderNode());

					if (sh->getMode() == Global && senderType == Client) { // propagate to every peer we know.
						LOG(outlog(), GLASS_INFO, "Propagating  " << varname << " to peers");
						std::set<nodeId> targets; // copy
						for (set_iter = sh->owners.begin(); set_iter != sh->owners.end(); set_iter++) {
							if (this->getGlass()->getNodeRelation(*set_iter) == Peer) {
								targets.insert(*set_iter);
							}
						}
						p.setTargetNodes(targets);
						p.setBlackMask(p.getSenderNode());
						sendPacket(p);
					}
				}
				break;
			  case SharedPacket::GLASS_DELETE: /* deleted */
	// todo: check if sync signal should be sent!
				if (it == vars.end()) {
					LOG(outlog(), GLASS_WARNING, "Inexistent variable " << varname << " deleted.");
					break;
				}
				sh = it->second;
				set_iter = sh->owners.find(p.getSenderNode());
				if (set_iter == sh->owners.end()) {
					LOG(outlog(), GLASS_WARNING, "Node deleted variable that it doesn't have.");
					break;
				}
				sh->owners.erase(set_iter);
				LOG(outlog(), GLASS_DEBUG, "Node " << NODE_FORMAT p.getSenderNode() << " removed variable " << varname);
				/* avoid a lock */
				if (p.getSenderNode() == sh->lastchanged) {
					sh->lastchanged = nobodyId;
				}
				if (sh->owners.empty()) {
					LOG(outlog(), GLASS_DEBUG, "Nobody has " << varname << " anymore. Deleted.");
					if (sh->getMode() == Global) {
						// we are a peer who received the message from a client
						// propagate it to other peers.
						LOG(outlog(), GLASS_INFO, "Notifying peers of delete " << varname);
						 // updated sender, because peers only see peers. Otherwise we'd keep the original sender id.
						p.setSenderNode(this->getGlass()->getNodeId());
						p.setTargetNodes(peersId); // send to peers
						p.setBlackMask(p.getSenderNode()); // but not to ourselves
						sendPacket(p);
					}
					vars.erase(it);
					delete sh;
				}
				break;
			  case SharedPacket::GLASS_SYNC: /* sync */
				LOG(outlog(), GLASS_DEBUG, "GLASS_SYNC from " << NODE_FORMAT p.getSenderNode());
				it = vars.find(varname);
				if (it == vars.end()) {
					LOG(outlog(), GLASS_WARNING, "Inexistent variable " << varname << " sync'ed.");
					break;
				}
				sh = it->second;

	/*TODO: THIS IS A ROUGH LOGIC THAT IS NOT 100% CORRECT. In short, what we
	are doing here is an asynchronous update that is not completely deterministic;
	that is, the order of the updates of the nodes may have different results.
	The best way to deal with this now is to either make sure that only one guy
	does the update, or that all those that do the update set the same value.
	Note that this is not a "fatal bug", since everything will work perfectly,
	but a logic bug, since the behaviour is not completely deterministic, and
	in my opinion this is unacceptable. */
				if (p.getSenderNode() == sh->lastchanged) {
					sh->lastchanged = nobodyId;
				}

				{
					SharedPacket sp2(SharedPacket::GLASS_CHANGED, *sh);
					LOG(outlog(), GLASS_DEBUG, "Sending " << varname << " to " <<
							NODE_FORMAT p.getSenderNode());
					sp2.setTargetNodes(p.getSenderNode());
					sendPacket(sp2);
				}
				break;
			  default:
				LOG(outlog(), GLASS_ERROR, "Invalid packet in SharedBase.");
			  	return false;
			}
		}
		else if (senderType == Server || senderType == None) {
			switch (sp.getType()) {
			  case SharedPacket::GLASS_CHANGED: /* changed. */
				if (it == vars.end()) {
					// variable doesn't exist. This shouldn't happen.
					LOG(outlog(), GLASS_WARNING, "Got update of a variable that I don't have.");
					break;
				}
				sh = it->second;
				if (sh->unpack(sp.getSharedData(), used) == false) {
					LOG(outlog(), GLASS_ERROR, "Could not unpack variable " << varname);
				}
				LOG(outlog(), GLASS_DEBUG, "Posting trigger to " << varname);
				sh->trigger.notify_all();
				break;
			  default:
				LOG(outlog(), GLASS_ERROR, "Invalid packet in SharedBase.");
			  	return false;
			}
		}
		else {
			LOG(outlog(), GLASS_ERROR, "Invalid senderType in SharedBase.");
			return false;
		}
	}
	catch (Exception e) {
		LOG(outlog(), GLASS_ERROR, "Decoding packet in shared:" << e.getMessage());
		return false;
	}
	return true;
}

_Shared *SharedBase::getByName(const string &name) const {
	// TODO: possible race condition with un|registerShared/getByName/un. Mutex this.
	std::map<const string, _Shared *>::const_iterator it;
	it = vars.find(name);
	if (it == vars.end()) {
		return NULL;
	}
	return it->second;
}

void SharedBase::unregisterNode(nodeId id) {
	nodeRelation			relation = getNodeRelation(id);
	_Shared				*sh;

	if (relation == Server) {
		; // server should never unregister
	}
	else if (relation == Client || relation == Peer) {
		std::map<const string, _Shared *>::iterator	it;
		for (it = vars.begin(); it != vars.end(); ) {
			sh = it->second;
			sh->owners.erase(id);
			if (id == sh->lastchanged) {
				sh->lastchanged = nobodyId;
			}
			if (sh->owners.empty()) {
				// TODO: possible race condition with un|registerShared/getByName/un. Mutex this.
				LOG(outlog(), GLASS_DEBUG, "Nobody has " << sh->name << " anymore. Deleted.");
				std::map<const string, _Shared *>::iterator	it2(it++);
				vars.erase(it2);
				delete sh;
			}
			else {
				it++;
			}
		}
	}
}

bool SharedBase::registerShared(_Shared *v) {
	LOG(outlog(), GLASS_DEBUG, CodeDebug);
	if (!v) {
		LOG(outlog(), GLASS_WARNING, CodeDebug << "NULL shared");
		return false;
	}

	// TODO: possible race condition with un|registerShared/getByName/un. Mutex this.

	std::map<const string, _Shared *>::iterator it;
	it = vars.find(v->name);
	if (it != vars.end()) {
		LOG(outlog(), GLASS_WARNING, CodeDebug << "Shared already exists.");
		return false;
	}
	vars[v->name] = v;
	return true;
}

bool SharedBase::unregisterShared(_Shared *v) {
	if (!v) {
		return false;
	}

	// TODO: possible race condition with un|registerShared/getByName/un. Mutex this.

	std::map<const string, _Shared *>::iterator it;
	it = vars.find(v->name);
	if (it == vars.end()) {
		return false;
	}
	vars.erase(it);
	return true;
}

bool SharedBase::sendUpdate(_Shared *s) {
	if (s->updatemode == _Shared::NEVER) {
		return true;
	}

	if (s->changed == true || s->updatemode == _Shared::FORCED) {
		LOG(outlog(), GLASS_DEBUG, "Updating " << s->name);

		s->nextdata = s->pack();
		SharedPacket			sp(SharedPacket::GLASS_UPDATE, *s);

		sp.setTargetNodes(serversId);
//		sp.setTargetNodes(peersId);

		return sendPacket(sp);
	}
	return true;
}


bool SharedBase::getUpdate(_Shared *s) {
	LOG(outlog(), GLASS_DEBUG, "Syncing " << s->name);

	/* get response */
	SharedPacket			sp(SharedPacket::GLASS_SYNC, *s);
	if (sendPacket(sp) == false) {
		return false;
	}

	if (s->updatemode == _Shared::NEXT_ONLY) {
		s->updatemode = _Shared::NEVER;
	}

	/* we must lock while we don't get the response */
	// TODO s->trigger.wait();
	s->changed = false;

	return true;
}

bool SharedBase::sendUpdateAll(void) {
	std::map<const string, _Shared *>::iterator it;

// TODO!!!!!!
#ifdef OLD
	_Shared				*s;
	Packet				p;

	p.setData("U", 1);

	/* send our updates */
	for (it = vars.begin(); it != vars.end(); it++) {
		s = it->second;

		if (s->updatemode == _Shared::NEVER) {
			continue;
		}
		if (s->changed == true || s->updatemode == _Shared::FORCED) {
			// send it
			p.appendData(s->name);
			chain c = s->pack();
			p.appendData(c);
		}
	}
	sendPacket(p);
#endif

	/* get server updates */
	// todo

	return true;
}

/*
 * _Shared
 */
_Shared::_Shared(const string &name, propagationMode mode) throw(Exception) :
		PluginInterface(mode),
		name(name),
		changed(false),
		updatemode(FORCED),
		lastchanged(nobodyId),
		currentdata(),
		nextdata() {
	if (!name.length()) {
		throw Exception("NULL remote name");
	}

	sb = dynamic_cast<SharedBase *>(getPluginBase("SharedBase"));
	if (!sb) {
		throw Exception("No Shared Base class.");
	}

	if (sb->registerShared(this) == false) {
		throw Exception("Error registering variable.");
	}

	SharedPacket::SharedPacketType type = SharedPacket::GLASS_CREATE_LOCAL;
	if (mode == Local)
		type = SharedPacket::GLASS_CREATE_LOCAL;
	else if (mode == Global)
		type = SharedPacket::GLASS_CREATE_GLOBAL;
	SharedPacket		sp(type, *this);
//	sp.setTargetNode(peersId);
	sp.setTargetNodes(serversId);
	if (sb->sendPacket(sp) == false) {
		sb->unregisterShared(this);
		throw Exception("Error sending packet.");
	}
}

_Shared::~_Shared() {
	LOG(outlog(), GLASS_DEBUG, CodeDebug);

	SharedPacket		sp(SharedPacket::GLASS_DELETE, *this);

//	sp.setTargetNodes(peersId);
	sp.setTargetNodes(serversId);
	if (sb->sendPacket(sp) == false) {
		// we can't do much here, can we?
	}

	LOG(outlog(), GLASS_DEBUG, CodeDebug << "CALL UNREGISTERED");
	sb->unregisterShared(this);
}

bool _Shared::getUpdate(void) {
	return sb->getUpdate(this);
}

bool _Shared::sendUpdate(void) {
	return sb->sendUpdate(this);
}

bool _Shared::sendUpdateAll(void) {
	PluginBase *p = getPluginManager()->getPlugin("SharedBase");
	if (!p) {
		/* highly unlikely, but... */
		return false;
	}
	SharedBase *sb = dynamic_cast<SharedBase *>(p);
	if (!sb) {
		/* impossible, but... */
		return false;
	}
	/* all that was to call SharedBase, but we can't access the sb member
	   directly since this is a static function. Say a nice little curse to
	   C++, now, and keep reading the code. */
	return sb->sendUpdateAll();
}
