/*  Glass - a distributed computing library
    Copyright (C) 2003-2009 Bruno Barberi Gnecco
    Copyright (C) 2009-2010 Corollarium Technologies


    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <errno.h>
#include <string.h>
#include "_glass.h"
#include "log.h"

#define MAXIMUM_PACKET_SIZE 16*1024*1024
#ifndef WIN32
#  include <sys/types.h>
#  include <sys/select.h>
#  include <sys/time.h>
#  include <unistd.h>
#  include <sys/socket.h>
#  include <netinet/in.h>
#  include <netinet/tcp.h>
#  include <arpa/inet.h>
#  include <netdb.h>
#  include <errno.h>
#else
// #include <winsock.h>
#  include <winsock2.h>
#  include <io.h>
#  include <process.h>
#endif // WIN32

#if defined(ENABLE_PACKET_LOG)
#  include "log.h"
#  if !defined(GLASS_STATIC_PACKET_LOG_FILENAME)
#    include <sstream>
#  endif
#endif

#include <boost/thread.hpp>
#include "tcp.h"

using namespace libglass;

// Define portablity macros
#if defined(_MSC_VER)
#  define GLASS_CLOSESOCKET closesocket
#  define GLASS_SETSOCKOPT_ARG4_T char*
#  define GLASS_RECV_ARG2_T char*
#  define GLASS_SOCKETFD_T SOCKET
#  define GLASS_SD_SEND SD_SEND
#  define GLASS_PROTOCOL_NUM IPPROTO_TCP

// #  define GLASS_SOCKET_ERROR SOCKET_ERROR - now defined in types.h

// Global character buffer for storing Windows Sockets API error message
#  define GLASS_ERRMSG_BUF_SIZE 512
char _errmsg_buf[GLASS_ERRMSG_BUF_SIZE];

// Function declarations
static char* libglass_strerror_wsa(int errnum);
static int libglass_strerror_r_wsa(int errnum, char *buf, size_t n);

#else
#  define GLASS_CLOSESOCKET close
#  define GLASS_SETSOCKOPT_ARG4_T void*
#  define GLASS_RECV_ARG2_T void*
#  define GLASS_SOCKETFD_T int
#  define GLASS_SD_SEND SHUT_WR
#  define GLASS_PROTOCOL_NUM 0

#endif


#if !HAVE_SOCKLEN_T
#  if defined(_MSC_VER)
  typedef int socklen_t;
#  else
  typedef unsigned long socklen_t;
#  endif // _MSC_VER
#endif

/*! Define static packet log filename
  Leave undefined if you want to use dynamic filename based on process ID
*/
// #define GLASS_STATIC_PACKET_LOG_FILENAME "packet.log"

#ifdef ENABLE_PACKET_LOG
// #  define PACKET_LOG(t, x)	if (plog) {LOG((*plog), t, x)}
#  define PACKET_LOG(t, x)	{LOG(outlog(), t, "TCP PACKET   " << x);}
#else
#  define PACKET_LOG(t, x)
#endif // ENABLE_PACKET_LOG


// Define maximum length of pending connections to listening server socket
#define GLASS_LISTEN_BACKLOG_LIMIT 6


const struct in_addr TCP::gethostbyname_r(const char *host) throw(Exception) {
#ifdef HAVE_GETNAMEBYHOST_R
	struct hostent		hostbuf, *hp;
	size_t			hstbuflen;
	char			*tmphstbuf;
	struct in_addr		addr;
	int			res;
	int			herr;

	hstbuflen = 1024;

	/* Allocate buffer, remember to free it to avoid memory leakage.  */
	tmphstbuf = (char *)malloc(hstbuflen);

	while ((res = ::gethostbyname_r (host, &hostbuf, tmphstbuf, hstbuflen,
                                           &hp, &herr)) == ERANGE) {
		/* Enlarge the buffer.  */
		hstbuflen *= 2;
		tmphstbuf = (char *)realloc(tmphstbuf, hstbuflen);
	}
	/*  Check for errors.  */
	if (res || hp == NULL) {
#if defined(_MSC_VER)
		PACKET_LOG(GLASS_ERROR, "gethostbyname(): " << libglass_strerror_wsa(WSAGetLastError()));
#else
		PACKET_LOG(GLASS_ERROR, "gethostbyname(): " << strerror(errno));
#endif // _MSC_VER
		throw Exception("gethostbyname");
	}

	addr = *(struct in_addr *)hp->h_addr;
	free(tmphstbuf);

	return addr;
#else
	struct hostent		*hp;
	struct in_addr		addr;

	// I had problems with these in multithreaded programs.
	this->mutex.lock();
	if ((hp = ::gethostbyname(host)) == NULL) {
#if defined(_MSC_VER)
		PACKET_LOG(GLASS_ERROR, "gethostbyname(): " << libglass_strerror_wsa(WSAGetLastError()));
#else
		PACKET_LOG(GLASS_ERROR, "gethostbyname(): " << strerror(errno));
#endif // _MSC_VER
		this->mutex.unlock();
		throw Exception("gethostbyname");
	}

	addr = *(struct in_addr *)hp->h_addr;
	this->mutex.unlock();

	return addr;
#endif
}

// Constructor
TCP::TCP(unsigned int port) : Protocol(), PROTOCOL(AF_INET), listenport(port), listenfd(-1), fdmax(0), no_timeout(false),
		plog(NULL) {
	FD_ZERO(&socklist);

#if defined(ENABLE_PACKET_LOG)
#  if defined(GLASS_STATIC_PACKET_LOG_FILENAME)
	// Create dynamic string object and initialise to packet log filename
	string plogname = GLASS_STATIC_PACKET_LOG_FILENAME;
#  else
#    if defined(_MSC_VER)
	int pid = _getpid();
#    else
	int pid = getpid();
#    endif // _MSC_VER
	// Create empty dynamic string object
	std::ostringstream oss;
	oss << pid;
	string plogname = "tcp" + oss.str() + ".log";
#  endif // GLASS_STATIC_PACKET_LOG_FILENAME

	// Contruct PacketLog object
	plog = new PacketLog(plogname.c_str());

	// Set verbosity level
	plog->setLevel(Log::GLASS_DEBUG);
	// Update log
	LOG(outlog(), GLASS_DEBUG, "Packet logging started: " << plogname);
	// Update packet log
	PACKET_LOG(GLASS_DEBUG, CodeDebug);
#endif // ENABLE_PACKET_LOG


#if defined(_MSC_VER)
	// Initialise windows sockets DLL
	WORD wVersionRequested;
	WSADATA wsaData;

	wVersionRequested = MAKEWORD( 2, 2 );

	if (WSAStartup( wVersionRequested, &wsaData ) != 0) {
		// Failed to initialise Winsock DLL
		LOG(outlog(), GLASS_ERROR, "WSAStartup(): Failed to initialise Winsock DLL");
	}

	// Confirm that the WinSock DLL supports v2.2
	// Note that if the DLL supports versions >= 2.2 it will still
	// return 2.2 in wVersion since that is the version we requested.                                        */
	if ( (LOBYTE( wsaData.wVersion ) != 2) || (HIBYTE( wsaData.wVersion ) != 2) ) {
			// Could not find a usable WinSock DLL
			WSACleanup( ); // Shutdown WinSock DLL
			LOG(outlog(), GLASS_ERROR, "WSAStartup(): Failed to find suitable Winsock DLL version");
	}
#endif // _MSC_VER
}

nodeId TCP::getUniqueId(void) {
	// Get local host name
	char hostName[256] = "";

	if(::gethostname(hostName, sizeof(hostName))) {
		// TODO: error
	}

	struct in_addr addr;
	try {
		addr = this->gethostbyname_r(hostName);
	}
	catch (Exception e) {
		return nobodyId;
	}

	return (nodeId)addr.s_addr;
}

TCP::~TCP() {
	// Iterate over all file descriptor values between 0 and fdmax
	this->_disconnect(everybodyAddress);

	// Close the server listen socket file descriptor
	_safely_close_socket(listenfd);

	// Clear file descriptor set
	FD_ZERO(&socklist);

#if defined(_MSC_VER)
	// The WSACleanup() call will cause any pending blocking or asynchronous calls issued
	// by any thread in this process to be canceled without posting any notification
	// messages or without signaling any event objects
	// Any calls to the WinSock API following this call (in any thread) will cause
	// an unhandled execption error.
	// WSACleanup( ); // Shutdown WinSock DLL
#endif // _MSC_VER


#ifdef ENABLE_PACKET_LOG
	{
		// Copy pointer to packet log object
		PacketLog *plog2 = plog;
		// Reset member attribute to ensure that PACKET_LOG macros don't try to access
		// the PacketLog object after we have freed it.
		plog = NULL;
		// Delete the dynamically allocaed packet log object
		delete plog2;
	}
#endif
}

#if 0
nodeAddress TCP::FDToAddress(nodeFD fd) {
	nodeAddress address;
	socklen_t len;

	// TODO: listenport
	getpeername(fd, (struct sockaddr*)&address.in_addr, &len);
	return address;
}
#endif

nodeAddress TCP::getLocalAddress(void) {
	nodeAddress address(this->getLocalHostname());
	struct hostent *host_entry;

	throw Exception("NOT READY");
	host_entry = gethostbyname(address.host.c_str());
	// local = inet_ntoa(*(struct in_addr *)*host_entry->h_addr_list);

	return nobodyAddress; // TODO
}

string TCP::getLocalHostname(void) {
	// Get the local hostname
	char hostname[255];
	gethostname(hostname, 255);

	return string(hostname);
}

int TCP::bind(int fd) {
	struct sockaddr_in addr;
	int    sockopt_arg;

	/* non blocking ? if so, have to update send() and recv() accordingly.
	fcntl(listenfd, F_SETFL, O_NONBLOCK); */

	/* get rid of "address already in use" messages */
	sockopt_arg = 1;
	::setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (GLASS_SETSOCKOPT_ARG4_T)&sockopt_arg, sizeof(int));

	/* TCP NODELAY */
	sockopt_arg = 1;
	::setsockopt(fd, IPPROTO_TCP, TCP_NODELAY, (GLASS_SETSOCKOPT_ARG4_T)&sockopt_arg, sizeof(int));

	/* Fill out the struct sockaddr_in structure */
	addr.sin_port = htons(listenport);
	addr.sin_family = this->PROTOCOL;
	addr.sin_addr.s_addr = INADDR_ANY;	/* Use my IP address */
	memset((void *)&addr.sin_zero, 0, sizeof(addr.sin_zero));

	return ::bind(fd, (struct sockaddr *)&addr, sizeof(addr));
}


int TCP::_listen(unsigned int nport) {
	PACKET_LOG(GLASS_DEBUG, CodeDebug);

#ifndef WIN32
	/* first of all, handle SIGTERM */
//	signal(SIGTERM, );
#endif

	if (listenfd >= 0) {
		PACKET_LOG(GLASS_ERROR, "Protocol already initialized.");
		return PROTOCOL_ALREADY_INITIALIZED;
	}

	if (!nport) {
		nport = listenport;
	}
	else {
		listenport = nport;
	}

	LOG(outlog(), GLASS_DEBUG, "Listening at " << nport);

	/* Attempt to open listenfd */
	GLASS_SOCKETFD_T	newfd = ::socket(PROTOCOL, SOCK_STREAM, GLASS_PROTOCOL_NUM);

	// Copy to socket file descriptor to member attribute
	// Cast is necessary under Win32 as socket fd is unsigned int
	listenfd = (int)newfd;

	// Check for error condition
#if defined(_MSC_VER)
	if (newfd == INVALID_SOCKET) {
		PACKET_LOG(GLASS_ERROR, "socket(): " << libglass_strerror_wsa(WSAGetLastError()));
		return PROTOCOL_NETWORK_ERROR;
	}
#else
	if (newfd < 0) {
		PACKET_LOG(GLASS_DEBUG, "socket(): " << strerror(errno));
		return PROTOCOL_NETWORK_ERROR;
	}
#endif // _MSC_VER

	/* Attempt to bind to the listenport */
	int bind_retval = this->bind(listenfd);

	// Check for error condition
	if (bind_retval == GLASS_SOCKET_ERROR) {
#if defined(_MSC_VER)
		PACKET_LOG(GLASS_ERROR, "bind(" << nport << "): " << libglass_strerror_wsa(WSAGetLastError()));
#else
		PACKET_LOG(GLASS_ERROR, "bind(" << nport << "): " << strerror(errno));
#endif // _MSC_VAR
		::GLASS_CLOSESOCKET(listenfd);
		return PROTOCOL_NETWORK_ERROR;
	}

	PACKET_LOG(GLASS_DEBUG, "Bound.");

	/* Limit max number of pending connections to listening server socket  */
	int listen_retval = ::listen(listenfd, GLASS_LISTEN_BACKLOG_LIMIT);

	// Check for error condition
	if (listen_retval == GLASS_SOCKET_ERROR) {
#if defined(_MSC_VER)
		PACKET_LOG(GLASS_ERROR, "listen(): " << libglass_strerror_wsa(WSAGetLastError()));
#else
		PACKET_LOG(GLASS_ERROR, "listen(): " << strerror(errno));
#endif // _MSC_VER
		::GLASS_CLOSESOCKET(listenfd); // Hard close socket
		return PROTOCOL_NETWORK_ERROR;
	}

	no_timeout = false;
	default_timeout.tv_sec = 3;
	default_timeout.tv_usec = 0;

	return 0;
}

nodeAddress TCP::_accept(void) {
	nodeAddress addr;
	socklen_t	sin_size = sizeof(struct sockaddr_in);

	PACKET_LOG(GLASS_DEBUG, CodeDebug);

	if (this->listenfd < 0) {
		PACKET_LOG(GLASS_ERROR, "Not listening");
		addr.status = PROTOCOL_NOT_SERVER;
		return addr;
	}

	/* Wait for a connection */
	// TODO: support IPV6
	addr.fd = ::accept(listenfd, (struct sockaddr *)&addr.in_addr, &sin_size);

#if _POSIX_C_SOURCE >= 1 || _XOPEN_SOURCE || _POSIX_SOURCE
	char host[1025];
	if (getnameinfo((struct sockaddr *)&addr.in_addr, sizeof(addr.in_addr),	host, sizeof(host), NULL, 0, NI_NUMERICHOST) == 0) {
		addr.host = host;
	}

	PACKET_LOG(GLASS_DEBUG, "connection from " << addr.host);
#else
	struct hostent *host_entry;
	host_entry = gethostbyaddr((char *)&addr.in_addr.sin_addr.s_addr, sizeof(addr.in_addr.sin_addr.s_addr), AF_INET);
	if (host_entry)
		addr.host = host_entry->h_name;
#endif

	// Check for error condition
#if defined(_MSC_VER)
	if (addr.fd == INVALID_SOCKET) {
		PACKET_LOG(GLASS_ERROR, "accept(): " << libglass_strerror_wsa(WSAGetLastError()));
		addr.status = PROTOCOL_NETWORK_ERROR;
		return addr;
	}
#else
	if (addr.fd == -1) {
		PACKET_LOG(GLASS_ERROR, "accept(): " << strerror(errno));
		addr.status = PROTOCOL_NETWORK_ERROR;
		return addr;
	}
#endif

	// Extract from select(2) documentation...
	// An fd_set is a fixed size buffer. Executing FD_CLR  or  FD_SET  with  a
	// value  of  fd that is negative or is equal to or larger than FD_SETSIZE
	// will result in undefined behavior.
	// Hence, check we do not accept more connections than we can safely store
	// in the fd_set buffer.
#if defined(_MSC_VER)
	if (socklist.fd_count >= FD_SETSIZE) {
#else
	if (addr.fd >= FD_SETSIZE) {
#endif
		// TODO: support more clients. This is a must if we are dealing
		// with large clusters. FD_SETSIZE is usually at least 32,
		// sometimes 64 (windows), sometimes 1024. poll(2) is a
		// possible alternative.
		PACKET_LOG(GLASS_ERROR, "Exceeded max number of clients determined by FD_SET (see socket(2))");
		close(addr.fd);
		addr.status = PROTOCOL_NETWORK_ERROR;
		return addr;
	}

	FD_SET(addr.fd, &socklist);
	if ((int)addr.fd > fdmax)
		fdmax = (int)addr.fd;

	PACKET_LOG(GLASS_DEBUG, "Accepted connection with fd " << addr.fd);

	return addr;
}

nodeAddress TCP::_connect(const char *host, unsigned int nport) {
	nodeAddress			address(host);
	address.setPort(nport);
	return this->_connect(address);
}

nodeAddress TCP::_connect(const nodeAddress &original) {
	nodeAddress	address(original);
	struct in_addr		in_a;

	if (address.host == "") {
		PACKET_LOG(GLASS_ERROR, "NULL host");
		address.status = PROTOCOL_ALREADY_INITIALIZED;
		return address;
	}

	try {
		in_a = this->gethostbyname_r(address.host.c_str());
	}
	catch (Exception e) {
		LOG(outlog(), GLASS_ERROR, "Cannot get hostname");
		address.status = PROTOCOL_NETWORK_ERROR;
		return address;
	}
	address.in_addr.sin_addr = in_a;

	/* Attempt to open client socket */
	address.fd = ::socket(this->PROTOCOL, SOCK_STREAM, GLASS_PROTOCOL_NUM);

	// Check for error condition
#if defined(_MSC_VER)
	if (address.fd == INVALID_SOCKET) {
		PACKET_LOG(GLASS_ERROR, "socket(): " << libglass_strerror_wsa(WSAGetLastError()));
		address.status = PROTOCOL_NETWORK_ERROR;
		return address;
	}
#else
	if (address.fd < 0) {
		PACKET_LOG(GLASS_DEBUG, "socket(): " << strerror(errno));
		address.status = PROTOCOL_NETWORK_ERROR;
		return address;
	}
#endif // _MSC_VER

	/* TCP NODELAY */
	int sockopt_arg = 1;
	::setsockopt(address.fd, IPPROTO_TCP, TCP_NODELAY, (GLASS_SETSOCKOPT_ARG4_T)&sockopt_arg, sizeof(int));

	if (address.port == 0)
		address.setPort(htons(this->listenport));
	address.in_addr.sin_family = this->PROTOCOL;
	memset((void *)&address.in_addr.sin_zero, 0, sizeof(address.in_addr.sin_zero));

	if (::connect(address.fd, (struct sockaddr *)&address.in_addr, sizeof(struct sockaddr)) == -1) {
#if defined(_MSC_VER)
		PACKET_LOG(GLASS_ERROR, "connect(): " << libglass_strerror_wsa(WSAGetLastError()));
#else
		PACKET_LOG(GLASS_ERROR, "connect(): " << strerror(errno));
#endif
		close(address.fd);
		address.fd = 0;
		address.status = PROTOCOL_NETWORK_ERROR;
		return address;
	}

	LOG(outlog(), GLASS_DEBUG, "connected, fd=" << address.fd);

	// add to list
	FD_SET(address.fd, &socklist);
	if ((int)address.fd > fdmax)
		fdmax = (int)address.fd;

	address.status = PROTOCOL_OK;
	return address;
}

// Note that node argument is only examined if (this->type==Client)
void TCP::_disconnect(const nodeAddress &node) {
	// Check if we should disconnect all clients
	if (node == everybodyAddress) {
		// Disconnect all clients
		// Iterate over all file descriptors in the set
		for (int i = 0; i <= fdmax; i++) {
			if (FD_ISSET(i, &socklist)) {
				// Try to close the socket safely
				_safely_close_socket(i);
			}
		}
		// Reset the set of file descriptors
		FD_ZERO(&socklist);
	}
	else {
		// Disconnect only one client
		// Check if the client is connected
		if (FD_ISSET(node.fd, &socklist)) {
			// Try to close the socket safely
			_safely_close_socket(node.fd);
			// Remove the file descriptor from the set
			FD_CLR((GLASS_SOCKETFD_T)node.fd, &socklist);
		}
	}
}

/*!
  Writes a string to a fd.

  \param fd The node FD.
  \param m The message.
  \param length The message length.
  \return The number of bytes actually written.
*/
unsigned int TCP::writePacket(const nodeAddress &addr, chain &m, unsigned int length) {
	unsigned int 		total = 0, partial = 0;
	chain_iterator		it;
	int			n;

	for (it = m.begin(); it != m.end(); it++) {
		partial = 0;
		while (partial < it.size()) {
			// Send data and return number of bytes sent
			n = ::send(addr.fd, it.data()+partial, it.size()-partial, MSG_NOSIGNAL);

// #define REALLY_HARD_DEBUG
#ifdef REALLY_HARD_DEBUG
			std::cerr << "packet 0x";
			for (unsigned int i = partial; i < it.size(); i++) {
				fprintf(stderr, "%02x-", (int)(it.data()[i] & 0xFF));
			}
			std::cerr << std::endl;
#endif

			// Check for error condition
			if (n == GLASS_SOCKET_ERROR) {
				if (errno == EPIPE) { // SIGPIPE
					PACKET_LOG(GLASS_ERROR, "Got a SIGPIPE, fd=" << addr.fd);
					// TODO: is the socket closed correctly?
					break;
				}

#if defined(_MSC_VER)
				PACKET_LOG(GLASS_ERROR, "Error in write: " << addr.fd << ", " << libglass_strerror_wsa(WSAGetLastError()));
#else
				PACKET_LOG(GLASS_ERROR, "Error in write: " << addr.fd << ", " << sys_errlist[errno]);
#endif // _MSC_VAR
				PACKET_LOG(GLASS_ERROR, "Fd: " << addr.fd << "; it.data() = " << (void *)it.data() <<
						"; it.size() = " << it.size() << "; partial " <<
						partial);
				break;
			}
			partial += n;
		}
		total += partial;
	}
	PACKET_LOG(GLASS_DEBUG, "Wrote " << total << " of expected " << length);
	return total;
}

bool TCP::_send(Packet &p) {
	PACKET_LOG(GLASS_DEBUG, "Starting send.");

	mutex.lock();
	std::set<nodeAddress> &dest(p.getTargetAdresses());
	std::set<nodeAddress>::const_iterator it;
	unsigned int 		length;
	chain 			s = p.encode(&length);

	// Iterate over all target file descriptors in the set
	for (it = dest.begin(); it != dest.end(); it++) {
		if (writePacket(*it, s, length) < length) {
			mutex.unlock();
			PACKET_LOG(GLASS_ERROR, "Writing packet");
			return false;
		}
		else {
			PACKET_LOG(GLASS_DEBUG, "Sent " << length << " bytes to [" << AddressToId(*it) << "]: " << p);
		}
	}
	mutex.unlock();
	return true;
}

#define BUFFERSIZE		1024

/*!
  Reads a packet from a fd.

  \param p A reference to the packet.
  \param sock The nodeFD.
  \param buffer The buffer to read from. It must be BUFFERSIZE bytes long.
  \retval 3 Error decoding packet
  \retval 1 Connection closed.
  \retval 0 OK.
  \retval -1 Error reading socket
  \retval -2 Unregistered sender.
*/
int TCP::readPacket(Packet &p, nodeAddress &addr) {
	unsigned char		lengthbuffer[6]; // Temp buffer in which to store packet header
	unsigned int		length = 0, l, headerlen = 0;
	int 				retval = 0; // Return code from this function

	p.setSenderNodeAddress(addr);

	// Read the beginning of the packet header (6 bytes which describe the length of the packet)
	int r = ::recv(addr.fd, (GLASS_RECV_ARG2_T)lengthbuffer, 6, 0);

	// Trap all error conditions here
	if ((r == 0) || (r == GLASS_SOCKET_ERROR)) {
		switch (r) {
			case 0:
				// Socket connection was closed unexpectedly
				PACKET_LOG(GLASS_WARNING, "Socket connection has been closed unexpectedly");
				retval = PROTOCOL_CONNECTION_CLOSED; // Set return value for this function
				break;
			case GLASS_SOCKET_ERROR:
				// There was an error reading socket
#if defined(_MSC_VER)
				PACKET_LOG(GLASS_ERROR, "::recv(): " << libglass_strerror_wsa(WSAGetLastError()));
#endif // _MSC_VER
				PACKET_LOG(GLASS_ERROR, "Error reading socket: " << strerror(errno) << addr);
				retval = PROTOCOL_NOT_INITIALIZED; // Set return value for this function
				break;
			default:
				PACKET_LOG(GLASS_ERROR, "UNHANDLED ERROR CODE");
				retval = PROTOCOL_NOT_INITIALIZED; // Just to make sure that it corresponds to an error code
				break;
		} // switch

		// Handle errors gracefully before returning
		// Shutdown sends on the socket - this will unblock any
		// recv() call waiting on the far end of the socket.
		if (::shutdown(addr.fd, GLASS_SD_SEND) == GLASS_SOCKET_ERROR)
			PACKET_LOG(GLASS_DEBUG, "Failed to shutdown sends on socket");
		// Remove client socket file descriptor from set
		FD_CLR((GLASS_SOCKETFD_T)addr.fd, &socklist);

		unregisterNode(AddressToId(addr));

		// Return error code
		return retval; // Error
	}

	// Get header and total length of packet (stored in first 6 bytes of packet header)
	headerlen = charToShort(lengthbuffer);
	length = charToNum(lengthbuffer+2);

	if (length < 6) {
		PACKET_LOG(GLASS_ERROR, "Packet length is too small - must be invalid (" << std::hex << headerlen << "," << std::hex << length << ")");
		return PROTOCOL_NOT_INITIALIZED; // Error reading from socket
	}
	else if (length > MAXIMUM_PACKET_SIZE) {
		PACKET_LOG(GLASS_ERROR, "Packet length is too large: (" << std::hex << headerlen << "," << std::hex << length << ")");
		return PROTOCOL_NOT_INITIALIZED; // Error reading from socket
	}

	// Allocate a temporary buffer in which to read the packet data
	// TODO: use iovecs and internal memory management
	char *headerbuffer = new char[headerlen];
	char *databuffer = new char[length - headerlen];
	// LOG(outlog(), GLASS_DEBUG, "Allocated bufferH at " << (void *)headerbuffer << " with length " << headerlen);
	// LOG(outlog(), GLASS_DEBUG, "Allocated bufferD at " << (void *)databuffer << " with length " << length - headerlen);

	// Copy packet header to the packet data buffer
	memcpy(headerbuffer, lengthbuffer, 6);
	PACKET_LOG(GLASS_DEBUG, length << " bytes to read from " << addr.fd);

	memset(databuffer, 0, length - headerlen);

	// Initialise buffer pointer (we have already read the packet header)
	bool readingdata = false;
	char *s = headerbuffer + 6;
	l = headerlen - 6; // i.e. We have already read the packet header

	// Iterate until the whole packet has been read or error occurs
	while (1) {
		// Read from socket
		r = ::recv(addr.fd, s, l, 0);

		PACKET_LOG(GLASS_DEBUG, "Loop: l = " << l << "; readingdata == " << readingdata);

		// Trap all error conditions here
		if ((r == 0) || (r == GLASS_SOCKET_ERROR)) {
			PACKET_LOG(GLASS_DEBUG, "Error in main recv loop: " << r);

			switch (r) {
				case 0:
					// Socket connection was closed unexpectedly
					PACKET_LOG(GLASS_DEBUG, "Socket connection has been closed unexpectedly");
					retval = PROTOCOL_CONNECTION_CLOSED; // Set return value for this function
					break;
				case GLASS_SOCKET_ERROR:
					// There was an error reading socket
	#if defined(_MSC_VER)
					PACKET_LOG(GLASS_ERROR, "::recv(): " << libglass_strerror_wsa(WSAGetLastError()));
	#endif // _MSC_VER
					PACKET_LOG(GLASS_ERROR, "Error reading socket");
					retval = PROTOCOL_NOT_INITIALIZED; // Set return value for this function
					break;
				default:
					PACKET_LOG(GLASS_ERROR, "UNHANDLED ERROR CODE");
					retval = PROTOCOL_NOT_INITIALIZED; // Just to make sure that it corresponds to an error code
					break;
			} // switch

			// Handle errors gracefully before returning
			// Shutdown sends on the socket - this will unblock any
			// recv() call waiting on the far end of the socket.
			if (::shutdown(addr.fd, GLASS_SD_SEND) == GLASS_SOCKET_ERROR)
				PACKET_LOG(GLASS_DEBUG, "Failed to shutdown sends on socket");

			// Remove client socket file descriptor from set
			FD_CLR((GLASS_SOCKETFD_T)addr.fd, &socklist);
			unregisterNode(AddressToId(addr));

			// Delete the dynamically allocated data buffer before returning
			delete[] headerbuffer;
			delete[] databuffer;

			// Return error code
			return retval; // Error
		}

		// Increment data buffer pointer
		s += r;
		// Decrement the length counter
		l -= r;

		if (l == 0) {
			if (readingdata == true) {
				break;
			}
			l = length - headerlen;
			s = databuffer;
			readingdata = true;
		}
		PACKET_LOG(GLASS_DEBUG, "Endloop: l = " << l << "; r = " << r << "; readingdata == " << readingdata);
	}

// #define REALLY_HARD_DEBUG2
#if defined ENABLE_GLASS_LOG && defined REALLY_HARD_DEBUG2
	std::cerr << "recv packet 0x";
	for (unsigned int i = 0; i < length - headerlen; i++) {
		fprintf(stderr, "%02x-", (int)(databuffer[i] & 0xFF));
	}
	std::cerr << std::endl;
#endif

	// Now that the whole packet has been read we can decode it
	if (p.decode((unsigned char *)headerbuffer, headerlen,
			(unsigned char *)databuffer, length - headerlen) == false) {
		PACKET_LOG(GLASS_ERROR, "Error decoding packet.");
		// Delete the dynamically allocated data buffer before returning
		delete[] headerbuffer;
		delete[] databuffer;
		// Return error code
		return PROTOCOL_INVALID_PACKET; // Error - decoding packet
	}
	delete[] headerbuffer; // Delete dynamically allocated data buffer before returning

	PACKET_LOG(GLASS_DEBUG, "Recv " << length << " bytes from [" << NODE_FORMAT p.getSenderNode() << "]: " << p);

	if (this->IdToAddress(p.getSenderNode()).relation == Unregistered) {
#ifdef OLD
		PACKET_LOG(GLASS_WARNING, "Error finding packet sender.");
		PACKET_LOG(GLASS_WARNING, "Got " << p << "bytes");
		delete[] databuffer;
#endif
		PACKET_LOG(GLASS_DEBUG, "From unregistered node " << NODE_FORMAT p.getSenderNode() << addr);

		// Return error code
		return PROTOCOL_UNREGISTERED_NODE;  // unregistered sender
	}

	return PROTOCOL_OK; // Success
}


// Return values are:
// -2 = Unregistered sender
// -1 = Error
// 0 = Success
// 1 = Connection closed
// 2 = Timeout
int TCP::_recv(Packet &p) {
	Packet			*o;
	bool			first = true;
	int			retval = 0;

	// from which client?
	if (p.getSenderNode() == everybodyId || p.getSenderNode() == nobodyId)  { /* get from anybody */
		if (!queue.empty()) {
			PACKET_LOG(GLASS_DEBUG, "QUEUE.");
			o = queue.front();
			p = *o;
			queue.pop_front();
			delete o; // TODO: fix reference problem with _mem!

			PACKET_LOG(GLASS_DEBUG, "Address " << p.getSenderNodeAddress());
			return PROTOCOL_OK;
		}
		struct timeval timeout;
		int sel;
		fd_set fdtemp;

		timeout = default_timeout;

#if defined(_MSC_VER)
		// Under MS windows select() returns -1 if the file descriptor
		// set is empty. Hence, if the set is empty simply sleep
		// rather than call select()

		// Check if we have any clients connected
		if (socklist.fd_count == 0) {
			// PACKET_LOG(GLASS_DEBUG, "No clients connected");
			if (!no_timeout)
				Sleep( timeout.tv_sec * 1000 );
			return PROTOCOL_TIMEOUT; // i.e. Timeout condition
		}
#endif // _MSC_VER

		fdtemp = socklist;
		sel = select(fdmax+1, &fdtemp, NULL, NULL,
			no_timeout == true ? NULL : &timeout);

		if (sel == 0) {
			// PACKET_LOG(GLASS_DEBUG, "Timeout: Select = 0.");
			return PROTOCOL_TIMEOUT; // Timeout
		}
		else if (sel == -1) {
#if defined(_MSC_VER)
			PACKET_LOG(GLASS_ERROR, "select(): " << libglass_strerror_wsa(WSAGetLastError()));
#else
			PACKET_LOG(GLASS_ERROR, "select(): " << strerror(errno));
#endif // _MSC_VER
			if (errno == EINTR) { // interrupted system call. Just ignore it.
				return PROTOCOL_TIMEOUT; // Timeout
			}

			LOG(outlog(), GLASS_ERROR, "Select = -1.");
			PACKET_LOG(GLASS_ERROR, "Select = -1.");

			return PROTOCOL_NETWORK_ERROR; // Error
		}
		for (int i = 0; i <= fdmax; i++ ) {
			if (FD_ISSET(i, &fdtemp)) {
				FD_CLR((GLASS_SOCKETFD_T)i, &fdtemp);
				/* with select, we must read
				   all available descriptors */

				nodeAddress addr(this->FDToAddress(i));
				if (addr.fd < 0) {
					// we have not registered this node yet.
					addr.fd = i;
				}

				if (queue.empty() && first == true) {
					PACKET_LOG(GLASS_DEBUG, "Read packet: " << addr);
					retval = readPacket(p, addr);
					first = false;
				}
				else {
					PACKET_LOG(GLASS_DEBUG, "Saving to queue: " << addr);
					o = new Packet(this->owner);
					retval = readPacket(*o, addr);
					queue.push_back(o);
				}

			}
		}
		PACKET_LOG(GLASS_DEBUG, "Returning from " << CodeDebug);
		// todo: NOT GOOD, we have several retvals. how to return?
		if (retval != PROTOCOL_OK && retval != PROTOCOL_UNREGISTERED_NODE) {
			LOG(outlog(), GLASS_ERROR, "Return readPacket " << retval);
		}
		return retval;
	}
	else {
		if (!queue.empty()) {
			std::list<Packet *>::iterator	it;

			for (it = queue.begin(); it != queue.end(); it++) {
				o = *it;
				if (o->getSenderNode() == p.getSenderNode()) {
					o = *queue.begin();
					p = *o;
					queue.erase(it);
					delete o;
					return 0; // Success
				}
			}
		}

		nodeAddress address(IdToAddress(p.getSenderNode()));
		fd_set fdtemp = socklist;
		if (!FD_ISSET(address.fd, &fdtemp)) {
			LOG(outlog(), GLASS_ERROR, "FD_ISSET not set.");
			PACKET_LOG(GLASS_ERROR, "FD_ISSET.");
			return -1;
		}
		FD_CLR((GLASS_SOCKETFD_T)address.fd, &fdtemp);
		/* note that this blocks indefinitely */
		return readPacket(p, address);
	}
	return 0;
}



// Close socket connection safely
// Returns 0 on success
// Return -1 on failure
int TCP::_safely_close_socket(nodeFD fd)
{
	const int bufsize = 1024;
	const int exit_success = 0;
	const int exit_failure = -1;

	if (fd <= 0) {
		return exit_success;
	}

	PACKET_LOG(GLASS_DEBUG, "Shutting down fd " << fd);
	// Disallow any further data sends.  This will tell the other side
	// that we want to go away now.  If we skip this step, we don't
	// shut the connection down nicely.
	if (::shutdown(fd, GLASS_SD_SEND) == GLASS_SOCKET_ERROR) {
		return exit_failure;	// Failed to shutdown sends
	}

	// Receive any extra data still sitting on the socket.  After all
	// data is received, this call will block until the remote host
	// acknowledges the TCP control packet sent by the shutdown above.
	// Then we'll get a 0 back from recv, signalling that the remote
	// host has closed its side of the connection.
	char buf[bufsize];
	while (1) {
		int bytes =::recv(fd, buf, bufsize, 0);
		if (bytes == GLASS_SOCKET_ERROR) {
			return exit_failure;	// Failed to receive data still waiting
		}
		else if (bytes != 0) {
			PACKET_LOG(GLASS_ERROR,
				   "Received " << bytes <<
				   " unexpected bytes during socket shutdown."
				  );
		}
		else {
			// Okay, we're done!
			break;
		}
	}

	// Close the socket.
	if (::GLASS_CLOSESOCKET(fd) == GLASS_SOCKET_ERROR) {
		return exit_failure;	// Failed to close the socket
	}

	return exit_success;	// Sucsess
}



#if defined(_MSC_VER)

// Get system error message and copy to global character buffer
// Wrapper around libglass_strerror_r_wsa().
// NOT THREAD SAFE
static char* libglass_strerror_wsa(int errnum)
{
	// Modify the global error message buffer
	int returnval = libglass_strerror_r_wsa(errnum, _errmsg_buf, GLASS_ERRMSG_BUF_SIZE);

	// Return pointer to global error message buffer
	return _errmsg_buf;
}


// Returns  the  appropriate  error description string, or an unknown
// error message if the error code is unknown. Upon success, returns 0
// and does not change the value of "errno".
// If error number is unrecognised then returns -1.
// Similar behaviour to strerror_r function in glibc.
// WINDOWS SOCKETS ERROR MESSAGES ONLY
static int libglass_strerror_r_wsa(int errnum, char *buf, size_t n)
{
	// Check for invalid arguments
	if ((!buf) || (n==0))
		return -1;

	int returnval = 0;

	switch (errnum) {
		case WSANOTINITIALISED:
			strncpy( buf, "A successful WSAStartup call must occur before using this function", n );
			break;
		case WSAENETDOWN:
			strncpy( buf, "The network subsystem or the associated service provider has failed", n );
			break;
		case WSAEAFNOSUPPORT:
			strncpy( buf, "The specified address family is not supported", n );
			break;
		case WSAEINPROGRESS:
			strncpy( buf, "A blocking Windows Sockets 1.1 call is in progress, or the service provider is still processing a callback function", n );
			break;
		case WSAEMFILE:
			strncpy( buf, "No more socket descriptors are available", n );
			break;
		case WSAENOBUFS:
			strncpy( buf, "No buffer space is available. The socket cannot be created", n );
			break;
		case WSAEPROTONOSUPPORT:
			strncpy( buf, "The specified protocol is not supported", n );
			break;
		case WSAEPROTOTYPE:
			strncpy( buf, "The specified protocol is the wrong type for this socket", n );
			break;
		case WSAESOCKTNOSUPPORT:
			strncpy( buf, "The specified socket type is not supported in this address family", n );
			break;
		case WSAECONNRESET:
			strncpy( buf, "An incoming connection was indicated, but was subsequently terminated by the remote peer prior to accepting the call", n );
			break;
		case WSAEFAULT:
			strncpy( buf, "The addrlen parameter is too small or addr is not a valid part of the user address space", n );
			break;
		case WSAEINTR:
			strncpy( buf, "A blocking Windows Sockets 1.1 call was canceled through WSACancelBlockingCall", n );
			break;
		case WSAEINVAL:
			strncpy( buf, "The listen function was not invoked prior to accept", n );
			break;
		case WSAENOTSOCK:
			strncpy( buf, "The descriptor is not a socket", n );
			break;
		case WSAEOPNOTSUPP:
			strncpy( buf, "The referenced socket is not a type that supports connection-oriented service", n );
			break;
		case WSAEWOULDBLOCK:
			strncpy( buf, "The socket is marked as nonblocking and no connections are present to be accepted", n );
			break;
		case WSAEACCES:
			strncpy( buf, "Attempt to connect datagram socket to broadcast address failed because setsockopt option SO_BROADCAST is not enabled", n );
			break;
		case WSAEADDRINUSE:
			strncpy( buf, "A process on the computer is already bound to the same fully-qualified address and the socket has not been marked to allow address reuse with SO_REUSEADDR. For example, the IP address and port are bound in the af_inet case). (See the SO_REUSEADDR socket option under setsockopt", n );
			break;
		case WSAEADDRNOTAVAIL:
			strncpy( buf, "The specified address is not a valid address for this computer", n );
			break;
		case WSAEISCONN:
			strncpy( buf, "The socket is already connected", n );
			break;
		case WSAENOTCONN:
			strncpy( buf, "The socket is not connected", n );
			break;
		case WSAENETRESET:
			strncpy( buf, "The connection has been broken due to the keep-alive activity detecting a failure while the operation was in progress", n );
			break;
		case WSAESHUTDOWN:
			strncpy( buf, "The socket has been shut down; it is not possible to receive on a socket after shutdown has been invoked with how set to SD_RECEIVE or SD_BOTH", n );
			break;
		case WSAEMSGSIZE:
			strncpy( buf, "The message was too large to fit into the specified buffer and was truncated", n );
			break;
		case WSAECONNABORTED:
			strncpy( buf, "The virtual circuit was terminated due to a time-out or other failure. The application should close the socket as it is no longer usable", n );
			break;
		case WSAETIMEDOUT:
			strncpy( buf, "The connection has been dropped because of a network failure or because the peer system failed to respond", n );
			break;
		default:
			// Unable to translate error code - Create dynamic error message
			_snprintf( buf, n, "Windows sockets error code %d", errnum );
			returnval = -1;
			break;
	}

	// Ensure that the error string is NULL terminated
	buf[n-1] = '\0';

	return returnval;
}


#endif // _MSC_VER
