/*  Glass - a distributed computing library
    Copyright (C) 2003-2009 Bruno Barberi Gnecco
    Copyright (C) 2009-2010 Corollarium Technologies


    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <errno.h>
#include <string.h>
#include "_glass.h"
#include "log.h"

#define MAXIMUM_PACKET_SIZE 16*1024*1024
#ifndef WIN32
#  include <sys/types.h>
#  include <sys/select.h>
#  include <sys/time.h>
#  include <unistd.h>
#  include <sys/socket.h>
#  include <netinet/in.h>
#  include <netinet/tcp.h>
#  include <arpa/inet.h>
#  include <netdb.h>
#  include <errno.h>
#else
// #include <winsock.h>
#  include <winsock2.h>
#  include <io.h>
#  include <process.h>
#endif // WIN32

#if defined(ENABLE_PACKET_LOG)
#  include "log.h"
#  if !defined(GLASS_STATIC_PACKET_LOG_FILENAME)
#    include <sstream>
#  endif
#endif

#define MAXIMUM_UDP_PACKET_SIZE		65536 /// theoretical, smaller in practice

#include <boost/thread.hpp>
#include "tcpudp.h"

using namespace libglass;

// Define portablity macros
#if defined(_MSC_VER)
#  define GLASS_CLOSESOCKET closesocket
#  define GLASS_SETSOCKOPT_ARG4_T char*
#  define GLASS_RECV_ARG2_T char*
#  define GLASS_SOCKETFD_T SOCKET
#  define GLASS_SD_SEND SD_SEND
#  define GLASS_PROTOCOL_UDP IPPROTO_UDP

// #  define GLASS_SOCKET_ERROR SOCKET_ERROR - now defined in types.h

// Global character buffer for storing Windows Sockets API error message
#  define GLASS_ERRMSG_BUF_SIZE 512
char _errmsg_buf[GLASS_ERRMSG_BUF_SIZE];

// Function declarations
static char* libglass_strerror_wsa(int errnum);
static int libglass_strerror_r_wsa(int errnum, char *buf, size_t n);

#else
#  define GLASS_CLOSESOCKET close
#  define GLASS_SETSOCKOPT_ARG4_T void*
#  define GLASS_RECV_ARG2_T void*
#  define GLASS_SOCKETFD_T int
#  define GLASS_SD_SEND SHUT_WR
#  define GLASS_PROTOCOL_UDP IPPROTO_UDP

// #  define GLASS_SOCKET_ERROR (-1) - now defined in types.h
#endif


#if !HAVE_SOCKLEN_T
#  if defined(_MSC_VER)
  typedef int socklen_t;
#  else
  typedef unsigned long socklen_t;
#  endif // _MSC_VER
#endif

/*! Define static packet log filename
  Leave undefined if you want to use dynamic filename based on process ID
*/
// #define GLASS_STATIC_PACKET_LOG_FILENAME "packet.log"

#ifdef ENABLE_PACKET_LOG
// #  define PACKET_LOG(t, x)	if (plog) {LOG((*plog), t, x)}
#  define PACKET_LOG(t, x)	{LOG(outlog(), t, "UDP PACKET   " << x);}
#else
#  define PACKET_LOG(t, x)
#endif // ENABLE_PACKET_LOG


// Define maximum length of pending connections to listening server socket
#define GLASS_LISTEN_BACKLOG_LIMIT 6

// Constructor
TCPUDP::TCPUDP(unsigned int porttcp, unsigned int portudp) :
		TCP(porttcp), listenportudp(portudp), listenfdudp(0) {


	this->udpoutsocket = ::socket(AF_INET, SOCK_DGRAM, GLASS_PROTOCOL_UDP);

	// Check for error condition
#if defined(_MSC_VER)
	if (this->udpoutsocket == INVALID_SOCKET) {
		PACKET_LOG(GLASS_ERROR, "socket(): " << libglass_strerror_wsa(WSAGetLastError()));
		// TODO return -2;
	}
#else
	if (this->udpoutsocket < 0) {
		PACKET_LOG(GLASS_DEBUG, "socket(): " << strerror(errno));
		// TODO return -2;
	}
#endif // _MSC_VER

	/* non blocking ? if so, have to update send() and recv() accordingly.
	fcntl(listenfd, F_SETFL, O_NONBLOCK); */

	default_timeout.tv_sec = 3;
	default_timeout.tv_usec = 0;
}

TCPUDP::~TCPUDP() {
	// Close the server listen socket file descriptor
	_safely_close_socket(listenfdudp);
}

int TCPUDP::_listen(unsigned int nport) {
	struct sockaddr_in 	addr;

	PACKET_LOG(GLASS_DEBUG, CodeDebug);

#ifndef WIN32
	/* first of all, handle SIGTERM */
//	signal(SIGTERM, );
#endif

	if (listenfdudp > 0) {
		PACKET_LOG(GLASS_ERROR, "Protocol already initialized.");
		return -1;
	}

	nport = listenportudp;

	/* Attempt to open listenfd */
	GLASS_SOCKETFD_T	newfd = ::socket(AF_INET, SOCK_DGRAM, GLASS_PROTOCOL_UDP);

	// Copy to socket file descriptor to member attribute
	// Cast is necessary under Win32 as socket fd is unsigned int
	listenfdudp = (int)newfd;

	// Check for error condition
#if defined(_MSC_VER)
	if (newfd == INVALID_SOCKET) {
		PACKET_LOG(GLASS_ERROR, "socket(): " << libglass_strerror_wsa(WSAGetLastError()));
		return -2;
	}
#else
	if (newfd < 0) {
		PACKET_LOG(GLASS_DEBUG, "socket(): " << strerror(errno));
		return -2;
	}
#endif // _MSC_VER

	/* non blocking ? if so, have to update send() and recv() accordingly.
	fcntl(listenfd, F_SETFL, O_NONBLOCK); */

#ifdef LATER
	/* get rid of "address already in use" messages */
	sockopt_arg = 1;
	::setsockopt(listenfdudp, SOL_SOCKET, SO_REUSEADDR, (GLASS_SETSOCKOPT_ARG4_T)&sockopt_arg, sizeof(int));
#endif

	/* Fill out the struct sockaddr_in structure */
	addr.sin_port = htons(nport);
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = INADDR_ANY;	/* Use my IP address */
	memset((void *)&addr.sin_zero, 0, sizeof(addr.sin_zero));

	/* Attempt to bind to the listenport */
	int bind_retval = ::bind(listenfdudp, (struct sockaddr *)&addr, sizeof(addr));

	// Check for error condition
	if (bind_retval == GLASS_SOCKET_ERROR) {
#if defined(_MSC_VER)
		PACKET_LOG(GLASS_ERROR, "bind(" << nport << "): " << libglass_strerror_wsa(WSAGetLastError()));
#else
		PACKET_LOG(GLASS_ERROR, "bind(" << nport << "): " << strerror(errno));
#endif // _MSC_VAR
		::GLASS_CLOSESOCKET(listenfdudp);
		return -2;
	}

	FD_SET(listenfdudp, &socklist);
	if ((int)listenfdudp > fdmax)
		fdmax = (int)listenfdudp;

	PACKET_LOG(GLASS_DEBUG, "Bound UDP on port " << nport);

	return TCP::_listen(0); // TODO: this does not let anybody override the port.
}

// Note that node argument is only examined if (this->type==Client)
void TCPUDP::_disconnect(nodeAddress &addr) {
	// Check if we should disconnect all clients
	if (addr == everybodyAddress) {
		// TODO
	}
	TCP::_disconnect(addr);
}

/*!
  Writes a string to a fd.

  \param fd The node FD.
  \param m The message.
  \param length The message length.
  \return The number of bytes actually written.
*/
unsigned int TCPUDP::UDPwritePacket(const nodeAddress &address, chain &m, unsigned int length) {
	unsigned int 		total = 0, partial = 0;
	chain_iterator		it;
	int			n;

	// TODO: we shouldn't call getpeername for every loop.
	struct sockaddr_in addr;
	socklen_t len;
	if ((n == getpeername(address.fd, (struct sockaddr*)&addr, &len)) < 0) {
		PACKET_LOG(GLASS_ERROR, "Cannot find address for " << address.fd << ". Result: "  << n);
		return PROTOCOL_NOT_INITIALIZED;
	}
	addr.sin_port = htons(this->listenportudp);
	addr.sin_family = AF_INET;

	for (it = m.begin(); it != m.end(); it++) {
		partial = 0;
		while (partial < it.size()) {
			// Send data and return number of bytes sent
			n = ::sendto(this->udpoutsocket, it.data()+partial, it.size()-partial, MSG_NOSIGNAL, (struct sockaddr*)&addr, sizeof(addr));

// #define REALLY_HARD_DEBUG
#ifdef REALLY_HARD_DEBUG
			std::cerr << "packet 0x";
			for (unsigned int i = partial; i < it.size(); i++) {
				fprintf(stderr, "%02x-", (int)(it.data()[i] & 0xFF));
			}
			std::cerr << std::endl;
#endif
			// PACKET_LOG(GLASS_ERROR, "Error in write: " << fd << ", " << libglass_strerror_wsa(WSAGetLastError()));

			// Check for error condition
			if (n == GLASS_SOCKET_ERROR) {
				if (errno == EPIPE) { // SIGPIPE
					PACKET_LOG(GLASS_ERROR, "Got a SIGPIPE, fd=" << address.fd);
					// TODO: is the socket closed correctly?
					break;
				}

#if defined(_MSC_VER)
				PACKET_LOG(GLASS_ERROR, "Error in udp write: " << address.fd << ", " << libglass_strerror_wsa(WSAGetLastError()));
#else
				PACKET_LOG(GLASS_ERROR, "Error in udp write: " << address.fd << ", " << sys_errlist[errno]);
#endif // _MSC_VAR
				PACKET_LOG(GLASS_ERROR, "UDP Fd: " << address.fd << "; it.data() = " << (void *)it.data() <<
						"; it.size() = " << it.size() << "; partial " <<
						partial);
				break;
			}
			partial += n;
		}
		total += partial;
	}
	PACKET_LOG(GLASS_DEBUG, "UDP Wrote " << total << " of expected " << length);
	return total;
}

bool TCPUDP::_send(Packet &p) {
	if (p.isReliable()) {
		return TCP::_send(p);
	}

	PACKET_LOG(GLASS_INFO, "Unreliable packet send");

	mutex.lock();
	std::set<nodeAddress> &dest(p.getTargetAdresses());
	std::set<nodeAddress>::iterator it;
	unsigned int 		length;
	chain 			s = p.encode(&length);

	// Iterate over all target file descriptors in the set
	for (it = dest.begin(); it != dest.end(); it++) {
		if (UDPwritePacket(*it, s, length) < length) {
			mutex.unlock();
			PACKET_LOG(GLASS_ERROR, "Writing packet");
			return false;
		}
		else {
			PACKET_LOG(GLASS_DEBUG, "UDP sent " << length << " bytes to [" << AddressToId(*it) << "]: " << p);
		}
	}
	mutex.unlock();
	return true;
}

#define BUFFERSIZE		1024

/*!
  Reads a packet from a fd.

  \param p A reference to the packet.
  \param sock The nodeFD.
  \param buffer The buffer to read from. It must be BUFFERSIZE bytes long.
  \retval 3 Error decoding packet
  \retval 1 Connection closed.
  \retval 0 OK.
  \retval -1 Error reading socket
  \retval -2 Unregistered sender.
*/
int TCPUDP::UDPreadPacket(Packet &p, nodeAddress &address) {
	unsigned char		lengthbuffer[MAXIMUM_UDP_PACKET_SIZE]; // Temp buffer in which to store packet
	unsigned int		length = 0, l, headerlen = 0;
	int 				retval = 0, n = 0; // Return code from this function

	p.setSenderNodeAddress(address);

	struct sockaddr_in remote;
	socklen_t remotelen;

	// Read the beginning of the packet header (6 bytes which describe the length of the packet)
	int r = ::recvfrom(this->listenfdudp, (GLASS_RECV_ARG2_T)lengthbuffer, MAXIMUM_UDP_PACKET_SIZE, 0, (struct sockaddr*)&remote, &remotelen);

	// Trap all error conditions here
	if ((r == 0) || (r == GLASS_SOCKET_ERROR)) {
		switch (r) {
			case 0:
				// Socket connection was closed unexpectedly
				PACKET_LOG(GLASS_DEBUG, "Socket connection has been closed unexpectedly");
				retval = PROTOCOL_CONNECTION_CLOSED; // Set return value for this function
				break;
			case GLASS_SOCKET_ERROR:
				// There was an error reading socket
#if defined(_MSC_VER)
				PACKET_LOG(GLASS_ERROR, "::recv(): " << libglass_strerror_wsa(WSAGetLastError()));
#endif // _MSC_VER
				PACKET_LOG(GLASS_ERROR, "Error reading socket: " << strerror(errno));
				retval = PROTOCOL_NOT_INITIALIZED; // Set return value for this function
				break;
			default:
				PACKET_LOG(GLASS_ERROR, "UNHANDLED ERROR CODE");
				retval = PROTOCOL_NOT_INITIALIZED; // Just to make sure that it corresponds to an error code
				break;
		} // switch

		// Return error code
		return retval; // Error
	}

	// Get header and total length of packet (stored in first 6 bytes of packet header)
	headerlen = charToShort(lengthbuffer);
	length = charToNum(lengthbuffer+2);

	if (length < 6) {
		PACKET_LOG(GLASS_ERROR, "Packet length is too small - must be invalid (" << std::hex << headerlen << "," << std::hex << length << ")");
		return PROTOCOL_NOT_INITIALIZED; // Error reading from socket
	}
	else if (length > MAXIMUM_PACKET_SIZE) {
		PACKET_LOG(GLASS_ERROR, "Packet length is too large: (" << std::hex << headerlen << "," << std::hex << length << ")");
		return PROTOCOL_NOT_INITIALIZED; // Error reading from socket
	}

	// Allocate a temporary buffer in which to read the packet data
	// TODO: use iovecs and internal memory management
	char *headerbuffer = new char[headerlen];
	char *databuffer = new char[length - headerlen];
	LOG(outlog(), GLASS_DEBUG, "Allocated bufferH at " << (void *)headerbuffer << " with length " << headerlen);
	LOG(outlog(), GLASS_DEBUG, "Allocated bufferD at " << (void *)databuffer << " with length " << length - headerlen);

	// Copy packet header to the packet data buffer
	memcpy(headerbuffer, lengthbuffer, 6);
	PACKET_LOG(GLASS_DEBUG, length << " bytes to read.");

	memset(databuffer, 0, length - headerlen);

	// Initialise buffer pointer (we have already read the packet header)
	bool readingdata = false;
	char *s = headerbuffer + 6;
	l = headerlen - 6; // i.e. We have already read the packet header

	// Iterate until the whole packet has been read or error occurs
	while (1) {
		// Read from socket
		int r = ::recvfrom(this->listenfdudp, s, l, 0, (struct sockaddr*)&remote, &remotelen);

		PACKET_LOG(GLASS_DEBUG, "Loop: l = " << l << "; readingdata == " << readingdata);

		// Trap all error conditions here
		if ((r == 0) || (r == GLASS_SOCKET_ERROR)) {
			PACKET_LOG(GLASS_DEBUG, "Error in main recv loop: " << r);

			switch (r) {
				case 0:
					// Socket connection was closed unexpectedly
					PACKET_LOG(GLASS_DEBUG, "Socket connection has been closed unexpectedly");
					retval = PROTOCOL_CONNECTION_CLOSED; // Set return value for this function
					break;
				case GLASS_SOCKET_ERROR:
					// There was an error reading socket
	#if defined(_MSC_VER)
					PACKET_LOG(GLASS_ERROR, "::recv(): " << libglass_strerror_wsa(WSAGetLastError()));
	#endif // _MSC_VER
					PACKET_LOG(GLASS_ERROR, "Error reading socket");
					retval = PROTOCOL_NOT_INITIALIZED; // Set return value for this function
					break;
				default:
					PACKET_LOG(GLASS_ERROR, "UNHANDLED ERROR CODE");
					retval = PROTOCOL_NOT_INITIALIZED; // Just to make sure that it corresponds to an error code
					break;
			} // switch

			// Delete the dynamically allocated data buffer before returning
			delete[] headerbuffer;
			delete[] databuffer;

			// Return error code
			return retval; // Error
		}

		// Increment data buffer pointer
		s += r;
		// Decrement the length counter
		l -= r;

		if (l == 0) {
			if (readingdata == true) {
				break;
			}
			l = length - headerlen;
			s = databuffer;
			readingdata = true;
		}
		PACKET_LOG(GLASS_DEBUG, "Endloop: l = " << l << "; r = " << r << "; readingdata == " << readingdata);
	}

// #define REALLY_HARD_DEBUG2
#if defined ENABLE_GLASS_LOG && defined REALLY_HARD_DEBUG2
	std::cerr << "recv packet 0x";
	for (unsigned int i = 0; i < length - headerlen; i++) {
		fprintf(stderr, "%02x-", (int)(databuffer[i] & 0xFF));
	}
	std::cerr << std::endl;
#endif

	// Now that the whole packet has been read we can decode it
	if (p.decode((unsigned char *)headerbuffer, headerlen,
			(unsigned char *)databuffer, length - headerlen) == false) {
		PACKET_LOG(GLASS_ERROR, "Error decoding packet.");
		// Delete the dynamically allocated data buffer before returning
		delete[] headerbuffer;
		delete[] databuffer;
		// Return error code
		return PROTOCOL_INVALID_PACKET; // Error - decoding packet
	}
	delete[] headerbuffer; // Delete dynamically allocated data buffer before returning

	PACKET_LOG(GLASS_DEBUG, "Recv " << length << " bytes from [0x" << NODE_FORMAT p.getSenderNode() << "]: " << p);

	// TODO: the code below should go to Protocol
	nodeId id = AddressToId(address);
	if (id == nobodyId) {
#ifdef OLD
		PACKET_LOG(GLASS_WARNING, "Error finding packet sender.");
		PACKET_LOG(GLASS_WARNING, "Got " << p << "bytes");
		delete[] databuffer;
#endif
		PACKET_LOG(GLASS_DEBUG, "From unregistered node");

		// Return error code
		return PROTOCOL_UNREGISTERED_NODE;  // unregistered sender
	}

	// Set packet sender node ID
	p.setSenderNode(id);

	return PROTOCOL_OK; // Success
}


// Return values are:
// -2 = Unregistered sender
// -1 = Error
// 0 = Success
// 1 = Connection closed
// 2 = Timeout
int TCPUDP::_recv(Packet &p) {
	Packet			*o;
	bool			first = true;
	int			retval = 0;

	PACKET_LOG(GLASS_INFO, "Unreliable packet receive loop");

	// from which client?
	if (p.getSenderNode() == everybodyId || p.getSenderNode() == nobodyId)  { /* get from anybody */
		if (!queue.empty()) {
			PACKET_LOG(GLASS_DEBUG, "QUEUE.");
			o = queue.front();
			p = *o;
			queue.pop_front();
			delete o;
			return PROTOCOL_OK;
		}
		struct timeval timeout;
		int sel;
		fd_set fdtemp;

		timeout = default_timeout;

#if defined(_MSC_VER)
		// Under MS windows select() returns -1 if the file descriptor
		// set is empty. Hence, if the set is empty simply sleep
		// rather than call select()

		// Check if we have any clients connected
		if (socklist.fd_count == 0) {
			// PACKET_LOG(GLASS_DEBUG, "No clients connected");
			if (!no_timeout)
				Sleep( timeout.tv_sec * 1000 );
			return PROTOCOL_TIMEOUT; // i.e. Timeout condition
		}
#endif // _MSC_VER

		fdtemp = socklist;
		sel = select(fdmax+1, &fdtemp, NULL, NULL,
			no_timeout == true ? NULL : &timeout);

		if (sel == 0) {
			PACKET_LOG(GLASS_DEBUG, "Timeout: Select = 0. no_timeout:" << no_timeout);
			return 2; // Timeout
		}
		else if (sel == -1) {
			LOG(outlog(), GLASS_ERROR, "Select = -1.");
			PACKET_LOG(GLASS_ERROR, "Select = -1.");
#if defined(_MSC_VER)
			PACKET_LOG(GLASS_ERROR, "select(): " << libglass_strerror_wsa(WSAGetLastError()));
#else
			PACKET_LOG(GLASS_ERROR, "select(): " << strerror(errno));
#endif // _MSC_VER
			if (errno == EINTR) { // interrupted system call. Just ignore it.
				return PROTOCOL_TIMEOUT; // Timeout
			}
			return -1; // Error
		}
		PACKET_LOG(GLASS_INFO, "Got");
		for (int i = 0; i <= fdmax; i++ ) {
			if (FD_ISSET(i, &fdtemp)) {
				FD_CLR((GLASS_SOCKETFD_T)i, &fdtemp);

				nodeAddress addr(this->FDToAddress(i));
				if (addr.fd < 0) {
					// we have not registered this node yet.
					addr.fd = i;
				}

				/* with select, we must read
				   all available descriptors */
				if (queue.empty() && first == true) {
					if (i == this->listenfdudp) {
						LOG(outlog(), GLASS_DEBUG, "Got a UDP read.");
						retval = UDPreadPacket(p, addr);
					}
					else {
						retval = readPacket(p, addr); // TCP
					}
					first = false;
				}
				else {
					PACKET_LOG(GLASS_DEBUG, "Saving to queue.");
					o = new Packet(this->owner);
					if (i == this->listenfdudp) {
						LOG(outlog(), GLASS_ERROR, "Got a UDP read in queue.");
						retval = UDPreadPacket(*o, addr);
					}
					else {
						retval = readPacket(*o, addr); // TCP
					}
					queue.push_back(o);
				}
			}
		}
		PACKET_LOG(GLASS_DEBUG, "Returning from " << CodeDebug);
		// todo: NOT GOOD, we have several retvals. how to return?
		if (retval != PROTOCOL_OK && retval != PROTOCOL_UNREGISTERED_NODE) {
			LOG(outlog(), GLASS_ERROR, "Return readPacket " << retval);
		}
		return retval;
	}
	else {
		if (!queue.empty()) {
			std::list<Packet *>::iterator	it;

			for (it = queue.begin(); it != queue.end(); it++) {
				o = *it;
				if (o->getSenderNode() == p.getSenderNode()) {
					o = *queue.begin();
					p = *o;
					queue.erase(it);
					delete o;
					return 0; // Success
				}
			}
		}

		nodeAddress address(IdToAddress(p.getSenderNode()));
		fd_set fdtemp = socklist;
		if (!FD_ISSET(address.fd, &fdtemp)) {
			LOG(outlog(), GLASS_ERROR, "FD_ISSET not set.");
			PACKET_LOG(GLASS_ERROR, "FD_ISSET.");
			return -1;
		}
		FD_CLR((GLASS_SOCKETFD_T)address.fd, &fdtemp);
		/* note that this blocks indefinitely */
		return readPacket(p, address);
	}
	return 0;
}


#if defined(_MSC_VER)

// Get system error message and copy to global character buffer
// Wrapper around libglass_strerror_r_wsa().
// NOT THREAD SAFE
static char* libglass_strerror_wsa(int errnum)
{
	// Modify the global error message buffer
	int returnval = libglass_strerror_r_wsa(errnum, _errmsg_buf, GLASS_ERRMSG_BUF_SIZE);

	// Return pointer to global error message buffer
	return _errmsg_buf;
}


// Returns  the  appropriate  error description string, or an unknown
// error message if the error code is unknown. Upon success, returns 0
// and does not change the value of "errno".
// If error number is unrecognised then returns -1.
// Similar behaviour to strerror_r function in glibc.
// WINDOWS SOCKETS ERROR MESSAGES ONLY
static int libglass_strerror_r_wsa(int errnum, char *buf, size_t n)
{
	// Check for invalid arguments
	if ((!buf) || (n==0))
		return -1;

	int returnval = 0;

	switch (errnum) {
		case WSANOTINITIALISED:
			strncpy( buf, "A successful WSAStartup call must occur before using this function", n );
			break;
		case WSAENETDOWN:
			strncpy( buf, "The network subsystem or the associated service provider has failed", n );
			break;
		case WSAEAFNOSUPPORT:
			strncpy( buf, "The specified address family is not supported", n );
			break;
		case WSAEINPROGRESS:
			strncpy( buf, "A blocking Windows Sockets 1.1 call is in progress, or the service provider is still processing a callback function", n );
			break;
		case WSAEMFILE:
			strncpy( buf, "No more socket descriptors are available", n );
			break;
		case WSAENOBUFS:
			strncpy( buf, "No buffer space is available. The socket cannot be created", n );
			break;
		case WSAEPROTONOSUPPORT:
			strncpy( buf, "The specified protocol is not supported", n );
			break;
		case WSAEPROTOTYPE:
			strncpy( buf, "The specified protocol is the wrong type for this socket", n );
			break;
		case WSAESOCKTNOSUPPORT:
			strncpy( buf, "The specified socket type is not supported in this address family", n );
			break;
		case WSAECONNRESET:
			strncpy( buf, "An incoming connection was indicated, but was subsequently terminated by the remote peer prior to accepting the call", n );
			break;
		case WSAEFAULT:
			strncpy( buf, "The addrlen parameter is too small or addr is not a valid part of the user address space", n );
			break;
		case WSAEINTR:
			strncpy( buf, "A blocking Windows Sockets 1.1 call was canceled through WSACancelBlockingCall", n );
			break;
		case WSAEINVAL:
			strncpy( buf, "The listen function was not invoked prior to accept", n );
			break;
		case WSAENOTSOCK:
			strncpy( buf, "The descriptor is not a socket", n );
			break;
		case WSAEOPNOTSUPP:
			strncpy( buf, "The referenced socket is not a type that supports connection-oriented service", n );
			break;
		case WSAEWOULDBLOCK:
			strncpy( buf, "The socket is marked as nonblocking and no connections are present to be accepted", n );
			break;
		case WSAEACCES:
			strncpy( buf, "Attempt to connect datagram socket to broadcast address failed because setsockopt option SO_BROADCAST is not enabled", n );
			break;
		case WSAEADDRINUSE:
			strncpy( buf, "A process on the computer is already bound to the same fully-qualified address and the socket has not been marked to allow address reuse with SO_REUSEADDR. For example, the IP address and port are bound in the af_inet case). (See the SO_REUSEADDR socket option under setsockopt", n );
			break;
		case WSAEADDRNOTAVAIL:
			strncpy( buf, "The specified address is not a valid address for this computer", n );
			break;
		case WSAEISCONN:
			strncpy( buf, "The socket is already connected", n );
			break;
		case WSAENOTCONN:
			strncpy( buf, "The socket is not connected", n );
			break;
		case WSAENETRESET:
			strncpy( buf, "The connection has been broken due to the keep-alive activity detecting a failure while the operation was in progress", n );
			break;
		case WSAESHUTDOWN:
			strncpy( buf, "The socket has been shut down; it is not possible to receive on a socket after shutdown has been invoked with how set to SD_RECEIVE or SD_BOTH", n );
			break;
		case WSAEMSGSIZE:
			strncpy( buf, "The message was too large to fit into the specified buffer and was truncated", n );
			break;
		case WSAECONNABORTED:
			strncpy( buf, "The virtual circuit was terminated due to a time-out or other failure. The application should close the socket as it is no longer usable", n );
			break;
		case WSAETIMEDOUT:
			strncpy( buf, "The connection has been dropped because of a network failure or because the peer system failed to respond", n );
			break;
		default:
			// Unable to translate error code - Create dynamic error message
			_snprintf( buf, n, "Windows sockets error code %d", errnum );
			returnval = -1;
			break;
	}

	// Ensure that the error string is NULL terminated
	buf[n-1] = '\0';

	return returnval;
}


#endif // _MSC_VER
