/*  Glass - a distributed computing library
    Copyright (C) 2003-2009 Bruno Barberi Gnecco
    Copyright (C) 2009-2010 Corollarium Technologies


    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <errno.h>
#include <stdio.h>
#include <string.h>
#include "_glass.h"
#include "log.h"

#define MAXIMUM_PACKET_SIZE 16*1024*1024
#ifndef WIN32
#  include <sys/types.h>
#  include <sys/select.h>
#  include <sys/time.h>
#  include <unistd.h>
#  include <sys/socket.h>
#  include <netinet/in.h>
#  include <netinet/tcp.h>
#  include <arpa/inet.h>
#  include <netdb.h>
#  include <errno.h>
#  include <sys/un.h>
#else
// #include <winsock.h>
#  include <winsock2.h>
#  include <io.h>
#  include <process.h>
#endif // WIN32

#if defined(ENABLE_PACKET_LOG)
#  include "log.h"
#  if !defined(GLASS_STATIC_PACKET_LOG_FILENAME)
#    include <sstream>
#  endif
#endif

#include <boost/thread.hpp>
#include "unixsocket.h"

using namespace libglass;

// Define portablity macros
#if defined(_MSC_VER)
#  define GLASS_CLOSESOCKET closesocket
#  define GLASS_SETSOCKOPT_ARG4_T char*
#  define GLASS_RECV_ARG2_T char*
#  define GLASS_SOCKETFD_T SOCKET
#  define GLASS_SD_SEND SD_SEND
#  define GLASS_PROTOCOL_NUM IPPROTO_TCP

// #  define GLASS_SOCKET_ERROR SOCKET_ERROR - now defined in types.h

// Global character buffer for storing Windows Sockets API error message
#  define GLASS_ERRMSG_BUF_SIZE 512
char _errmsg_buf[GLASS_ERRMSG_BUF_SIZE];

// Function declarations
static char* libglass_strerror_wsa(int errnum);
static int libglass_strerror_r_wsa(int errnum, char *buf, size_t n);

#else
#  define GLASS_CLOSESOCKET close
#  define GLASS_SETSOCKOPT_ARG4_T void*
#  define GLASS_RECV_ARG2_T void*
#  define GLASS_SOCKETFD_T int
#  define GLASS_SD_SEND SHUT_WR
#  define GLASS_PROTOCOL_NUM 0

#endif


#if !HAVE_SOCKLEN_T
#  if defined(_MSC_VER)
  typedef int socklen_t;
#  else
  typedef unsigned long socklen_t;
#  endif // _MSC_VER
#endif

/*! Define static packet log filename
  Leave undefined if you want to use dynamic filename based on process ID
*/
// #define GLASS_STATIC_PACKET_LOG_FILENAME "packet.log"

#ifdef ENABLE_PACKET_LOG
// #  define PACKET_LOG(t, x)	if (plog) {LOG((*plog), t, x)}
#  define PACKET_LOG(t, x)	{LOG(outlog(), t, "Unix PACKET   " << x);}
#else
#  define PACKET_LOG(t, x)
#endif // ENABLE_PACKET_LOG


// Define maximum length of pending connections to listening server socket
#define GLASS_LISTEN_BACKLOG_LIMIT 6

// Constructor
UnixSocket::UnixSocket() : TCP() {
	PACKET_LOG(GLASS_DEBUG, "UnixSocket");
	this->PROTOCOL = AF_UNIX;

}

nodeId UnixSocket::getUniqueId(void) {
	return nobodyId; // TODO
}

UnixSocket::~UnixSocket() {
}

int UnixSocket::bind(int fd) {
	struct sockaddr_un addr;

	int    sockopt_arg;
	/* non blocking ? if so, have to update send() and recv() accordingly.
	fcntl(listenfd, F_SETFL, O_NONBLOCK); */

	/* Fill out the struct sockaddr_in structure */
	addr.sun_family = this->PROTOCOL;
	snprintf(addr.sun_path, sizeof(addr.sun_path), "/tmp/glass_socket"); // TODO: hardcoded path

	/* get rid of "address already in use" messages */
	sockopt_arg = 1;
	::setsockopt(listenfd, SOL_SOCKET, SO_REUSEADDR, (GLASS_SETSOCKOPT_ARG4_T)&sockopt_arg, sizeof(int));
	unlink(addr.sun_path);

	PACKET_LOG(GLASS_DEBUG, "Binding unix");
	return ::bind(fd, (struct sockaddr *)&addr, sizeof(addr));
}

nodeAddress UnixSocket::_connect(const nodeAddress &original) {
	nodeAddress	address(original);

	PACKET_LOG(GLASS_DEBUG, "UnixSocket connect");

	if (address.host != "localhost" && address.host != "127.0.0.1") {
		PACKET_LOG(GLASS_ERROR, "Invalid host: unix sockets only accept localhost");
		address.status = PROTOCOL_ALREADY_INITIALIZED;
		return address;
	}

	/* Attempt to open client socket */
	address.fd = ::socket(this->PROTOCOL, SOCK_STREAM, GLASS_PROTOCOL_NUM);

	// Check for error condition
#if defined(_MSC_VER)
	if (address.fd == INVALID_SOCKET) {
		PACKET_LOG(GLASS_ERROR, "socket(): " << libglass_strerror_wsa(WSAGetLastError()));
		address.status = PROTOCOL_NETWORK_ERROR;
		return address;
	}
#else
	if (address.fd < 0) {
		PACKET_LOG(GLASS_DEBUG, "socket(): " << strerror(errno));
		address.status = PROTOCOL_NETWORK_ERROR;
		return address;
	}
#endif // _MSC_VER

	address.un_addr.sun_family = this->PROTOCOL;
	snprintf(address.un_addr.sun_path, sizeof(address.un_addr.sun_path), "/tmp/glass_socket"); // TODO: hardcoded path

	if (::connect(address.fd, (struct sockaddr *)&address.un_addr, sizeof(struct sockaddr_un)) == -1) {
#if defined(_MSC_VER)
		PACKET_LOG(GLASS_ERROR, "connect(): " << libglass_strerror_wsa(WSAGetLastError()));
#else
		PACKET_LOG(GLASS_ERROR, "connect(): " << strerror(errno));
#endif
		close(address.fd);
		address.fd = 0;
		address.status = PROTOCOL_NETWORK_ERROR;
		return address;
	}

	LOG(outlog(), GLASS_DEBUG, "connected, fd=" << address.fd);

	// add to list
	FD_SET(address.fd, &socklist);
	if ((int)address.fd > fdmax)
		fdmax = (int)address.fd;

	return address;
}
