/*  Glass - a distributed computing library
    Copyright (C) 2003-2009 Bruno Barberi Gnecco
    Copyright (C) 2009-2010 Corollarium Technologies


    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/******************************************************************************
LibGlass event plugin tester

Program description:
	Event tests

Dependencies:
	all basic Glass structure.

Authors:
	Bruno Barberi Gnecco <brunobg@users.sourceforge.net>

History:
	31 August 2011: Initial implementation (Bruno)

Return codes:
0	Test ended
 ******************************************************************************/


#include <iostream>
#include <unistd.h>
#define BOOST_TEST_DYN_LINK
#define BOOST_TEST_MODULE event
#include <boost/test/unit_test.hpp>
#include "libglass.h"

#define NUM_LOOPS 5
#define TOTAL_CLIENTS 5

using namespace libglass;

boost::barrier b1(TOTAL_CLIENTS), b2(TOTAL_CLIENTS);

void client_thread(unsigned int id, unsigned int totalclients) {
	GlassClient	*g;
	Barrier *b;
	Event<int, rawPack, rawUnpack> *eventint;
	std::string nodename;

	BOOST_TEST_MESSAGE("Client trying to connect to localhost");

	try {
		std::ostringstream oss;
		oss << id;

		nodename = "node" + oss.str();
		g = new GlassClient(new PROTOCOLCLASS(), "localhost", nodename);
		BOOST_REQUIRE(g != NULL);
	}
	catch (Exception &e) {
		BOOST_FAIL("Client constructor failed: " + e.getMessage());
		return;
	}
	BOOST_TEST_MESSAGE("Client " << boost::this_thread::get_id() << " " << nodename << " connected, glass: " << (void *)g);

	try {
		b = new Barrier(0, 0, true);
	} catch (Exception &e) {
		BOOST_FAIL("Barrier constructor failed: " + e.getMessage());
		return;
	}

	try {
		eventint = new Event<int, rawPack, rawUnpack> ("eventint");
	}
	catch (Exception &e) {
		BOOST_FAIL("Event constructor failed: " + e.getMessage());
		return;
	}

	if (g->isMaster()) {
		BOOST_TEST_MESSAGE(nodename << " is master");
	}

	sleep(1);

	BOOST_MESSAGE("Testing");
	BOOST_REQUIRE_MESSAGE(totalclients+1 == g->getTotalNodes(),  "Expected " << totalclients+1 << " got " << g->getTotalNodes());

	BOOST_REQUIRE(eventint->isEmpty() == true);

	b1.wait();

	for (int i = 0; i < NUM_LOOPS; i++) {
		if (g->isMaster()) {
			BOOST_TEST_MESSAGE("Enqueueing " << i);
			bool b = eventint->enqueueEvent(i, everybodyId, 1, true);
			sleep (2); 			//o bastante para o evento ser enviado
			BOOST_REQUIRE(b == true);
		}

		b1.wait();
		b->sync();

		BOOST_CHECK_MESSAGE(eventint->isEmpty() == false, "isEmpty for node " << nodename);
		BOOST_CHECK_MESSAGE(eventint->queueSize() == 1, "queueSize for node " << nodename);

		try {
			int e = eventint->getEvent();
			sleep (2); 			//o bastante para o evento ser recebido
			BOOST_CHECK_MESSAGE(e == i, "eventint->getEvent() == " << e <<
					" should be " << i << " for client " << nodename);
		}
		catch (Exception e) {
			BOOST_FAIL("Exception in getEvent  " << e.getMessage());
		}

	}

	b2.wait();

	delete b;
	delete eventint;
	delete g;
	BOOST_TEST_MESSAGE("Client finished");

	return;
}

void server_thread() {
	GlassServer		*g;

	BOOST_TEST_MESSAGE("Server starting");

	try {
		g = new GlassServer(new PROTOCOLCLASS());
		BOOST_CHECK(g != NULL);
	}
	catch (Exception &e) {
		BOOST_FAIL("Server constructor failed: " + e.getMessage());
		return;
	}
	delete g;
	BOOST_TEST_MESSAGE("Server finished");

	return;
}

BOOST_AUTO_TEST_CASE(eventCase) {
	unsigned int clientnum = 5; // amount of clients

	BOOST_TEST_MESSAGE("Total clients: " << clientnum);

	// open server
	boost::thread server(&server_thread);
	boost::thread_group clients;
	unsigned int i;

	sleep(2);
	for (i = 0; i < clientnum; i++) {
		clients.create_thread(boost::bind(client_thread, i, clientnum));
	}

	BOOST_TEST_MESSAGE("Joining: " << clientnum);

	clients.join_all();

	BOOST_TEST_MESSAGE("Joined");
}
