/*  Glass - a distributed computing library
    Copyright (C) 2003-2009 Bruno Barberi Gnecco
    Copyright (C) 2009-2010 Corollarium Technologies


    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/******************************************************************************
LibGlass event plugin tester

Program description:
	Test event sending by nickname

Dependencies:
	all basic Glass structure.

Authors:
	Bruno Barberi Gnecco <brunobg@users.sourceforge.net>

History:
	31 August 2011: Initial implementation (Bruno)

Return codes:
0	Test ended
******************************************************************************/


#include <iostream>
#include <unistd.h>
#define BOOST_TEST_DYN_LINK
#define BOOST_TEST_MODULE event
#include <boost/test/unit_test.hpp>
#include "libglass.h"

#define NUM_LOOPS 5

using namespace libglass;

void client_thread(unsigned int id, unsigned int totalclients) {
	GlassClient	*g;
	Barrier *b;
	Event<int, rawPack, rawUnpack> *eventint;
	std::string nodename;

	BOOST_TEST_MESSAGE("Client trying to connect to localhost");

	// quick way to ensure node0 is master.
	if (id != 0) {
		sleep(1);
	}

	try {
		std::ostringstream oss;
		oss << id;

		nodename = "node" + oss.str();
		g = new GlassClient(new PROTOCOLCLASS(), "localhost", nodename);
		BOOST_REQUIRE(g != NULL);
	}
	catch (Exception &e) {
		BOOST_FAIL("Client constructor failed: " + e.getMessage());
		return;
	}
	BOOST_TEST_MESSAGE("Client " << boost::this_thread::get_id() << " " << nodename << " connected, glass: " << (void *)g);

	try {
		b = new Barrier(0, 0, true);
	} catch (Exception &e) {
		BOOST_FAIL("Barrier constructor failed: " + e.getMessage());
		return;
	}

	try {
		eventint = new Event<int, rawPack, rawUnpack> ("eventint");
	}
	catch (Exception &e) {
		BOOST_FAIL("Event constructor failed: " + e.getMessage());
		return;
	}

	if (g->isMaster()) {
		BOOST_TEST_MESSAGE(nodename << " is master");
	}

	BOOST_REQUIRE(eventint->isEmpty() == true);

	// avoids a small race condition due to the initial sleep.
	if (g->isMaster()) {
		sleep(1);
	}

	b->sync();

	for (int i = 0; i < NUM_LOOPS; i++) {
		if (g->isMaster()) {
			BOOST_TEST_MESSAGE("Enqueueing " << i);
			bool b = eventint->enqueueEvent(i, "node1");
			BOOST_REQUIRE(b == true);
		}

// 		b->sync();

		sleep(1); // should be enough for event be transmitted
		if (nodename == "node1") {
			BOOST_CHECK_MESSAGE(eventint->isEmpty() == false, "isEmpty for node " << nodename);
			BOOST_CHECK_MESSAGE(eventint->queueSize() > 0, "queueSize for node " << nodename);
			try {
				int e = eventint->getEvent();
				BOOST_CHECK_MESSAGE(e == i, "eventint->getEvent() == " << e <<
						" should be " << i << " for client " << nodename);
			}
			catch (Exception e) {
				// BOOST_FAIL("Exception in getEvent  " << e.getMessage());
			}
		}
		else {
			BOOST_CHECK_MESSAGE(eventint->isEmpty() == true, "isEmpty for node " << nodename);
			BOOST_CHECK_MESSAGE(eventint->queueSize() == 0, "queueSize for node " << nodename);
		}
	}

	delete b;
	delete eventint;
	delete g;
	BOOST_TEST_MESSAGE("Client finished");

	return;
}

void server_thread() {
	GlassServer		*g;

	BOOST_TEST_MESSAGE("Server starting");

	try {
		g = new GlassServer(new PROTOCOLCLASS());
		BOOST_CHECK(g != NULL);
	}
	catch (Exception &e) {
		BOOST_FAIL("Server constructor failed: " + e.getMessage());
		return;
	}
	delete g;
	BOOST_TEST_MESSAGE("Server finished");

	return;
}

BOOST_AUTO_TEST_CASE(eventnicknameCase) {
	unsigned int clientnum = 4; // amount of clients

	BOOST_TEST_MESSAGE("Total clients: " << clientnum);

	// open server
	boost::thread server(&server_thread);
	boost::thread_group clients;
	unsigned int i;

	sleep(2);
	for (i = 0; i < clientnum; i++) {
		clients.create_thread(boost::bind(client_thread, i, clientnum));
	}

	BOOST_TEST_MESSAGE("Joining: " << clientnum);

	clients.join_all();

	BOOST_TEST_MESSAGE("Joined");
}
