/*  Glass - a distributed computing library
    Copyright (C) 2003-2009 Bruno Barberi Gnecco
    Copyright (C) 2009-2010 Corollarium Technologies


    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/******************************************************************************
LibGlass shared plugin tester

Program description:
	Shared tests 

Dependencies:
	all basic Glass structure.

Authors:
	Bruno Barberi Gnecco <brunobg@users.sourceforge.net>
	Diego Roberto Colombo Dias <diegocolombo@users.sourceforge.net>

History:
	31 August 2011: Initial implementation (Bruno)

Return codes:
0	Test ended
 ******************************************************************************/


#include <iostream>
#include <unistd.h>
#define BOOST_TEST_DYN_LINK
#define BOOST_TEST_MODULE shared
#include <boost/test/unit_test.hpp>
#include "libglass.h"

#define NUM_LOOPS 5
#define TOTAL_CLIENTS 5

using namespace libglass;

boost::barrier b1(TOTAL_CLIENTS), b2(TOTAL_CLIENTS);

void client_thread(unsigned int id, unsigned int totalclients) {
	GlassClient	*g;
	Barrier *b;
	Shared<int, rawPack, rawUnpack> *sharedint;
	std::string nodename;

	BOOST_TEST_MESSAGE("Client trying to connect to localhost");

	try {
		std::ostringstream oss;
		oss << id;

		nodename = "node" + oss.str();
		g = new GlassClient(new PROTOCOLCLASS(), "localhost", nodename);
		BOOST_REQUIRE(g != NULL);
	}
	catch (Exception &e) {
		BOOST_FAIL("Client constructor failed: " + e.getMessage());
		return;
	}
	BOOST_TEST_MESSAGE("Client " << boost::this_thread::get_id() << " connected, glass: " << (void *)g);

	try {
		b = new Barrier(0, 0, true);
	} catch (Exception &e) {
		BOOST_FAIL("Barrier constructor failed: " + e.getMessage());
		return;
	}

	try {
		sharedint = new Shared<int, rawPack, rawUnpack> ("sharedint");
	}
	catch (Exception &e) {
		BOOST_FAIL("Shared constructor failed: " + e.getMessage());
		return;
	}

	*sharedint = 0;
	BOOST_REQUIRE(sharedint->getData() == 0);

	if (g->isMaster()) {
		BOOST_TEST_MESSAGE(nodename << " is master");
	}

	b1.wait();
	b->sync();

	for (int i = 0; i < NUM_LOOPS; i++) {
		if (g->isMaster()) {
			*sharedint = i;
			sharedint->sendUpdate();
			sleep(2);
		}

		b1.wait();

		sharedint->getUpdate();

		b->sync();


		BOOST_CHECK_MESSAGE(sharedint->getData() == i, "sharedint->getData() == " << sharedint->getData() <<
				" should be " << i << " for client " << nodename);

		b1.wait();
	}

	delete b;
	delete sharedint;
	delete g;
	BOOST_TEST_MESSAGE("Client finished");

	return;
}

void server_thread() {
	GlassServer		*g;

	BOOST_TEST_MESSAGE("Server starting");

	try {
		g = new GlassServer(new PROTOCOLCLASS());
		BOOST_CHECK(g != NULL);
	}
	catch (Exception &e) {
		BOOST_FAIL("Server constructor failed: " + e.getMessage());
		return;
	}
	delete g;
	BOOST_TEST_MESSAGE("Server finished");

	return;
}

BOOST_AUTO_TEST_CASE(sharedCase) {
	unsigned int clientnum = 5; // amount of clients

	BOOST_TEST_MESSAGE("Total clients: " << clientnum);

	// open server
	boost::thread server(&server_thread);
	sleep (4);

	boost::thread_group clients;
	unsigned int i;

	for (i = 0; i < clientnum; i++) {
		clients.create_thread(boost::bind(client_thread, i, clientnum));
	}

	BOOST_TEST_MESSAGE("Joining: " << clientnum);

	clients.join_all();

	BOOST_TEST_MESSAGE("Joined");
}
