/*  Glass - a distributed computing library
    Copyright (C) 2003-2009 Bruno Barberi Gnecco
    Copyright (C) 2009-2010 Corollarium Technologies


    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/******************************************************************************
LibGlass barrier plugin tester

Program description:
	Tests the barrier plugin.

Dependencies:
	all basic Glass structure, pthreads.

Authors:
	Bruno Barberi Gnecco <brunobg@users.sourceforge.net>
	Marcelo de Paiva Guimaraes<marpaiva@users.sourceforge.net>

History:
	02 Nov 2004: Initial implementation (Marcelo)

Return codes:
0	Test ended
******************************************************************************/


#include <iostream>
#include <unistd.h>
#include "libglass.h"

using namespace libglass;

//#define MANUAL_DEBUGGING

using std::cout;
using std::endl;

#define NUM_BARRIERS	5
#define NUM_LOOPS	10

unsigned int clientnum = 2; // amount of clients
unsigned int TOTAL_TESTS = 2; // amount of tests

// Struct used to store the tests
struct Test {
	char			*description; // test executed
	unsigned int		status; //   0=ok  1 = there is a problem
	string                  error; // error returned
};

// tests executed
Test tests[] = {
	{"Server Initialization", 0, },
	{"Client Initialization - TCP", 0},
	{"Barrier Creation", 0},
	{"Barrier Test", 0}
};

//Struct used to check if the lock is been done correctly.
struct Barrier_Tests {
	pthread_t		thread_number;
	enum {
		LOCKED, UNLOCKED
	} status;
//	unsigned int		status; // 0 - unlocked // 1 - locked
};
Barrier_Tests *barrier_test;

// The status value of this thread is changed to LOCKED
void set_locked(pthread_t xthread){
	unsigned int i;
	for (i = 0; i < clientnum; i++) {
		if (barrier_test[i].thread_number == xthread){
			if (barrier_test[i].status == Barrier_Tests::UNLOCKED){
				barrier_test[i].status = Barrier_Tests::LOCKED;
#ifdef MANUAL_DEBUGGING
			cout << "Thread number :" <<  pthread_self() << " LOCKED "<< endl;
#endif

			}
			// if the status is already LOCKED means that the same thread
			// sent another sync sinal. It should do nothing.
		}
	}
}

void clean_locked(pthread_t xthread){
	for (unsigned int i = 0; i < clientnum; i++) {
		if (barrier_test[i].thread_number == xthread){
			barrier_test[i].status = Barrier_Tests::UNLOCKED;
#ifdef MANUAL_DEBUGGING
		cout << "Thread number :" <<  barrier_test[i].thread_number << " UNLOCKED "<< endl;
#endif
		}
	}
}

// It tests if the sync sinal is being executed in order
void check_barrier_ifsync(){
	for (unsigned int i = 0; i < clientnum; i++) {
		if (barrier_test[i].status == Barrier_Tests::UNLOCKED) { // It is not syncronized
			tests[3].status = 1;
			tests[3].error = "Barrier was not syncronized correctly";

#ifdef MANUAL_DEBUGGING
			cout << "Barrier error...." << endl;
#endif

		}
	}
}


void  test_barrier(){
	/* For first dish, a fresh Barrier. */
	Barrier			**b;

	unsigned int		i, j;

	cout << "Thread number: "<< pthread_self()<< " Client - Executing test: " << tests[2].description<< endl;
	try {
		cout << "Thread number: "<< pthread_self()<< " Creating Barries"<< endl;
		b = new Barrier *[NUM_BARRIERS];
		for (i = 0; i < NUM_BARRIERS; i++) {
			cout << "Thread number: "<< pthread_self() <<" Creating Barrier: "<< i << endl;
			b[i] = new Barrier(i,false);
		}
	}
	catch (Exception e) {
		cout << "Error: Barrier initialization"<< endl;
		tests[2].status = 1;
		//tests[2].error = "Thread " + pthread_self() + e.getMessage();
		return;
	}
	/* The main course is 60 synchronizations of the barrier. */
	cout << "Client - Executing test number: 3 " << tests[3].description<< endl;
	sleep(5);
	for (i = 0; i < NUM_LOOPS; i++) {
		for (j = 0; j < NUM_BARRIERS; j++) {
#ifdef MANUAL_DEBUGGING
			cout << "Thread number :" <<  pthread_self() << "  locking!  step " << j << endl;
#endif
			b[j]->sync();

			// The status value of this thread is changed to LOCKED
			set_locked(pthread_self());
			b[j]->sync();
#ifdef MANUAL_DEBUGGING
			cout << "Checking status " << endl;
#endif

			// Check if there is UNLOCKED barrier.
			// If there is UNLOCKED barrier, the sync system is wrong.
			check_barrier_ifsync();
			b[j]->sync();
			// The status value of this thread is changed to UNLOCKED
  		        clean_locked(pthread_self());
		}

	}
	cout << "Deleting Client barries - test number: 3 " << endl;
	for (i = 0; i < NUM_BARRIERS; i++) {
		delete b[i];
	}
	cout << "Client barries deleted - test number: 3 " << endl;
	delete[] b;
	return;
}


void* server_thread(void *arg) {
	cout << "Server - Executing test number: 0 " << tests[0].description << endl;
	GlassServer		*g;
	try {
		g = new GlassServer(new TCP());
	}
	catch (Exception e) {
		cout <<" Failed" << endl;
		tests[0].status=1;
		tests[0].error = e.getMessage();
		return NULL;
	}
	delete g;
	return NULL;
}

void* client_thread(void *arg) {
	GlassClient		*g;
	long barrier_number = (long)arg;
	cout << "Executing test number: " << barrier_number << tests[1].description << endl;
#ifdef MANUAL_DEBUGGING
	cout << "Client trying to connect to localhost" << endl;
#endif
	try {
		g = new GlassClient(new TCP(), "localhost");
		g->setCurrent();
	}
	catch (Exception e) {
		tests[1].status = 1;
		tests[1].error = e.getMessage();
		return NULL;
	}
	// add this barrier
	barrier_test[barrier_number].thread_number = pthread_self();
	barrier_test[barrier_number].status = Barrier_Tests::UNLOCKED;
	barrier_number++;
	test_barrier();
	delete g;
	return NULL;

}


int main(int argc, char **argv) {
	unsigned int i;
	pthread_t 		server, *clients;

	if (argc > 1) {
		clientnum = atoi(argv[1]);
	}
	cout << "Total of tests : " << TOTAL_TESTS << "\tTotal of clients: " << clientnum << endl;
	barrier_test = (Barrier_Tests *)malloc(sizeof(Barrier_Tests)*clientnum+1);
	clients = new pthread_t[clientnum];
	pthread_create(&server, NULL, server_thread, NULL);

	// create threads
	// for barrier < barrier_number;
	for (i = 0; i < clientnum; i++) {
		pthread_create(&clients[i], NULL, client_thread, 0);
	}

	for (i = 0; i < clientnum; i++) {
		pthread_join(clients[i], NULL);
	}

	cout << endl;

	int flag = 0;
	for (i = 0; i < TOTAL_TESTS; i++) {
		if (tests[i].status == 1) {
			cout << "Error: " << tests[i].description << "\t\tTest number: "<<i<< endl;
			cout << "  Message: " << tests[i].error <<endl;
			flag = 1;
		}else	{
			cout << tests[i].description << ".....OK!" <<endl;
		}
	}
	if (flag == 0)
		cout << endl << "No errors detected: " << TOTAL_TESTS << " tests done" << endl;
	cout <<endl;

	return 0;
}
