/*  Glass - a distributed computing library
    Copyright (C) 2003-2009 Bruno Barberi Gnecco
    Copyright (C) 2009-2010 Corollarium Technologies


    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/******************************************************************************
LibGlass event plugin tester

Program description:
	Tests the event plugin.

Dependencies:
	Glass

Authors:
	Bruno Barberi Gnecco <brunobg@users.sourceforge.net>

History:
	26 dec 2004: Initial implementation
	09 jan 2006: Fixed bugs (Marcelo Paiva)

Return codes:
0	Test ended successfully.
1	Could not instantiate Glass.
******************************************************************************/

#include <iostream>
#include <stdio.h>
#include "libglass.h"
#include <boost/thread/barrier.hpp>

#if defined(_MSC_VER)
#  define SLEEP(a) Sleep(a * 1000)
#else
#  define SLEEP(a) sleep(a)
#endif // _MSC_VER

using namespace libglass;
using std::cout;
using std::endl;

unsigned int numsenders = 1;
unsigned int numreceivers = 3;
unsigned int numevents = 10;
unsigned int strsize = 35;
char *teststr = NULL;
boost::barrier senderbarrier(numsenders);
boost::barrier receiverbarrier(numreceivers);

class TestServer : public GlassThread  {
private:
	GlassServer			*g;
public:
	TestServer() : GlassThread(false) {
		start(); // start thread

		// we have to ensure that the server is up, so
		// hang on a couple of seconds.
		sleep(2);
	}

	~TestServer() {
		delete g;
	}

	void cleanup(void) {
//		delete g;
	}

	void execute(void) {
		try {
			g = new GlassServer(new TCP());
		}
		catch (Exception e) {
			printf("Server: Initializaiont error \n");			
			return;
		}
	}
};

class TestClient : public GlassThread  {
private:
	GlassClient			*g;
	Event<int> 			*eventint;
	Event<std::string> 		*eventstr;
	bool				send;
	bool				ready;
	unsigned int			eventnum;


public:
	TestClient(bool send) : GlassThread(false), send(send), ready(false),
			eventnum(0) {
		start(); // start thread
	}
	~TestClient() {
		delete g;
	}

	void cleanup(void) {
//		delete g;
	}

	void execute(void) {
//		cout << "Executing test number: 1 " << tests[1].description << endl;
#ifdef MANUAL_DEBUGGING
		cout << "Client trying to connect to localhost" << endl;
#endif
		try {
			g = new GlassClient(new TCP(), "localhost");
		}
		catch (Exception e) {
			printf("Error: Client Initialization \n");
			return;
		}
		// create events
		try {
			eventint = new Event<int>("integer");
			eventstr = new Event<std::string>("string");
		}
		catch (Exception e) {
			printf("Error: event creating \n");		
			return;
		}

		ready = true; // this should work even in concurrent accesses

		// wait for signal
		if (send) {
			senderbarrier.wait();
		}
		else {
			receiverbarrier.wait();
		}

		while (eventnum < numevents) {
			if (send == true) {
				eventnum++;				
				if (eventint->enqueueEvent(eventnum) == false) {
					printf("Error enqueuing int \n");
				}
				else {
					printf("Enqueuing int value = %d \n",eventnum);
				}
				SLEEP(1); // seconds
			}
			else{				
				if (eventint->isEmpty()) {
					SLEEP(1); // seconds
				}
				else {
					int i = eventint->getEvent();
					std::cout << "  Thread " << this->t.get_id() << " received int value = " << i << std::endl;
					eventnum++;
				}
			}
		}
		eventnum=0;
		int charvalue=96;
		while (eventnum < numevents) {
			if (send == true) {
				eventnum++;		
				charvalue++;
				if (charvalue>172)
				     charvalue=97;
				memset(teststr, charvalue , strsize);		
				if (eventstr->enqueueEvent(teststr) == false) {
					printf("Error enqueuing string \n");
				}	
				else {
					printf("Enqueuing str %i  value = %s \n",eventnum ,teststr);
				}
				SLEEP(1); // seconds
			}
			else{				
				if (eventstr->isEmpty()) {
					SLEEP(1); // seconds
				}
				else {
					std::cout << "  Thread " << this->t.get_id() << " event=" << eventnum <<
							" received str =" <<  eventstr->getEvent() << std::endl;
					eventnum++;
				}	
			}
		}		
		
		delete g;
		pthread_exit(NULL);
	}

	bool isReady(void) const {
		return ready;
	}

	void check(void) const {
		// todo!
	}
};

int main(int argc, char **argv) {
	unsigned int		i;
	TestServer		*ts;
	TestClient		*senders[numsenders], *receivers[numreceivers];

	/* TODO: args
	if (argc > 1) {
		clientnum = atoi(argv[1]);
	} */
	ts = new TestServer();
	teststr = new char [strsize];

	cout << "Number of client senders: " << numsenders << endl <<
		"Number of client receivers: " << numreceivers << endl <<
		"Number of events per sender: " << numevents << endl <<
		"String test size: " << strsize <<endl;

	for (i = 0; i < numsenders; i++) {
		senders[i]= new TestClient(true);
	}
	for (i = 0; i < numreceivers; i++) {
		receivers[i]= new TestClient(false);
	}

	/* wait for all threads to be ready */
	for (i = 0; i < numsenders; i++) {
		while (senders[i]->isReady() == false) {
			sleep(1);
		}
	}
	for (i = 0; i < numreceivers; i++) {
		while (receivers[i]->isReady() == false) {
			sleep(1);
		}
	}

	/* ready, set, GO! */
	for (i = 0; i < numsenders; i++) {
		/// automatic now senders[i]->signal();
	}
	for (i = 0; i < numreceivers; i++) {
		/// automatic now receivers[i]->signal();
	}

	cout << "started" << endl;

	/* Wait for all tasks to complete */
	for (i = 0; i < numsenders; i++) {
		senders[i]->join();
		cout << "senders[" << i << "] done" << endl;
	}
	for (i = 0; i < numreceivers; i++) {
		receivers[i]->join();
		cout << "receivers[" << i << "] done" << endl;
	}

	/* check results */
	for (i = 0; i < numreceivers; i++) {
		receivers[i]->check();
	}
	delete [] teststr;	
	return 0;
}
