/*  Glass - a distributed computing library
 Copyright (C) 2003-2009 Bruno Barberi Gnecco
 Copyright (C) 2009-2010 Corollarium Technologies


 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/******************************************************************************
 LibGlass barrier plugin tester

 Program description:
 Tests the barrier plugin.

 Dependencies:
 all basic Glass structure, pthreads.

 Authors:
 Marcelo de Paiva Guimaraes <marpaiva@users.sourceforge.net>

 History:
 27 Fev 2005: Initial implementation (Marcelo)

 Return codes:
 0	Test ended
 ******************************************************************************/

#include <iostream>
#include <unistd.h>
#include "libglass.h"

using namespace libglass;

//#define MANUAL_DEBUGGING

using std::cout;
using std::endl;

#define NUM_BARRIERS	5
#define NUM_LOOPS	50

unsigned int clientnum = 2; // amount of clients
unsigned int TOTAL_TESTS = 4; // amount of tests

// Struct used to store the tests
struct Test {
	char * description; // test executed
	unsigned int status; //   0=ok  1 = there is a problem
	string error; // error returned
};

// tests executed
Test tests[] = { { "Server Initialization", 0, }, {
		"Client Initialization - TCP", 0 }, { "Shared Creation", 0 }, {
		"Shared Test", 0 } };

void test_shared(bool ismaster) {
	/* For first dish, a fresh Barrier. */
	Barrier * b;
	unsigned int i, j;
	int testvalue = 0;
	Shared<int, rawPack, rawUnpack> *sharedint = new Shared<int, rawPack,
			rawUnpack> ("sharedint");
	cout << "Executing test number: 2: " << tests[2].description << endl;
	try {
		b = new Barrier(0,false);
	} catch (Exception e) {
		cout << "Error : Test number 2: creating barrier" << endl;
		tests[2].status = 1;
		tests[2].error = "Thread " + pthread_self() + e.getMessage();
		return;
	}
	cout << "Test number: 2 finished" << endl;
	/* The main course is 60 synchronizations of the barrier. */
	cout << "Executing test number: 3 " << tests[3].description << endl;
	cout << "aqui 1 " << endl;
	sleep(5);
	cout << "aqui 2 " << endl;
	*sharedint = 0;
	cout << "aqui 3 " << endl;
	for (i = 0; i < NUM_LOOPS; i++) {
		for (j = 0; j < NUM_BARRIERS; j++) {
#ifdef MANUAL_DEBUGGING
			cout << "Thread number :" << pthread_self () << "  locking!  step " << j << endl;
#endif
			if (ismaster) {
#ifdef MANUAL_DEBUGGING
				cout << "Thread number :" << pthread_self () << " Before: " <<sharedint->getData () << endl;
#endif
				*sharedint = sharedint->getData() + 1;
				testvalue = sharedint->getData();
				sharedint->sendUpdate();
#ifdef MANUAL_DEBUGGING
				cout << "Thread number :" << pthread_self () << " After : " <<sharedint->getData () << endl;
#endif
			}
			b->sync();
			sharedint->getUpdate();
			if (ismaster) {
#ifdef MANUAL_DEBUGGING
				cout << "Thread number :" << pthread_self () << " testvalue: "
				<< testvalue << " sharedint: " << sharedint->getData () << endl;
#endif
				if (testvalue != sharedint->getData()) {
					tests[3].status = 1;
					tests[3].error
							= " The Master(client) value is not being updated";
				}
			}
#ifdef MANUAL_DEBUGGING
			cout << "Thread number :" << pthread_self () <<
			"  Shared Variable Value: " << sharedint->getData () << endl;
#endif
		}
	}
	cout << "Test number: 3 finished - deleting instance" << endl;
	delete b;
	return;
}

void *
server_thread(void *arg) {
	cout << "Executing test number: 0 " << tests[0].description << endl;
	GlassServer * m;
	try {
		m = new GlassServer(new TCP());
	} catch (Exception e) {
		cout << " Failed" << endl;
		tests[0].status = 1;
		tests[0].error = e.getMessage();
		return NULL;
	}
	delete m;
	return NULL;
}

void *
client_thread(void *arg) {
	GlassClient * g;
	cout << "Executing test number: 1 " << tests[1].description << endl;
#ifdef MANUAL_DEBUGGING
	cout << "Client trying to connect to localhost" << endl;
#endif
	try {
		g = new GlassClient(new TCP(), "localhost");
	} catch (Exception e) {
		tests[1].status = 1;
		tests[1].error = e.getMessage();
		return NULL;
	}

	if (g->isMaster()) {
		cout << "Thread number :" << pthread_self() << "-> Client Master "
				<< endl;
		test_shared(true);
	} else {
		cout << "Thread number :" << pthread_self() << "-> Client " << endl;
		test_shared(false);
	}
	delete g;
	return NULL;
}

int main(int argc, char **argv) {
	unsigned int i;
	pthread_t server, *clients;

	if (argc > 1) {
		clientnum = atoi(argv[1]);
	}
	cout << "Total of tests : " << TOTAL_TESTS << "\tTotal of clients: "
			<< clientnum << endl;
	clients = new pthread_t[clientnum];
	pthread_create(&server, NULL, server_thread, NULL);

	for (i = 0; i < clientnum; i++) {
		pthread_create(&clients[i], NULL, client_thread, NULL);
	}

	for (i = 0; i < clientnum; i++) {
		pthread_join(clients[i], NULL);
	}

	cout << endl;

	int flag = 0;
	for (i = 0; i < TOTAL_TESTS; i++) {
		if (tests[i].status == 1) {
			cout << "Error: " << tests[i].description << "\t\tTest number: "
					<< i << endl;
			cout << "  Message: " << tests[i].error << endl;
			flag = 1;
		} else {
			cout << tests[i].description << ".....OK!" << endl;
		}
	}
	if (flag == 0)
		cout << endl << "No errors detected: " << TOTAL_TESTS << " tests done"
				<< endl;
	cout << endl;

	return 0;
}
