/*  Glass - a distributed computing library
    Copyright (C) 2003-2009 Bruno Barberi Gnecco
    Copyright (C) 2009-2010 Corollarium Technologies


    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/******************************************************************************
LibGlass packager system tester

Program description: 
	Tests the serialization and unserialization functions.

Dependencies:
	chain

Authors:
	Bruno Barberi Gnecco <brunobg@users.sourceforge.net>

History: 
	02 Jul 2004: XXXXXXXXXXXXX

Return codes:
XXXXXXXXXXXXXXXXXXXXXXX

******************************************************************************/


#include <iostream>
#include "types.h"
#include "packager.h"

using namespace std;
using namespace libglass;

namespace libglass {

#define serializer toString
#define unserializer fromString
	
bool test_short(void) {
	short values[] = { 256, 0xFFFF, 1234, -4321 };
	short s;
	unsigned int i;
	unsigned int length;

	cout << "Testing short." << endl;
	for (i = 0; i < sizeof(values)/sizeof(short); i++) {
		chain c;
		if (serializer<short>(c, values[i]) == false) {
			cout << "Short: error encoding " << values[i] << endl;
			return false;
		}
			
		if (unserializer<short>(c, 0, s, length) == false) {
			cout << "Short: error decoding " << values[i] << endl;
			return false;
		}
			
		if (s != values[i]) {
			cout << "Short: " << s << " != " << values[i] << endl;
			return false;
		}
	}
	cout << "Passed." << endl;
	return true;
}

bool test_int(void) {
	int values[] = { 0, 0xFFFFFFFF, 123456789, -987654321 };
	int s;
	unsigned int i;
	unsigned int length;

	cout << "Testing int." << endl;
	for (i = 0; i < sizeof(values)/sizeof(int); i++) {
		chain c;
		if (serializer<int>(c, values[i]) == false) {
			cout << "Int: error encoding " << values[i] << endl;
			return false;
		}
		if (unserializer<int>(c, 0, s, length) == false) {
			cout << "Int: error decoding " << values[i] << endl;
			return false;
		}
		
		if (s != values[i]) {
			cout << "Int: " << s << " != " << values[i] << endl;
			return false;
		}
	}
	cout << "Passed." << endl;
	return true;
}

bool test_int_array(void) {
	int values[] = { 0, 0xFFFFFFFF, 123456789, -987654321 };
	int s[sizeof(values)/sizeof(int)];
	unsigned int i;
	unsigned int length;

	cout << "Testing int array." << endl;
	chain c;
	if (serializer<int>(c, values, sizeof(values)/sizeof(int)) == false) {
		cout << "Int: error encoding array. " << endl;
		return false;
	}
	if (unserializer<int>(c, 0, s, length, sizeof(values)/sizeof(int)) == false) {
		cout << "Int: error decoding array. " << endl;
		return false;
	}

	for (i = 0; i < sizeof(values)/sizeof(int); i++) {
		if (s[i] != values[i]) {
			cout << "Int " << i << ": " << s[i] << " != " << values[i] << endl;
			return false;
		}
	}
	cout << "Passed." << endl;
	return true;
}

bool test_float(void) {
	float values[] = { 0.0f, 1.23456f, 2.2348f, -19395.2392f, 18592.1f };
	float s;
	unsigned int i;
	unsigned int length;

	cout << "Testing float." << endl;
	for (i = 0; i < sizeof(values)/sizeof(float); i++) {
		chain c;
		if (serializer<float>(c, values[i]) == false) {
			cout << "Float: error encoding " << values[i] << endl;
			return false;
		}
		if (unserializer<float>(c, 0, s, length) == false) {
			cout << "Float: error decoding " << values[i] << endl;
			return false;
		}
		if (s - values[i] > 0.1 || s - values[i] < -0.1 ) {
			cout << "Float: " << s << " != " << values[i] << endl;
			return false;
		}
	}
	cout << "Passed." << endl;
	return true;
}

bool test_stdstring(void) {
	std::string s("This is a string."), s2;
	unsigned int length;

	cout << "Testing std::string." << endl;
	chain c;
	if (serializer<std::string>(c, s) == false) {
		cout << "std::string: error encoding " << s << endl;
		return false;
	}
	if (unserializer<std::string>(c, 0, s2, length) == false) {
		cout << "std::string: error decoding " << s << ". Length " << length << endl;
		return false;
	}
	if (s != s2) {
		cout << "std::string: " << s << " != " << s2 << endl;
		return false;
	}
	cout << "Passed." << endl;
	return true;
}

} // namespace

int main() {
	unsigned int	 	i;
	bool (*tests[])(void) = { 
		test_short, 
		test_int, 
		test_float,
		test_int_array,
		test_stdstring
	};
	int		errors = 0;

	cout << "Executing : " << sizeof(tests)/sizeof(*tests) << " tests." << endl;
	for (i = 0; i < sizeof(tests)/sizeof(*tests); i++) {
		if (tests[i]() == false) {
			errors++;
		}
	}
	cout << errors << " errors detected." << endl;

	return 0;
}
