/*  Glass - a distributed computing library
    Copyright (C) 2003-2009 Bruno Barberi Gnecco
    Copyright (C) 2009-2010 Corollarium Technologies


    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/******************************************************************************
LibGlass protocol class tester

Program description: 
	This program tests a protocol class to ensure that it's working. 
	Most of the Protocol methods, including the virtual ones, are called and
	the result validated against the expected behavior.

Dependencies:
	Packet, Glass and chain classes must be working correctly.

Authors:
	Bruno Barberi Gnecco <brunobg@users.sourceforge.net>

History: 
	12 Sep 2003: Initial implementation (BBG)
	02 Jul 2004: More thorough tests (BBG)

Return codes:
- Server
0	Successful.
1	Error listening.
2	Error accepting connection.
3	Error receiving data.
4	Error sending data.

- Client
0	Successful.
1	Error connecting.
2	Error sending data.
3	Error receiving data.

******************************************************************************/


#include "libglass.h"

using namespace libglass;

class GlassDummy : public Glass {
private:

protected:
	virtual bool rootPacket(Packet &p) {
		return true;
	}

	// very simple, only one node at a time.
	virtual nodeId registerNode(nodeAddress &address, nodeId id, nodeRelation r, bool force) {
		if (setNode(address, baseServerId, Server) == false) {
			return nobodyId;
		}
		return baseServerId;
	}
public:
	GlassDummy(Protocol *p) throw(Exception) : Glass(p) {
		;
	}

	unsigned int getTotalNodes(void) {
		return 0;
	}

	// Glass::sendPacket is protected, we need it public here
	bool sendPacket(Packet &p) {
		return Glass::sendPacket(p);
	}
};

int main(int argc, char **argv) {
	nodeAddress		address;
	Protocol 		*p = new TCP();
	const char		*data = "1234567890";
	int			retval;
	GlassDummy		gd(p);
	Packet			pack(&gd);

	if (argc == 2) { // client
		std::cout << "Opening client." << std::endl;
		std::cout << "Trying to connect to " << argv[1] << std::endl;
		address = p->connect(argv[1]);
		if (address.status != PROTOCOL_OK) {
			std::cout << "Connect returned " << address.status << std::endl;
			return 1;
		}
		std::cout << "Connect successful: " << address.fd << std::endl;

		// TODO gd->setNode(this->serveraddr, nobodyId, Server, true);

		std::cout << "Sending data..." << std::endl;
		pack.setData(data, strlen(data));
		pack.setTargetNodes(baseServerId);
		if (gd.sendPacket(pack) == false) {
			std::cout << "Error sending data:" << std::endl;
			return 2;
		}
		std::cout << "Sent." << std::endl;
		
		std::cout << "Receiving data..." << std::endl;
		retval = p->recv(pack);
		if (retval) {
			std::cout << "Error receiving data: " << retval << std::endl;
			return 3;
		}
		std::cout << "Received." << std::endl;

		std::cout << "Disconnecting." << std::endl;
		p->disconnect(address);
		std::cout << "Test successful." << std::endl;
	}
	else { // server
		std::cout << "Opening server." << std::endl;
		if (p->listen() != 0) {
			std::cout << "Could not listen." << std::endl;
			return 1;
		}	

		while (1) {
			std::cout << "Listening for connections." << std::endl;
			address = p->accept();
			if (address.status < 0 || address.fd < 0) {
				std::cout << "Error: accept returned " << address.status << std::endl;
				return 2;
			}
			std::cout << "Accepted a new connection." << std::endl;

			std::cout << "Receiving data..." << std::endl;
			pack.setSenderNode(everybodyId);
			retval = p->recv(pack);
			if (retval) {
				std::cout << "Error receiving data: " << retval << std::endl;
				return 3;
			}
			std::cout << "Received." << std::endl;

			sleep(1);

			std::cout << "Sending data..." << std::endl;
			pack.setData(data, strlen(data));
			pack.setTargetNodes(baseServerId);
			if (gd.sendPacket(pack) == false) {
				std::cout << "Error sending data:" << std::endl;
				return 4;
			}
			std::cout << "Sent." << std::endl;
			std::cout << "Disconnecting." << std::endl;
			p->disconnect(address);
			std::cout << "Test successful." << std::endl;
		}	
	}
	return 0;
}
